#define INCL_DOSMODULEMGR
#include <os2.h>

#include "language.h"

#include "addinqu.h"
#include "newstr.h"
#include "misc.h"

AddInQueue::AddInQueue( char *dll, char *par ) :
  name( newStr( dll ) ),
  params( newStr( par ) ),
  handle( 0 ),
  init( 0 ),
  done( 0 ),
  scanInit( 0 ),
  scanDone( 0 ),
  scan( 0 ),
  sendInit( 0 ),
  sendDone( 0 ),
  send( 0 )
{
  char buf[80];
  APIRET rc = DosLoadModule( buf, sizeof(buf), name, &handle );

  if( rc )
  {
    log( Msg_ErrDllNotLoaded, name, rc );
    os2Message( rc );
NotFound:
    handle = 0;
    return;
  }

  union SearchAddr
  {
    PFN                pfn;
    AddInQueueInit     *init;
    AddInQueueDone     *done;
    AddInQueueScanInit *scanInit;
    AddInQueueScanDone *scanDone;
    AddInQueueScan     *scan;
    AddInQueueSend     *send;
    AddInQueueSendInit *sendInit;
    AddInQueueSendDone *sendDone;
  } procedureAddress;

  char *search = "INIT";

  rc = DosQueryProcAddr( handle, 0, search, &procedureAddress.pfn );

  if( rc )
  {
ErrorFind:
    log( Msg_UnableFindProc, search, name );
    os2Message( rc );
FreeModule:
    DosFreeModule( handle );
    goto NotFound;
  }

  init = procedureAddress.init;

  search = "DONE";

  rc = DosQueryProcAddr( handle, 0, search, &procedureAddress.pfn );

  if( rc )
    goto ErrorFind;

  done = procedureAddress.done;

  search = "SCAN";

  rc = DosQueryProcAddr( handle, 0, search, &procedureAddress.pfn );

  if( rc )
    goto ErrorFind;

  scan = procedureAddress.scan;

  search = "SCANINIT";

  rc = DosQueryProcAddr( handle, 0, search, &procedureAddress.pfn );

  if( rc )
    goto ErrorFind;

  scanInit = procedureAddress.scanInit;

  search = "SCANDONE";

  rc = DosQueryProcAddr( handle, 0, search, &procedureAddress.pfn );

  if( rc )
    goto ErrorFind;

  scanDone = procedureAddress.scanDone;

  search = "SEND";

  rc = DosQueryProcAddr( handle, 0, search, &procedureAddress.pfn );

  if( rc )
    goto ErrorFind;

  send = procedureAddress.send;

  search = "SENDINIT";

  rc = DosQueryProcAddr( handle, 0, search, &procedureAddress.pfn );

  if( rc )
    goto ErrorFind;

  sendInit = procedureAddress.sendInit;

  search = "SENDDONE";

  rc = DosQueryProcAddr( handle, 0, search, &procedureAddress.pfn );

  if( rc )
    goto ErrorFind;

  sendDone = procedureAddress.sendDone;

  rc = (*init)( params, &log );

  if( rc )
  {
    log( Msg_ErrInitDll, name, rc );
    goto FreeModule;
  }
}

AddInQueue::~AddInQueue()
{
  if( handle )
  {
    (*done)();
    DosFreeModule( handle );
  }

  delete name;
  delete params;
}

int AddInQueue::equal( const BaseObject &t ) const
{
  return !stricmp( name, (*(const AddInQueue *)&t).name );
}

char *AddInQueue::nameOfClass() const
{
  return "AddInQueue";
}

void AddInQueue::print( ostream& o ) const
{ 
  o << name << char(':')
    << (params==NULL ? "<NULL>" : params) << "->"
    << "h="  << handle   << char(',')
    << "i="  << init     << char(',')
    << "d="  << done     << char(',')
    << "si=" << scanInit << char(',')
    << "sd=" << scanDone << char(',')
    << "s="  << scan     << char(',')
    << "ti=" << sendInit << char(',')
    << "td=" << sendDone << char(',')
    << "t="  << send;
}

AddInQueueList::AddInQueueList() :
  Collection( 10, 5, SemaphoreModeNoNeed )
{
}
  
AddInQueue* AddInQueueList::Find( AddInQueue& item, LockMode lock ) const
{
  return (AddInQueue*)Collection::Find( item, lock );
}
    
AddInQueue& AddInQueueList::operator [] ( const int pos ) const
{
  return *(AddInQueue *)&(Collection::operator[]( pos ));
}

AddInQueueIterator::AddInQueueIterator( AddInQueueList& ini ) :
  CollectionIterator( *(Collection *)&ini )
{
}
  
AddInQueue& AddInQueueIterator::operator () () const
{
  return *(AddInQueue *)&(CollectionIterator::operator()());
}

int _System scanCallBack( Queue *que, Address *address, Traffic *traf )
{
  FIDOAddress addr( *address );
  NodeConfig *addObj( new NodeConfig( &addr ) ),
             *inList( systemList->Find( *addObj ) );

  if( inList!=NULL )
  {
    delete addObj;
    addObj = new NodeConfig( inList );
  }

  QueueEntry *cur( new QueueEntry( *addObj ) );

  memcpy( cur->getTrafficPointer(), traf, sizeof( Traffic ) );

  QueueEntry *found( que->Find( *cur, LockModeNoNeed ) );

  if( found==NULL )
    que->add( *cur );
  else
  {
    delete cur;
    found->addTraffic( *traf );
  }

  delete addObj;

  return 0;
}

void AddInQueue::scanQueue( Queue *que )
{
  if( !handle )
    return;

  (*scanInit)();
  (*scan)    ( que, scanCallBack );
  (*scanDone)();
}

void AddInQueueList::scan( Queue *que )
{
  for( AddInQueueIterator list( *this ); list; list++ )
    list().scanQueue( que );
}
