#define INCL_DOS
#define INCL_NOPMAPI
#include <os2.h>

#include <process.h>
#include <io.h>
#include <stdlib.h>
#include <builtin.h>

#include "except.h"
#include "misc.h"
#include "language.h"

class Beep
{
public:
  long freq, dur;

  Beep( long _freq, long _dur ) : freq( _freq ), dur( _dur ) {}
};

static HMODULE hmoduleMMPM = 0;
static HQUEUE queue=0xffffffffl;

union SearchAddr
{
  PFN pfn1;
  ULONG (* APIENTRY pfn)( PSZ    lpstrCommandBuf,
                          PSZ    lpstrReturnString,
                          USHORT wReturnLength,
                          HWND   hCallBack,
                          USHORT wUserParm );

  ULONG (* APIENTRY pfn2)( ULONG  dwError,
                           PSZ    lpstrBuffer,
                           USHORT wLength );
};

static SearchAddr procedureAddress = { (PFN)NULL },
                  errorAddress = { (PFN)NULL };

int SendString( char *cmd )
{
  APIRET rc=(*procedureAddress.pfn)( cmd, NULL, 0, 0, 0 );

  if( rc )
  {
    char buf[128] = { 0 };

    if( errorAddress.pfn==NULL || errorAddress.pfn2( rc, buf, sizeof(buf)-1 ) )
      *buf = 0;

    buf[ sizeof(buf)-1 ] = 0;

    log( Msg_ErrUnablePlay, cmd, rc, *buf ? ": " : "", buf );
  }

  return int( rc );
}

void ProcessWave( char *name )
{
  if( procedureAddress.pfn==NULL )
    return;

  char drive[_MAX_DRIVE],
       dir[_MAX_DIR],
       nn[_MAX_FNAME],
       ext[_MAX_EXT];

  _splitpath( name, drive, dir, nn, ext );

  if( !*ext )
    strcpy( ext, ".Wav" );

  if( miscInfo.wavePath!=NULL && !*dir && !*drive )
  {
    name = (char *)_alloca( strlen(miscInfo.wavePath)+strlen(nn)+strlen(ext)+1 );
    strcpy( name, miscInfo.wavePath );
  }
  else
  {
    name = (char *)_alloca( strlen(drive)+strlen(dir)+strlen(nn)+strlen(ext)+1 );
    strcpy( name, drive );
    strcat( name, dir );
  }

  strcat( name, nn );
  strcat( name, ext );

  capitalize( name );

  if( access( name, 0 ) )
  {
    log( Msg_ErrUnableFind, name, strError() );
    return;
  }

  static char c1w[] = "WAVEAUDIO",
              c1s[] = "SEQUENCER";
  char *device;

  if( !stricmp( ext, ".Wav" ) || !stricmp( ext, ".Wave" ) )
    device = c1w;
  else if( !stricmp( ext, ".Mid" ) || !stricmp( ext, ".Midi" ) )
    device = c1s;
  else
  {
    log( Msg_ErrIllegalPlayType, name );
    return;
  }

  static char *cmd2 = NULL,
              *cmd3 = NULL,
              *cmd4 = NULL,
              *cmd5 = NULL,
              wait[] = " Wait",
              alias[32];

  if( cmd2==NULL )
  {
    char vol[16];
    static char c2[]  = "Load ",
                c3[]  = "Set ",
                c4[]  = "Play ",
                c5[]  = "Close ",
                set[] = " Audio Volume ";

    sprintf( alias, "BrakePlayAlias%d", miscInfo.taskNumber );
    sprintf( vol, "%d", miscInfo.volume );
    cmd2 = new char [ sizeof(c2)+strlen(alias) ];
    cmd3 = new char [ sizeof(c3)+sizeof(set)+sizeof(wait)+strlen(alias)-2+strlen(vol) ];
    cmd4 = new char [ sizeof(c4)+sizeof(wait)+strlen(alias)-1 ];
    cmd5 = new char [ sizeof(c5)+sizeof(wait)+strlen(alias)-1 ];
    strcpy( cmd2, c2 );
    strcat( cmd2, alias );
    strcpy( cmd3, c3 );
    strcat( cmd3, alias );
    strcat( cmd3, set );
    strcat( cmd3, vol );
    strcat( cmd3, wait );
    strcpy( cmd4, c4 );
    strcat( cmd4, alias );
    strcat( cmd4, wait );
    strcpy( cmd5, c5 );
    strcat( cmd5, alias );
    strcat( cmd5, wait );
  }

  static char c1[]  = "Open ",
              c1a[] = " Alias ",
              sha[] = " Shareable";
  char *cmd1 = (char *)_alloca( sizeof(c1)+sizeof(wait)+sizeof(sha)+
                                sizeof(c1a)+
                                strlen(alias)+strlen(device)-3 );

  strcpy( cmd1, c1 );
  strcat( cmd1, device );
  strcat( cmd1, c1a );
  strcat( cmd1, alias );
  strcat( cmd1, sha );
  strcat( cmd1, wait );

  char *cmd = (char *)_alloca( strlen(name)+strlen(cmd2)+sizeof(wait)+3 );

  strcpy( cmd, cmd2 );
  strcat( cmd, " \"" );
  strcat( cmd, name );
  strcat( cmd, "\"" );
  strcat( cmd, wait );

  if( !SendString( cmd1 ) )
  {
    int rc = SendString( cmd );

    if( !rc && miscInfo.volume )
      rc = SendString( cmd3 );

    if( !rc )
      SendString( cmd4 );

    SendString( cmd5 );
  }
}

void _Optlink beepThread( void * )
{
  registerExceptionHandler
//  DosSetPriority( PRTYS_THREAD, PRTYC_FOREGROUNDSERVER, 0x1f, 0 );
  DosSetPriority( PRTYS_THREAD, PRTYC_REGULAR, -0x1f, 0 );

  while( 1 )
  {
    Beep *beep = NULL;
    REQUESTDATA req;
    ULONG len;
    BYTE prior;

    APIRET rc = DosReadQueue( queue, &req, &len, (PPVOID)&beep,
			      0, DCWW_WAIT, &prior, 0 );

    if( req.ulData==255 )
    {
      deRegisterExceptionHandler
      return;
    }

    if( miscInfo.noSound==NULL || !miscInfo.noSound->isWorked() )
      switch( req.ulData )
      {
      default:
        log( "~BUG~ beepThread: illegal req.ulData" );
        delete (void *)beep;
        break;

      case 1:
        if( procedureAddress.pfn!=NULL )
          ProcessWave( (char *)(void *)beep );

        delete (char *)beep;
        break;

      case 0:
        if( !rc && beep!=NULL && len==sizeof(Beep) )
          if( beep->freq )
            DosBeep( beep->freq, beep->dur );
          else
            sleep( beep->dur );

        delete beep;
        break;
      }
  }
}

static long beepTID = -1;

void deInitBeepProcess()
{
  if( beepTID==-1 )
    return;

  DosWriteQueue( queue, 255, 1, (PVOID)deInitBeepProcess, 0 );
  DosWaitThread( (PTID)&beepTID, DCWW_WAIT );
  deleteQueue( &queue );

  if( hmoduleMMPM )
    DosFreeModule( hmoduleMMPM );
}

char needSoundThread = 0;

void initBeepProcess()
{
  if( !needSoundThread )
    return;

  APIRET rc = createQueue( &queue, "Beep" );

  if( rc )
  {
    log( Msg_ErrBeepQueue, rc );
    os2Message( rc );
Abort:
    queue = 0xffffffffl;
    return;
  }

  char MMPM = 0;

  if( !miscInfo.enableMMPM )
    log( Msg_MMPMNotEnabled );
  else if( noPMFeatures )
    log( Msg_MMPMNotAvail );
  else if( needSoundThread&2 )
  {
    char buf[80];
    char *dll = "MDM";

    rc = DosLoadModule( buf, sizeof(buf), dll, &hmoduleMMPM );

    if( rc )
      log( Msg_MMPMNotFound );
    else
    {
      MMPM = 1;
      rc = DosQueryProcAddr( hmoduleMMPM, 0, "mciSendString",
                             &procedureAddress.pfn1 );

      if( rc )
      {
        log( Msg_UnableFindProc, "mciSendString", dll );
        os2Message( rc );
        procedureAddress.pfn = NULL;
        MMPM = 0;
      }

      rc = DosQueryProcAddr( hmoduleMMPM, 0, "mciGetErrorString",
                             &errorAddress.pfn1 );

      if( rc )
      {
        log( Msg_UnableFindProc, "mciGetErrorString", dll );
        os2Message( rc );
        procedureAddress.pfn = NULL;
        errorAddress.pfn = NULL;
      }

      if( procedureAddress.pfn != NULL )
      {
        char buf[128] = { 0 };

        rc = (*procedureAddress.pfn)( "info waveaudio product wait",
                                      buf, sizeof(buf), 0, 0 );

        if( !rc && *buf )
          log( Msg_MMPMDevice, buf );
      }
    }
  }

  beepTID = -1;

  if( !MMPM && !(needSoundThread&(~2)) )
    goto DoNotStartThread;

  beepTID = _beginthread( beepThread, 0, 40960, 0 );

  if( beepTID==-1 )
  {
    log( Msg_ErrBeepServ, strError() );

DoNotStartThread:

    deleteQueue( &queue );

    if( hmoduleMMPM )
      DosFreeModule( hmoduleMMPM );

    goto Abort;
  }

  log( Msg_ThrBeepServ, beepTID );
}

void makeBeep( int f, int d )
{
  if( queue!=0xffffffffl )
    DosWriteQueue( queue, 0, sizeof(Beep), (PVOID)new Beep( f, d ), 0 );
}

void playMelody( int *a )
{
/*
;   ࢮ ᫮  ⢥ન  -  ⮢ ⮭ (Hz) (0 砥 砭)
;   2-                       -  饭 ⮭  ⠪
;   3-                       -  ⥫쭮 ⠪ (ᥪ)
;   4-                       -  ᫮ ⠪⮢
;           Case     Rep.     Pattern(s)
  int q0[]={ 2, 1, 1000, 100, 100, 4, 1300, -100, 100, 3 },  // connect
      q1[]={ 1, 3, 1000, 500, 100, 2 },			     // beg sess
      q2[]={ 1, 5, 4000, 1000, 10, 5 },			     // end sess
      q3[]={ 1, 1, 900, 10, 20, 25 },			     // abort
      q4[]={ 1, 5, 300, 150, 6, 10 },			     // human
      q5[]={ 2, 1, 100, 10, 10, 100,  1090, -10, 10, 99 },   // error
      q6[]={ 3, 4, 3500, 0, 50, 1, 3000, 0, 50, 1, 4000, 0, 5, 1 }; //ring
      q8[]={ 1, 1, 2048, 0, 249, 1 } 				//alarm
      q9[]={}							//start
      q10[]={ 1, 1, 1024, 0, 249, 1 };				//bell

*/

  if( queue!=0xffffffffl )
    for( int repeatCount=0; repeatCount<a[1]; repeatCount++ )
      for( int taktCount=0; taktCount<a[0]; taktCount++ )
        for( int i=0, tone = a[2+taktCount*4]; i<a[2+taktCount*4+3];
             i++, tone+=a[2+taktCount*4+1] )
          makeBeep( tone, a[2+taktCount*4+2] );
}

void playWaveFile( char *name )
{
  if( queue!=0xffffffffl )
    DosWriteQueue( queue, 1, strlen(name)+1, newStr( name ), 0 );
}

