#include <process.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <io.h>
#include <string.h>
#include <sys\stat.h>
#include <builtin.h>

#define INCL_DOSEXCEPTIONS
#define INCL_DOSPROCESS
#include <os2.h>

#include "except.h"

#include "language.h"

#include "misc.h"
#include "32base.h"
#include "direct.h"

void processFile( char *full, char *name, int& count, int task )
{
  char *unl = (char *)_alloca( strlen(full)+strlen(name)+1 );

  strcpy( unl, full );
  strcat( unl, name );
  capitalize( unl );

  struct stat st;
  char needUnlink( 0 );
  time_t curTime = time( NULL );

  if( !task )
    if( !stat( unl, &st ) )
    {
      if( st.st_ctime+miscInfo.bsyAge < curTime )
        needUnlink = 1;
    }
    else
      log( Msg_ErrUnableStat, unl, strError() );
  else
  {
    struct stat st;

    if( stat( unl, &st ) )
      log( Msg_ErrUnableStat, unl, strError() );
    else if( st.st_size )
    {
      FILE *bsy = _fsopen( unl, "r" );

      if( bsy!=NULL )
      {
        int n( int(0xffffffffl) );

        fscanf( bsy, "%d", &n );
        fclose( bsy );

        if( miscInfo.taskNumber==n )
          needUnlink = 2;
      }
      else
        log( Msg_ErrUnableOpen, unl, strError() );
    }
  }

  if( needUnlink )
    if( !forceDelete( unl ) )
    {
      count++;

      long age = curTime - st.st_ctime;

      if( needUnlink==1 )
        log( Msg_InfoBSYDeleteAge, age/60/60, (age/60)%60, age%60, unl );
      else
        log( Msg_InfoBSYDeleteTask, unl );
    }
}

void processPoint( char *fullPath, char *nodePath, int& count, int task )
{
  char *mask( "\\*.bsy" ),
       *path = (char *)_alloca( strlen(fullPath)+strlen(nodePath)+strlen( mask )+1 ),
       *full = (char *)_alloca(strlen(fullPath)+strlen(nodePath)+2 );

  strcpy( path, fullPath );
  strcat( path, nodePath );
  strcat( path, mask );

  strcpy( full, fullPath );
  strcat( full, nodePath );
  strcat( full, "\\" );

  for( Directory rc( path ); rc; rc++ )
  {
    uShort point( 0 ), fict( 0 );
    char *name( rc.getName() );

    if( strlen( name )==12 && split16base( name, fict, point ) )
      processFile( full, name, count, task );
  }
}

void processZone( int outLen, char *zoneStr, int& count,
                  Domain *dom, int task )
{
  uShort zone( 0 );
  char ext[_MAX_EXT];

  {                            //  ᪮ ᯮ짮 fnsplit
    char drive[_MAX_DRIVE],
	 dir[_MAX_DIR],
	 file[_MAX_FNAME];

    _splitpath( zoneStr, drive, dir, file, ext );

    if( *ext )
    {
      if( val16base( ext+1, zone ) )
	return;
    }
    else
      zone = dom->getDefZone();
  }

  char mask[]( "\\*" ),
    *path = (char *)_alloca( outLen+strlen(ext)+sizeof( mask ) ),
    *full = (char *)_alloca( outLen+strlen(ext)+20 ),
    *outbound( dom->getOutbound() );

  strcpy( path, outbound );
  strcpy( path+outLen-1, ext );
  strcat( path, mask );

  strcpy( full, outbound );
  strcpy( full+outLen-1, ext );
  strcat( full, "\\" );

  for( Directory rc( path, SearchAttributes( MayBeArchived|MayBeReadOnly|MayBeDirectory ) );
       rc; rc++ )
  {
    uShort net( 0 ), node( 0 );
    char *name( rc.getName() );

    if( strlen( name )==12 && split16base( name, net, node ) )
      if( (rc.getAttrib()&MayBeDirectory)==MayBeDirectory && !stricmp( name+8, ".pnt" ) )
	processPoint( full, name, count, task );
      else
	if( !stricmp( name+8, ".bsy" ) )
	  processFile( full, name, count, task );
  }

  delete outbound;
}

char volatile processBsyActive = 0;

void resetBsy( int task )
{
  processBsyActive = 1;

  long start( getTimeMark() );

  static char mask[]( ".*" );
  int count( 0 );

  for( DomainIterator list( *outboundInfo.domains ); list; list++ )
  {
    Domain *dom = (Domain *)&list();
    char *outbound( dom->getOutbound() );
    int outLen( strlen(outbound) );
    char *path = new char [ outLen+sizeof(mask) ];

    strcpy( path, outbound );
    strcpy( path+outLen-1, mask );

    for( Directory rc( path, SearchAttributes( MayBeArchived|MayBeReadOnly|
                                               MayBeDirectory ) );
         rc; rc++ )
    {
      char *name( rc.getName() );

      if( (rc.getAttrib()&MayBeDirectory)==MayBeDirectory )
        if( strcmp( name, "." ) && strcmp( name, ".." ) )
	  processZone( outLen, name, count, dom, task );
    }

    delete outbound;
    delete path;
  }

  if( miscInfo.longOutbound!=NULL )
  {
    static char msk[]( "*.Busy" );
    char *search( (char *)_alloca( strlen(miscInfo.longOutbound)+sizeof(msk) ) );

    strcpy( search, miscInfo.longOutbound );
    strcat( search, msk );

    for( Directory rc( search, SearchAttributes(MayBeArchived|MayBeReadOnly) );
         rc; rc++ )
    {
      char *name( newStr( rc.getName() ) );
      uShort zone,
             net,
             node,
             point;
      char *domain;
      Flavour flavour;
      LongNameType type;
      
      if( splitLongName( name, zone, net, node, point,
                         &domain, flavour, type ) && type==busyLongName )
        processFile( miscInfo.longOutbound, rc.getName(), count, task );

      delete name;
    }
  }

  if( count )
  {
    start = getTimeMark() - start;
    log( Msg_InfoBSY, start/60, start%60, count );
  }

  processBsyActive = 0;
}

void _Optlink bsyProcessor( void * )
{
  registerExceptionHandler

  DosSetPriority( PRTYS_THREAD, 0, -miscInfo.priorityDelta, 0 );

  while( 1 )
  {
    resetBsy( 0 );
    sleep( (miscInfo.bsyInterval>1 ? miscInfo.bsyInterval : 1)*1000 );
  }
}
