#define INCL_DOS
#include <os2.h>

#include <fstream.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "misc.h"
#include "buffer.h"
#include "config.h"
#include "cfgparse.h"

#include "language.h"

Macro::Macro( char *n, char *v, char im ) :
  name( newStr( n ) ),
  val( newStr( v ) ),
  imported( im )
{
}

Macro::~Macro()
{
  delete name;
  delete val;
}

int Macro::equal( const BaseObject &t ) const
{
  return stricmp( name, (*(const Macro*)&t).name ) ? 0 : 1;
}

void Macro::print( ostream& s ) const
{
  s << '"' << name << char('=') << val << "\", "
    << (imported ? "imported" : "defined");
}

char *Macro::nameOfClass() const
{
  return "Macro";
}

MacroList::MacroList() :
  Collection( 50, 10, SemaphoreModeNoNeed, DeleteModeDelete )
{
}

Macro* MacroList::Find( Macro& item, LockMode lock ) const
{
  return (Macro*)Collection::Find( item, lock );
}
    
Macro& MacroList::operator [] ( const int pos ) const
{
  return *(Macro *)&(Collection::operator[]( pos ));
}

MacroIterator::MacroIterator( MacroList& ini ) :
  CollectionIterator( *(Collection *)&ini )
{
}
  
Macro& MacroIterator::operator () () const
{
  return *(Macro *)&(CollectionIterator::operator()());
}

ParserStateStackEntry::ParserStateStackEntry( int l, ParserState st, char skip ) :
  beginLine( l ),
  state( st ),
  inSkipMode( skip )
{
}

int ParserStateStackEntry::equal( const BaseObject &t ) const
{
  return this == &t;
}

void ParserStateStackEntry::print( ostream& s ) const
{
}

char *ParserStateStackEntry::nameOfClass() const
{
  return "ParserStateStackEntry";
}

ParserStateStack::ParserStateStack() :
  Collection( 10, 5, SemaphoreModeNoNeed )
{
}
  
ParserStateStackEntry* ParserStateStack::Find( ParserStateStackEntry& item,
                                               LockMode lock ) const
{
  return (ParserStateStackEntry*)Collection::Find( item, lock );
}
    
ParserStateStackEntry& ParserStateStack::operator [] ( const int pos ) const
{
  return *(ParserStateStackEntry *)&(Collection::operator[]( pos ));
}

MacroList macroList;

Buffer inputWord;
short cfgLine = 1;
static short internalCfgLine = 0;
CfgStack *cfgStack = NULL;
ifstream *currentCfg = NULL;
char listConfig = 0;

int need( char *who )
{
  log( Msg_ErrCfgExpect, who, inputWord.str, cfgLine );
  return 2;
}

int isValidDigit( int ch, char base, char& curDigit, char &left )
{
  if( !left )
    return 0;

  left--;
  ch = toupper( ch );

  for( curDigit=0; curDigit<base; curDigit++ )
    if( hexString[curDigit]==ch )
      return 1;

  return 0;
}

int getChar( ifstream *cfg )
{
  char base( 10 ),
       state( 0 ),
       ctlChar( 0 ),
       left( 3 ),
       curDigit( 0 );

  register int ch;

  for( int done(0); !done; )
  {
    ch=cfg->get();

    if( ch == '\n' )
      internalCfgLine++;
    else
      if( ch == EOF )
	break;

    switch( state )
    {
    case 0:
      if( ch=='^' )
	state = 1;
      else
       done = 1;

      break;

    case 1:
      switch( ch )
      {
      case '\n':
	state = 0;
	break;

      case 'r':
	ch = '\r';
	done = 1;
	break;

      case 'n':
	ch = '\n';
	done = 1;
	break;

      case 'b':
	ch = '\b';
	done = 1;
	break;

      case '^':
	ch = '^';
	done = 1;
	break;

      case 't':
	ch = '\t';
	done = 1;
	break;

      case 'f':
	ch = '\f';
	done = 1;
	break;

      case '0':
	base = 8;
	state = 2;
	break;

      case 'x':
      case 'X':
	base = 16;
	left = 2;
	state = 2;
	break;

      default:
	cfg->putback( (char)ch );

        if( ch=='\n' )
          internalCfgLine--;

	if( isDigit(ch) )
	  state = 2;
	else
	{
	  ch = '^';
	  done = 1;
	}

      }

      break;

    case 2:
      if( isValidDigit( ch, base, curDigit, left ) )
	ctlChar = char( ctlChar*base + curDigit );
      else
      {
	cfg->putback( char( ch ) );

        if( ch=='\n' )
          internalCfgLine--;

	ch = ctlChar;
	done = 1;
      }
      break;

    default:
      done = 1;

    }

  }

  return ch;
}

char parserNoRead = 0;
static char wordIsQuoted = 0;
static char needRescan = 1;

char *expandVars( char *strIni )
{
  Buffer ret( strlen(strIni), 10 );

  for( char *str = strIni; *str; str++ )
    if( *str!='#' )
      ret.putInto( *str );
    else if( str[1]=='#' )
    {
      ret.putInto( '#' );
      str++;
    }
    else
    {
      char *str1 = strchr( str+1, '#' );

FindAgain:
      if( str1==NULL )
      {
        ret.putString( str );
        break;
      }

      if( str1[1]=='#' )
      {
        str1 = strchr( str1+2, '#' );
        goto FindAgain;
      }

      char *var = newStr( str+1 );

      var[ str1-str-1 ] = 0;

      Macro test( var, NULL ),
            *found = macroList.Find( test );

      if( found!=NULL )
      {
        delete var;
        var = newStr( found->getValue() );
      }
      else
        log( Msg_ErrMacroNotFound, cfgLine, var );

      if( needRescan )
      {
        char *nvar = expandVars( var );

        delete var;
        var = nvar;
      }

      ret.putString( var );

      delete var;
      str = str1;
    }

  ret.putInto( 0 );

  return newStr( ret.str );
}

int getWordTrivial( ifstream **cfg )
{
  int ch;
  char first;

  if( parserNoRead )
  {
    parserNoRead = 0;
    ch = 0;
    return ch;
  }

  if( !internalCfgLine )
    internalCfgLine = 1;

get_again:

  wordIsQuoted = 0;
  inputWord.reset();
  first = 1;

  do
  {
    for( ch = ' '; isSpace(ch) && ch!=EOF; )
      ch = getChar( *cfg ) ;

    for( char quote( 0 ), done( 0 ); ch!=EOF; ch = getChar( *cfg ) )
    {
      switch( ch )
      {
      case ';':
	if( !quote )
	{
	  while( ch!=EOF && ch!='\n' )
            ch = getChar( *cfg );

	  done = 1;
	}
	else
        {
          if( first )
          {
            first = 0;
            cfgLine = internalCfgLine;
          }

	  inputWord.putInto( (char)ch );
        }
        break;

      case '"':
	ch = getChar( *cfg );

	if( quote )
	  switch( ch )
	  {
	  case '"':
            if( first )
            {
              first = 0;
              cfgLine = internalCfgLine;
            }

	    inputWord.putInto( '"' );
	    break;
	  default:
	    (*cfg)->putback( (char)ch );

            if( ch=='\n' )
              internalCfgLine--;

	  case EOF:
	    done = 1;
	  }
	else
	  switch( ch )
	  {
	  case '"':
	  case EOF:
            if( first )
            {
              first = 0;
              cfgLine = internalCfgLine;
            }
	    inputWord.putInto( '"' );
	    break;
	  default:
	    quote = 1;
            wordIsQuoted = 1;
	    (*cfg)->putback( (char)ch );

            if( ch=='\n' )
              internalCfgLine--;
	  }

	break;

      default:
	if( isSpace(ch) && !quote )
	  done = 1;
	else
        {
          if( first )
          {
            first = 0;
            cfgLine = internalCfgLine;
          }

	  inputWord.putInto( (char)ch );
        }
      }

      if( done )
	break;

    }

  } while( !inputWord.length() && ch!=EOF );

  inputWord.putInto( 0 );

  ch = ch==EOF && inputWord.length()==1 ? 1 : 0;

  if( ch )
    if( !cfgStack->end() )
    {
      delete *cfg;
      *cfg = NULL;
      Config& n = (*cfgStack)[0];

      log( Msg_InfoCfgReturn, cfgLine );
      *cfg = n.getCfg();
      currentCfg = *cfg;
      internalCfgLine = (short)n.getLine();

      cfgStack->Remove( 0, DeleteModeDelete );

      goto get_again;
    }
    else
      return ch;

  char *ret = expandVars( inputWord.str );

  inputWord.reset();
  inputWord.putString( ret );
  inputWord.putInto( 0 );
  delete ret;

  return 0;
}

static ParserStateStack parserStateStack;
static ParserStateStackEntry *curState = NULL;

int getWordPacked( ifstream **cfg )
{
  if( curState==NULL )
  {
    curState = new ParserStateStackEntry( 1, ParserStateNormal, 0 );
    parserStateStack.AddAtHead( *curState );
  }

  while( 1 )
  {
    ParserState macroMode;
    int rc = getWordTrivial( cfg );

    if( rc )
    {
ReturnWithError:
      if( rc==1 && parserStateStack.GetItemsInContainer()>1 )
      {
        log( Msg_ErrCfgUnExpectedHere, cfgLine, "EOF" );

        for( int i = 0; i<parserStateStack.GetItemsInContainer()-1; i++ )
        {
          ParserStateStackEntry& c = *(ParserStateStackEntry *)&(parserStateStack[i]);
          char *what,
               *stt;

          switch( c.getState() )
          {
          case ParserStateIfDef:
            stt = "#IfDef";
            what = "#EndIf";
            break;

          case ParserStateIfNDef:
            stt = "#IfNDef";
            what = "#EndIf";
            break;

          case ParserStateFor:
            stt = "#For";
            what = "#EndFor";
            break;

          default:
            log( "~BUG~ Unknown parser state at line %d -> %d",
                 cfgLine, c.getState() );
            stt = "~BUG~";
            what = "~BUG~";
            break;
          }

          log( Msg_ErrCfgExpectedHere, cfgLine, what, stt, c.getLine() );
        }

        rc = 2;
      }

      return rc;
    }

    if( !stricmp( inputWord.str, "#EndFor" ) )
      if( curState->getState()==ParserStateFor )
      {
DropStack:
        if( parserStateStack.GetItemsInContainer()<2 )
        {
          log( "~BUG~ Corrupted parser state stack at line %d", cfgLine );
          rc = 2;
          goto ReturnWithError;
        }

        parserStateStack.Remove( 0, DeleteModeDelete );
        curState = (ParserStateStackEntry*)&parserStateStack[0];
      }
      else
      {
NotExpected:
        log( Msg_ErrCfgUnExpectedHere, cfgLine, inputWord.str );
        rc = 2;
        goto ReturnWithError;
      }
    else if( !stricmp( inputWord.str, "#EndIf" ) )
      if( curState->getState()==ParserStateIfDef || curState->getState()==ParserStateIfNDef )
        goto DropStack;
      else
        goto NotExpected;
    else if( !stricmp( inputWord.str, "#IfDef" ) )
    {
      macroMode = ParserStateIfDef;

ProcessMacro:
      rc = getWordTrivial( cfg );

      if( rc )
        goto ReturnWithError;

      Macro test( inputWord.str, NULL );
      int found = macroList.HasMember( test );
      char skip = curState->getSkipMode();

      if( !skip )
        if( !found && macroMode==ParserStateIfDef )
          skip = 1;
        else if( found && macroMode==ParserStateIfNDef )
          skip = 1;

      curState = new ParserStateStackEntry( cfgLine, macroMode, skip );
      parserStateStack.AddAtHead( *curState );
    }
    else if( !stricmp( inputWord.str, "#IfNDef" ) )
    {
      macroMode = ParserStateIfNDef;
      goto ProcessMacro;
    }
    else if( !stricmp( inputWord.str, "#For" ) )
    {
      char isForUs = 0;
      char first = 1;

      while( 1 )
      {
        rc = getWordTrivial( cfg );

        if( rc )
          goto ReturnWithError;

        char *str = inputWord.str;

        if( *str=='!' )
          str++;

        char *end;
        long parserTask = strtoul( str, &end, 10 );

        if( *end )
          if( first )
          {
            need( "TaskNumber" );
            rc = 2;
            goto ReturnWithError;
          }
          else
            break;

          first = 0;

        if( parserTask==miscInfo.taskNumber && str==inputWord.str )
          isForUs = 1;
        else if( parserTask!=miscInfo.taskNumber && str!=inputWord.str )
          isForUs = 1;
      }

      parserNoRead = 1;
      char skip = curState->getSkipMode();

      if( !skip )
        skip = char( !isForUs );

      curState = new ParserStateStackEntry( cfgLine, ParserStateFor, skip );
      parserStateStack.AddAtHead( *curState );
    }
    else
      if( !curState->getSkipMode() )
        return 0;
  }
}

int builtinMacro()
{
  char val[32];
  ULONG b[26];

  DosQuerySysInfo( QSV_MAX_PATH_LENGTH, 26, b, sizeof(b) );

  sprintf( val, "%d", miscInfo.taskNumber );
  macroList.Replace( *new Macro( "TASK", val, 0 ) );
  
  sprintf( val, "%d", b[QSV_VERSION_MAJOR-1] );
  macroList.Replace( *new Macro( "OSVMAJOR", val, 0 ) );
  
  sprintf( val, "%d", b[QSV_VERSION_MINOR-1] );
  macroList.Replace( *new Macro( "OSVMINOR", val, 0 ) );
  
  sprintf( val, "%d", b[QSV_VERSION_REVISION-1] );
  macroList.Replace( *new Macro( "OSVREVISION", val, 0 ) );
  
  sprintf( val, "%d", b[QSV_TOTPHYSMEM-1] );
  macroList.Replace( *new Macro( "TOTPHYSMEM", val, 0 ) );
  
  sprintf( val, "%d", b[26-1] ? b[26-1] : 1 );
  macroList.Replace( *new Macro( "CPU", val, 0 ) );
  
  macroList.Replace( *new Macro( "VERSION", (char *)M_VERSION, 0 ) );

  return 0;
}

int importMacro( ifstream **cfg )
{
  int rc = getWordPacked( cfg );

  if( rc )
    return rc;

  int all = !stricmp( inputWord.str, "#All" ),
      found = 0;
  extern char **envPtr;

  for( char **ptr=envPtr; *ptr!=NULL; ptr++ )
  {
    char *name = newStr( *ptr ),
         *val = strchr( name, '=' );

    if( val!=NULL )
    {
      *(val++) = 0;

      if( all || (!all && !stricmp( inputWord.str, name ) ) )
      {
        macroList.Replace( *new Macro( name, val, 1 ) );
        found = 1;
      }
    }

    delete name;
  }

  if( !found )
    log( Msg_ErrEnvNotFound, cfgLine, inputWord.str );

  return 0;
}

int exportMacro( ifstream **cfg )
{
  int rc = getWordPacked( cfg );

  if( rc )
    return rc;

  int all = !stricmp( inputWord.str, "#All" ),
      found = 0;

  for( MacroIterator list( macroList ); list; list++ )
  {
    Macro& cur( *(Macro *)&(list()) );

    if( !cur.isImported() )
    {
      char *name = cur.getName(),
           *val  = cur.getValue();

      if( all || (!all && !stricmp( inputWord.str, name ) ) )
      {
        char *str  = new char [ strlen(name)+strlen(val)+2 ];

        strcpy( str, name );
        strupr( str );
        strcat( str, "=" );
        strcat( str, val );

        if( putenv( str ) )
          log( Msg_ErrUnableAddEnv, cfgLine, str );

        found = 1;
      }
    }
  }

  if( !found )
    log( Msg_ErrMacroNotFound, cfgLine, inputWord.str );

  return 0;
}

int getWordAndNoList( ifstream **cfg )
{
  int rc=0;

  while( 1 )
  {
    rc = getWordPacked( cfg );

    if( rc )
      return rc;

    if( !stricmp( inputWord.str, "#Include" ) )
    {
      rc = getWordPacked( cfg );

      if( rc )
      {
        log( Msg_ErrCfgNeedInclude, cfgLine );
        return 2;
      }

      ifstream *s = new ifstream( inputWord.str, ios::in|ios::nocreate );
      cfgFileList.registerFile( inputWord.str );

      if( !s->good() )
      {
        log( Msg_ErrCfgNotOpen, cfgLine, inputWord.str, strError() );
        delete s;
      }
      else
      {
        log( Msg_InfoCfgInclude, cfgLine, inputWord.str );
        cfgStack->AddAtHead( *new Config( *cfg, cfgLine ) );
        *cfg = s;
        currentCfg = s;
        internalCfgLine = 1;
      }
    }
    else if( !stricmp( inputWord.str, "#Rescan" ) )
      needRescan = 1;
    else if( !stricmp( inputWord.str, "#NoRescan" ) )
      needRescan = 0;
    else if( !stricmp( inputWord.str, "#DumpMacro" ) )
      cout << macroList;
    else if( !stricmp( inputWord.str, "#Builtin" ) )
      rc = builtinMacro();
    else if( !stricmp( inputWord.str, "#Import" ) )
      rc = importMacro( cfg );
    else if( !stricmp( inputWord.str, "#Export" ) )
      rc = exportMacro( cfg );
    else if( !stricmp( inputWord.str, "#Display" ) )
    {
      rc = getWordPacked( cfg );

      if( rc )
      {
        log( Msg_ErrCfgNeedDisplay, cfgLine );
        return 2;
      }

      log( Msg_InfoCfgComment, cfgLine, inputWord.str );
    }
    else if( !stricmp( inputWord.str, "#Error" ) )
    {
      rc = getWordPacked( cfg );

      if( rc )
      {
        log( Msg_ErrCfgNeedError, cfgLine );
        return 2;
      }

      log( Msg_InfoCfgError, cfgLine, inputWord.str );
      return 2;
    }
    else if( !stricmp( inputWord.str, "#Remove" ) )
    {
      rc = getWordPacked( cfg );

      if( rc )
      {
        log( Msg_ErrCfgNeedMacroName, cfgLine );
        return 2;
      }

      Macro rem( inputWord.str, NULL );

      macroList.Remove( rem );
    }
    else if( !stricmp( inputWord.str, "#Macro" ) )
    {
      rc = getWordPacked( cfg );

      if( rc )
      {
        log( Msg_ErrCfgNeedMacroName, cfgLine );
        return 2;
      }

      char *name = newStr( inputWord.str );

      rc = getWordPacked( cfg );

      if( rc )
      {
        log( Msg_ErrCfgNeedMacroDef, cfgLine );
        delete name;
        return 2;
      }

      macroList.Add( *new Macro( name, inputWord.str ) );
      delete name;
    }
    else
      break;

  }

  return 0;
}

int getWord( ifstream **cfg )
{
  static int line = 0;
  int oldSkip = parserNoRead;
  int rc = getWordAndNoList( cfg );

  if( listConfig && !oldSkip )
  {
    if( line!=cfgLine )
    {
      line = cfgLine;
      cout << char('\n') << line << char('\t');
    }

    if( rc )
      if( rc==1 )
        cout << "\n*EOF*\n";
      else
        cout << "\n*Error " << rc << "*\n";
    else
    {
      char *ch = wordIsQuoted ? "\"" : "";

      cout << ch << inputWord.str << ch << char(' ');
    }
  }

  return rc;
}

void closeAllConfig( ifstream *cfg )
{
  delete cfg;
  currentCfg = NULL;

  for( CfgStackIterator list( *cfgStack ); list; list++ )
     delete list().getCfg();

  cfgStack->Flush();
  parserStateStack.Flush();
  curState = NULL;
  macroList.Flush();
  internalCfgLine = 0;
}
