#include <sys/stat.h>

#include "language.h"

#include "config.h"
#include "misc.h"

Config::Config( ifstream *c, int l ) :
  cfg( c ),
  line( l )
{
}

int Config::equal( const BaseObject &t ) const
{
  return cfg == (*(const Config *)&t).cfg;
}

void Config::print( ostream& s ) const
{
}

char *Config::nameOfClass() const
{
  return "Config";
}

CfgStack::CfgStack() :
  Collection( 5, 5, SemaphoreModeNoNeed )
{
}

Config* CfgStack::Find( Config& item, LockMode lock ) const
{
  return (Config*)Collection::Find( item, lock );
}
    
Config& CfgStack::operator [] ( const int pos ) const
{
  return *(Config *)&(Collection::operator[]( pos ));
}

int CfgStack::end() const
{
  if( GetItemsInContainer() )
    return 0;

  if( currentCfg==NULL )
    return 1;

  return !currentCfg->good();
}

CfgStackIterator::CfgStackIterator( CfgStack& ini ) :
  CollectionIterator( *(Collection *)&ini )
{
}
  
Config& CfgStackIterator::operator () () const
{
  return *(Config *)&(CollectionIterator::operator()());
}

CfgFile::CfgFile( char *n ) :
  name( newStr( n ) )
{
  struct stat st;

  if( stat( n, &st ) )
    time = 0;
  else
    time = st.st_mtime;
}

CfgFile::~CfgFile()
{
  delete name;
}

int CfgFile::needReRead() const
{
  struct stat st;
  int ret = 0;

  if( stat( name, &st ) )
    st.st_mtime = 0;

  if( time != st.st_mtime )
    ret = 1;

  if( ret )
    log( Msg_InfoCtlChanged, name );

  return ret;
}

int CfgFile::equal( const BaseObject &t ) const
{
  return !stricmp( name, (*(const CfgFile *)&t).name );
}

void CfgFile::print( ostream& s ) const
{
  s << name << ",MTime:" << time;
}

char *CfgFile::nameOfClass() const
{
  return "CfgFile";
}

CfgFileList::CfgFileList() :
  Collection( 5, 5, SemaphoreModeNoNeed )
{
}

CfgFile* CfgFileList::Find( CfgFile& item, LockMode lock ) const
{
  return (CfgFile*)Collection::Find( item, lock );
}
    
CfgFile& CfgFileList::operator [] ( const int pos ) const
{
  return *(CfgFile *)&(Collection::operator[]( pos ));
}

int CfgFileList::needReRead() const
{
  int ret = 0;

  for( CfgFileListIterator list(*this); list && !ret; list++ )
    ret = list().needReRead();

  return ret;
}

void CfgFileList::registerFile( char *name )
{
  CfgFile *item = new CfgFile( name );

  Add( *item );
}

CfgFileListIterator::CfgFileListIterator( const CfgFileList& ini ) :
  CollectionIterator( *(const Collection *)&ini )
{
}
  
CfgFile& CfgFileListIterator::operator () () const
{
  return *(CfgFile *)&(CollectionIterator::operator()());
}
