#ifndef __CONFIG_H

#define __CONFIG_H

#include <fstream.h>
#include <string.h>

#include "buffer.h"
#include "misc.h"
#include "class.h"

int getWord( ifstream **cfg );
void closeAllConfig( ifstream *cfg );
int need( char *who );

extern Buffer inputWord;
extern short cfgLine;
extern ifstream *currentCfg;
extern char parserNoRead;
extern char listConfig;

class Config : public BaseObject
{
private:
  ifstream *cfg;
  int line;

  int equal( const BaseObject &t ) const;
  void print( ostream& s ) const;
  char *nameOfClass() const;

public:

  Config( ifstream *c, int l );

  ifstream *getCfg() const { return cfg; }
  int getLine() const { return line; }
};

class CfgStack : public Collection
{
public:
  CfgStack();

  Config* Find( Config& item, LockMode lock=LockModeNeed ) const;
  Config& operator [] ( const int pos ) const;

  int end() const;
};

class CfgStackIterator : public CollectionIterator
{
public:
  CfgStackIterator( CfgStack& ini );
  
  Config& operator () () const;
};


extern CfgStack *cfgStack;

class CfgFile : public BaseObject
{
private:
  char *name;
  long time;

  int equal( const BaseObject &t ) const;
  void print( ostream& s ) const;
  char *nameOfClass() const;

public:

  CfgFile( char *n );
  ~CfgFile();

  int needReRead() const;
};

class CfgFileList : public Collection
{
public:

  CfgFileList();

  CfgFile* Find( CfgFile& item, LockMode lock=LockModeNeed ) const;
  CfgFile& operator [] ( const int pos ) const;

  int needReRead() const;
  void registerFile( char *name );
};

class CfgFileListIterator : public CollectionIterator
{
public:
  CfgFileListIterator( const CfgFileList& ini );
  
  CfgFile& operator () () const;
};

extern CfgFileList cfgFileList;

int storeNumber( ifstream **cfg, int &number, char *msg );
int storeLong( ifstream **cfg, long &number, char *msg );
int storePath( ifstream **cfg, char **where, char *what, int flag=1 );

int processKeys( ifstream **cfg );
int processBothFlag( ifstream **Cfg, Schedule **list, int *count );
int processExitStartUp( ifstream **cfg, Schedule **list, int *count );
int processActions( ifstream **cfg );

extern char needSoundThread;

#endif