#include <io.h>
#include <ctype.h>
#include <stdlib.h>
#include <builtin.h>

#include "language.h"

#include "misc.h"
#include "fidoaddr.h"
#include "queue.h"
#include "direct.h"

int stringEnter( int len, char *str, char *prompt )
{
  if( detachedSession )
    return 1;

  char *work = (char *)_alloca( len+1 );

  memset( work, 0, len+1 );
  strcpy( work, str );

  Color buttonTxt( colorSet.popUpItem ),
	buttonHig( colorSet.popUpBold );

  DLG_CONTROL Control[] =
  {
    { "-> ",
      CSF_DATAENTRY, CSS_ACTIVATE|CSS_CALLNOTHING,
      k_Alt_F,
      1,						//was: 0
      2, 2,
      1, (len<<8)|33,
      buttonTxt.fore, buttonTxt.back,
      buttonTxt.fore, buttonTxt.back,
      work, "", NULL
    },
    {
      Msg_ScrPrmtOK, CSF_BUTTON, CSS_ACTIVATE|CSS_EXIT,
      k_Alt_D,
      1,
      6, 4,
      1, 12,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      NULL, "", NULL
    },
    {
      Msg_ScrPrmtCancel, CSF_BUTTON, CSS_ACTIVATE|CSS_EXIT,
      k_Esc,
      1,
      24, 4,
      1, 12,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      NULL, "", NULL
    },
    {
      NULL, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, "", NULL
    }
  };

  DIALOG dial =
  {
    (Z_ScreenWidth-40)/2+1, (Z_ScreenHeight-7)/2+1,
    7, 40, DBF_NOHINT,
    (int(popUpColor)<<8)|0x20,
    popUpColor.fore, popUpColor.back,
    1, 1, 1,
    BS_SINGLELINE, popUpColor.fore, popUpColor.back,
    DS_LOWERRIGHT, 0x07,
    prompt,
    NULL,
    Control
  };

  int rc = Z_HandleDialog( &dial );

  strncpy( str, work, len );

  for( int i( strlen(str)-1 ); i>=0 && str[i]==' '; i-- ) continue;

  str[++i]=0;

  return rc!=1;
}

static Flavour flavourTable[5] =
{
  immedFlavour, crashFlavour,
  directFlavour, normalFlavour,
  holdFlavour
};

char flavourChar( int flavour )
{
  switch( flavour )
  {
  case 0:
    return 'I';

  case 1:
    return 'C';

  case 2:
    return 'D';

  case 3:
    return 'F';

  case 4:
    return 'H';
  }

  return '|';
}


void mailerChange()
{
  if( detachedSession )
    return;

  QueueEntry *node = queue->getCurrent();

  if( node==NULL )
  {
    Z_InformUser( Msg_ScrError, Msg_ScrQueEmpty, Msg_ScrContinue, NULL );
    return;
  }

  char *bink  = node->getBinkName();

  if( bink==NULL )
  {
    Z_InformUser( Msg_ScrError, Msg_ScrIllAddress, Msg_ScrContinue, NULL );
    return;
  }

  Color buttonTxt( colorSet.popUpItem ),
	buttonHig( colorSet.popUpBold );

  int fromFlavour(-1),
      toFlavour(-1);

  DLG_CONTROL Control[] =
  {
    {
      Msg_ScrPrmtFromTxt, CSF_TEXT, CSS_NOACTIVATE|CSS_CALLNOTHING,
      0,
      0,
      6, 2,
      1, strlen(Msg_ScrPrmtFromTxt),
      popUpColor.fore, popUpColor.back,
      popUpColor.fore, popUpColor.back,
      NULL, NULL, NULL
    },
    {
      Msg_ScrPrmtToTxt, CSF_TEXT, CSS_NOACTIVATE|CSS_CALLNOTHING,
      0,
      0,
      25, 2,
      1, strlen(Msg_ScrPrmtToTxt),
      popUpColor.fore, popUpColor.back,
      popUpColor.fore, popUpColor.back,
      NULL, NULL, NULL
    },
// from
    {
      Msg_ScrPrmtImmed, CSF_ONEOFBUTTON, CSS_EXIT,
      k_Alt_M,
      2,
      6, 3,
      1, 0,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      &fromFlavour, "", NULL
    },
    {
      Msg_ScrPrmtCrash, CSF_ONEOFBUTTON, CSS_EXIT,
      k_Alt_A,
      3,
      6, 4,
      1, 1,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      &fromFlavour, "", NULL
    },
    {
      Msg_ScrPrmtDirect, CSF_ONEOFBUTTON, CSS_EXIT,
      k_Alt_E,
      4,
      6, 5,
      1, 2,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      &fromFlavour, "", NULL
    },
    {
      Msg_ScrPrmtNormal, CSF_ONEOFBUTTON, CSS_EXIT,
      k_Alt_O,
      2,
      6, 6,
      1, 3,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      &fromFlavour, "", NULL
    },
    {
      Msg_ScrPrmtHold, CSF_ONEOFBUTTON, CSS_EXIT,
      k_Alt_L,
      3,
      6, 7,
      1, 4,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      &fromFlavour, "", NULL
    },
// to
    {
      Msg_ScrPrmtImmed, CSF_ONEOFBUTTON, CSS_EXIT,
      k_Alt_I,
      1,
      25, 3,
      1, 0,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      &toFlavour, "", NULL
    },
    {
      Msg_ScrPrmtCrash, CSF_ONEOFBUTTON, CSS_EXIT,
      k_Alt_C,
      1,
      25, 4,
      1, 1,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      &toFlavour, "", NULL
    },
    {
      Msg_ScrPrmtDirect, CSF_ONEOFBUTTON, CSS_EXIT,
      k_Alt_R,
      3,
      25, 5,
      1, 2,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      &toFlavour, "", NULL
    },
    {
      Msg_ScrPrmtNormal, CSF_ONEOFBUTTON, CSS_EXIT,
      k_Alt_N,
      1,
      25, 6,
      1, 3,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      &toFlavour, "", NULL
    },
    {
      Msg_ScrPrmtHold, CSF_ONEOFBUTTON, CSS_EXIT,
      k_Alt_H,
      1,
      25, 7,
      1, 4,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      &toFlavour, "", NULL
    },
    {
      Msg_ScrPrmtOK, CSF_BUTTON, CSS_ACTIVATE|CSS_EXIT,
      k_Alt_D,
      1,
      8, 9,
      1, 12,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      NULL, "", NULL
    },
    {
      Msg_ScrPrmtCancel, CSF_BUTTON, CSS_ACTIVATE|CSS_EXIT,
      k_Esc,
      1,
      27, 9,
      1, 12,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      NULL, "", NULL
    },
    {
      NULL, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, "", NULL
    }
  };

  char buf[65];

  node->toString( buf, sizeof(buf) );

  char *title = (char *)_alloca( strlen(buf)+strlen(Msg_ScrPrmtChange)+2 );

  strcpy( title, Msg_ScrPrmtChange );
  strcat( title, buf );
  strcat( title, " " );

  DIALOG dial =
  {
    (Z_ScreenWidth-43)/2+1, (Z_ScreenHeight-12)/2+1,
    12, 43, DBF_NOHINT,
    (int(popUpColor)<<8)|0x20,
    popUpColor.fore, popUpColor.back,
    1, 1, 1,
    BS_SINGLELINE, popUpColor.fore, popUpColor.back,
    DS_LOWERRIGHT, 0x07,
    title,
    NULL,
    Control
  };

  char mask[] = ".?lo",
       *full  = (char *)_alloca( strlen(bink)+sizeof(mask) );
  char immed(0),
       crash(0),
       direct(0),
       normal(0),
       hold(0);

  strcpy( full, bink );
  strcat( full, mask );

  for( Directory iter( full ); iter; iter++ )
  {
    char *name = iter.getName();

    if( strlen(name)==12 && ( toupper(name[10])=='L' && toupper(name[11])=='O' ) )
      switch( toupper(name[9]) )
      {
      case 'I':
        immed = 1;
        break;

      case 'C':
        crash = 1;
        break;

      case 'D':
        direct = 1;
        break;

      case 'F':
      case 'N':
	normal = 1;
        break;

      case 'H':
        hold = 1;
        break;
      }
  }

  if( immed )
  {
    if( fromFlavour==-1 )
      fromFlavour = 0;

    Control[2].Flags |= CSS_ACTIVATE;
    Control[7].Style  = CSF_TEXT;
    Control[7].Text   = "";
  }
  else
  {
    if( toFlavour==-1 )
      toFlavour = 0;

    Control[7].Flags |= CSS_ACTIVATE;
    Control[2].Style  = CSF_TEXT;
    Control[2].Text   = "";
  }

  if( crash )
  {
    if( fromFlavour==-1 )
      fromFlavour = 1;

    Control[3].Flags |= CSS_ACTIVATE;
    Control[8].Style  = CSF_TEXT;
    Control[8].Text   = "";
  }
  else
  {
    if( toFlavour==-1 )
      toFlavour = 1;

    Control[8].Flags |= CSS_ACTIVATE;
    Control[3].Style  = CSF_TEXT;
    Control[3].Text   = "";
  }

  if( direct )
  {
    if( fromFlavour==-1 )
      fromFlavour = 2;

    Control[4].Flags |= CSS_ACTIVATE;
    Control[9].Style  = CSF_TEXT;
    Control[9].Text   = "";
  }
  else
  {
    if( toFlavour==-1 )
      toFlavour = 2;

    Control[9].Flags |= CSS_ACTIVATE;
    Control[4].Style  = CSF_TEXT;
    Control[4].Text   = "";
  }

  if( normal )
  {
    if( fromFlavour==-1 )
      fromFlavour = 3;

    Control[5].Flags |= CSS_ACTIVATE;
    Control[10].Style  = CSF_TEXT;
    Control[10].Text   = "";
  }
  else
  {
    if( toFlavour==-1 )
      toFlavour = 3;

    Control[10].Flags |= CSS_ACTIVATE;
    Control[5].Style  = CSF_TEXT;
    Control[5].Text   = "";
  }

  if( hold )
  {
    if( fromFlavour==-1 )
      fromFlavour = 4;

    Control[6].Flags |= CSS_ACTIVATE;
    Control[11].Style  = CSF_TEXT;
    Control[11].Text   = "";
  }
  else
  {
    if( toFlavour==-1 )
      toFlavour = 4;

    Control[11].Flags |= CSS_ACTIVATE;
    Control[6].Style  = CSF_TEXT;
    Control[6].Text   = "";
  }

  int rc;

  if( fromFlavour==-1 )
  {
    char *bb = (char *)_alloca( strlen(Msg_ScrNoMail)+strlen(buf)+2 );

    strcpy( bb, Msg_ScrNoMail );
    strcat( bb, buf );
    Z_InformUser( Msg_ScrError, bb, Msg_ScrContinue, NULL );
    goto Done;
  }

  if( toFlavour==-1 )
  {
    char *bb = (char *)_alloca( strlen(Msg_ScrNoFreeMail)+strlen(buf)+2 );

    strcpy( bb, Msg_ScrNoFreeMail );
    strcat( bb, buf );
    Z_InformUser( Msg_ScrError, bb, Msg_ScrContinue, NULL );
    goto Done;
  }

  +screenCursor;
  rc = Z_HandleDialog( &dial );
  -screenCursor;

  if( rc==12 )
  {
    int len = strlen(bink)+sizeof(mask);
    char *from = (char *)_alloca( len ),
         *to   = (char *)_alloca( len );

    strcpy( from, bink );
    strcat( from, mask );
    strcpy( to, bink );
    strcat( to, mask );

    from[ len-4 ] = flavourChar( fromFlavour );
    to[ len-4 ]   = flavourChar( toFlavour );
    capitalize( from );
    capitalize( to );

    if( rename( from, to ) )
      log( Msg_ErrUnableRename, from, to, strError() );
  }

Done:
  delete bink;
}

void illegalAddress( char *str )
{
  if( detachedSession )
    return;

  char *msg = (char *)_alloca( strlen(str)+strlen(Msg_ScrIllAddress)+1 );

  strcpy( msg, Msg_ScrIllAddress );
  strcat( msg, str );

  Z_InformUser( Msg_ScrError, msg, Msg_ScrContinue, NULL );
}

void mailerPoll()
{
  if( detachedSession )
    return;

  initFileInfo( 0 );

  char *work = (char *)_alloca( stringLen+1 );

  memset( work, 0, stringLen+1 );
  strcpy( work, pollString );

  Color buttonTxt( colorSet.popUpItem ),
	buttonHig( colorSet.popUpBold );

  DLG_CONTROL Control[] =
  {
    { Msg_ScrPrmtAddr,
      CSF_DATAENTRY, CSS_ACTIVATE|CSS_CALLNOTHING,
      k_Alt_A,
      0,
      2, 2,
      1, (stringLen<<8)|33,
      buttonTxt.fore, buttonTxt.back,
      buttonTxt.fore, buttonTxt.back,
      work, "", NULL
    },
    {
      Msg_ScrPrmtOK, CSF_BUTTON, CSS_ACTIVATE|CSS_EXIT,
      k_Alt_D,
      1,
      6, 5,
      1, 12,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      NULL, "", NULL
    },
    {
      Msg_ScrPrmtCancel, CSF_BUTTON, CSS_ACTIVATE|CSS_EXIT,
      k_Esc,
      1,
      6, 7,
      1, 12,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      NULL, "", NULL
    },
    {
      Msg_ScrPrmtImmed, CSF_ONEOFBUTTON, CSS_ACTIVATE|CSS_EXIT,
      k_Alt_I,
      1,
      24, 4,
      1, 0,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      &pollFlavour, "", NULL
    },
    {
      Msg_ScrPrmtCrash, CSF_ONEOFBUTTON, CSS_ACTIVATE|CSS_EXIT,
      k_Alt_C,
      1,
      24, 5,
      1, 1,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      &pollFlavour, "", NULL
    },
    {
      Msg_ScrPrmtDirect, CSF_ONEOFBUTTON, CSS_ACTIVATE|CSS_EXIT,
      k_Alt_R,
      3,
      24, 6,
      1, 2,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      &pollFlavour, "", NULL
    },
    {
      Msg_ScrPrmtNormal, CSF_ONEOFBUTTON, CSS_ACTIVATE|CSS_EXIT,
      k_Alt_N,
      1,
      24, 7,
      1, 3,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      &pollFlavour, "", NULL
    },
    {
      Msg_ScrPrmtHold, CSF_ONEOFBUTTON, CSS_ACTIVATE|CSS_EXIT,
      k_Alt_H,
      1,
      24, 8,
      1, 4,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      &pollFlavour, "", NULL
    },
    {
      NULL, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, "", NULL
    }
  };

  DIALOG dial =
  {
    (Z_ScreenWidth-43)/2+1, (Z_ScreenHeight-11)/2+1,
    11, 43, DBF_NOHINT,
    (int(popUpColor)<<8)|0x20,
    popUpColor.fore, popUpColor.back,
    1, 1, 1,
    BS_SINGLELINE, popUpColor.fore, popUpColor.back,
    DS_LOWERRIGHT, 0x07,
    Msg_ScrPrmtPoll,
    NULL,
    Control
  };

  +screenCursor;
  int rc = Z_HandleDialog( &dial );
  -screenCursor;

  strncpy( pollString, work, stringLen );

  for( int i( strlen(pollString)-1 ); i>=0 && pollString[i]==' '; i-- ) continue;

  pollString[++i]=0;

  if( rc!=1 )
    return;

  initFileInfo( 1 );

  FIDOAddress adr( pollString );

  if( !adr.isValid() )
    illegalAddress( pollString );
  else
    if( pollFlavour<0 || pollFlavour>4 )
      log( "~BUG~ mailerPoll() unknown flavour index - %d", pollFlavour );
    else
      adr.makeLO( flavourTable[pollFlavour] );
}

void mailerDrop()
{
  if( detachedSession )
    return;

  initFileInfo( 0 );

  char *work = (char *)_alloca( stringLen+1 );

  memset( work, 0, stringLen+1 );
  strcpy( work, dropString );

  Color buttonTxt( colorSet.popUpItem ),
	buttonHig( colorSet.popUpBold );

  DLG_CONTROL Control[] =
  {
    {
      Msg_ScrPrmtAddr,
      CSF_DATAENTRY, CSS_ACTIVATE|CSS_CALLNOTHING,
      k_Alt_A,
      0,
      2, 2,
      1, (stringLen<<8)|33,
      buttonTxt.fore, buttonTxt.back,
      buttonTxt.fore, buttonTxt.back,
      work, "", NULL
    },
    {
      Msg_ScrPrmtOK, CSF_BUTTON, CSS_ACTIVATE|CSS_EXIT,
      k_Alt_D,
      1,
      6, 5,
      1, 12,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      NULL, "", NULL
    },
    {
      Msg_ScrPrmtCancel, CSF_BUTTON, CSS_ACTIVATE|CSS_EXIT,
      k_Esc,
      1,
      6, 8,
      1, 12,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      NULL, "", NULL
    },
    {
      Msg_ScrPrmtImmed, CSF_ONEOFBUTTON, CSS_ACTIVATE|CSS_EXIT,
      k_Alt_I,
      1,
      24, 4,
      1, 0,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      &dropFlavour, "", NULL
    },
    {
      Msg_ScrPrmtCrash, CSF_ONEOFBUTTON, CSS_ACTIVATE|CSS_EXIT,
      k_Alt_C,
      1,
      24, 5,
      1, 1,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      &dropFlavour, "", NULL
    },
    {
      Msg_ScrPrmtDirect, CSF_ONEOFBUTTON, CSS_ACTIVATE|CSS_EXIT,
      k_Alt_R,
      3,
      24, 6,
      1, 2,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      &dropFlavour, "", NULL
    },
    {
      Msg_ScrPrmtNormal, CSF_ONEOFBUTTON, CSS_ACTIVATE|CSS_EXIT,
      k_Alt_N,
      1,
      24, 7,
      1, 3,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      &dropFlavour, "", NULL
    },
    {
      Msg_ScrPrmtHold, CSF_ONEOFBUTTON, CSS_ACTIVATE|CSS_EXIT,
      k_Alt_H,
      1,
      24, 8,
      1, 4,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      &dropFlavour, "", NULL
    },
    {
      Msg_ScrPrmtAll, CSF_ONEOFBUTTON, CSS_ACTIVATE|CSS_EXIT,
      k_Alt_L,
      2,
      24, 9,
      1, 5,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      &dropFlavour, "", NULL
    },
    {
      NULL, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, "", NULL
    }
  };

  DIALOG dial =
  {
    (Z_ScreenWidth-43)/2+1, (Z_ScreenHeight-12)/2+1,
    12, 43, DBF_NOHINT,
    (int(popUpColor)<<8)|0x20,
    popUpColor.fore, popUpColor.back,
    1, 1, 1,
    BS_SINGLELINE, popUpColor.fore, popUpColor.back,
    DS_LOWERRIGHT, 0x07,
    Msg_ScrPrmtDrop,
    NULL,
    Control
  };

  +screenCursor;
  int rc = Z_HandleDialog( &dial );
  -screenCursor;

  strncpy( dropString, work, stringLen );

  for( int i( strlen(dropString)-1 ); i>=0 && dropString[i]==' '; i-- ) continue;

  dropString[++i]=0;

  if( rc!=1 )
    return;

  initFileInfo( 1 );

  FIDOAddress adr( dropString );

  if( !adr.isValid() )
    illegalAddress( dropString );
  else
    if( dropFlavour==5 )
      for( int i=0; i<5; i++ )
        adr.dropLO( flavourTable[i] );
    else
      if( dropFlavour<0 || dropFlavour>5 )
        log( "~BUG~ mailerDrop() unknown flavour index - %d", dropFlavour );
      else
        adr.dropLO( flavourTable[dropFlavour] );
}

int ioDialog( char *string, Flavour& flavour, char *file,
			    char *title, int check )
{
  if( detachedSession )
    return 1;

  char *work  = (char *)_alloca( stringLen+1 ),
       *work1 = (char *)_alloca( stringLen+1 );

  memset( work, 0, stringLen+1 );
  strcpy( work, string );

  memset( work1, 0, stringLen+1 );
  strcpy( work1, file );

  Color buttonTxt( colorSet.popUpItem ),
	buttonHig( colorSet.popUpBold );

  int mode;

  switch( flavour )
  {
  case immedFlavour:
    mode = 0;
    break;

  case crashFlavour:
    mode = 1;
    break;

  case directFlavour:
    mode = 2;
    break;

  case normalFlavour:
    mode = 3;
    break;

  default:
    mode = 4;
  }

  DLG_CONTROL Control[] =
  {
    { Msg_ScrPrmtAddr,
      CSF_DATAENTRY, CSS_ACTIVATE|CSS_CALLNOTHING,
      k_Alt_A,
      0,
      2, 2,
      1, (stringLen<<8)|33,
      buttonTxt.fore, buttonTxt.back,
      buttonTxt.fore, buttonTxt.back,
      work, "", NULL
    },
    { Msg_ScrPrmtFile,
      CSF_DATAENTRY, CSS_ACTIVATE|CSS_CALLNOTHING,
      k_Alt_F,
      0,
      2, 3,
      1, (stringLen<<8)|33,
      buttonTxt.fore, buttonTxt.back,
      buttonTxt.fore, buttonTxt.back,
      work1, "", NULL
    },
    {
      Msg_ScrPrmtOK, CSF_BUTTON, CSS_ACTIVATE|CSS_EXIT,
      k_Alt_D,
      1,
      6, 6,
      1, 12,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      NULL, "", NULL
    },
    {
      Msg_ScrPrmtCancel, CSF_BUTTON, CSS_ACTIVATE|CSS_EXIT,
      k_Esc,
      1,
      6, 8,
      1, 12,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      NULL, "", NULL
    },
    {
      Msg_ScrPrmtImmed, CSF_ONEOFBUTTON, CSS_ACTIVATE|CSS_EXIT,
      k_Alt_I,
      1,
      24, 5,
      1, 0,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      &mode, "", NULL
    },
    {
      Msg_ScrPrmtCrash, CSF_ONEOFBUTTON, CSS_ACTIVATE|CSS_EXIT,
      k_Alt_C,
      1,
      24, 6,
      1, 1,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      &mode, "", NULL
    },
    {
      Msg_ScrPrmtDirect, CSF_ONEOFBUTTON, CSS_ACTIVATE|CSS_EXIT,
      k_Alt_R,
      3,
      24, 7,
      1, 2,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      &mode, "", NULL
    },
    {
      Msg_ScrPrmtNormal, CSF_ONEOFBUTTON, CSS_ACTIVATE|CSS_EXIT,
      k_Alt_N,
      1,
      24, 8,
      1, 3,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      &mode, "", NULL
    },
    {
      Msg_ScrPrmtHold, CSF_ONEOFBUTTON, CSS_ACTIVATE|CSS_EXIT,
      k_Alt_H,
      1,
      24, 9,
      1, 4,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      &mode, "", NULL
    },
    {
      NULL, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, "", NULL
    }
  };

  DIALOG dial =
  {
    (Z_ScreenWidth-43)/2+1, (Z_ScreenHeight-12)/2+1,
    12, 43, DBF_NOHINT,
    (int(popUpColor)<<8)|0x20,
    popUpColor.fore, popUpColor.back,
    1, 1, 1,
    BS_SINGLELINE, popUpColor.fore, popUpColor.back,
    DS_LOWERRIGHT, 0x07,
    title,
    NULL,
    Control
  };

  +screenCursor;

  int rc;

  while( 1 )
  {
    rc = Z_HandleDialog( &dial );

    if( rc!=2 )
      break;

    strncpy( string, work, stringLen );
    for( int i( strlen(string)-1 ); i>=0 && string[i]==' '; i-- ) continue;
    string[++i]=0;

    strncpy( file, work1, stringLen );
    for( i = strlen(file)-1; i>=0 && file[i]==' '; i-- ) continue;
    file[++i]=0;

    FIDOAddress adr( string );

    if( !adr.isValid() )
    {
      illegalAddress( string );
      continue;
    }

    if( check )
    {
      Directory rc( file );

      if( !rc )
      {
        char *msg = (char *)_alloca( strlen(file)+strlen(Msg_ScrNotFound)+2 );

	strcpy( msg, Msg_ScrNotFound );
	strcat( msg, " " );
	strcat( msg, file );

	Z_InformUser( Msg_ScrError, msg, Msg_ScrContinue, NULL );

	continue;
      }
    }

    break;
  }

  -screenCursor;

  flavour = flavourTable[mode];

  return  rc!=2 ? 1 : 0;
}

void mailerGet()
{
  if( detachedSession )
    return;

  initFileInfo( 0 );

  if( ioDialog( getString, getFlavour, getFile, Msg_ScrPrmtFREQ, 0 ) )
    return;

  leaveOneBlank( getFile );
  initFileInfo( 1 );

  FIDOAddress adr( getString );

  for( char *str = tmpStr( getFile ), *tmp = NULL;
       str!=NULL && *str;
       str = tmp==NULL ? NULL : tmp+1 )
  {
    tmp = strchr( str, ' ' );

    if( tmp!=NULL && tmp[1]=='!' )
      tmp = strchr( tmp+1, ' ' );

    if( tmp!=NULL )
      *tmp = 0;

    adr.makeFreq( str, getFlavour );
  }
}

void mailerSend()
{
  if( detachedSession )
    return;

  initFileInfo( 0 );

  if( ioDialog( sendString, sendFlavour, sendFile, Msg_ScrPrmtSend, 1 ) )
    return;

  initFileInfo( 1 );

  FIDOAddress adr( sendString );

  char drive[ _MAX_DRIVE ],
       dir[ _MAX_DIR ],
       fname[ _MAX_FNAME ],
       ext[ _MAX_EXT ];

  _splitpath( sendFile, drive, dir, fname, ext );

  for( Directory rc( sendFile ); rc; rc++ )
  {
    char name[ _MAX_PATH ];

    strcpy( name, drive );
    strcat( name, dir );
    strcat( name, rc.getName() );
    adr.makeSend( name, sendFlavour );
  }
}
