#include <stdio.h>
#include <strstrea.h>

#include "misc.h"
#include "domain.h"

Domain::Domain( char *_name, short *_zlist,
                int _zlsize, short _defzone,
                char *_outb, NDLList *_ndl,
                char *_inb, char *_prot ) :
  name( newStr(_name) ),
  outbound( newStr( _outb==NULL ? _name : _outb ) ),
  inbound( newStr( _inb ) ),
  protect( newStr( _prot ) ),
  zlist( _zlist ),
  zlsize( _zlsize ),
  defzone( _defzone ),
  ndl( _ndl )
{}

Domain::~Domain() 
{
  delete name;
  delete outbound;
  delete ndl;
  delete zlist;
  delete protect;
  delete inbound;
}

int Domain::equal( const BaseObject &t ) const
{
  return !stricmp( name, (*(const Domain *)&t).name );
}

void Domain::print( ostream& s ) const
{
  s << name << ",Out:" << outbound
    << ",DefZone:" << defzone << ",ZLSize:" << zlsize
    << ",Inbound:";

  strItem( s, inbound );

  s << ",Protected:";

  strItem( s, protect );

  s << ",Zones:";

  if( !zlsize || zlist==NULL )
    s << "<None>";
  else
    for( int i=0; i<zlsize; i++ )
    {
      s << zlist[i];

      if( i!=zlsize-1 )
        s << char(',');
    }

  s << ",\nNodeLists:";

  if( ndl==NULL )
    s << "<None>";
  else
    s << endl << *ndl;
}

char *Domain::nameOfClass() const
{
  return "Domain";
}

char *Domain::getOutbound() const
{
  char *ret = new char [ strlen( outboundInfo.root ) + strlen( outbound ) + 1 ];

  strcpy( ret, outboundInfo.root );
  strcat( ret, outbound );

  return ret;
}

char *Domain::makePath( unsigned short zone ) const
{
  int len = strlen( outboundInfo.root ) + strlen( outbound ) + 1 + 4;

  zone &= 0xfff;

  char *ret = new char [len];

  strcpy( ret, outboundInfo.root );
  strcat( ret, outbound );

  if( !defzone || zone!=defzone )
  {
    char a[8];

    sprintf( a, ".%03x\\", zone );
    ret[ strlen(ret)-1 ] = 0;
    strcat( ret, a );
  }

  return ret;
}

DomainList::DomainList() :
  Collection( 5, 5, SemaphoreModeNoNeed )
{
}

Domain* DomainList::Find( Domain& item, LockMode lock ) const
{
  return (Domain*)Collection::Find( item, lock );
}
    
Domain& DomainList::operator [] ( const int pos ) const
{
  return *(Domain*)&(Collection::operator[]( pos ));
}

DomainIterator::DomainIterator( DomainList& ini ) :
  CollectionIterator( *(Collection*)&ini )
{
}
  
Domain& DomainIterator::operator () () const
{
  return *(Domain*)&(CollectionIterator::operator()());
}

NDLItem::NDLItem( char *_name, char *_diff, NDLItemType _type,
                  short dz, short dnt, short dnd ) :
  name( newStr( _name ) ),
  diff( newStr( _diff ) ),
  type( _type ),
  defZone( dz ),
  defNet( dnt ),
  defNode( dnd )
{}

NDLItem::~NDLItem()
{
  delete name;
  delete diff;
}

int NDLItem::equal( const BaseObject &t ) const
{
  return !stricmp( name, (*(const NDLItem *)&t).name );
}

void NDLItem::print( ostream& s ) const
{
  s << name
    << ",Diff:" << (diff==NULL ? "<None>" : diff)
    << ",DefZone:" << defZone
    << ",DefNet:" << defNet
    << ",DefNode:" << defNode
    << ",Type:";

  switch( type )
  {
  case NDLItemAdd:
    s << "Add";
    break;

  case NDLItemReplace:
    s << "Replace";
    break;

  default:
    s << "*Unknown*";
  }
}

char *NDLItem::nameOfClass() const
{
  return "NDLItem";
}

char *NDLItem::construct( char *nm )
{
  if( nm==NULL )
    return NULL;

  char *p = fileName( nm );

  if( p!=nm )
    return newStr( nm );

  p = new char [ strlen(nm)+strlen(outboundInfo.path)+1 ];
  strcpy( p, outboundInfo.path );
  strcat( p, nm );
  return p;
}

char *NDLItem::toString( char *buf, size_t bufSize ) const
{
  memset( buf, 0, bufSize );

  ostrstream o( buf, bufSize );
  
  switch( type )
  {
  case NDLItemAdd:
    o << "Add";
    break;
  
  case NDLItemReplace:
    o << "Replace";
    break;

  default:
    o << "Unknown";
    break;
  }
  
  o << " Default " << defZone << char( ':' ) << defNet << char( '/' ) << defNode 
    << name;
    
  if( diff!=NULL )
    o << " Diff " << diff;

  return buf;
}

NDLList::NDLList() :
  Collection( 5, 5, SemaphoreModeNoNeed )
{
}

NDLItem* NDLList::Find( NDLItem& item, LockMode lock ) const
{
  return (NDLItem*)Collection::Find( item, lock );
}
    
NDLItem& NDLList::operator [] ( const int pos ) const
{
  return *(NDLItem *)&(Collection::operator[]( pos ));
}

NDLIterator::NDLIterator( NDLList& ini ) : 
  CollectionIterator( *(Collection*)&ini )
{
}
  
NDLItem& NDLIterator::operator () () const
{
  return *(NDLItem*)&(CollectionIterator::operator()());
}
