#ifndef __DOMAIN_H

#define __DOMAIN_H

#include <string.h>

#include "newstr.h"
#include "class.h"

enum NDLItemType { NDLItemAdd, NDLItemReplace };

class NDLItem : public BaseObject
{
private:
  NDLItemType type;
  char *name,
       *diff;
  short defZone,
        defNet,
        defNode;

  int equal( const BaseObject &t ) const;
  void print( ostream& s ) const;
  char *nameOfClass() const;

protected:
  static char *construct( char *nm );

public:
  NDLItem( char *_name, char *_diff, NDLItemType _type,
           short dz=0, short dnt=0, short dnd=0 );
  ~NDLItem();

  char *constructName() const { return construct( name ); }
  char *constructDiffName() const { return construct( diff ); }
  
  char *toString( char *, size_t ) const;

  char        *getName() const { return name; }
  char        *getDiff() const { return diff; }
  NDLItemType  getType() const { return type; }
  short        getDefZone() const { return defZone; }
  short        getDefNet() const { return defNet; }
  short        getDefNode() const { return defNode; }
};

class NDLList : public Collection
{
public:
  NDLList();

  NDLItem* Find( NDLItem& item, LockMode lock=LockModeNeed ) const;
  NDLItem& operator [] ( const int pos ) const;
};

class NDLIterator : public CollectionIterator
{
public:
  NDLIterator( NDLList& ini );
  
  NDLItem& operator () () const;
};

class Domain : public BaseObject
{
private:
  char *name,
       *outbound,
       *inbound,
       *protect;
  NDLList *ndl;
  short *zlist;
  int zlsize;
  short defzone;

  int equal( const BaseObject &t ) const;
  void print( ostream& s ) const;
  char *nameOfClass() const;

public:
  Domain( char *_name, short *_zlist=NULL,
          int _zlsize=0, short _defzone=0,
          char *_outb=NULL, NDLList *_ndl=NULL,
          char *_inb=NULL, char *_prot=NULL );
  ~Domain();

  char *getOutbound() const;

  char *getOutboundRel() const { return outbound; }
  char *getInbound()     const { return inbound;  }
  char *getProtected()   const { return protect;  }

  char    *getName()     const { return name;     }
  NDLList *getNdl()      const { return ndl;      }
  short   *getZList()    const { return zlist;    }
  int      getZLSize()   const { return zlsize;   }
  short    getDefZone()  const { return defzone;  }

  char *makePath( unsigned short zone ) const;
};

class DomainList : public Collection
{
public:
  DomainList();

  Domain* Find( Domain& item, LockMode lock=LockModeNeed ) const;
  Domain& operator [] ( const int pos ) const;
};

class DomainIterator : public CollectionIterator
{
public:
  DomainIterator( DomainList& ini );
  
  Domain& operator () () const;
};

struct OutboundInfo
{
  char *path,
       *root,
       *index;
  DomainList *domains;
  Domain *defDomain;
};

extern OutboundInfo outboundInfo;

#endif
