#define INCL_DOSFILEMGR   /* File Manager values */
#define INCL_DOSERRORS    /* DOS error values    */
#include <os2.h>

#include <string.h>
#include <stdio.h>
#include <io.h>

#include "language.h"
#include "misc.h"

#pragma pack(4)

class binbuf
{
public:
  char *b;

  binbuf( const int len ) { b = new char [len]; }
  ~binbuf() { delete [] b; }
};

void setExtendedAttribute( const char *file_name,
                           const char *ea_name,
                           const char *ea_data_i,
                           const char msgFlag )
{
  if( miscInfo.state&NoEA )
    return; 

  int ea_data_len_i = strlen( ea_data_i );

  APIRET rc = NO_ERROR;      // Return code
  EAOP2  op;

  binbuf bbb( ea_data_len_i+4 );
  char *ea_data = bbb.b;

  *((short*)ea_data) = short( EAT_ASCII );
  ea_data += sizeof(short);
  *((short*)ea_data) = short( ea_data_len_i );
  ea_data += sizeof(short);

  int ea_data_len = ea_data_len_i + ea_data-bbb.b;

  memcpy( ea_data, ea_data_i, ea_data_len_i );
  ea_data = bbb.b;

  int ea_name_len = strlen( ea_name );

  if( ea_name_len > 255 )
  {
    log( "~BUG~ setExtendedAttribute: too long attribute name: %s", ea_name );
    return;
  }
    
  binbuf databuf( (64*2*1024)+1024 ); 	// twice 64K for EA data +
					//  1024 for any case
    
  op.fpGEA2List = (PGEA2LIST)0;
  op.fpFEA2List = (PFEA2LIST)databuf.b;

  char *databufp = databuf.b + sizeof(long);

  *((long*)databufp) = 0; 			// Next field offset is zero -
						//  just one field here
  databufp += sizeof(long);
  *databufp++ = 0; 				// not critical
  *databufp++ = (char)ea_name_len;
  *((short*)databufp) = short( ea_data_len );
  databufp += sizeof(short);
  memcpy( databufp, ea_name, ea_name_len+1 ); 	// with trailing zero
  databufp += ea_name_len+1;
  memcpy( databufp, ea_data, ea_data_len ); 	// with trailing zero
  databufp += ea_data_len;
    
  *((long*)(databuf.b)) = databufp-(databuf.b); // Size of all that stuff
    
  rc = DosSetPathInfo( file_name, FIL_QUERYEASIZE, &op, sizeof(op), 0);

  if( rc != NO_ERROR && msgFlag )
  {
//    log( Msg_ErrUnableSetEA, ea_name, file_name, rc );
//    os2Message( rc );
  }
}

void setExtendedAttributeIcon( const char *file_name,
                               const char *icon_name )
{
  if( miscInfo.state&NoEA )
    return; 

  APIRET rc = NO_ERROR;      // Return code
  EAOP2  op;
  const char *ea_name = ".ICON";

  FILE *icon = fopen( icon_name, "rb" );

  if( icon==NULL )
  {
    log( Msg_ErrUnableOpen, icon_name, strError() );
    return;
  }

  int ea_data_len_i( int( filelength( fileno( icon ) ) ) );
  char *ea_data_i = new char [ ea_data_len_i ];

  if( fread( ea_data_i, 1, ea_data_len_i, icon )!=ea_data_len_i )
  {
    log( Msg_ErrUnableRead, icon_name, strError() );
    delete [] ea_data_i;
    fclose( icon );
    return;
  }

  fclose( icon );

  binbuf bbb( ea_data_len_i+sizeof(short)+sizeof(short) );
  char *ea_data = bbb.b;

  *((short*)ea_data) = short( EAT_ICON );
  ea_data += sizeof(short);
  *((short*)ea_data) = short( ea_data_len_i );
  ea_data += sizeof(short);

  int ea_data_len = ea_data_len_i + ea_data-bbb.b;

  memcpy( ea_data, ea_data_i, ea_data_len_i );
  ea_data = bbb.b;

  int ea_name_len = strlen( ea_name );

  binbuf databuf( ea_data_len+ea_name_len+100 );
    
  op.fpGEA2List = (PGEA2LIST)0;
  op.fpFEA2List = (PFEA2LIST)databuf.b;

  char *databufp = databuf.b + sizeof(long);

  *((long*)databufp) = 0; 			// Next field offset is zero -
						//  just one field here
  databufp += sizeof(long);
  *databufp++ = 0; 				// not critical
  *databufp++ = (char)ea_name_len;
  *((short*)databufp) = short( ea_data_len );
  databufp += sizeof(short);
  memcpy( databufp, ea_name, ea_name_len+1 ); 	// with trailing zero
  databufp += ea_name_len+1;
  memcpy( databufp, ea_data, ea_data_len ); 	// with trailing zero
  databufp += ea_data_len;
    
  *((long*)(databuf.b)) = databufp-(databuf.b); // Size of all that stuff
    
  rc = DosSetPathInfo( file_name, FIL_QUERYEASIZE, &op, sizeof(op), 0);
  delete [] ea_data_i;

  if( rc != NO_ERROR )
  {
    log( Msg_ErrUnableSetICON, icon_name, file_name, rc );
    os2Message( rc );
  }
}

/*
void set_ea_ASCII( const char *fn, const char *ea_name, string data )
{
  if( miscInfo.state&NoEA )
    return; 

  binbuf b(64*1024);
  char *buf = b.b;

  *((short*)buf) = EAT_ASCII;
  buf += sizeof(short);
  *((short*)buf) = data.length();
  buf += sizeof(short);
  strcpy( buf, data.c_str() );
    
  set_ea( fn, ea_name, b.b, data.length() + 4 );
}


void set_ea_MVMT_ASCII( const char *fn, const char *ea_name,
                        const vector <string> &data )
{
  if( miscInfo.state&NoEA )
    return; 

  binbuf b(64*1024);
  char *buf = b.b;
    
  *((short*)buf) = EAT_MVMT;       buf += sizeof(short);
  *((short*)buf) = 0;              buf += sizeof(short); // Default CodePage == 0
  *((short*)buf) = data.size();    buf += sizeof(short);

  int len = data.size();

  for( int i = 0; i < len; i++ )
  {
    if( (64*1024-1) < ((buf-b.b) + data[i].length() + 4) )
      throw Ex_Arg("set_ea_MVMT_ASCII","vector too big to fit in EA","");
        
    *((short*)buf) = EAT_ASCII;          buf += sizeof(short);
    *((short*)buf) = data[i].length();   buf += sizeof(short);
    strcpy( buf, data[i].c_str() );      buf += data[i].length();
  }
    
  set_ea( fn, ea_name, b.b, buf-b.b );
}

void set_ea_MVST_ASCII( const char *fn, const char* ea_name, const vector <string> &data )
{
  if( miscInfo.state&NoEA )
    return; 

  binbuf b(64*1024);
    
  char *buf = b.b;
    
  *((short*)buf) = EAT_MVST;       buf += sizeof(short);
  *((short*)buf) = 0;              buf += sizeof(short); // Default CodePage == 0
  *((short*)buf) = data.size();    buf += sizeof(short);
  *((short*)buf) = EAT_ASCII;      buf += sizeof(short);
    
  int len = data.size();

  for( int i = 0; i < len; i++ )
  {
    if( (64*1024-1) < ((buf-b.b) + data[i].length() + 4) )
      throw Ex_Arg("set_ea_MVST_ASCII","vector too big to fit in EA","");
        
    *((short*)buf) = data[i].length();   buf += sizeof(short);
    strcpy( buf, data[i].c_str() );      buf += data[i].length();
  }
    
  set_ea( fn, ea_name, b.b, buf-b.b );
}

#include <stdio.h>

void main( int, char **argc )
{
  char *text = "My SUBJ";

  try
  {
    set_ea( argc[1], argc[2], argc[3], strlen(argc[3]) );
  }
  catch( Ex_Arg( a ) )
  {
    printf( "Ex_Arg\n" );
  }
  catch( Ex_Errno( a ) )
  {
    printf( "Ex_Errno\n" );
  }
}
*/
