#include <process.h>
#include <io.h>
#include <time.h>
#include <stdlib.h>
#include <builtin.h>

#define INCL_DOSEXCEPTIONS
#define INCL_DOSSEMAPHORES
#define INCL_DOSQUEUES
#define INCL_DOSSESMGR
#define INCL_DOSERRORS
#define INCL_DOSPROCESS
#include <os2.h>

#include "language.h"

#include "misc.h"
#include "session.h"
#include "execute.h"
#include ".\process.h"
#include "sched.h"
#include "except.h"
#include "key.h"

void changeTitleAndSwitch( char * );

static long startBBS = 0;
HQUEUE  SessionList::termQueue = 0;
char   *SessionList::queueName = NULL;

Session::Session( unsigned long _sema, unsigned long _ses ) :
  sema(_sema),
  session(_ses)
{
}

int Session::equal( const BaseObject &t ) const
{
  return session == (*(const Session *)&t).session;
}

void Session::print( ostream& s ) const
{
  s << sema << char(':') << session;
}

char *Session::nameOfClass() const
{
  return "Session";
}

static SessionList *sessionList = NULL;

class SesMgrUnableStart {} sesMgrUnableStart;

SessionList::SessionList() : Collection( 10, 5, SemaphoreModeNeed )
{
  if( termQueue!=0 )
  {
    log( Msg_ErrSesMgrAlreadyExist );
    return;
  }

  unsigned long rc=createQueue( &termQueue, "TermQ", &queueName );

  if( rc )
  {
    log( Msg_ErrSesMgrQueue, rc );
    os2Message( rc );
    termQueue = 0;
    queueName = NULL;
    throw sesMgrUnableStart;
  }

  int tid = _beginthread( sessionManager, 0, 40960, 0 );

  if( tid==-1 )
  {
    log( Msg_ErrSesMgrThrErr, strError() );
    delete queueName;
    deleteQueue( &termQueue );
    termQueue = 0;
    queueName = NULL;
    throw sesMgrUnableStart;
  }
  else
    log( Msg_ErrSesMgrThrRun, tid );
}

SessionList::~SessionList()
{
  delete queueName;
  deleteQueue( &termQueue );
}

Session* SessionList::Find( Session& item, LockMode lock ) const
{
  return (Session*)Collection::Find( item, lock );
}
    
Session& SessionList::operator [] ( const int pos ) const
{
  return *(Session *)&(Collection::operator[]( pos ));
}

void _Optlink sessionManager( void * )
{
  while( 1 )
  {
    USHORT *buf = NULL;
    REQUESTDATA req;
    ULONG len = 0;
    BYTE prior = 0;
    APIRET rc = 1;

    rc = DosReadQueue( SessionList::termQueue, &req, &len, (PPVOID)&buf,
	               0, DCWW_WAIT, &prior, 0 );

    if( !rc && buf!=NULL )
    {
      Session se( 0, buf[0] ),
              *found;

      for( int i=0; i<5; i++ )
      {
        found = sessionList->Find( se );

        if( found==NULL )
          sleep( 500 );
        else
          break;
      }

      if( found!=NULL )
      {
        APIRET rc = DosPostEventSem( found->getSema() );

        if( rc )
        {
          log( Msg_ErrSesMgrUnablePost, rc );
          os2Message( rc );
        }
      }
      else
        log( Msg_ErrSesMgrNotFound );

      DosFreeMem( buf );
    }
  }
}

struct StartSessionData
{
  STARTDATA *data;
  RunOption opt;
  HEV sema;
  int prio;
};

void clearStartData( StartSessionData *start )
{
  delete start->data->IconFile;
  delete start->data->PgmTitle;
  delete start->data->PgmInputs;
  delete start->data->ObjectBuffer;
  delete start->data->Environment;
  delete start->data;
  delete start;
}

void _Optlink startSessionThread( void *ptr )
{
  registerExceptionHandler

  StartSessionData *start = (StartSessionData *)ptr;

  if( (start->opt&RunOptNoClip)==RunOptNoClip )
    suspendClipServer();

  log( Msg_LogRun, start->data->PgmInputs+3 );

  if( !start->prio )
    DosSetPriority( PRTYS_THREAD, PRTYC_REGULAR, PRTYD_MINIMUM, 0 );
  else
    DosSetPriority( PRTYS_THREAD, (start->prio>>8)&0xff, start->prio&0xff, 0 );

  ULONG sesID;
  PID pID;
  APIRET rc = DosStartSession( start->data, &sesID, &pID );
  char started = 1;

  if( (start->opt&RunOptNoClip)==RunOptNoClip )
    resumeClipServer();

  if( rc )
    if( rc==ERROR_SMG_START_IN_BACKGROUND )
    {
      log( Msg_LogBackSession );
      rc = 0;
    }
    else
    {
      log( Msg_LogStartSessFailed, rc, start->data->ObjectBuffer );
      os2Message( rc );
      started = 0;
    }

  char wait( char( (start->opt&RunOptWait)==RunOptWait && 
                   (start->opt&RunOptChild)==RunOptChild ) );

  if( sessionList==NULL && wait && started )
  {
    log( Msg_ErrSesMgrNoWait );
    wait = 0;
  }

  if( wait && started )
    sessionList->AddAtTail( *new Session( start->sema, sesID ) );
  else
  {
    APIRET rc = DosPostEventSem( start->sema );

    if( rc )
    {
      log( Msg_ErrSesMgrUnablePost, rc );
      os2Message( rc );
    }
  }

  clearStartData( start );
  deRegisterExceptionHandler
}

volatile static char execCloseActive = 0;

void preExecute( int close )
{
  comPort->lock( Msg_InfoModemLockExec );

  comPort->clearin();
  comPort->suspend();
  
  if( close )
    comPort->closePort();
}

void postExecute( int close )
{
  int rc( close ? comPort->openPort() : 0 );

  if( !close )
    comPort->setup();

  if( !rc )
  {
    comPort->resume();
    comPort->init();
  }
  else
  {
    log( Msg_ErrNoCOMAccess, int(comPort->getPort()), rc );
    os2Message( rc );
    modemOs2Rc = 1;
  }

  comPort->unlock();
}

int executeCommand( int wait, char **cmd, RunOption o, char *title,
                    char *env, Schedule *sh  )
{
  if( *cmd==NULL )
  {
    log( "executeCommand: *cmd==NULL" );
    return 0;
  }

  char *environ = NULL;
  int rc = 0;

  if( env!=NULL )
  {
    int len = strlen( env )+2;

    environ = new char [ len ];
    memset( environ, 0, len );

    char *ptr;

    for( ptr = environ; *env; env++, ptr++ )
      if( *env==',' )
        *ptr = 0;
      else
        *ptr = *env;

    *ptr = 0;
  }

  int len( 0 );

  for( int i=0; *(cmd+i)!=NULL; i++ )
    len += strlen( *(cmd+i) )+1;

  char *l = (char *)_alloca( len+1 );

  *l = 0;

  for( i=0; *(cmd+i)!=NULL; i++ )
  {
    strcat( l, *(cmd+i) );
    strcat( l, " " );
  }

  l[ len-1 ] = 0;

  const failureObjectSize( 256 );

  if( wait==P_WAIT )
  {
    char *failureObject = (char *)_alloca( failureObjectSize );
    *failureObject = 0;

    log( Msg_LogExecute, l );

    if( o&RunOptClose || o&RunOptPort )
      preExecute( o&RunOptClose );

    l[ strlen( *cmd ) ] = 0;

    if( o&RunOptNoClip )
      suspendClipServer();

    Window *scr;

    if( o&RunOptClear )
    {
      Color cl( WHITE, BLACK );

      disableScreenSaver();
      scr =  new Window( 1, 1, Z_ScreenWidth, Z_ScreenHeight, cl );
      +(*scr);
      (*scr)( 1, 1 );
      scr->setPhysicalCursor( 1 );
      +screenCursor;
      VioScreenDisable();
    }

    RESULTCODES ret;

    long startExec( getTimeMark() );
    long beginExec( time( 0 ) );

    len = (int)DosExecPgm( failureObject, failureObjectSize,
                           EXEC_SYNC,
                           l,
                           (PBYTE)environ,
                           &ret,
                           l );

    long endExec = getTimeMark()-startExec;

    if( ret.codeTerminate!=TC_EXIT )
      ret.codeResult = 0xffffffffl;

    rc = int( ret.codeResult );

    changeTitleAndSwitch( 0 );
    changeIcon();

    if( !terminalActive )
      -screenCursor;
    else
      +screenCursor;

    if( o&RunOptClear )
    {
      enableScreenSaver();
      VioScreenEnable();
      delete scr;
    }

    if( len )
    {
      log( Msg_LogExecErr, len, failureObject );
      os2Message( len );
    }

    if( o&RunOptNoClip )
      resumeClipServer();

    if( !startBBS && getMyTid()==mainTid )
      writeToBinaryLog( 4, &endExec, sizeof( startExec ), beginExec );

    if( o&RunOptClose || o&RunOptPort )
      postExecute( o&RunOptClose );

    delete environ;
  }
  else
  {
    *l = 0;

    for( i=1; *(cmd+i)!=NULL; i++ )
    {
      strcat( l, *(cmd+i) );
      strcat( l, " " );
    }

    l[ len-1 ] = 0;

    STARTDATA *startData = new STARTDATA;

    memset( startData, 0, sizeof( *startData ) );

    startData->Length = (unsigned short)(noPMFeatures ? 30 
                                                      : sizeof( *startData ) ); 

    if( o&RunOptChild )
      startData->Related = SSF_RELATED_CHILD;
    else
      startData->Related = SSF_RELATED_INDEPENDENT;

    if( o&RunOptFore )
      startData->FgBg = SSF_FGBG_FORE;
    else
      startData->FgBg = SSF_FGBG_BACK;

    startData->TraceOpt = SSF_TRACEOPT_NONE;

    if( sh!=0 && sh->title!=0 )
      startData->PgmTitle = newStr( sh->title );
    else
      startData->PgmTitle = newStr( title==0 ? (PSZ)l+3 : (PSZ)title );

    if( sh!=0 && sh->icon!=0 )
      startData->IconFile = newStr( sh->icon );

    startData->PgmInputs = (PBYTE)newStr( l );
    startData->Environment = (PBYTE)environ;

    if( o&RunOptParent )
      startData->InheritOpt = SSF_INHERTOPT_PARENT;
    else
      startData->InheritOpt = SSF_INHERTOPT_SHELL;

    startData->SessionType = SSF_TYPE_DEFAULT;

    if( (o&RunOptFs)!=RunOptFs && (o&RunOptWin)!=RunOptWin )
      o = RunOption( o | ( isFullScreen ? RunOptFs : RunOptWin ) );

    if( o&RunOptDos )
    {
      if( o&RunOptFs )
        startData->SessionType = SSF_TYPE_VDM;

      if( o&RunOptWin )
        startData->SessionType = SSF_TYPE_WINDOWEDVDM;
    }
    else
    {
      if( o&RunOptFs )
        startData->SessionType = SSF_TYPE_FULLSCREEN;

      if( o&RunOptWin )
        startData->SessionType = SSF_TYPE_WINDOWABLEVIO;
    }

    if( o&RunOptInVisible )
      startData->PgmControl |= SSF_CONTROL_INVISIBLE;

    if( o&RunOptMax )
      startData->PgmControl |= SSF_CONTROL_MAXIMIZE;

    if( o&RunOptMin )
      startData->PgmControl |= SSF_CONTROL_MINIMIZE;

    if( o&RunOptNoClose )
      startData->PgmControl |= SSF_CONTROL_NOAUTOCLOSE;

    if( sh!=0 )
    {
      startData->InitXPos  = (USHORT)sh->xPos;
      startData->InitYPos  = (USHORT)sh->yPos;
      startData->InitXSize = (USHORT)sh->wPos;
      startData->InitYSize = (USHORT)sh->hPos;

      if( startData->InitXPos  || startData->InitYPos  ||
          startData->InitXSize || startData->InitYSize    )
        startData->PgmControl |= SSF_CONTROL_SETPOS;
    }

    startData->ObjectBuffer = new char [ failureObjectSize ];
    *startData->ObjectBuffer = 0;
    startData->ObjectBuffLen = failureObjectSize;

    if( (o&RunOptWait)==RunOptWait && 
        (o&RunOptChild)==RunOptChild )
    {
      if( sessionList==NULL )
      {
        try { sessionList = new SessionList; }
        catch( SesMgrUnableStart ) {}
      }

      startData->TermQ = sessionList->getQueueName();
    }

    StartSessionData *startSessionData = new StartSessionData;

    startSessionData->data = startData;
    startSessionData->opt = o;

    if( sh!=0 )
      startSessionData->prio = sh->prior;
    else
      startSessionData->prio = 0;

    HEV sesSema( 0 );

    len = int( createEventSemaphore( &sesSema, "TermQueSema" ) );

    startSessionData->sema = sesSema;

    if( len )
    {
      log( Msg_ErrSesMgrErrSesSemaCr, len );
      os2Message( len );
      clearStartData( startSessionData );
    }
    else if( _beginthread( startSessionThread, 0, 40960,
                           (void *)startSessionData )==-1 )
    {
      log( Msg_ErrRunThr, strError() );
      clearStartData( startSessionData );
      deleteEventSemaphore( sesSema );
    }
    else
    {
      if( getMyTid()==mainTid && (o&RunOptWait)==RunOptWait && 
                                 (o&RunOptChild)==RunOptChild )
        showAction( Msg_ScrRunWait );

      len = int( DosWaitEventSem( sesSema, (ULONG)SEM_INDEFINITE_WAIT ) );

      if( len )
      {
        log( Msg_ErrSesMgrErrSesSemaWt, len );
        os2Message( len );
      }

      deleteEventSemaphore( sesSema );
    }
  }

  return rc;
}

char **parseMacro( FIDOAddress& address, long elapsed,
                   char *name, char *actIni, int wait, char *title )
{
  int allocLen = strlen(actIni)+10;
  char **cmd( new char * [ allocLen ] ),
       adr[65],
       *actStart( tmpStr( actIni ) ),
       *act( actStart );

  memset( cmd, 0, sizeof(*cmd)*allocLen );

  int i( 2 );

  while( act!=NULL && *act )
  {
    char *ch( strchr( act, ' ' ) );

    if( ch!=NULL )
      *ch = 0;

    if( !stricmp( act, "%Port" ) )
    {
      sprintf( adr, "COM%d", comPort->getPort() );
      *(cmd+i) = newStr( adr );
      i++;
    }
    else if( !stricmp( act, "%PortNumber" ) )
    {
      sprintf( adr, "%d", comPort->getPort() );
      *(cmd+i) = newStr( adr );
      i++;
    }
    else if( !stricmp( act, "%Mode" ) )
    {
      switch( wait )
      {
      case P_WAIT:
        *(cmd+i) = newStr( "Exec" );
        break;

      case P_NOWAIT:
        *(cmd+i) = newStr( "Run" );
        break;

      default:
        *(cmd+i) = newStr( "Rexx" );
        break;
      }

      i++;
    }
    else if( !stricmp( act, "%Handle" ) )
    {
      sprintf( adr, "%d", comPort->getHandle() );
      *(cmd+i) = newStr( adr );
      i++;
    }
    else if( !stricmp( act, "%Baud" ) )
    {
      sprintf( adr, "%lu", comPort->getLastBaud() );
      *(cmd+i) = newStr( adr );
      i++;
    }
    else if( !stricmp( act, "%IsARQ" ) )
    {
      *(cmd+i) = newStr( comPort->isArq() ? "YES" : "NO" );
      i++;
    }
    else if( !stricmp( act, "%Address" ) )
    {
      address.toString( adr, 65, FIDOAddressShowModeEMSI );
      *(cmd+i) = newStr( adr );
      i++;
    }
    else if( !stricmp( act, "%Elapsed" ) )
    {
      sprintf( adr, "%lu", elapsed );
      *(cmd+i) = newStr( adr );
      i++;
    }
    else if( !stricmp( act, "%Name" ) )
      if( name!=NULL )
      {
        *(cmd+i) = newStr( name );
        i++;
      }
      else
        goto Not_Available;
    else if( !stricmp( act, "%Title" ) )
      if( title!=NULL )
      {
        *(cmd+i) = newStr( title );
        i++;
      }
      else
        goto Not_Available;
    else if( !stricmp( act, "%Box" ) )
      if( miscInfo.tbox==NULL )
      {
Not_Available:
        *(cmd+i) = newStr( Msg_InfoNotSpec );
        i++;
      }
      else
      {
        char *tm( address.getTBoxName() );

        if( tm==NULL )
          goto Not_Available;

        *(cmd+i) = new char [ strlen( tm )+
                              strlen( miscInfo.tbox )+1 ];
        strcpy( *(cmd+i), miscInfo.tbox );
        strcat( *(cmd+i), tm );
        i++;

        delete tm;
      }
    else if( !stricmp( act, "%List" ) )
    {
      char *bink( address.getBinkName() );

      if( bink==NULL )
       goto Not_Available;

      char mask[]( ".Qlo" );

      *(cmd+i) = new char [ strlen( bink )+
                            sizeof( mask ) ];
      strcpy( *(cmd+i), bink );
      strcat( *(cmd+i), mask );
      i++;
    }
    else if( !stricmp( act, "%Temp" ) )
    {
      char *bink( address.getBinkName() );

      if( bink==NULL )
        goto Not_Available;

      char *pathname( new char [ strlen(bink)+16 ] );
      int len = strlen(bink);

      strcpy( pathname, bink );

      for( time_t i=getTimeMark(); ; i++ )
      {
        sprintf( pathname+len, ".%03x", i&0xfff  );

        if( access(pathname, 0) )
          break;
      }

      *(cmd+i) = pathname;
      i++;

      delete bink;
    }
    else if( !stricmp( act, "%Task" ) )
    {
      sprintf( adr, "%lu", miscInfo.taskNumber );
      *(cmd+i) = newStr( adr );
      i++;
    }
    else
    {
      *(cmd+i) = newStr( act );
      i++;
    }

    act = ch == NULL ? (char *)NULL : ch+1;
  }

  *(cmd+i) = NULL;

  *cmd = newStr( getenv( "COMSPEC" ) );

  if( *cmd==NULL )
    *cmd = newStr( "Cmd.Exe" );

  *(cmd+1) = newStr( "/c" );

  return cmd;
}

void freeMacro( char **cmd )
{
  for( int i=0; *(cmd+i)!=NULL; i++ )
    delete cmd[i];

  delete cmd;
}

int makeCmd( FIDOAddress& address, long elapsed,
             char *name, char *actIni, int wait,
             RunOption o, char *title,
             char *dosSetting,
             Schedule *sh )
{

  char **cmd = parseMacro( address, elapsed, name, actIni, wait, title );
  int rc = executeCommand( wait, cmd, o, title, dosSetting, sh );

  freeMacro( cmd );
  return rc;
}

void executeBBS()
{
  char sep[3] = { '\r', '\n', '\n' };

  execSchedule( melodies.human );
  comPort->putOnLine( sep, sizeof(sep), 5000 );

  if( processBBS!=NULL && bbsAtThisCall )
  {
    log( Msg_BBSCallAvail );

    long beginBBS( time( 0 ) );

    startBBS = getTimeMark();

    comPort->putMessage( bbsInfo.loading );
    comPort->putOnLine( sep, sizeof(sep), 5000 );

    doEventKey( *processBBS, -1 );

    long endBBS( getTimeMark()-startBBS );

    writeToBinaryLog( 3, &endBBS, sizeof( startBBS ), beginBBS );
    startBBS = 0;
  }
  else
  {
    log( Msg_BBSCallNotAvail );
    comPort->putMessage( bbsInfo.goodBye );
    comPort->putOnLine( sep, sizeof(sep), 2000 );

    for( Timer t(2000); !comPort->outEmpty() && !t.isTimeOut(); )
      checkPoint();
  }

  comPort->hangUp();
}
