#define INCL_DOSPROCESS
#define INCL_DOSEXCEPTIONS
#include <os2.h>

#include <io.h>
#include <builtin.h>

#include "except.h"

#include "language.h"

#include "timer.h"
#include "misc.h"
#include "fax.h"

char faxActive = 0;

void _Optlink faxMonitor( void *pars )
{
  registerExceptionHandler

  FaxReceiver& fax = *(FaxReceiver *)pars;

  Window& wnd = *fax.wnd;
  int width = wnd.getWidth()-2-7;
  const idSize( sizeof(fax.response.remote_id) );
  char id[ idSize+1 ];
  unsigned long wasTotal(0xffffffff),
                wasPage(0xffffffff),
                wasElapsed(0xffffffff),
                wasSize(0xffffffff);
  char *wasFileName = newStr( "Empty" ),
       *wasModemState( 0 );

  for( ; fax.monitor; sleep( 500 ) )
  {
    char buf[80];

    if( fax.modemState!=wasModemState )
    {
      wasModemState = (char *)fax.modemState;
      wnd.drawTransact( 9, 2, stdColor, wasModemState );
    }

    if( strnicmp( id, (char *)fax.response.remote_id, idSize ) )
    {
      char tmp[ idSize+1 ];

      strncpy( id, (char *)fax.response.remote_id, idSize );
      strncpy( tmp, id, idSize );
      tmp[ idSize ] = 0;

      if( strlen(tmp)>width )
        tmp[ width ] = 0;

      wnd.drawTransact( 9, 3, stdColor, tmp );
    }

    if( stricmp( (char *)fax.fileName, wasFileName ) )
    {
      delete wasFileName;
      wasFileName = newStr( (char *)fax.fileName );

      char *tmp = newStr( wasFileName );

      if( strlen(tmp)>width )
        tmp[ width ] = 0;

      wnd.drawTransact( 9, 4, stdColor, tmp );

      delete tmp;
    }

    if( fax.faxsize!=wasSize )
    {
      wasSize = fax.faxsize;
      sprintf( buf, "%d", wasSize );
      wnd.drawTransact( 9, 5, stdColor, buf );
    }

    if( fax.totalsize!=wasTotal || fax.filesWritten!=wasPage )
    {
      wasTotal = fax.totalsize;
      wasPage = fax.filesWritten;
      sprintf( buf, "%d -> %lu", wasPage, wasTotal );
      wnd.drawTransact( 9, 6, stdColor, buf );
    }

    unsigned long elapsed = getTimeMark()-fax.startReceive;

    if( wasElapsed!=elapsed )
    {
      wasElapsed = elapsed;
      sprintf( buf, "%02d:%02d", wasElapsed/60, wasElapsed%60 );
      wnd.drawTransact( 9, 7, stdColor, buf );

      if( !elapsed )
        strcpy( buf, "0" );
      else
        sprintf( buf, "%d", fax.totalsize/elapsed );

      wnd.drawTransact( 9, 8, stdColor, buf );
    }
  }

  delete wasFileName;

  deRegisterExceptionHandler
}

FaxReceiver::FaxReceiver( Modem *_port ) :
  faxsize( 0 ),
  totalsize( 0 ),
  filesWritten( 0 ),
  port( _port ),
  fileName( newStr( "*None*" ) ),
  modemState( "Init" )
{

  int delta = (miscInfo.state1&StatusLine)==StatusLine ? 0 : 1;

  wnd = new Window( Z_ScreenWidth-25, 3-delta, 25, 9, stdColor );

  wnd->drawTransact( 2, 2, stdColor, "Modem:" );	//done
  wnd->drawTransact( 2, 3, stdColor, "   ID:" );	//done
  wnd->drawTransact( 2, 4, stdColor, " File:" );
  wnd->drawTransact( 2, 5, stdColor, "Bytes:" );	//done
  wnd->drawTransact( 2, 6, stdColor, "Total:" );	//done
  wnd->drawTransact( 2, 7, stdColor, " Time:" );	//done
  wnd->drawTransact( 2, 8, stdColor, "  CPS:" );	//done

  +(*wnd);

  monitor = 1;
  tid = _beginthread( faxMonitor, 0, 40960, (void *)this );

  if( tid==-1 )
    log( "Unable to start FAX monitor thread: %s", strError() );

  // swap the low order 4 bits with the high order

#ifdef INVERTED
  for( int i(0); i<256; i++ )
    swaptable[i] = (unsigned char)( ((i & 0x01) << 7) |
                                    ((i & 0x02) << 5) |
                                    ((i & 0x04) << 3) |
                                    ((i & 0x08) << 1) |
                                    ((i & 0x10) >> 1) |
                                    ((i & 0x20) >> 3) |
                                    ((i & 0x40) >> 5) |
                                    ((i & 0x80) >> 7) );
#endif

  response.remote_id[0] = '\0';
  response.fcon = FALSE;
  response.connect = FALSE;
  response.ok = FALSE;
  response.error = FALSE;
  response.hangup_code = -1;
  response.post_page_response_code = -1;
  response.post_page_message_code = -1;
  response.T30.ec = response.T30.bf = 0;
}

FaxReceiver::~FaxReceiver()
{
  monitor = 0;

  if( tid!=-1 )
    DosWaitThread( (PTID)&tid, DCWW_WAIT );

  delete (Window *)wnd;
  showAction( Msg_InfoActionHang );
  port->hangUp();
}

int FaxReceiver::receive( char *path, int taskNumber ) 
{
  logPtr->empty();

  if( path==NULL )
  {
    log( "Internal FAX receiver is disabled" );
    logPtr->empty();
    logPtr->flush();
    return 0;
  }

  startReceive = getTimeMark();

  if( faxConfig.baud )
    port->setbaud( faxConfig.baud );

  gEnd_of_document = FALSE;
  response.fcon = TRUE;			// we already connected

  int result = 0;

  for( int page( 0 ); gEnd_of_document==FALSE; page++ )
  {
    result = get_fax_file( path, page, taskNumber );
    log( "?FAX get_fax_file() = %d", result );

    switch( result )
    {
    case PageRCGood:
      continue;

    case PageRCHangUp:
      log( "Total received %d FAX pages", page );
      result = 1;
      gEnd_of_document = TRUE;
      break;

    default:
      log( "Error during FAX receiving" );
      result = page;
      gEnd_of_document = TRUE;
      break;
    }
  }

  if( faxConfig.baud )
    port->setbaud();

  unsigned long start = getTimeMark()-startReceive;

  if( !start )
    start = 1;

  log( "FAX elapsed time %02d:%02d, %lu bytes, %d pages, %lu bytes/sec",
       start/60, start%60, totalsize, filesWritten, totalsize/start );
  logPtr->empty();
  logPtr->flush();

  return result;
}

// This executes the +FDR receive page command, and looks for
// the proper CONNECT response, or, if the document is finished,
// looks for the FHNG/FHS code.
//
// returns:
//  PAGE_GOOD         no error conditions occured during reception
//  PAGE_HANGUP       normal end of transmission
//  PAGE_ERROR        something's wrong

PageRC FaxReceiver::get_fax_file( char *path, int page, int taskNumber )
{
  char *buf = (char *)_alloca( strlen(path)+32 );
  int opage( page );

  log( "?FAX get_fax_file" );

  do
    sprintf( buf, "%s%04x%04x.Fax", path, taskNumber&0xffff, opage++ );
  while( !access( buf, 0 ) && opage<0x10000 );

  if( opage == 256 )
  {
    log( "Couldn't construct output FAX file name" );
    return PageRCError;
  }

  PageRC result = faxmodem_receive_page( page );

  if( result == 0 )
  {
    makeFilePath( buf );
    FILE *fp = _fsopen( buf, "wb" );

    if( fp == NULL)
    {
      log( "Couldn't create output FAX file %s: %s", buf, strError() );
      return PageRCError;
    }

    delete (char *)fileName;
    fileName = newStr( ::fileName( buf ) );
    filesWritten++;

    if( !page )
      log( "Receiving FAX from %s", response.remote_id );

    log( "Receive page %d to file %s", page, ::fileName(buf) );
    log( "#Format is VR:%d BR:%d WD:%d LN:%d DF:%d EC:%d BF:%d ST:%d",
      response.T30.vr, response.T30.br, response.T30.wd,
      response.T30.ln, response.T30.df, response.T30.ec,
      response.T30.bf, response.T30.st );

    result = read_g3_stream( fp, page );

    fclose (fp);

    if( faxsize <= 256L )
    {
      log( "Too short FAX file (%db), remove it" );

      if( unlink( buf ) )
        log( Msg_ErrUnableDelete, ::fileName( buf ), strError() );
    }
    else
      log( "FAX received, %lu bytes", faxsize );
  }

  return result;
}

// Reads a data stream from the faxmodem, unstuffing DLE characters.
// Returns the +FET value (2 if no more pages) or 4 if hangup.

PageRC FaxReceiver::read_g3_stream( FILE * fp, int page )
{
  log( "?FAX read_g3_stream");

  response.post_page_response_code = -1;	// reset page codes
  response.post_page_message_code = -1;

  port->clearin();

  const SecBufSize = 1024;
  unsigned char *secbuf = (unsigned char *)_alloca( SecBufSize ),
                *p = secbuf;

  memset( secbuf, 1, SecBufSize );

  int pseudo_carrier = !port->isOnline();	// test if modem sets DCD

  if( pseudo_carrier )
    log( "?FAX read_g3_stream: modem doesn't assert DCD");

  log( "?FAX read_g3_stream: DC2");
  modemState = "-> DC2";

  port->putWait( (unsigned char)DC2, !pseudo_carrier );
						// Send DC2 to start phase C
						//  data stream

  Timer ltimer;
  register int c = 0;

  while( pseudo_carrier || port->isOnline() )	// data only when carrier high
  {
    c = port->get();				// get a character

    if( c==-1 )					// if nothing ready,
    {
      if( pseudo_carrier )			// process timeout if modem
						//  does not set DCD, this is
						// only a kludge, but it could
						// prevent an endless loop
        if( !ltimer.isStarted() )
          ltimer.startTimer( 15000 ); 		// 15 secs timeout
        else if( ltimer.isTimeOut() )
        {
          log( "?FAX read_g3_stream: ltimer timed out");
          goto fax_error;			// Houston, we lost the downlink
        }

      continue;					// process timeouts
    }
    else
      ltimer.endTimer();			// reset no char waiting timer

    c = c&0xff;

    if( c==DLE )				// DLE handling
    {
      Timer ltimer2( 4000 );

      log( "?FAX read_g3_stream: DLE handling");

      for( c=port->get(); c==-1; c=port->get() )
        if( ltimer2.isTimeOut() )
        {
          faxsize = 0L;
          log( "?FAX read_g3_stream: ltimer2 timed out");
          goto fax_error;			// give up
        }

      if( c == ETX )				// end of stream
      {
        log( "?FAX read_g3_stream: ETX");
        goto end_page;
      }

      // DLE DLE gives DLE. We don't know what to do if it
      // isn't ETX (above) or DLE. So we'll just always treat
      // DLE (not ETX) as (not ETX). 
      //
      // Fall out of here into storage. */
    }

    if( !faxsize )
      modemState = "<- Data";

#ifdef INVERTED
    *p++ = swaptable[ (unsigned char)c ];
#else
    *p++ = (unsigned char)c;
#endif

    faxsize++;
    totalsize++;

    if( !(faxsize % SecBufSize) )
    {
      if( fwrite( secbuf, 1, SecBufSize, fp) != SecBufSize )
      {
UnableWrite:
        log( Msg_ErrUnableWrite, "FAX file", strError() );
        goto fax_error;
      }

      p = secbuf;
    }
  }

end_page:

  if( faxsize % SecBufSize )
    if( fwrite( secbuf, 1, faxsize%SecBufSize, fp ) != faxsize%SecBufSize )
      goto UnableWrite;

  log("?FAX read_g3_stream: waiting +FET/+FHNG");

  c = 0;

  modemState = "<- EndOfPage";

  while( response.post_page_message_code == -1)		// wait for +FET
  {
    get_modem_result_code();
    c++;

    if( !response.post_page_response_code || c>5 || response.error )
      return PageRCError;

    if( response.hangup_code != -1 )
      return PageRCHangUp;
  }

  return PageRCGood;

fax_error:
  get_modem_result_code();
  return PageRCError;
}

/*--------------------------------------------------------------------------*/
/* Class 2 Faxmodem Protocol Functions                                      */
/*                                                                          */
/* Taken from EIA Standards Proposal No. 2388: Proposed New Standard        */
/* "Asynchronous Facsimile DCE Control Standard" (if approved,              */
/* to be published as EIA/TIA-592)                                          */
/*--------------------------------------------------------------------------*/

/* reads a line of characters, terminated by a newline */

void FaxReceiver::get_faxline( char *p, int nbytes, int wtime )
{
  int count = 1;			// character count (+null)

  Timer t( wtime );
  char *resp( p );

  while( count<nbytes && !t.isTimeOut() )
  {
    int c = port->get();

    if( c==-1 )
      continue;

    if( c=='\n' )
      continue;

    if( c=='\r' )
      if(count>1 )
        break;				// get out
      else
        continue;			// otherwise just keep going

    *p++ = (char)c;			// store the character
    ++count;				// increment the counter
  }

  *p = '\0';				// terminate the new string
  log("?FAX get_faxline: %s", resp );
}

// This function parses numeric responses from the faxmodem.
// It fills in any relevant slots of the faxmodem_response structure.

void FaxReceiver::get_modem_result_code()
{
  log( "?FAX get_modem_result_code" );

  for( Timer t( 4000 ); !t.isTimeOut(); )
  {
    char buf[256];

    buf[0] = '\0';
    get_faxline( buf, sizeof(buf)-1, 1000 );

    if( buf[0] )
    {
      parse_text_response( buf );
      return;
    }
  }
}

void FaxReceiver::parse_text_response( char *str )
{
  // Look for +FCON, +FDCS, +FDIS, +FHNG, +FHS, +FPTS, +FK, +FTSI

  log( "?FAX parse_text_response: %s", str );

  if( !strnicmp( "+FCO", str, 4 ) )
  {
    modemState = "+FCO";
    response.fcon = TRUE;
    return;
  }

  if( !strnicmp( str, "OK", 2 ) )
  {
    modemState = "OK";
    response.ok = TRUE;
    return;
  }

  if( !strnicmp( str, "CONNECT", 7 ) )
  {
    modemState = "CONNECT";
    response.connect = TRUE;
    return;
  }

  if( !strnicmp( str, "NO CARRIER", 10 ) )
  {
    modemState = "NO CARRIER";
    response.error = TRUE;
    response.hangup_code = 0;
    return;
  }

  if( !strnicmp( str, "ERROR", 5 ) )
  {
    modemState = "ERROR";
    response.error = TRUE;
    response.hangup_code = 0;
    return;
  }

  if( !strnicmp( str, "+FDCS", 5 ) )
  {
    modemState = "+FDCS";
    sscanf( str+6, "%d,%d,%d,%d,%d,%d,%d,%d",
            &response.T30.vr, &response.T30.br, &response.T30.wd,
            &response.T30.ln, &response.T30.df, &response.T30.ec,
            &response.T30.bf, &response.T30.st );
    return;
  }

  if( !strnicmp( str, "+FHNG", 5 ) )
  {
    modemState = "+FHNG";
    sscanf( str+6, "%d", &response.hangup_code );
    return;
  }

  if( !strnicmp( str, "+FPTS", 5 ) )
  {
    modemState = "+FPTS";
    sscanf( str+6, "%d", &response.post_page_response_code );
    return;
  }

  if( !strnicmp( str, "+FTSI", 5 ) )
  {
    modemState = "+FTSI";
    strncpy( (char *)response.remote_id, str+6, sizeof(response.remote_id)-1 );
    response.remote_id[ sizeof(response.remote_id)-1 ] = 0;
    return;
  }

  if( !strnicmp( str, "+FET", 4 ) )
  {
    modemState = "+FET";
    sscanf( str+5, "%d", &response.post_page_message_code );
    return;
  }

  if( !strnicmp( str, "+FHS", 4 ) )		// Class 2.0
  {
    modemState = "+FHS";
    sscanf( str+5, "%d", &response.hangup_code );
    return;
  }

  if( !strnicmp( str, "+FCS", 4 ) )		// Class 2.0
  {
    modemState = "+FCS";
    sscanf( str+5, "%d,%d,%d,%d,%d,%d,%d,%d",
            &response.T30.vr, &response.T30.br, &response.T30.wd,
            &response.T30.ln, &response.T30.df, &response.T30.ec,
            &response.T30.bf, &response.T30.st );
    return;
  }

  if( !strnicmp( str, "+FPS", 4 ) )		// Class 2.0
  {
    modemState = "+FPS";
    sscanf( str+5, "%d", &response.post_page_response_code );
    return;
  }

  if( !strnicmp( str, "+FTI", 4 ) )		// Class 2.0
  {
    modemState = "+FTI";
    strncpy( (char *)response.remote_id, str+5, sizeof(response.remote_id)-1 );
    response.remote_id[ sizeof(response.remote_id)-1 ] = 0;
    return;
  }
}

// Action Commands

// Receive a page
// after receiving OK,
// send +FDR
// This is somewhat ugly, because the FDR command can return
// a couple of possible results;
// If the connection is valid, it returns something like
//  +FCFR
//  +FDCS: <params>
//  CONNECT
//
// If, on the other hand, the other machine has hung up, it returns
// +FHNG: <code>  or
// +FHS: <code>
//
// and if the connection was never made at all, it returns ERROR (actually numeric
// code 4)
//
// faxmodem_receive_page returns values:
// PAGE_GOOD     page reception OK, data coming
// PAGE_HANGUP   normal hangup
// PAGE_ERROR    page error

PageRC FaxReceiver::faxmodem_receive_page( int page )
{
  faxsize = 0L;
  response.connect = response.ok = FALSE;

  // We wait until a string "OK" is seen
  // or a "+FHNG"
  // or a "ERROR" or "NO CARRIER"
  // or until 10 seconds for a response.

  Timer t( 10000 );

  log( "?FAX faxmodem_receive_page: Waiting OK");
  modemState = "<- OK";

  while( !t.isTimeOut() && !response.ok )
  {
    char buf[100];

    get_faxline( buf, sizeof(buf), 1000 );
    parse_text_response( buf );

    if( response.hangup_code!=-1 )
      return PageRCHangUp;

    if( response.error )
      return PageRCError;
  }

  if( !response.ok )
    return PageRCError;

  t.startTimer( 10000 );

  log( "?FAX faxmodem_receive_page: AT+FDR");
  modemState = "-> +FDR";
  port->putOnLine( "AT+FDR\r", 7, t.rest() );

  // We wait until either a string "CONNECT" is seen
  // or a "+FHNG"
  // or until 10 seconds for a response.

  log( "?FAX faxmodem_receive_page: Waiting CONNECT" );
  modemState = "<- CONNECT";

  while( !t.isTimeOut() )
  {
    char buf[100];

    get_faxline( buf, sizeof(buf), 1000 );
    parse_text_response( buf );

    if( response.connect == TRUE )
      return PageRCGood;

    if( response.hangup_code != -1 )
      return PageRCHangUp;

    if( response.error )
      return PageRCError;
  }

  return PageRCError;
}

void receiveFax()
{
  FaxReceiver fax( comPort );

  faxActive = 1;
  fax.receive( faxConfig.inbound, faxConfig.baud );
  faxActive = 0;
}