#include <string.h>
#include <direct.h>
#include <stdlib.h>
#include <strstrea.h>
#include <iostream.h>
#include <iomanip.h>
#include <time.h>
#include <ctype.h>
#include <io.h>
#include <sys\stat.h>
#include <share.h>
#include <fcntl.h>
#include <builtin.h>
#include <errno.h>

#include "language.h"

#include "fidoaddr.h"
#include "queue.h"
#include "misc.h"
#include "translat.h"
#include "32base.h"
#include "calls.h"
#include "rexxhook.h"
#include "ea.h"
#include "zmodem.h"

// FIDOAddress

int FIDOAddress::equal( const BaseObject &t ) const
{
  if( stricmp( addr.domain, (*(const FIDOAddress *)&t).addr.domain ) )
    return 0;

  if( addr.zone != (*(const FIDOAddress *)&t).addr.zone )
    return 0;

  if( addr.net != (*(const FIDOAddress *)&t).addr.net )
    return 0;

  if( addr.node != (*(const FIDOAddress *)&t).addr.node )
    return 0;

  if( addr.point != (*(const FIDOAddress *)&t).addr.point )
    return 0;

  return 1;
}

int FIDOAddress::less( const FIDOAddress &t ) const
{
  register int r = stricmp( addr.domain, t.addr.domain );

  if( r )
    return r<0 ? 1 : 0;

  if( addr.zone != t.addr.zone )
    return addr.zone < t.addr.zone;

  if( addr.net != t.addr.net )
    return addr.net < t.addr.net;

  if( addr.node != t.addr.node )
    return addr.node < t.addr.node;

  if( addr.point != t.addr.point )
    return addr.point < t.addr.point;

  return 0;
}

void FIDOAddress::print( ostream& s ) const
{
  char buf[128];

  s << (valid ? "Valid" : "NotValid") << char(':')
    << toString( buf, sizeof(buf), FIDOAddressShowModeEMSI );
}

char *FIDOAddress::nameOfClass() const
{
  return "FIDOAddress";
}

void item( ostream& o, char *s1, char *s2 )
{
  o << endl << "    ";
  strItem( o, s1 );
  o << char(':');
  strItem( o, s2 ? s2 : "<None>" );
}

void itemSched( ostream& o, char *str, Schedule *sch )
{
  o << endl << "    ";
  strItem( o, str );
  o << char(':');

  if( !sch )
    strItem( o, "<None>" );
  else
    o << *sch;
}

void FIDOAddress::makeDomain()
{
  const HOW=6;
  static short *zones = new short [HOW];
  static Domain std( "FidoNet", zones, HOW, 2 );
  static char first = 1;

  if( dom==NULL )
  {
    if( addr.domain==NULL )
      addr.domain = newStr( "" );

    char *search = tmpStr( addr.domain );

Again:
    for( DomainIterator list( *outboundInfo.domains ); list; list++ )
    {
      Domain& cur = list();

      if( !stricmp( search, cur.getName() ) )
      {
        dom = &cur;
        goto Done;
      }
    }

    for( register char *q=search+strlen(search)-1; q>=search && *q!='.'; q-- )
      continue;

    if( q>=search && *q=='.' )
    {
      *q = 0;
      goto Again;
    }

    for( list.restart(); list; list++ )
    {
      Domain& cur = list();
      short *list = cur.getZList();
      int size = cur.getZLSize();

      for( int i=0; i<size; i++, list++ )
        if( addr.zone==*list )
        {
          dom = &cur;
          goto Done;
        }
    }

    dom = outboundInfo.defDomain;
    unknowDomain = 1;
  }
Done:

  if( dom==NULL )
  {
    if( first )
    {
      first = 0;

      for( int i=0; i<HOW; i++ )
        zones[i] = short( i+1 );
    }

    dom = &std;
    unknowDomain = 1;
  }

  delete addr.domain;
  addr.domain = newStr( dom->getName() );
}

FIDOAddress::FIDOAddress( uShort iZone, uShort iNet, uShort iNode,
			  uShort iPoint, char *iDomain, Domain *idom ) :
  valid( 1 ),
  dom( idom ),
  unknowDomain( 0 )
{
  addr.zone   = iZone;
  addr.net    = iNet;
  addr.node   = iNode;
  addr.point  = iPoint;
  addr.domain = newStr( iDomain );
  makeDomain();
}

FIDOAddress::FIDOAddress( const Address& a ) :
  valid( 1 ),
  dom( NULL ),
  unknowDomain( 0 )
{
  addr.zone   = a.zone;
  addr.net    = a.net;
  addr.node   = a.node;
  addr.point  = a.point;
  addr.domain = newStr( a.domain );
  makeDomain();
}

FIDOAddress::FIDOAddress( char *s, Domain *idom ) :
  dom( idom ),
  unknowDomain( 0 )
{
  if( s==NULL || !checkAddress( s, addr.zone, addr.net, addr.node, addr.point, &addr.domain ) )
  {
    addr.zone   = 0;
    addr.net    = 0;
    addr.node   = 0;
    addr.point  = 0;
    addr.domain = NULL;
    valid  = 0;
  }
  else
  {
    addr.domain = newStr( addr.domain );
    valid = 1;
  }

  makeDomain();
}

FIDOAddress::FIDOAddress( const FIDOAddress *a )
{
  if( a!=NULL )
  {
    addr.zone   = a->addr.zone;
    addr.net    = a->addr.net;
    addr.node   = a->addr.node;
    addr.point  = a->addr.point;
    addr.domain = newStr( a->addr.domain );
    valid  = a->valid;
    dom    = a->dom;
    unknowDomain = a->unknowDomain;
  }
  else
  {
    addr.zone   = 0;
    addr.net    = 0;
    addr.node   = 0;
    addr.point  = 0;
    addr.domain = newStr( "" );
    valid  = 0;
    dom    = outboundInfo.defDomain;
    unknowDomain = 1;
  }
}

FIDOAddress::~FIDOAddress()
{
  delete addr.domain;
}

char *FIDOAddress::getBsySemaName() const
{
  char *name = getLongName(),
       *full = new char [ strlen(miscInfo.bsySemaPath)+strlen(name)+1 ];

  strcpy( full, miscInfo.bsySemaPath );
  strcat( full, name );
  capitalize( full );
  delete name;

  return full;
}

int FIDOAddress::makeBsySemaphore( int create )
{
  char *full = getBsySemaName();

  int rc( 0 );

  if( create )
    if( busySemaphoreExist( full ) )
      rc = 1;
    else
      busySemaphoreCreate( full );
  else
    if( busySemaphoreExist( full ) )
      busySemaphoreRemove( full );

  delete full;
  return rc;
}

int FIDOAddress::createBusy( char *bsyFull )
{
  makeFilePath( bsyFull );

  int h( _sopen( bsyFull, O_CREAT|O_EXCL|O_WRONLY|O_TEXT,
                 SH_DENYRW,
                 S_IREAD|S_IWRITE ) );

  if( h==-1 )
  {
    if( errno!=EEXIST )
      log( Msg_ErrUnableCreate, bsyFull, strError() );

    return 1;
  }

  char buf[32];

  sprintf( buf, "%d", miscInfo.taskNumber );

  int len( strlen(buf) );

  if( write( h, buf, len )!=len )
    log( Msg_ErrUnableWrite, bsyFull, strError() );

  close( h );
  setExtendedAttribute( bsyFull, ".TYPE", Msg_EABSY );

  return 0;
}

void FIDOAddress::deleteBusy( char *bsyFull )
{
  if( !access( bsyFull, 0 ) )
    forceDelete( bsyFull );
}

char *FIDOAddress::getLongBsyName() const
{
  char *bink( getLongBinkName() );

  if( bink==NULL )
    return NULL;

   char busy[]( ".Busy" ),
        *path( new char[ strlen(bink)+sizeof(busy) ] );

  strcpy( path, bink );
  strcat( path, busy );
  delete bink;
  capitalize( path );

  return path;
}

char *FIDOAddress::getBsyName() const
{
  char *bsyPath( getBinkName() );

  if( bsyPath==NULL )
    return NULL;

  char busy[]( ".Bsy" ),
       *bsyFull( new char [ strlen(bsyPath)+sizeof(busy) ] );

  strcpy( bsyFull, bsyPath );
  strcat( bsyFull, busy );
  delete bsyPath;
  capitalize( bsyFull );

  return bsyFull;
}

int FIDOAddress::makeLongBsy( int create )
{
  char *bink( getLongBsyName() );
  int rc(0);

  if( bink!=NULL )
  {
    if( create )
      rc = createBusy( bink );
    else
      deleteBusy( bink );

    delete bink;
  }

  return rc;
}

int FIDOAddress::makeBinkBsy( int create )
{
  char *bsyFull( getBsyName() );

  if( bsyFull==NULL )
    return 0;

  int rc(0);

  if( create )
    rc = createBusy( bsyFull );
  else 					// NO CREATE! delete requested
  {
    deleteBusy( bsyFull );

    char drive[_MAX_DRIVE],
         dir[_MAX_DIR],
         file[_MAX_FNAME],
         ext[_MAX_EXT];

    _splitpath( bsyFull, drive, dir, file, ext );

    char *path = (char *)_alloca( strlen(drive)+strlen(dir)+1 );

    strcpy( path, drive );
    strcat( path, dir );

    if( addr.point )
    {
      path[ strlen(path)-1 ] = 0;

//      if( !access( path, 0 ) )
        rmdir( path );

      _splitpath( path, drive, dir, file, ext );
      strcpy( path, drive );
      strcat( path, dir );
    }

    if( addr.zone!=dom->getDefZone() )
    {
      path[ strlen(path)-1 ] = 0;

//      if( !access( path, 0 ) )
        rmdir( path );
    }
  }

  delete bsyFull;
  return rc;
}

int FIDOAddress::makeBsy( int create )
{
  if( (miscInfo.state2&SemaphoreBsy)==SemaphoreBsy )
    return makeBsySemaphore( create );

// we no need a semaphore :(

  if( !create )		//delete requested
  {
    makeBinkBsy( 0 );
    makeLongBsy( 0 );
    return 0;
  }

// CREATE busy flags

  int binkRc( makeBinkBsy(1) ),
      longRc( makeLongBsy(1) );

  if( binkRc && longRc )
    return 1;

  if( !binkRc && !longRc )
    return 0;

  if( longRc )
    makeBinkBsy( 0 );

  if( binkRc )
    makeLongBsy( 0 );

  return 1;
}

int FIDOAddress::checkBsy() const
{
  if( (miscInfo.state2&SemaphoreBsy)==SemaphoreBsy )
  {
    char *full = getBsySemaName();
    int rc = busySemaphoreExist( full );

    delete full;
    return rc;
  }

// we no need a semaphore :(

  int binkBsyExist( 0 ),
      longBsyExist( 0 );

// check bink busy

  char *bsyPath( getBsyName() );

  if( bsyPath!=NULL )
  {
    binkBsyExist = access( bsyPath, 0 )==0;
    delete bsyPath;
  }

// check long busy

  bsyPath = getLongBsyName();

  if( bsyPath!=NULL )
  {
    longBsyExist = access( bsyPath, 0 )==0;
    delete bsyPath;
  }

  return binkBsyExist || longBsyExist;
}

char *FIDOAddress::toString( char *buf, int len, FIDOAddressShowMode mode ) const
{
  memset( buf, 0, len );

  ostrstream o( buf, len-1 );

  if( mode==FIDOAddressShowModeCfg && systemList!=NULL )
  {
    NodeConfig test( (FIDOAddress *)this ),
               *found( systemList->Find( test ) );

    if( found!=NULL && found->getAlias()!=NULL  )
    {
      o << found->getAlias();
      return buf;
    }
  }

  if( mode==FIDOAddressShowModeCfg && 
      miscInfo.addrForm[0]!='@' &&
      miscInfo.addrForm[0]!='I' )
    o << addr.domain << char( miscInfo.addrForm[0] );

  o << addr.zone << char(':') << addr.net << char('/') << addr.node;

  if( is4D() || miscInfo.addrForm[1]=='P' )
    o << char('.') << addr.point;

  if( miscInfo.addrForm[0]=='@' || mode==FIDOAddressShowModeEMSI )
    o << char('@') << addr.domain;

  return buf;
}

char *FIDOAddress::getLongName() const
{
  char *s = new char [ strlen(addr.domain)+64 ];

  sprintf( s, "%s.%u.%u.%u.%u", addr.domain, int(addr.zone), int(addr.net), int(addr.node), int(addr.point) );
  return s;
}

char *FIDOAddress::getLongBox() const
{
  if( miscInfo.longBox==NULL )
    return NULL;

  char *name=getLongName(),
       *s = new char [ strlen(name)+strlen(miscInfo.longBox)+1 ];

  strcpy( s, miscInfo.longBox );
  strcat( s, name );
  delete name;
  return s;
}

char *FIDOAddress::getLongBinkName() const
{
  if( miscInfo.longOutbound==NULL )
    return NULL;

  char *name= getLongName(),
       *ret = new char [ strlen(miscInfo.longOutbound)+strlen(name)+1 ];

  strcpy( ret, miscInfo.longOutbound );
  strcat( ret, name );
  delete name;
  return ret;
}

char *FIDOAddress::getBinkName() const
{
  if( addr.zone>4095 )
    return NULL;

  char a[32];

  sprintf( a, "%04x%04x", addr.net, addr.node );

  if( addr.point )
    sprintf( a+8, ".Pnt\\%08x", addr.point );

  char *out = dom->makePath( addr.zone ),
       *ret = new char [ strlen(out)+strlen(a)+1 ];

  strcpy( ret, out );
  strcat( ret, a );
  delete out;
  return ret;
}

char *FIDOAddress::getTBoxName() const
{
  char *ret = new char [16];

  if( addr.zone>1023 || addr.point>1023 || addr.net>32767 || addr.node>32767 )
    return NULL;

  to32base( addr.zone, 2, ret );
  to32base( addr.net, 3, ret+2 );
  to32base( addr.node, 3, ret+5 );
  ret[ 8 ] = '.';
  to32base( addr.point, 2, ret+9 );
  ret[ 11 ] = 0;

  return ret;
}

void FIDOAddress::updateLoName( Flavour flavourLO, char *xlo )
{
  static char fLO[]( ".!lo" );

  strcpy( xlo, fLO );

  switch( flavourLO )
  {
  case immedFlavour:
    xlo[1] = 'I';
    break;

  case crashFlavour:
    xlo[1] = 'C';
    break;

  case directFlavour:
    xlo[1] = 'D';
    break;

  case normalFlavour:
    xlo[1] = 'F';
    break;

  default:
    xlo[1] = 'H';
    break;
  }
}

char *FIDOAddress::buildOutboundName( char *sh, char *lo ) const
{
  char *loPath( getLongBinkName() ),
       *suf( lo );

  if( loPath==NULL )
  {
    loPath = getBinkName();
    suf = sh;
  }

  if( loPath )
  {
    char *ret( new char [ strlen( loPath )+strlen( suf )+1 ] );

    strcpy( ret, loPath );
    strcat( ret, suf );
    capitalize( ret );

    delete loPath;
    return ret;
  }

  return NULL;
}

char *FIDOAddress::buildListName( Flavour flavourLO ) const
{
  static char list[]( ".List" );
  char *flav( flavourString( flavourLO ) );
  char *lo( (char *)_alloca( strlen(flav)+sizeof(list)+1 ) );
  char sh[5];

  strcpy( lo, "." );
  strcat( lo, flav );
  strcat( lo, list );

  updateLoName( flavourLO, sh );

  return buildOutboundName( sh, lo );
}

void FIDOAddress::makeLO( Flavour flavourLO, int needCheckBsy ) const
{
  if( flavourLO == unknownFlavour )
    return;

  char buf[65];

  toString( buf, sizeof buf );

  if( needCheckBsy && checkBsy() )
  {
    log( Msg_LogPollMakeBsy, buf, flavourName( flavourLO ) );
    return;
  }

  char *loFull = buildListName( flavourLO );

  if( !loFull )
    return;

  if( access( loFull, 0 ) )
  {
    log( Msg_LogPollMake, buf, flavourName( flavourLO ) );

    makeFilePath( loFull );

    FILE *hd( _fsopen( loFull, "a" ) );

    if( !hd )
      log( Msg_ErrUnableCreate, loFull, strError() );
    else
    {
      fclose( hd );
      setExtendedAttribute( loFull, ".TYPE", Msg_EALO );
    }
  }

  delete loFull;
}

void FIDOAddress::makeHold( unsigned long delta ) const
{
  char *loFull( buildOutboundName( ".Hld", ".Freeze" ) );

  if( loFull==NULL )
    return;

  makeFilePath( loFull );

  FILE *hld = _fsopen( loFull, "r+" );
  unsigned long to = 0;

  if( hld==NULL )
    hld = _fsopen( loFull, "w" );
  else
  {
    fscanf( hld, "%lu", &to );

    if( ferror( hld ) )
    {
      log( Msg_ErrUnableRead, loFull, strError() );
      fclose( hld );
      hld = 0;
    }

    if( hld )
      fseek( hld, 0, SEEK_SET );
  }

  if( hld==NULL )
  {
    log( Msg_ErrUnableCreate, loFull, strError() );
    delete loFull;
    return;
  }

  unsigned long how = delta==0xffffffffl ? 0xffffffffl :
                                           (unsigned long)time( NULL )+delta;

  if( to!=0xffffffffl && (to<how || how==0xffffffffl) )
  {
    fprintf( hld, "%lu", how );

    if( ferror(hld) )
    {
      fclose( hld );

      log( Msg_ErrUnableWrite, loFull, strError() );

      if( unlink( loFull ) )
        log( Msg_ErrUnableDelete, loFull, strError() );

      delete loFull;
      return;
    }

    char buf[65];

    log( Msg_LogHoldMake, toString( buf, sizeof buf ), delta/60, delta%60 );
  }

  fclose( hld );
  setExtendedAttribute( loFull, ".TYPE", Msg_EAHold );
  delete loFull;
}

void FIDOAddress::dropLO( Flavour flavourLO ) const
{
  char *loFull = buildListName( flavourLO );

  if( !loFull )
    return;

  struct stat st;

  if( !stat( loFull, &st ) && !st.st_size )
  {
    char buf[65];

    log( Msg_LogPollDrop, toString( buf, sizeof buf ),
	 flavourName( flavourLO ) );

    if( unlink( loFull) )
      log( Msg_ErrUnableDelete, loFull, strError() );
  }

  delete loFull;
}

void FIDOAddress::makeSend( char *param, Flavour fla ) const
{
  char *full = buildListName( fla );

  if( full==NULL )
    return;

  char buf[65];

  toString( buf, sizeof buf );

  if( checkBsy() )
  {
    log( Msg_LogMakeSendBsy, buf, param, flavourName( fla ) );
    return;
  }

  log( Msg_LogMakeSend, param, buf, flavourName( fla ) );

  makeFilePath( full );

  FILE *out = _fsopen( full, "a" );

  if( out==NULL )
    log( Msg_ErrUnableOpen, full, strError() );
  else
  {
    fprintf( out, "%s\n", param );
    fclose( out );
    setExtendedAttribute( full, ".TYPE", Msg_EALO );
  }

  delete full;
}

char *FIDOAddress::concatenateWithDot( const char *s1, const char *s2 )
{
  if( !s1 )
    return 0;

  char *ret( new char [ strlen(s1)+strlen(s2)+2 ] );

  strcpy( ret, s1 );
  strcat( ret, "." );
  strcat( ret, s2 );
  return ret;
}

void FIDOAddress::removeFreqFile( char *bink ) const
{
  if( bink!=NULL )
  {
    capitalize( bink );

    if( !access( bink, 0 ) )
    {
      log( Msg_SesFREQRemove, bink );

      if( unlink( bink ) )
        log( Msg_ErrUnableDelete, bink, strError() );
    }

    delete bink;
  }
}

void FIDOAddress::dropFreq() const
{
  removeFreqFile( concatenateWithDot( getLongBinkName(), 
                                      flavourString( requestFlavour ) ) );
  removeFreqFile( concatenateWithDot( getBinkName(), "Req" ) );
}

void FIDOAddress::makeFreq( char *param, Flavour fla ) const
{
  char buf[65];

  toString( buf, sizeof buf );

  if( checkBsy() )
  {
    log( Msg_LogMakeFREQBsy, buf, param, flavourName( fla ) );
    return;
  }

  char *loReq1( flavourString( requestFlavour ) ),
       *loReq( (char *)_alloca( strlen( loReq1 )+2 ) );

  *loReq = '.';
  strcpy( loReq+1, loReq1 );

  char *full = buildOutboundName( ".Req", loReq );

  if( full!=NULL )
  {
    capitalize( full );
    makeFilePath( full );

    log( Msg_LogMakeFREQ, param, buf, flavourName( fla ) );

    FILE *out = _fsopen( full, "a" );

    if( out==NULL )
      log( Msg_ErrUnableOpen, full, strError() );
    else
    {
      fprintf( out, "%s\n", param );
      fclose( out );
      setExtendedAttribute( full, ".TYPE", Msg_EAREQ );
    }

    delete full;
    makeLO( fla );
  }
}

char *FIDOAddress::getTryFileName() const
{
  char *ss( buildOutboundName( ".Sta", ".Status" ) );

  if( ss )
    capitalize( ss );

  return ss;
}

FIDOAddressList::FIDOAddressList() :
  Collection( 10, 5, SemaphoreModeNoNeed )
{
}
  
FIDOAddress* FIDOAddressList::Find( FIDOAddress& item, LockMode lock ) const
{
  return (FIDOAddress*)Collection::Find( item, lock );
}
    
FIDOAddress& FIDOAddressList::operator [] ( const int pos ) const
{
  return *(FIDOAddress *)&(Collection::operator[]( pos ));
}

FIDOAddressIterator::FIDOAddressIterator( FIDOAddressList& ini ) :
  CollectionIterator( *(Collection *)&ini )
{
}
  
FIDOAddress& FIDOAddressIterator::operator () () const
{
  return *(FIDOAddress *)&(CollectionIterator::operator()());
}

// Hidden

int Hidden::equal( const BaseObject &t ) const
{
  return stricmp( phone, (*(const Hidden *)&t).phone ) == 0;
}

void Hidden::print( ostream& s ) const
{
  strItem( s, phone );

  s << ",Wait:" << wait
    << ",Dial:";
  strItem( s, dial );

  s << ",Suff:";
  strItem( s, suff );

  s << ",Work:";

  if( work==NULL )
    s << "<NONE>";
  else
    s << *work;
}

char *Hidden::nameOfClass() const
{
  return "Hidden";
}

char *findPref( char *flg, CallsList *listPref )
{
  if( listPref!=NULL )
  {
    char *ret( NULL );
    char *q = (char *)_alloca( strlen(flg)+3 );

    strcpy( q, "," );
    strcat( q, flg );
    strcat( q, "," );

    for( CallsIterator list( *listPref ); list && ret==NULL; list++ )
    {
      Calls& cur = list();
      char *patt = new char [ strlen( cur.getFlag() )+3 ];

      strcpy( patt, "," );
      strcat( patt, cur.getFlag() );
      strcat( patt, "," );

      if( includeStr( patt, q ) )
	ret = cur.getPref();

      delete patt;
    }

    return ret;
  }
  else
    return NULL;
}

Hidden::Hidden( int wat, char *ph, char *dil, WorkTime *wt, char *suf ) :
  wait( wat ),
  phone( newStr( ph ) ),
  dial( newStr( dil ) ),
  suff( newStr( suf ) )
{
  if( wt!=NULL )
    work = new WorkTime( wt );
  else
    work = NULL;
}

Hidden::Hidden( Hidden *i )
{
  if( i==NULL )
  {
    wait = 0;
    phone = NULL;
    dial = NULL;
    suff = NULL;
    work = NULL;
  }
  else
  {
    wait = i->wait;
    phone = newStr( i->phone );
    dial = newStr( i->dial );
    suff = newStr( i->suff );

    if( i->work!=NULL )
      work = new WorkTime ( i->work );
    else
      work = NULL;
  }
}

Hidden::~Hidden()
{
  delete phone;
  delete dial;
  delete suff;
  delete work;
}

int Hidden::isWorked( long stamp ) const
{
  if( work!=NULL )
    return work->isWorked( stamp );
/*
  else
    if( miscInfo.zmh!=NULL && miscInfo.zmh->isWorked( stamp ) )
      return 1;
*/

  return 0;
}

HiddenList::HiddenList() :
  Collection( 5, 5, SemaphoreModeNoNeed )
{
}

Hidden* HiddenList::Find( Hidden& item, LockMode lock ) const
{
  return (Hidden*)Collection::Find( item, lock );
}
    
Hidden& HiddenList::operator [] ( const int pos ) const
{
  return *(Hidden *)&(Collection::operator[]( pos ));
}

HiddenIterator::HiddenIterator( HiddenList& ini ) :
  CollectionIterator( *(Collection *)&ini )
{
}
  
Hidden& HiddenIterator::operator () () const
{
  return *(Hidden *)&(CollectionIterator::operator()());
}

// NodeConfig

int NodeConfig::equal( const BaseObject &t ) const
{
  if( mask==NULL && (*(const NodeConfig *)&t).mask==NULL )
    return FIDOAddress::equal( *(const FIDOAddress *)(const NodeConfig *)&t );
  else if( mask!=NULL && (*(const NodeConfig *)&t).mask!=NULL )
    return !strcmp( mask, (*(const NodeConfig *)&t).mask );
  else
    return 0;
}

void itemInt( ostream& o, char *s1, int s2 )
{
  o << endl << "    ";
  strItem( o, s1 );
  o << char(':') << s2;
}

char *NodeConfig::toString( char *buf, int len, FIDOAddressShowMode mode ) const
{
  if( mask!=NULL )
  {
    strncpy( buf, mask, len );
    buf[ len-1 ] = 0;
  }
  else
    FIDOAddress::toString( buf, len, mode );

  return buf;
}

void NodeConfig::print( ostream& s ) const
{
  s << "Node:" << (isValid() ? "Valid," : "NotValid,");

  if( mask==NULL )
    FIDOAddress::print( s );
  else
    s << "Mask:" << mask << char(',');

  item( s, "Alias", alias );

  s << endl << "    Mode:" << flavourName( getCfgFlavour() );

  s << endl << "    ";

  if( work!=NULL )
    s << *work;
  else
    s << "*No WorkTime*";

  s << endl << "    " << (isListed() ? "Listed:" : "NotListed");

  if( isListed() )
    s << endl << *ndl;

  itemInt( s, "Cost", cost );
  itemInt( s, "Baud", baud );

  itemSched( s, "Before", before );
  itemSched( s, "After", after );
  itemSched( s, "Do", doit );

  s << endl << "    ";

  if( aka==NULL )
    s << "No AKAs";
  else
    s << *aka;

  s << endl << "    ";

  if( hidden==NULL )
    s << "No Hidden";
  else
    s << "Hidden:" << *hidden;

  itemInt( s, "Current", current );
  item( s, "Flg", flags );
  item( s, "Pwd", getPassword() );
  item( s, "Phn", phone );
  item( s, "RXD", rexxDial );
  item( s, "Dil", dial );
  item( s, "Suf", suff );
  itemInt( s, "Car", wait );
  itemInt( s, "MaxIn", maxIn );
  itemInt( s, "MaxOut", maxOut );
  itemInt( s, "MaxTry", maxTry );
  itemInt( s, "Prio", priority );
  item( s, "Iut", getImmedBox() );
  item( s, "Cut", getCrashBox() );
  item( s, "Dut", getDirectBox() );
  item( s, "Nut", getNormalBox() );
  item( s, "Hut", getHoldBox() );
  item( s, "Inb", getInbound() );

  s << endl << "    "
    << "DisableDZA:" << (noDZAenabled ? "Yes" : "No") << char( ' ' ) 
    << "DisableZAP:" << (noZAPenabled ? "Yes" : "No") << char( ' ' ) 
    << "DisableHydra:" << (noHydraEnabled ? "Yes" : "No") << char( ' ' )
    << "Chain:" << (chain ? "Yes" : "No");
}

char *NodeConfig::nameOfClass() const
{
  return "NodeConfig";
}

extern int cfgLine;

void NodeConfig::changeAka( FIDOAddressList *list, int line )
{
  if( list!=NULL )
  {
    if( aka!=NULL )
    {
      log( Msg_CfgNodeDoubleAka, line );
      aka->Flush();
    }
    else
      aka = new FIDOAddressList;

    for( FIDOAddressIterator iter( *list ); iter; iter++ )
      aka->AddAtTail( *new FIDOAddress( &iter() ) );
  }
}

void NodeConfig::changeHidden( HiddenList *lines, int line )
{
  if( lines!=NULL )
  {
    if( hidden!=NULL )
    {
      log( Msg_CfgNodeDoubleHidden, line );
      hidden->Flush();
    }
    else
      hidden = new HiddenList;

    for( HiddenIterator iter( *lines ); iter; iter++ )
      hidden->AddAtTail( *new Hidden( &iter() ) );
  }
}

void NodeConfig::changeWork( WorkTime *iwork )
{
  delete work;
  work = iwork!=NULL ? new WorkTime( iwork ) : (WorkTime *)NULL;
}

void NodeConfig::changeMask( char *str )
{
  delete mask;
  mask = newStr( str );
}

void NodeConfig::changeImmedBox( char *str )
{
  delete immedBox;
  immedBox = newStr( str );
}

void NodeConfig::changeDirectBox( char *str )
{
  delete directBox;
  directBox = newStr( str );
}

void NodeConfig::changeHoldBox( char *str )
{
  delete holdBox;
  holdBox = newStr( str );
}

void NodeConfig::changeNormalBox( char *str )
{
  delete normalBox;
  normalBox = newStr( str );
}

void NodeConfig::changeCrashBox( char *str )
{
  delete crashBox;
  crashBox = newStr( str );
}

void NodeConfig::changeInbound( char *str )
{
  delete inbound;
  inbound = newStr( str );
}

void NodeConfig::changePassword( char *str )
{
  delete password;
  password = newStr( str );
}

void NodeConfig::changePhone( char *str )
{
  delete phone;
  phone = newStr( str );
}

void NodeConfig::changeDial( char *str )
{
  delete dial;
  dial = newStr( str );
}

void NodeConfig::changeDialSuff( char *str )
{
  delete suff;
  suff = newStr( str );
}

void NodeConfig::changeFlags( char *str )
{
  delete flags;
  flags = newStr( str );
}

void NodeConfig::changeRexxDial( char *str )
{
  delete rexxDial;
  rexxDial = newStr( str );
}

void NodeConfig::changeWait( int wt )
{
  wait=wt;
}

void NodeConfig::changeBaud( int wt )
{
  baud=wt;
}

void NodeConfig::changeFlavour( Flavour wt )
{
  flavour=wt;
}

void NodeConfig::changeMaxIn( short wt )
{
  maxIn=wt;
}

void NodeConfig::changeMaxOut( short wt )
{
  maxOut=wt;
}

void NodeConfig::changeMaxTry( int wt )
{
  maxTry=wt;
}

void NodeConfig::changeCost( int wt )
{
  cost=wt;
}

void NodeConfig::changeNoDZAenabled( int wt )
{
  noDZAenabled=wt;
}

void NodeConfig::changeNoZAPenabled( int wt )
{
  noZAPenabled=wt;
}

void NodeConfig::changeNoHydraEnabled( int wt )
{
  noHydraEnabled=wt;
}

void NodeConfig::changeAfter( Schedule *wt )
{
  deleteSched( after );
  after=newSched(wt);
}

void NodeConfig::changeBefore( Schedule *wt )
{
  deleteSched( before );
  before=newSched(wt);
}

void NodeConfig::changeDoit( Schedule *wt )
{
  deleteSched( doit );
  doit=newSched(wt);
}

void NodeConfig::changeChain( int ch )
{
  chain = ch;
}

void NodeConfig::changePriority( int prio )
{
  priority = prio;
}

void NodeConfig::changeAlias( char *str )
{
  delete alias;
  alias = newStr( str );
}

void NodeConfig::init( char *ph, char *pass,
                       char *norm, char *hold, char *crash,
                       char *immed, char *direct,
                       char *inb,
                       char *preff, int wt, char *fl,
                       FIDOAddressList *list, HiddenList *lines,
                       Schedule *iBefore, Schedule *iAfter,
                       Schedule *iDoit,
                       char *rD, char *suf, char *ali )
{
  phone = newStr( ph );
  password = newStr( pass );
  normalBox = newStr( norm );
  holdBox = newStr( hold );
  crashBox = newStr( crash );
  immedBox = newStr( immed );
  directBox = newStr( direct );
  inbound = newStr( inb );
  dial = newStr( preff );
  suff = newStr( suf );
  flags = newStr( fl );
  wait = wt;
  rexxDial = newStr( rD );
  alias = newStr( ali );

  before = newSched( iBefore );
  after  = newSched( iAfter );
  doit   = newSched( iDoit );

  changeAka( list, 0 );
  changeHidden( lines, 0 );

  if( mask==NULL && ndl==NULL && nodeListPtr!=NULL && nodeList.lock( 500 ) )
  {
    CacheItem *item = nodeList.find( getZone(), getNet(),
                                     getNode(), getPoint() );

    if( item!=NULL )
      ndl = new CacheItem( item );

    nodeList.unlock();
  }

  if( mask==NULL && ndl!=NULL && (phone==NULL || work==NULL) )
  {
    if( phone==NULL )
      phone = newStr( ndl->getPhone() );

    if( work==NULL )
    {
      char *tm = newStr( ndl->getTime() );

      if( tm!=NULL )
      {
        work = new WorkTime( tm );

        if( !work->isValid() )
        {
          delete work;
          work = NULL;
        }
      }

      delete tm;
    }
  }

  if( mask==NULL && phone!=NULL && *phone!='#' )
/*
    for( int i(0); phone[i]; i++ )
      if( !isdigit(phone[i]) && phone[i]!='-' )
*/
    if( !stricmp( phone, "-Unpublished-" ) )
    {
      delete phone;
      phone = NULL;
//      break;
    }

  char *flg = getFlags();

  if( mask==NULL && dial==NULL && flg!=NULL )
    dial = newStr( findPref( flg, preffixList ) );

  if( mask==NULL && suff==NULL && flg!=NULL )
    suff = newStr( findPref( flg, suffixList ) );

  if( mask==NULL && !cost && translationTable!=NULL && phone!=NULL )
    cost = translationTable->findCost( phone );

  if( mask==NULL && !baud && ndl!=NULL )
    baud = ndl->getBaud();

// get from mask

  resolveFromMask();

// post process

  if( !baud )
    baud = 300;
}

void NodeConfig::resolveFromMask()
{
  for( NodeConfigIterator listPos( *maskList ); listPos; listPos++ )
  {
    NodeConfig& testNode = listPos();
    char *itemMask = testNode.getMask();
    RegularExpresion expression( itemMask );
    char test[ 128 ];

    toString( test, sizeof(test)-1, FIDOAddressShowModeEMSI );

    if( expression.maskMach( test ) )
    {
      if( work==NULL )
        changeWork( testNode.work );

      if( holdBox==NULL )
        holdBox = newStr( testNode.holdBox );

      if( normalBox==NULL )
        normalBox = newStr( testNode.normalBox );

      if( crashBox==NULL )
        crashBox = newStr( testNode.crashBox );

      if( immedBox==NULL )
        immedBox = newStr( testNode.immedBox );

      if( directBox==NULL )
        directBox = newStr( testNode.directBox );

      if( inbound==NULL )
        inbound = newStr( testNode.inbound );

      if( password==NULL )
        password = newStr( testNode.password );

      if( dial==NULL )
        dial = newStr( testNode.dial );

      if( suff==NULL )
        suff = newStr( testNode.suff );

      if( phone==NULL )
        phone = newStr( testNode.phone );

      if( flags==NULL )
        flags = newStr( testNode.flags );

      if( alias==NULL )
        alias = newStr( testNode.alias );

      if( !wait )
        wait = testNode.wait;

      if( !baud )
        baud = testNode.baud;

      if( flavour==unknownFlavour )
        flavour = testNode.flavour;

      if( !priority )
        priority = testNode.priority;

      if( !maxIn )
        maxIn = testNode.maxIn;

      if( !maxOut )
        maxOut = testNode.maxOut;

      if( !maxTry )
        maxTry = testNode.maxTry;

      if( aka==NULL )
        changeAka( testNode.aka, 0 );

      if( hidden==NULL )
        changeHidden( testNode.hidden, 0 );

      if( after==NULL )
        after=newSched( testNode.after );

      if( before==NULL )
        before=newSched( testNode.before );

      if( doit==NULL )
        doit=newSched( testNode.doit );

      if( !noDZAenabled )
        noDZAenabled = testNode.noDZAenabled;

      if( !noZAPenabled )
        noZAPenabled = testNode.noZAPenabled;

      if( !noHydraEnabled )
        noHydraEnabled = testNode.noHydraEnabled;

      if( rexxDial==NULL )
        rexxDial = newStr( testNode.rexxDial );

      if( !cost )
        cost = testNode.cost;

      if( !chain )
        chain = testNode.chain;

      break;
    }
  }
}

char *NodeConfig::getTranslatedPhone( char *orig )
{
  if( orig==NULL )
    return NULL;

  char node[65];

  toString( node, sizeof(node)-1, FIDOAddressShowModeEMSI );

  char *rexxPhone = processRexxHook( "Translate", node, orig, NULL,
                                     NULL, NULL, NULL, 1 );

  if( rexxPhone==NULL || !*rexxPhone )
  {
    delete rexxPhone;

    if( translationTable!=NULL )
      translationTable->translate( orig, &rexxPhone );
    else
      rexxPhone = newStr( orig );
  }

  return rexxPhone;
}

NodeConfig::NodeConfig( char *iAddr, char *iWork, Flavour iF,
			char *ph, char *pass,
			char *norm, char *hold, char *crash,
			char *immed, char *direct,
			char *inbound, char *preff, int wt, char *fl,
			short iMaxIn, short iMaxOut, int iMaxTry,
			FIDOAddressList *list, HiddenList *lines,
			Schedule *iBefore, Schedule *iAfter,
			Schedule *iDoit, int noDZA, char *rexxDial,
			int cs, int ba, int noHydra, int chaIn,
                        char *ms, int prio, char *suf, char *alias,
                        int noZAP ) :
  FIDOAddress( iAddr ),
  flavour( iF ),
  ndl( NULL ),
  maxIn( iMaxIn ),
  maxOut( iMaxOut ),
  maxTry( iMaxTry ),
  aka( NULL ),
  current( -1 ),
  hidden( NULL ),
  noDZAenabled( noDZA ),
  noZAPenabled( noZAP ),
  noHydraEnabled( noHydra ),
  cost( cs ),
  baud( ba*10 ),
  chain( chaIn ),
  mask( newStr( ms ) ),
  priority( prio )
{
  if( iWork!=NULL )
    work = new WorkTime(iWork);
  else
    work = NULL;

  init( ph, pass,
        norm, hold, crash, immed, direct,
        inbound, preff, wt, fl, list, lines,
	iBefore, iAfter, iDoit, rexxDial, suf, alias );
}


NodeConfig::NodeConfig( FIDOAddress *iAddr, WorkTime *iWork, Flavour iF,
			char *ph, char *pass,
			char *norm, char *hold, char *crash,
			char *immed, char *direct,
			char *inbound, char *preff, int wt, char *fl,
			short iMaxIn, short iMaxOut, int iMaxTry,
			FIDOAddressList *list, HiddenList *lines,
			Schedule *iBefore, Schedule *iAfter,
			Schedule *iDoit, int noDZA, char *rexxDial,
			int cs, int ba, int noHydra, int chaIn,
                        char *ms, int prio, char *suf, char *alias,
                        int noZAP ) :
  FIDOAddress( iAddr ),
  flavour( iF ),
  ndl( NULL ),
  maxIn( iMaxIn ),
  maxOut( iMaxOut ),
  maxTry( iMaxTry ),
  aka( NULL ),
  current( -1 ),
  hidden( NULL ),
  noDZAenabled( noDZA ),
  noZAPenabled( noZAP ),
  noHydraEnabled( noHydra ),
  cost( cs ),
  baud( ba*10 ),
  chain( chaIn ),
  mask( newStr( ms ) ),
  priority( prio )
{
  if( iWork!=NULL )
    work = new WorkTime( iWork );
  else
    work = NULL;

  init( ph, pass,
        norm, hold, crash, immed, direct,
        inbound, preff, wt, fl, list, lines,
	iBefore, iAfter, iDoit, rexxDial, suf, alias );
}

NodeConfig::NodeConfig( const NodeConfig *i ) :
  FIDOAddress( (const FIDOAddress *)i ),
  aka( NULL ),
  hidden( NULL )
{
  if( i!=NULL )
  {

    if( i->work!=NULL )
      work = new WorkTime( i->work );
    else
      work = NULL;

    flavour = i->flavour;

    if( i->ndl!=NULL )
      ndl = new CacheItem( i->ndl );
    else
      ndl = NULL;

    maxIn = i->maxIn;
    maxOut = i->maxOut;
    maxTry = i->maxTry;
    current = i->current;
    noDZAenabled = i->noDZAenabled;
    noZAPenabled = i->noZAPenabled;
    noHydraEnabled = i-> noHydraEnabled;
    cost = i->cost;
    baud = i->baud;
    chain = i->chain;
    priority = i->priority;
    mask = newStr( i->mask );

    init( i->phone, i->password,
          i->normalBox, i->holdBox, i->crashBox, i->immedBox, i->directBox,
	  i->inbound, i->dial, i->wait, i->flags, i->aka, i->hidden,
	  i->before, i->after, i->doit, i->rexxDial, i->suff, i->alias );
  }
  else
  {
    work = NULL;
    phone = password = dial = suff = NULL;
    directBox = immedBox = normalBox = holdBox = crashBox = inbound = NULL;
    wait = 0;
    ndl = NULL;
    flags = NULL;
    maxIn = maxOut = 0;
    maxTry = 0;
    current = -1;
    after = before = doit = NULL;
    noDZAenabled = 0;
    noZAPenabled = 0;
    noHydraEnabled = 0;
    rexxDial = NULL;
    cost = 0;
    baud = 300;
    chain = 0;
    priority = 0;
    mask = NULL;
    alias = NULL;
  }
}

NodeConfig::~NodeConfig()
{
  delete mask;
  delete work;
  delete phone;
  delete password;
  delete normalBox;
  delete holdBox;
  delete crashBox;
  delete immedBox;
  delete directBox;
  delete dial;
  delete suff;
  delete inbound;
  delete ndl;
  delete flags;
  delete rexxDial;
  delete alias;

  deleteSched( before );
  deleteSched( after );
  deleteSched( doit );

  delete aka;
  delete hidden;
}

char *NodeConfig::getPhone() const
{
  if( current==-1 )
    return phone;

  Hidden *hi = getCurrentHidden();

  if( hi!=NULL )
    return hi->getPhone();

  return NULL;
}

char *NodeConfig::getDial() const
{
  if( current==-1 )
    return dial;

  Hidden *hi = getCurrentHidden();

  if( hi!=NULL )
    return hi->getDial();

  return NULL;
}

char *NodeConfig::getDialSuff() const
{
  if( current==-1 )
    return suff;

  Hidden *hi = getCurrentHidden();

  if( hi!=NULL )
    return hi->getDialSuff();

  return NULL;
}

int NodeConfig::getWaitCarrier() const
{
  if( current==-1 )
    return wait;

  Hidden *hi = getCurrentHidden();

  if( hi!=NULL )
    return hi->getWait();

  return 0;
}

int NodeConfig::isValid() const
{
  if( !FIDOAddress::isValid() )
    return 0;

  if( work==NULL )
    return 1;
  else if( work->isValid() )
    return 1;
  else if( hidden!=NULL )
    for( HiddenIterator list( *hidden ); list; list++ )
    {
      Hidden& cur = list();

      if( cur.getPhone()!=NULL )
        if( cur.getWorkTime()!=NULL )
          if( cur.getWorkTime()->isValid() )
            return 1;
    }

  return 0;
}

int NodeConfig::isPrimaryWorked( uLong stamp ) const
{
  if( !isValid() )
    return 0;

  if( work!=NULL )
    return work->isWorked( stamp );

  if( miscInfo.zmh!=NULL && !is4D() )
    return miscInfo.zmh->isWorked( stamp );

  return 0;
}

int NodeConfig::isWorked( uLong stamp ) const
{
  if( isPrimaryWorked( stamp ) )
    return 1;

  if( hidden!=NULL )
    for( HiddenIterator list( *hidden ); list; list++ )
    {
      Hidden& cur = list();

      if( cur.getWorkTime()!=NULL && cur.getWorkTime()->isWorked( stamp ) )
        return 1;
    }

  return 0;
}

Hidden *NodeConfig::getCurrentHidden() const
{
  if( current!=-1 && hidden!=NULL )
  {
    int i=0;

    for( HiddenIterator list( *hidden ); list; list++, i++ )
      if( i==current )
	return &list();
  }

  return NULL;
}

void NodeConfig::setCurrent()
{
  if( hidden!=NULL )
  {
    if( current==-1 )
      current = 0;
    else if( ++current>=hidden->GetItemsInContainer() )
      current = -1;

    NodeConfig test( this ),
               *found( systemList->Find( test ) );

    if( found!=NULL )
      found->current = current;

    QueueEntry test1( *this );

    queue->Lock();

    QueueEntry *found1( queue->Find( test1, LockModeNoNeed ) );

    if( found1!=NULL )
      ((NodeConfig*)found1)->current = current;

    queue->UnLock();
  }
}

void NodeConfig::findAvailable()
{
  if( hidden==NULL )
    return;

  long stamp = time( NULL );

  int i = current;

  do
  {
    if( current == -1 )
      if( isPrimaryWorked( stamp ) )
	return;
      else
      {
        setCurrent();

        if( i==current )
          break;
      }

    Hidden *hi =  getCurrentHidden();

    if( hi==NULL )
    {
      current = -1;
      return;
    }

    if( hi->isWorked( stamp ) )
      return;

    setCurrent();

  } while( i!=current );
}

char *NodeConfig::getFlags() const
{
  if( flags!=NULL )
    return flags;

  if( ndl!=NULL )
    return ndl->getFlags();

  return NULL;
}

short NodeConfig::getMaxInComing() const
{
  return maxIn ? maxIn : miscInfo.maxIn;
}

short NodeConfig::getMaxOutGoing() const
{
  return maxOut ? maxOut : miscInfo.maxOut;
}

int NodeConfig::getMaxTry() const
{
  return maxTry ? maxTry : miscInfo.maxTry;
}

// NodeConfigList

NodeConfigList::NodeConfigList() :
  Collection( 50, 50 )
{
}

NodeConfig* NodeConfigList::Find( NodeConfig& item, LockMode lock ) const
{
  return (NodeConfig*)Collection::Find( item, lock );
}
    
NodeConfig& NodeConfigList::operator [] ( const int pos ) const
{
  return *(NodeConfig *)&(Collection::operator[]( pos ));
}

void NodeConfigList::postProcess()
{
  for( NodeConfigIterator cur( *this ); cur; cur++ )
    cur().resolveFromMask();
}

NodeConfig *NodeConfigList::FindAlias( const char *alias )
{
  for( NodeConfigIterator cur( *this ); cur; cur++ )
    if( cur().getAlias()!=NULL && !stricmp( cur().getAlias(), alias ) )
      return &cur();

  return NULL;
}

// NodeConfigIterator

NodeConfigIterator::NodeConfigIterator( NodeConfigList& ini ) :
  CollectionIterator( *(Collection *)&ini )
{
}
  
NodeConfig& NodeConfigIterator::operator () () const
{
  return *(NodeConfig *)&(CollectionIterator::operator()());
}

// Flavours maintenance routine

char *flavourName( Flavour f )
{
  switch( f )
  {
  case immedFlavour:
    return Msg_FlavourImmed;

  case crashFlavour:
    return Msg_FlavourCrash;

  case directFlavour:
    return Msg_FlavourDirect;

  case normalFlavour:
    return Msg_FlavourNormal;

  case holdFlavour:
    return Msg_FlavourHold;
  }

  return Msg_FlavourUnknown;
}

char *flavourString( Flavour f )
{
  switch( f )
  {
  case immedFlavour:
    return "Immediate";

  case crashFlavour:
    return "Crash";

  case directFlavour:
    return "Direct";

  case normalFlavour:
    return "Normal";

  case onlineFlavour:
    return "Online";

  case requestFlavour:
    return "Request";
  }

  return "Hold";
}

char *realFlavour( Flavour flavour )
{
  switch( flavour )
  {
  case immedFlavour:
    return "Immed";

  case crashFlavour:
    return "Crash";

  case directFlavour:
    return "Direct";

  case normalFlavour:
    return "Normal";

  case holdFlavour:
    return "Hold";

  case onlineFlavour:
    return "Online";

  case requestFlavour:
    return "FREQ";
  }

  return "*UNKNOWN*";
}

int immedFlavourPriority  = immedFlavour;
int crashFlavourPriority  = crashFlavour;
int directFlavourPriority = directFlavour;
int normalFlavourPriority = normalFlavour;
int holdFlavourPriority   = holdFlavour;

int flavourPriority( Flavour fl )
{
  switch( fl )
  {
  case immedFlavour:
    return immedFlavourPriority;

  case crashFlavour:
    return crashFlavourPriority;

  case directFlavour:
    return directFlavourPriority;

  case normalFlavour:
    return normalFlavourPriority;

  case holdFlavour:
    return holdFlavourPriority;
  }

  return 0;
}

char *flavourNameByPriority( int prio )
{
  static char buf[32];

  if( prio==holdFlavourPriority )
    return Msg_FlavourHold;
  else if( prio==normalFlavourPriority )
    return Msg_FlavourNormal;
  else if( prio==directFlavourPriority )
    return Msg_FlavourDirect;
  else if( prio==crashFlavourPriority )
    return Msg_FlavourCrash;
  else if( prio==immedFlavourPriority )
    return Msg_FlavourImmed;

  sprintf( buf, "%d", prio );
  return buf;
}

FLO::FLO( char *_path ) :
  ok(1),
  xlo( 0 ),
  path( newStr( _path ) )
{
  if( access( path, 0 ) )
    return;

  xlo = _fsopen( path, "r+", SH_DENYWR );

  if( xlo==NULL )
  {
    log( Msg_ErrUnableOpen, path, strError() );
    return;
  }
}

FLO::~FLO()
{
  if( xlo )
    if( ok )
    {
      fclose( xlo );

      if( unlink( path ) )
        log( Msg_ErrUnableDelete, path, strError() );
    }
    else
    {
      purge();
      fclose( xlo );
      setExtendedAttribute( path, ".TYPE", Msg_EALO );
    }

  delete path;
}

void FLO::purge()
{
  rewind( xlo );
  
  long readPos( ftell(xlo) ),
       writePos( readPos );
  char err(0);
  
  while( 1 )
  {
    char name[ _MAX_PATH ];
    
    fseek( xlo, readPos, SEEK_SET );
    *name = 0;
    fgets( name, _MAX_PATH, xlo );
    
    if( ferror( xlo ) )
    {
      log( Msg_ErrUnableRead, path, strError() );
      err = 1;
      break;
    }
    
    readPos = ftell( xlo );
    
    allTrim( name );

    if( *name=='~' || !*name )
      if( feof( xlo ) )
        break;
      else
        continue;
      
    fseek( xlo, writePos, SEEK_SET );
    fprintf( xlo, "%s\n", name );
    
    if( ferror( xlo ) )
    {
      log( Msg_ErrUnableWrite, path, strError() );
      err = 1;
      break;
    }
    
    writePos = ftell( xlo );
  }
  
  fflush( xlo );
  
  if( _chsize( fileno(xlo), writePos ) )
  {
    log( Msg_ErrUnableChSize, path, strError() );
    err = 1;
  }
  
  if( err )
    log( Msg_ErrMayBeCorruptFlo, fileName( path ) );
}

int FLO::process( int delFile, int (*func)( char *, char, void * ), void *arg )
{
  if( !xlo )
    return OK;

  char name[ _MAX_PATH ];
  int zret( OK );

  while( !feof( xlo ) && !ferror( xlo ) && zret!=ERROR )
  {
    char *test( name );
    long before,
         after;

    *name = 0;
    before = ftell( xlo );
    fgets( name, _MAX_PATH, xlo );
    after = ftell( xlo );

    allTrim( name );

    if( name[0]=='~' || !name[0] )
      continue;

    char flag(0);

    if( *name=='#' || *name=='^' )
    {
      flag = *name;
      test++;
    }

    if( !*test )
      continue;

    makeFileCase( test );

    zret = func( test, flag, arg );

    if( zret==OK || zret==SPEC_COND )
    {
      fseek( xlo, before, SEEK_SET );
      fputc( '~', xlo );
      fseek( xlo, after, SEEK_SET );
      real_flush( xlo );
      
      if( delFile )
        postProcessAttach( flag, test );
    }
    else
    {
      ok = 0;

      if( zret==VERY_SPEC_COND )
        zret = OK;
      else
        zret = ERROR;
    }
  }

  if( ferror( xlo ) )
    log( Msg_ErrUnableRead, path, strError() );

  if( !feof( xlo ) )
    ok = 0;

  return zret;
}

void postProcessAttach( char flag, char *test )
{
  switch( flag )
  {
  case '^':
    chmod( test, S_IREAD|S_IWRITE );

    if( unlink( test ) )
      log( Msg_ErrUnableDelete, test, strError() );

    break;

  case '#':
    chmod( test, S_IREAD|S_IWRITE );

    FILE *tr = _fsopen( test, "w" );

    if( tr==NULL )
      log( Msg_ErrUnableTruncate, test, strError() );
    else
    {
      fclose( tr );
      setExtendedAttribute( test, ".TYPE", Msg_EATruncatedBundle );
    }

    break;
  }
}