#define INCL_DOSNMPIPES
#define INCL_DOSERRORS
#define INCL_KBD
#define INCL_DOS
#include <os2.h>

#include <signal.h>
#include <string.h>

#include "key.h"
#include "misc.h"
#include "except.h"

#include "language.h"

static ULONG lastKeyPressed = 0;
static volatile char screenSaverIsActive = 0;
static volatile int screenSaverThread = 0;
static volatile char screenSaverIsEnabled = 1;
HPIPE pipeHandle = 0xffffffffl;

void disableScreenSaver()
{
  screenSaverIsActive = 0;
  screenSaverIsEnabled = 0;
  
  if( screenSaverThread )
    DosWaitThread( (PTID)&screenSaverThread, DCWW_WAIT );
}

void enableScreenSaver()
{
  screenSaverIsEnabled = 1;
  
  DosQuerySysInfo( QSV_MS_COUNT, QSV_MS_COUNT, &lastKeyPressed,
                   sizeof lastKeyPressed );

  lastKeyPressed /= 1000;
}

void _Optlink screenSaverThreadProc( void * )
{
  registerExceptionHandler

  DosSetPriority( PRTYS_THREAD, PRTYC_FOREGROUNDSERVER, 0x1f, 0 );
  
  char errBuf[80] = {0};
  HMODULE h( 0 );
  char externSaver( 1 );
  extern volatile char VioScreenSaverActive;

  struct scr_stat
  {
    int xsize;
    int ysize;
  };

  void (* __cdecl load)( scr_stat * )( 0 );
  void (* __cdecl step)()( 0 );
  void (* __cdecl unload)()( 0 );

  if( !miscInfo.saverModule )
    externSaver = 0;
  else
  {
    APIRET rc( DosLoadModule( errBuf, sizeof(errBuf), miscInfo.saverModule, &h ) );

    if( rc )
    {
      log( Msg_ErrScrSaverDll,
           *errBuf ? " -> " : "",
           miscInfo.saverModule, errBuf, rc );
      os2Message( rc );
      externSaver = 0;
    }

    if( externSaver )
    {
      rc = DosQueryProcAddr( h, 1, 0, (PFN *)&load );

      if( rc )
      {
        log( Msg_ErrScrSaverNotFound, "load", 1, miscInfo.saverModule, rc );
        os2Message( rc );
        externSaver = 0;
      }
    }

    if( externSaver )
    {
      rc = DosQueryProcAddr( h, 2, 0, (PFN *)&step );

      if( rc )
      {
        log( Msg_ErrScrSaverNotFound, "step", 2, miscInfo.saverModule, rc );
        os2Message( rc );
        externSaver = 0;
      }
    }

    if( externSaver )
    {
      rc = DosQueryProcAddr( h, 3, 0, (PFN *)&unload );

      if( rc )
      {
        log( Msg_ErrScrSaverNotFound, "unload", 3, miscInfo.saverModule, rc );
        os2Message( rc );
        externSaver = 0;
      }
    }
  }

  APIRET16 rc16( KbdGetFocus( 1, 0 ) );

  if( rc16 )
    log( "*BUG* KbdGetFocus failed with RC=%d", rc16 );
  else
  {
    Window *scr;

    if( !externSaver )
    {
      Color cl( WHITE, BLACK );

      scr =  new Window( 1, 1, Z_ScreenWidth, Z_ScreenHeight, cl );
      +(*scr);
    }

    VioScreenSaverActive = 1;

    scr_stat c = { Z_ScreenWidth, Z_ScreenHeight };

    if( externSaver )
      load( &c );

    while( screenSaverIsActive )
    {
      KBDKEYINFO k;

      rc16 = KbdPeek( &k, 0 );

      if( rc16 ? 0 : k.fbStatus & KBDTRF_FINAL_CHAR_IN )
        break;

      if( externSaver )
        step();

      checkPoint();
    }

    if( externSaver )
      unload();

    rc16 = KbdFreeFocus( 0 );

    if( rc16 )
      log( "*BUG* KbdFreeFocus failed with RC=%d", rc16 );

    VioScreenSaverActive = 0;

    if( !externSaver )
      delete scr;
  }

  if( h )
    DosFreeModule( h );

  screenSaverIsActive = 0;
  screenSaverThread = 0;
  deRegisterExceptionHandler
}

extern "C"
{
  APIRET16 (* APIENTRY16 RealKbdCharIn)( PKBDKEYINFO, USHORT, HKBD );
  APIRET16 (* APIENTRY16 RealKbdPeek)( PKBDKEYINFO, HKBD );
}

int InitKbdSystem( void )
{
  char errBuf[80] = {0};
  HMODULE h;

  APIRET rc = DosLoadModule( errBuf, sizeof errBuf, "KBDCALLS", &h );

  if( rc )
  {
    log( Msg_ErrKbdNotLoad, *errBuf ? " -> " : "", errBuf, rc );
    os2Message( rc );
    return int( rc );
  }

  rc = DosQueryProcAddr( h, 0, "KBDCHARIN", (PFN *)&RealKbdCharIn );

  if( rc )
  {
    log( Msg_ErrKbdNotFound1, rc );
    os2Message( rc );
    return int( rc );
  }

  rc = DosQueryProcAddr( h, 0, "KBDPEEK", (PFN *)&RealKbdPeek );

  if( rc )
  {
    log( Msg_ErrKbdNotFound2, rc );
    os2Message( rc );
    return int( rc );
  }

  return 0;
}

extern char detachedSession;

int getch( void )
{
  if( detachedSession )
    return 0;

  KBDKEYINFO  info;
  static  unsigned char flag = 255;
  static  unsigned char code;

  if( (flag == 0) || (flag == 0xe0) )
  {
    flag = 255;
    return code;
  }

 if( KbdCharIn(&info, IO_WAIT, 0) )
   return -1;

  DosQuerySysInfo( QSV_MS_COUNT, QSV_MS_COUNT, &lastKeyPressed,
                   sizeof lastKeyPressed );

  lastKeyPressed /= 1000;

  if( (info.fbStatus & 2) && 
      ((info.chChar == 0) || 
       (info.chChar == 0xe0 && info.chScan) ) )
  {
    flag = info.chChar;
    code = info.chScan;
    return 0;
  }

  return info.chChar;
}

int kbhit( void )
{
  if( detachedSession )
    return 0;

  KBDKEYINFO k;
  APIRET16 ret = KbdPeek( &k, 0 );
  int pressed( ret ? 0 : k.fbStatus & KBDTRF_FINAL_CHAR_IN );
  ULONG mark(0);

  DosQuerySysInfo( QSV_MS_COUNT, QSV_MS_COUNT, &mark, sizeof mark );
  mark /= 1000;

  if( !lastKeyPressed )
    lastKeyPressed = mark;
  else if( !pressed &&
           screenSaverIsEnabled &&
           !screenSaverIsActive && 
           miscInfo.saverInterval &&
           mark > lastKeyPressed+miscInfo.saverInterval )
  {
    screenSaverIsActive = 1;

    screenSaverThread = _beginthread( screenSaverThreadProc, 0, 64*1024, NULL );

    if( screenSaverThread == -1 )
    {
      log( Msg_ErrScrSaverThr, strError() );
      screenSaverIsActive = 0;
    }
    else
      log( Msg_ThrScrSaver, screenSaverThread );
  }

  return pressed;
}

static volatile char CtrlC = 0;
static volatile int locked = 0;

void keyBoardLock()
{
  locked = 1;
}

void keyBoardUnLock()
{
  locked = 0;
}

int keyBoardIsLocked()
{
  return locked;
}

void keyClear( void )
{
  if( locked )
    return;

  while( keyPressed() )
    keyGet();
}

int connectNPipe( void )
{
  APIRET rc;

  if( pipeHandle==0xffffffffl )
    return 0;

connect_pipe:
  rc=DosConnectNPipe( pipeHandle );

  if( rc==ERROR_BROKEN_PIPE )
  {
    DosDisConnectNPipe( pipeHandle );
    goto connect_pipe;
  }

  return rc==0;
}

static int foundPipe=0;

int keyGet( void )
{
  if( locked )
    return 0;

  if( foundPipe )
  {
    int ch=0;
    ULONG readed=0;

    foundPipe = 0;

    if( !connectNPipe() )
      return 0;

    DosRead( pipeHandle, &ch, 1, &readed );

    if( !readed )
      return 0;

    if( ch )
      return ch;

    DosRead( pipeHandle, &ch, 1, &readed );

    if( !readed )
      return 0;

    return ch<<8;
  }
  else
  {
    int ch;

    if( CtrlC )
    {
      CtrlC = 0;
      ch = 3;
    }
    else
      ch = getch();

    if( ch )
      return ch;

    return getch()<<8;
  }
}

int keyPressed( void )
{
  if( locked )
    return 0;
  else if( CtrlC )
    return 3;
  else
  {
    register int rt;

    APIRET rc;
    char ch;
    ULONG readed = 0,
  	  state;
    AVAILDATA avail;

    rt = kbhit();

    if( !rt )
    {
      if( !connectNPipe() )
        return 0;

      rc = DosPeekNPipe( pipeHandle, &ch, 1, &readed, &avail, &state );
      rt = (rc || readed==0) ? 0 : 1;

      if( rt )
        foundPipe = 1;

      return rt;
    }

    return rt;
  }
}

void keyCtrlCFound()
{
  CtrlC = 1;
}
