#define INCL_DOSPROCESS
#define INCL_DOSQUEUES
#define INCL_DOSNMPIPES
#define INCL_DOSERRORS
#define INCL_DOSEXCEPTIONS
#include <os2.h>

#include <malloc.h>
#include <process.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <time.h>
#include <fcntl.h>
#include <io.h>

#include "except.h"

#include "language.h"

#include "misc.h"
#include "timer.h"
#include "queue.h"
#include "nodelist.h"
#include "translat.h"
#include "key.h"
#include "ea.h"
#include "nolog.h"
#include "rexxhook.h"

unsigned long LOG::queue;

static char volatile mayFlush = 1;

static char *logBuffer;
static volatile int logBufferPtr;
static char *logName;

extern HPIPE pipeHandle;

void flushBuffer()
{
  if( !logBufferPtr )
    return;

  const tryCount = 10;

  for( int i(0); i<tryCount; i++ )
  {
    FILE *f = _fsopen( logName, "a" );

    if( f!=NULL )
    {
      if( fwrite( logBuffer, 1, logBufferPtr, f )!=logBufferPtr )
        log( Msg_ErrUnableWrite, logName, strError() );

      fclose( f );
      break;
    }
    else if( i==tryCount-1 )
      log( Msg_ErrUnableOpen, logName, strError() );
    else
      sleep( 1000 );
  }

  logBufferPtr = 0;
}

void toBuffer( int ch )
{
  logBuffer[ logBufferPtr ] = char( ch );

  if( ++logBufferPtr==miscInfo.logBuffer )
    flushBuffer();
}

void toBuffer( register char *str )
{
  while( *str )
  {
    toBuffer( int( *str ) );
    str++;
  }
}

void toBufferAndFilter( unsigned long tid, register char *str )
{
  static int haveHook( 0 );
  static unsigned long myTid( 0 );

  if( !myTid )
  {
    myTid = getMyTid();
    haveHook = getRexxHook( "LogFilter" )==NULL ? 0 : 1;
  }

  if( !haveHook || tid == myTid )
    toBuffer( str );
  else
  {
    char *hook = processRexxHookNoLog( "LogFilter", str, NULL, NULL,
                                       NULL, NULL, NULL, 1 );

    if( hook==NULL )
      toBuffer( str );
    else if( *hook )
    {
      toBuffer( hook );
      delete hook;
    }
  }
}

void _Optlink logQueueFlushProcess( void *queue )
{
  registerExceptionHandler

  DosSetPriority( PRTYS_THREAD, 0, -miscInfo.priorityDelta, 0 );

  while( mayFlush )
  {
    sleep( 2000 );

    if( logBufferPtr )
      DosWriteQueue( *(unsigned long *)queue, 1, 1, logQueueFlushProcess, 0 );
  }

  deRegisterExceptionHandler
}

int connectNPipe();

static volatile long addTid = 0;

class LogEntry
{
public:
  char *msg;
  time_t stamp;
  unsigned long tid;

  LogEntry( char *_msg ) :
    msg( newStr( _msg ) ),
    stamp( time(0) ),
    tid( getMyTid() )
  {}

  ~LogEntry() { delete msg; }
};

void writeToHandle( int pipeHandle, char *bb, char *str )
{
  write( pipeHandle, bb, strlen(bb) );
  write( pipeHandle, str, strlen(str) );
  write( pipeHandle, "\n", 1 );
}

void _Optlink logQueueProcess( void * )
{
  registerExceptionHandler

  DosSetPriority( PRTYS_THREAD, 0, -miscInfo.priorityDelta, 0 );

  addTid = _beginthread( logQueueFlushProcess, 0, 40960, (void *)&LOG::queue );

  if( addTid==-1 )
    log(  Msg_ErrLogThr, strError() );

  Buffer scrBuf( 256, 128 );

  while( 1 )
  {
    char bb[80];
    char bbb[80];
    LogEntry *entry = NULL;
    REQUESTDATA req;
    ULONG len;
    BYTE prior;
    char pipeOnly( 0 );
    char scrOnly( 0 );

    req.ulData = 0;

    APIRET rc = DosReadQueue( LOG::queue, &req, &len, (PPVOID)&entry,
			      0, DCWW_WAIT, &prior, 0 );

    switch( req.ulData )
    {
    case 255:
      flushBuffer();
      mayFlush = 0;
      deRegisterExceptionHandler
      return;

    case 1:
      flushBuffer();
      break;

    default:
      if( rc || entry==NULL )
	break;

      struct tm *tm = localTime( &entry->stamp );
      char *buf = entry->msg;

      sprintf( bbb, "%02d:%02d ", tm->tm_hour, tm->tm_min );

      if( (miscInfo.state&BinkLog)==BinkLog )
        strftime( bb, sizeof(bb), "* %d %b %H:%M:%S BRK! ", tm );
      else
      {
        sprintf( bb, "%02d/%02d/%02d %02d:%02d:%02d",
  		     tm->tm_mday, tm->tm_mon+1, tm->tm_year%100,
		     tm->tm_hour, tm->tm_min, tm->tm_sec );
        
        if( (miscInfo.state2&TaskInLog)==TaskInLog )
          sprintf( bb+strlen(bb),
                   " [%lu,%02d]",
                   miscInfo.taskNumber,
                   entry->tid );

        strcat( bb, terminalActive ? "  " : "  " );
      }

      pipeOnly = 0;

      if( *buf=='#' )
      {
	buf++;

	if( (miscInfo.state2&FullLog)!=FullLog )
	  pipeOnly = 1;
      }

      if( *buf=='!' )
      {
	buf++;
	scrOnly = 1;
      }
      else
        scrOnly = 0;

      int scrDisable = 0;

      if( noScrList!=NULL )
        scrDisable =  noScrList->found( buf );

      int logDisable = 0;

      if( noLogList!=NULL )
        logDisable =  noLogList->found( buf );

      scrBuf.reset();

      for( int i( 0 ); buf[i]; i++ )
      {
        if( (unsigned char)buf[i]<0x20 )
        {
          scrBuf.putInto( '^' );
          buf[i] = char(buf[i]+'A'-1);
        }

        scrBuf.putInto( buf[i] );
      }

      scrBuf.putInto( 0 );

      if( pipeHandle!=0xffffffffl && connectNPipe() )
        writeToHandle( (int)pipeHandle, bb, scrBuf.str );

      if( daemonMode !=-1 )
        writeToHandle( daemonMode, bbb, scrBuf.str );

      if( !pipeOnly )
      {
	if( !scrOnly && !logDisable )
        {
          char *str( new char [strlen(bb)+strlen(scrBuf.str)+2] );

          strcpy( str, bb );
          strcat( str, scrBuf.str );
          strcat( str, "\n" );
	  toBufferAndFilter( entry->tid, str );
          delete str;
        }

        if( (!terminalActive || 
             (terminalActive && miscInfo.state&TerminalLog)
            ) &&
            !scrDisable )
  	  if( !terminalActive )
	  {
            int prfLen( strlen(bbb) );
            char *strIni = new char [ strlen(scrBuf.str)+prfLen+1 ],
                 *str( strIni ),
                 *prf = new char [ prfLen+1 ];
            int width( wndLog.getWidth() );
            char first(1);

            strcpy( str, bbb );
            strcat( str, scrBuf.str );
            memset( prf, ' ', prfLen );
            prf[ prfLen ] = 0;

            for( int done(0); !done; )
            { 
              char *ptr = str;
              int strLen( strlen(str) );
              int delta( first ? 0 : prfLen );

  	      if( strLen+delta>=width )
              {
                for( int i=strLen-1; i>=0; i-- )
                  if( isspace( str[i] ) && i+delta<width-1 )
                    break;

                if( i<=0 )
	        {
	          str[ width-1-delta ] = 0;
	          str[ width-2-delta ] = 26;
                  done = 1;
	        }
                else
                {
                  str[i] = 0;
                  str += i+1;
                }
              }
              else
                done = 1;

	      wndLog << (char *)"\r\n";

              if( !first )
                wndLog << prf;
              else
                first = 0;

              wndLog << ptr;
            }
            
            delete prf;
            delete strIni;
	  }
	  else
	  {
	    if( terminalWindow.getCurrentX()!=1 )
	      terminalWindow << "\r\n";

	    terminalWindow << (char *)bbb << scrBuf.str << "\r\n";
	    terminalWindow.setPhysicalCursor( 1 );
	  }
      }						//if(!pipeOnly)

      delete entry;
    }
  }
}

void LOG::operator () ( char *fmt )
{
  DosWriteQueue( queue, 0, sizeof( LogEntry* ), new LogEntry( fmt ), 0 );
}

void LOG::flush()
{
  DosWriteQueue( queue, 1, sizeof(*this), this, 0 );
}

LOG::LOG( char * name )
{
  logBuffer = new char [ miscInfo.logBuffer ];
  logBufferPtr = 0;

  makeFilePath( name );
  logName = newStr( name );

  FILE *log_file=_fsopen( name, "a" );

  if( log_file==NULL )
  {
    log( Msg_ErrLogCreate, name, strError() );

    if( systemIsReady )
      dataDeInit();

    exit( 2 );
  }

  fclose( log_file );
  log_file = NULL;
  setExtendedAttribute( name, ".TYPE", Msg_EALOG );

  APIRET rc;

  if( rc=createQueue( &queue, "Log" ), rc )
  {
    log( Msg_ErrLogQueue, rc );
    os2Message( rc );

    if( systemIsReady )
      dataDeInit();

    exit( 2 );
  }

  tid = _beginthread( logQueueProcess, 0, 40960, 0 );

  if( tid==-1 )
  {
    log( Msg_ErrLogMan, strError() );

    if( systemIsReady )
      dataDeInit();

    exit( 2 );
  }

  if( miscInfo.pipe!=NULL )
  {
    rc = DosCreateNPipe( miscInfo.pipe, &pipeHandle,
			 NP_WRITEBEHIND|NP_NOINHERIT|NP_ACCESS_DUPLEX,
			 NP_NOWAIT|NP_TYPE_BYTE|NP_READMODE_BYTE|1,
			 4096, 4096, 0 );

    if( rc )
    {
      pipeHandle = 0xffffffffl;
      log( Msg_ErrPipeFault, miscInfo.pipe, rc );
      os2Message( rc );
    }
    else
      if( detachedSession )
        dup2( int( pipeHandle ), 0 );
  }

 this->empty();

 char bb[80];

 sprintf( bb, Msg_LogReady, M_NAME, M_VERSION, M_NUMBER, getpid() );
 (*this)( bb );

 while( !addTid )
   sleep( 10 );

 sprintf( bb, Msg_ThrLogMan, tid, addTid );
 (*this)( bb );
}

LOG::~LOG()
{
  flush();

  DosWriteQueue( queue, 255, sizeof(*this), this, 0 );

  DosWaitThread( (PTID)&tid, DCWW_WAIT );
  deleteQueue( &queue );

  if( pipeHandle!=0xffffffffl )
    DosClose( pipeHandle );

  delete logName;
  delete logBuffer;
}

void LOG::empty()
{
  (*this)( " " );
}

