#define INCL_DOSSEMAPHORES
#define INCL_DOSPROCESS
#define INCL_DOSERRORS
#define INCL_WINSHELLDATA
#define INCL_DOSEXCEPTIONS
#include <os2.h>

#include <stdlib.h>
#include <io.h>
#include <ctype.h>
#include <fstream.h>
#include <time.h>
#include <process.h>
#include <iostream.h>
#include <new.h>
#include <share.h>
#include <fcntl.h>

#include "except.h"

#include "language.h"

#include "dinamic.h"
#include "misc.h"
#include "queue.h"
#include "translat.h"
#include "32base.h"
#include "key.h"
#include "timer.h"
#include "fidoaddr.h"
#include "pair.h"
#include "session.h"
#include "config.h"
#include "rexx.h"
#include "rexxhook.h"
#include "freq.h"

static Timer circleTimer( 0 );
static int needAnswer = 0,
	   theEnd = 0;
static int quickDial = 0;
static int queueDelta = 0;

void newCircle()
{
  circleTimer.startTimer( quickDial ? long(quickDial)*1000l
                                    : makeCircle()*1000 );
}

void finita()
{
  log( Msg_ScrQueKbdAX );
  theEnd = 1;
}

void manualAnsw()
{
  log( Msg_ScrQueKbdAA );
  needAnswer = 3;
}

void _Optlink viewQueue( void *args )
{
  registerExceptionHandler

  if( (unsigned long)args )
    keyBoardLock();

  queue->view();

  if( (unsigned long)args )
    keyBoardUnLock();

  deRegisterExceptionHandler
}

void doView()
{
  if( !detachedSession )
    _beginthread( viewQueue, 0, 40960, (void *)1 );
}

void doView1()
{
  if( !detachedSession )
    viewQueue( (void *)0 );
}

void forceCall()
{
  circleTimer.endTimer();
}

void termCall()
{
  void terminalEmulator();

  terminalEmulator();

  showAction( Msg_ScrQueKbdAI );
  comPort->hangUp();
  comPort->init();

  Queue::needForRescan();
  newCircle();
}

void initMdm()
{
  log( Msg_ScrQueKbdAI );
  comPort->init();
}

extern ULONG iniFileHandle;

void toModem()
{
  char stringIntoModem[65]( "AT|" );

  +screenCursor;

  if( !noPMFeatures && !detachedSession )
    BrakePrfQueryProfileString( iniFileHandle, pszAppBrake,
 			        (PSZ)"ModemString",
			        NULL,
			        (PVOID)stringIntoModem,
			        (ULONG)sizeof( stringIntoModem ) );

  if( !stringEnter( sizeof(stringIntoModem), stringIntoModem,
		    Msg_ScrQueKbdAMTtl ) )
  {
    if( !noPMFeatures && !detachedSession )
      BrakePrfWriteProfileString( iniFileHandle, pszAppBrake,
                                  (PSZ)"ModemString",
                                  (PSZ)stringIntoModem );
    log( Msg_ScrQueKbdAM, stringIntoModem );
    comPort->putOffLine( stringIntoModem );
  }

  -screenCursor;
}

void doHelp()
{
  help( Msg_ScrMailerHelp, 1 );
}

void doPageUp()
{
  queueDelta = -2;
}

void doPageDn()
{
  queueDelta = 2;
}

void doUpNext()
{
  queueDelta = -1;
}

void doDnNext()
{
  queueDelta = 1;
}

void doHomeNext()
{
  queueDelta = -queue->getCurrentPointer();

  if( !queueDelta )
    queueDelta = -1;
}

void doEndNext()
{
  queueDelta = queue->GetItemsInContainer()-queue->getCurrentPointer();
}

void doQuickDial()
{
  quickDial = quickDial ? 0 : 1;
  newCircle();
}

void localPoll()
{
  mailerPoll();
  Queue::needForRescan();
}

void localGet()
{
  mailerGet();
  Queue::needForRescan();
}

void localDrop()
{
  mailerDrop();
  Queue::needForRescan();
}

void localSend()
{
  mailerSend();
  Queue::needForRescan();
}

void localChange()
{
  mailerChange();
  Queue::needForRescan();
}

void doRexx()
{
  char string[65]( ".Brk" );

  +screenCursor;

  if( !noPMFeatures && !detachedSession )
    BrakePrfQueryProfileString( iniFileHandle, pszAppBrake,
	 		        (PSZ)"MailerRexxProcedure",
			        NULL,
			        (PVOID)string,
			        (ULONG)sizeof( string ) );

  int i = stringEnter( sizeof(string), string, Msg_ScrQueKbdARTtl );

  -screenCursor;
  capitalize( string );

  if( !i )
  {
    if( !noPMFeatures && !detachedSession )
      BrakePrfWriteProfileString( iniFileHandle, pszAppBrake,
			          (PSZ)"MailerRexxProcedure",
			          (PSZ)string );

    showAction( fileName( string ) );

    FIDOAddress *fAddr = (FIDOAddress *) queue->getCurrent();

    if( fAddr==NULL )
      fAddr = defaultAddress;

    doParamsRexx( *fAddr, 0, NULL, string, -1, NULL );
  }
}

int callThisNode( NodeConfig& cfg )
{
  NodeConfig *testCfg( systemList->Find( cfg ) ),
             &realCfg = testCfg==NULL ? cfg : *testCfg;

  QueueEntry test( realCfg ),
             *found( queue->Find( test ) ),
             &qentry = found==NULL ? test : *found;

  char addr[65];
  qentry.toString( addr, sizeof(addr) );

  if( qentry.getPhone()==NULL )
  {
    log( Msg_InfoNoPhone, addr );
    return 0;
  }

  log( Msg_ScrQueKbdAC, addr );

  int rc=0;

  for( int done=0; !done; )
  {
    rc = callNode( qentry, 0 );

    switch( rc )
    {
    case 255:
      log( Msg_InfoMdmErrCall );
      OSReturnCode = 1;

    // fall

    case 0:
    case 254:
    case 252:
    case 253:
    case 3:
    case 8:
      done = 1;
      break;
    }
  }

  return rc;
}

void manualCall()
{
  if( detachedSession )
    return;

  if( !comPort->getPort() )
    return;

  char addr[65]( "" );

  if( !noPMFeatures && !detachedSession )
    BrakePrfQueryProfileString( iniFileHandle, pszAppBrake,
			        (PSZ)"ManualCall",
			        NULL,
			        (PVOID)addr,
			        (ULONG)sizeof( addr ) );

  +screenCursor;

  if( stringEnter( sizeof(addr), addr, Msg_ScrQueKbdACTtl ) )
  {
    -screenCursor;
    return;
  }

  -screenCursor;

  NodeConfig cfg( addr );
  void  illegalAddress( char *str );

  if( !cfg.isValid() )
  {
    illegalAddress( addr );
    return;
  }

  if( !noPMFeatures && !detachedSession )
    BrakePrfWriteProfileString( iniFileHandle, pszAppBrake,
                                (PSZ)"ManualCall",
		                (PSZ)addr );

  callThisNode( cfg );

  queue->needForRescan();
  newCircle();
}

void dropFreq()
{
  if( detachedSession )
    return;

  char addr[65]( "" );

  if( !noPMFeatures && !detachedSession )
    BrakePrfQueryProfileString( iniFileHandle, pszAppBrake,
			        (PSZ)"DropFREQ",
			        NULL,
			        (PVOID)addr,
			        (ULONG)sizeof( addr ) );

  +screenCursor;

  if( stringEnter( sizeof(addr), addr, Msg_ScrQueKbdANTtl ) )
  {
    -screenCursor;
    return;
  }

  -screenCursor;

  FIDOAddress cfg( addr );
  void  illegalAddress( char *str );

  if( !cfg.isValid() )
  {
    illegalAddress( addr );
    return;
  }

  if( !noPMFeatures && !detachedSession )
    BrakePrfWriteProfileString( iniFileHandle, pszAppBrake,
                             (PSZ)"DropFREQ",
		             (PSZ)addr );

  cfg.toString( addr, sizeof(addr) );
  log( Msg_ScrQueKbdAN, addr );

  cfg.dropFreq();
  Queue::needForRescan();
}

int _Optlink nodeCmp( const void *n1, const void *n2 )
{
  if( ((const NodeFreqInfo *)n1)->zone > ((const NodeFreqInfo *)n2)->zone )
    return 1;

  if( ((const NodeFreqInfo *)n1)->zone < ((const NodeFreqInfo *)n2)->zone )
    return -1;

  if( ((const NodeFreqInfo *)n1)->net > ((const NodeFreqInfo *)n2)->net )
    return 1;

  if( ((const NodeFreqInfo *)n1)->net < ((const NodeFreqInfo *)n2)->net )
    return -1;

  if( ((const NodeFreqInfo *)n1)->node > ((const NodeFreqInfo *)n2)->node )
    return 1;

  if( ((const NodeFreqInfo *)n1)->node < ((const NodeFreqInfo *)n2)->node )
    return -1;

  if( ((const NodeFreqInfo *)n1)->point > ((const NodeFreqInfo *)n2)->point )
    return 1;

  if( ((const NodeFreqInfo *)n1)->point < ((const NodeFreqInfo *)n2)->point )
    return -1;

  return 0;
}

void _Optlink viewFreqBaseThr( void *args )
{
  registerExceptionHandler

  const width = 80,
        height = Z_ScreenHeight-2;

  PICKLIST pick =
  {
    (Z_ScreenWidth-width)/2,
    (Z_ScreenHeight-height)/2,
    height,
    width,
    width-2,
    PLF_VERTSCROLL,
    popUpColor.fore, popUpColor.back,
    stdColor.fore, stdColor.back,
    BS_SINGLELINE, popUpColor.fore, popUpColor.back,
    ((((popUpColor.back<<4)|popUpColor.fore)<<8)|0x20)&0xffff,
    DS_NONE, DARKGREY,
    Msg_ScrViewFreq
  };

  FILE *freq = _fsopen( miscInfo.freqBase, "rb" );

  if( freq==NULL )
  {
    log( Msg_ErrUnableOpen, miscInfo.freqBase, strError() );
    deRegisterExceptionHandler
    return;
  }

  int count = int( filelength(fileno(freq))/sizeof(NodeFreqInfo) );
  NodeFreqInfo *info = new NodeFreqInfo [ count ];
  int needread = count*sizeof(NodeFreqInfo),
      readed = fread( info, 1, needread, freq );

  if( readed!=needread )
  {
    log( Msg_ErrUnableRead, miscInfo.freqBase, strError() );
    fclose( freq );
    delete info;
    deRegisterExceptionHandler
    return;
  }

  fclose( freq );
  qsort( info, count, sizeof(NodeFreqInfo), nodeCmp );

  Z_AddPickListItem( &pick, Msg_ScrViewFreqT1 );
  Z_AddPickListItem( &pick, Msg_ScrViewFreqT2 );

  for( int i=0; i<count; i++ )
  {
    char buf[90],
         b[21];
    FIDOAddress adr( info[i].zone, info[i].net, info[i].node, info[i].point );

    adr.toString( b, sizeof(b) );

    char *bl = strchr( b, '@' );

    if( bl!=NULL )
      *bl = 0;

    sprintf( buf, "%-21s %02lu/%02lu/%02lu %6lu %9lu %6lu %9lu %6lu %9lu",
             b,
             info[i].lastDay & 0xff,
             (info[i].lastDay >> 8 ) & 0xff,
             (info[i].lastDay >> 16 ) & 0xffff,
             info[i].monTime, info[i].monSize,
             info[i].weekTime, info[i].weekSize,
             info[i].dayTime, info[i].daySize );

    Z_AddPickListItem( &pick, buf );
  }

  delete info;

  if( (unsigned long)args )
    keyBoardLock();

  Z_HandlePickList( &pick );

  if( (unsigned long)args )
    keyBoardUnLock();

  Z_PurgePickList( &pick );
  Z_DestroyPickList( &pick );

  deRegisterExceptionHandler
}

void viewFreqBase()
{
  if( miscInfo.freqBase==NULL )
    return;

  _beginthread( viewFreqBaseThr, 0, 40960, (void *)1 );
}

void viewFreqBase1()
{
  if( miscInfo.freqBase==NULL )
    return;

  viewFreqBaseThr( (void *)0 );
}

void _Optlink viewEntryThr( void *cur )
{
  registerExceptionHandler

  keyBoardLock();
  ((QueueEntry *)cur)->show();
  keyBoardUnLock();
  delete (QueueEntry *)cur;

  deRegisterExceptionHandler
}

void viewEntry()
{
  queue->Lock();

  QueueEntry *cur = queue->getCurrent();

  if( cur==NULL )
  {
    queue->UnLock();
    return;
  }

  _beginthread( viewEntryThr, 0, 40960, (void *)new QueueEntry( *cur ) );
  queue->UnLock();
}

void _Optlink infoEntryThr( void *cur )
{
  registerExceptionHandler

  keyBoardLock();
  ((QueueEntry *)cur)->info();
  keyBoardUnLock();
  delete (QueueEntry *)cur;

  deRegisterExceptionHandler
}

void infoEntry()
{
  queue->Lock();

  QueueEntry *cur = queue->getCurrent();

  if( cur==NULL )
  {
    queue->UnLock();
    return;
  }

  _beginthread( infoEntryThr, 0, 40960, (void *)new QueueEntry( *cur ) );
  queue->UnLock();
}

static char answerCurrentlyAvailable = 1;
char *answerDisabled = NULL;

void toggleAnswer()
{
  if( comPort->getPort() )
  {
    answerCurrentlyAvailable ^= 1;
    answerDisabled = answerCurrentlyAvailable ? NULL : Msg_ScrNoAnswer;
  }
}

void _Optlink processMenuThr( void * )
{
  registerExceptionHandler

  Color buttonHig( colorSet.popUpBold );

  MENU_ITEM actionMenuItems[] =
  {
    "Execute REXX  ", MIF_CALLFUNC,    'E', 0,  1, 1,  doRexx,        "",
    "View FREQ base", MIF_CALLFUNC,    'V', 0,  1, 2,  viewFreqBase1, "",
    "Debug DUMP    ", MIF_CALLFUNC,    'D', 0,  1, 3,  debugDump,     "",
    "Toggle answer ", MIF_CALLFUNC,    'T', 0,  1, 4,  toggleAnswer,  "",
    NULL, 0, 0, 0, 0, 0, NULL, ""
  };

  MENU actionMenu =
  {
    2, 1, 6, 18,
    16,
    MMF_VERTMENU,
    popUpColor.fore, popUpColor.back,
    buttonHig.fore,  buttonHig.back,
    buttonHig.fore,  buttonHig.back,
    popUpColor.fore, popUpColor.back,
    1, 1, 1,
    BS_SINGLELINE, popUpColor.fore, popUpColor.back,
    DS_LOWERRIGHT, 0x07,
    "",
    actionMenuItems
  };

  MENU_ITEM nodeMenuItems[] =
  {
    "Poll node     ", MIF_CALLFUNC, 'P', 0,  1, 1, localPoll,   "",
    "Send file(s)  ", MIF_CALLFUNC, 'S', 0,  1, 2, localSend,   "",
    "Get File      ", MIF_CALLFUNC, 'G', 0,  1, 3, localGet,    "",
    "Change flavour", MIF_CALLFUNC, 'C', 0,  1, 4, localChange, "",
    "Drop poll     ", MIF_CALLFUNC, 'D', 0,  1, 5, localDrop,   "",
    "Drop FREQ     ", MIF_CALLFUNC, 'F', 5,  1, 6, dropFreq,    "",
    NULL, 0, 0, 0, 0, 0, NULL, ""
  };

  MENU nodeMenu =
  {
    12, 1, 8, 18,
    16,
    MMF_VERTMENU,
    popUpColor.fore, popUpColor.back,
    buttonHig.fore,  buttonHig.back,
    buttonHig.fore,  buttonHig.back,
    popUpColor.fore, popUpColor.back,
    1, 1, 1,
    BS_SINGLELINE, popUpColor.fore, popUpColor.back,
    DS_LOWERRIGHT, 0x07,
    "",
    nodeMenuItems
  };

  MENU_ITEM queueMenuItems[] =
  {
    "Rescan outbound", MIF_CALLFUNC,    'R', 0,  1, 1, Queue::needForRescan,  "",
    "View queue     ", MIF_CALLFUNC,    'V', 0,  1, 2, doView1,               "",
    NULL, 0, 0, 0, 0, 0, NULL, ""
  };

  MENU queueMenu =
  {
    22, 1, 4, 19,
    17,
    MMF_VERTMENU,
    popUpColor.fore, popUpColor.back,
    buttonHig.fore,  buttonHig.back,
    buttonHig.fore,  buttonHig.back,
    popUpColor.fore, popUpColor.back,
    1, 1, 1,
    BS_SINGLELINE, popUpColor.fore, popUpColor.back,
    DS_LOWERRIGHT, 0x07,
    "",
    queueMenuItems
  };

  MENU_ITEM modemMenuItems[] =
  {
    "Initialize ", MIF_CALLFUNC,    'I', 0,  1, 1, initMdm,  "",
    "Send string", MIF_CALLFUNC,    'S', 0,  1, 2, toModem,  "",
    NULL, 0, 0, 0, 0, 0, NULL, ""
  };

  MENU modemMenu =
  {
    32, 1, 4, 15,
    13,
    MMF_VERTMENU,
    popUpColor.fore, popUpColor.back,
    buttonHig.fore,  buttonHig.back,
    buttonHig.fore,  buttonHig.back,
    popUpColor.fore, popUpColor.back,
    1, 1, 1,
    BS_SINGLELINE, popUpColor.fore, popUpColor.back,
    DS_LOWERRIGHT, 0x07,
    "",
    modemMenuItems
  };

  MENU_ITEM mainMenuItems[] =
  {
    " Action ", MIF_CALLMENU, 'A', 1, 2,                0, &actionMenu, "",
    " Node   ", MIF_CALLMENU, 'N', 1, 12,               0, &nodeMenu,  "",
    " Queue  ", MIF_CALLMENU, 'Q', 1, 22,               0, &queueMenu,  "",
    " Modem  ", MIF_CALLMENU, 'M', 1, 32,               0, &modemMenu,  "",
//  " Quit!  ", MIF_CALLFUNC, 'Q', 1, Z_ScreenWidth-10, 0, (void*)finita,     "",
    NULL, 0, 0, 0, 0, 0, NULL, ""
  };

  MENU mainMenu =
  {
    0, 0, 1, Z_ScreenWidth,
    8,
    MMF_HORIZMENU,
    popUpColor.fore, popUpColor.back,
    buttonHig.fore, buttonHig.back,
    buttonHig.fore, buttonHig.back,
    popUpColor.fore, popUpColor.back,
    1, 1, 1,
    BS_NOBORDER, popUpColor.fore, popUpColor.back,
    DS_NONE, 0x07,
    "",
    mainMenuItems
  };

  keyBoardLock();
  Z_HandleMenu( &mainMenu );
  keyBoardUnLock();

  deRegisterExceptionHandler
}

void processMenu()
{
  _beginthread( processMenuThr, 0, 40960, 0 );
}

void unixTime()
{
  log( "Current UNIX time: %d", time(0) );
}

static Pair remapTable[] =
{
  Pair( 'T'-'A'+1, unixTime ),       // Ctrl-T
  Pair( key_Ctrl_D, debugDump ),     // Ctrl-D
  Pair( key_Ctrl_Q, doQuickDial ),   // Ctrl-Q

  Pair( key_PgUp,  doPageUp ),       // PgUp
  Pair( key_PgDn,  doPageDn ),       // PgDn
  Pair( key_UpArr, doUpNext ),       // ^
  Pair( key_DnArr, doDnNext ),       // v
  Pair( key_Home,  doHomeNext ),     // Home
  Pair( key_End,   doEndNext ),      // Home

  Pair( key_Ins, doHelp  ),          // Ins
  Pair( key_F1, doHelp  ) ,          // F1
  Pair( key_Alt_M, toModem ),        // Alt-M
  Pair( key_Alt_T, termCall ),       // Alt-T
  Pair( key_Alt_I, initMdm ),        // Alt-I
  Pair( 0x20, forceCall ),                      // space
  Pair( key_Alt_Q, Queue::needForRescan ),      // Alt-Q
  Pair( key_Alt_D, localDrop ) ,                // Alt-D
  Pair( key_Alt_V, doView ),         // Alt-V
  Pair( key_Alt_A, manualAnsw ),     // Alt-A
  Pair( key_Alt_X, finita ),         // Alt-X
  Pair( key_Alt_P, localPoll ),      // Alt-P
  Pair( key_Alt_S, localSend ),      // Alt-S
  Pair( key_Alt_G, localGet ),       // Alt-G
  Pair( key_Alt_R, doRexx ),         // Alt-R
  Pair( key_Alt_C, manualCall ),     // Alt-C
  Pair( key_Alt_N, dropFreq ),       // Alt-N
  Pair( key_Alt_E, viewFreqBase ),   // Alt-E
  Pair( '?', viewEntry ),            // ?
  Pair( key_Alt_O, toggleAnswer ),   // Alt-O
  Pair( key_Alt_Enter, infoEntry ),  // Alt-Enter
  Pair( key_Ctrl_C, localChange ),   // Ctrl-C
  Pair( key_F10, processMenu )	     //F10
};

// ======================== End Keyboard function ==========================

static long rescanServerTID = -1;
static volatile char rescanServerNeedAbort = 0;
volatile char nowInScan = 0;

void _Optlink rescanServer( void * )
{
  registerExceptionHandler

  DosSetPriority( PRTYS_THREAD, miscInfo.reScanPriority,
		  miscInfo.reScanPriorityDelta, 0 );

  if( miscInfo.rescan<1 )
    miscInfo.rescan = 1;

  if( miscInfo.rescanPeriod<1 )
    miscInfo.rescanPeriod = 1;

  while( 1 )
  {
    if( rescanServerNeedAbort )
    {
      deRegisterExceptionHandler
      return;
    }

    long cur = getTimeMark();

    lockMutexSemaphore( Queue::globalSemaphore );

    if( cur>=Queue::sharedQueueHeader->lastScaned+miscInfo.rescan )
      Queue::sharedQueueHeader->needScan = 1;

    if( Queue::sharedQueueHeader->scanStart+60*5<getTimeMark() &&
        Queue::sharedQueueHeader->inScan )
    {
      log( Msg_BUGQueueDeadLock );
      Queue::sharedQueueHeader->inScan = 0;
    }

    if( Queue::sharedQueueHeader->needScan &&
        !Queue::sharedQueueHeader->inScan )
    {
      nowInScan = 1;
      Queue::sharedQueueHeader->needScan = 0;
      Queue::sharedQueueHeader->inScan = 1;
      Queue::sharedQueueHeader->scanStart = getTimeMark();
      unlockMutexSemaphore( Queue::globalSemaphore );
      queue->rescanQueue();
      lockMutexSemaphore( Queue::globalSemaphore );
      Queue::sharedQueueHeader->inScan = 0;
      Queue::sharedQueueHeader->scanStart = 0;
      unlockMutexSemaphore( Queue::globalSemaphore );
      nowInScan = 0;
    }
    else
      unlockMutexSemaphore( Queue::globalSemaphore );

    sleep( miscInfo.rescanPeriod );
  }
}

static long autoInit;
static Timer initTimer( 0 );

void abortRescanServer()
{
  rescanServerNeedAbort = 1;
  Queue::needForRescan();
  DosWaitThread( (unsigned long *)&rescanServerTID, DCWW_WAIT );
}

volatile char needInit = 0;
static int RC( 0 );

int initModem() 
{
  needInit = 0;
  showAction( Msg_ScrQueKbdAI );

  if( comPort->init() )
  {
    log( Msg_ErrUnableInitMdm );
    RC = 1;
    return 1;
  }

  if( autoInit )
    initTimer.startTimer( autoInit );

  return 0;
}

void mailerWait( int /* rest */ )
{
  setTitleMode( NULL );
  changeTitle( Msg_TtlWait );

/*
  char *msg( (char*)_alloca( strlen(Msg_TtlWait)+32 ) );

  if( rest )
    sprintf( msg, "%s: %d", Msg_TtlWait, rest );
  else
    strcpy( msg, Msg_TtlWait );

  changeTitle( msg );
*/

}

int mailer()
{
  time_t bbbMailerStart( getMSec() );

  mailerWait( 0 );

  autoInit = miscInfo.autoInit*1000;

  static char flavourChar[]( "ICFNH" );

  if( !onlineAnswer )
  {
    rescanServerTID = _beginthread( rescanServer, 0, 40960, 0 );

    if( rescanServerTID==-1 )
    {
      log( Msg_ErrRescanServ, strError() );
      return 2;
    }

    log( Msg_ThrRescanServ, rescanServerTID );
  }

  queue->getLast();

  if( startUpCount && startUpList!=NULL )
  {
    log( Msg_InfoStartUp );

    for( int i(0); i<startUpCount; i++ ) 		// Define StartUp
      doEventKey( startUpList[i], -1 );
  }

  long autoInit( miscInfo.autoInit*1000 );

  Timer configTimer( miscInfo.configCheck );

  initTimer.startTimer( autoInit );

  int tryExitAsync=0;
  RC = 0;

  if( onlineAnswer )
  {
    log( Msg_InfoAnswerBatchSession );
    comPort->setConnectInfo( onlineAnswerBaud );

    if( callOneNode==NULL )
    {
      if( answer( 2, 0 )==255 )
      {
ModemErrorFound:
        log( Msg_InfoMdmErrAnswer );
        RC = 1;
      }
    }
    else
    {
      NodeConfig cfg( callOneNode );

      if( callThisNode( cfg )==155 )
        goto ModemErrorFound;
    }

    goto return_to_caller;
  }

  void startSchedule();
  startSchedule();

  needAnswer = 0;
  theEnd = 0;
  Queue::needForRescan();

  if( !comPort->isOnline() )
    needInit = 1;
  else
    needInit = 0;

  dataInitMsg( getMSec()-bbbMailerStart, Msg_InitStat6 );

  while( 1 )
  {
    if( OSReturnCode )
      goto Done;

    if( todayMessage( 4, 30000, 500 )==2 )
      log( Msg_Joke1, flavourChar[ random(sizeof(flavourChar)-1) ],
	   random(6)+1, random(9999)+1, random(999)+1 );

    newCircle();

    int oldRest( 0 );

    if( !queue->GetItemsInContainer() && miscInfo.emptyExit )
      if( ++tryExitAsync==2 )
      {
        log( Msg_InfoEmtyQueue );
        RC = miscInfo.emptyExit;
        goto return_to_caller;
      }

    int connectString = 0;

    while( !circleTimer.isTimeOut() && !needAnswer && !connectString )
    {
      if( OSReturnCode )
        goto Done;

      queue->read();

      if( configTimer.isTimeOut() )
      {
        showAction( Msg_ScrCheckCtl );

        if( cfgFileList.needReRead() )
        {
  	  RC = 3;
          goto return_to_caller;
        }

        configTimer.startTimer( miscInfo.configCheck );
      }

      if( autoInit && initTimer.isTimeOut() )
        needInit = 1;

      if( needInit && !comPort->isLocked() )
        if( initModem() )
          goto return_to_caller;

      int newRest( circleTimer.secsLeft() );

      if( newRest!=oldRest )
      {
        char msg[64];

	oldRest = newRest;
	sprintf( msg, Msg_ScrIdleTime, newRest );
	showAction( msg );
        mailerWait( newRest );
      }

      if( !comPort->getPort() || comPort->isLocked() )
        checkPoint();

      if( !keyBoardIsLocked() && keyPressed() )
      {
	int ch( keyGet() );

        for( int i(0); i<keyCount; i++ )
          if( keyList[i].done==short(ch) )
          {
            ch = 0;
            RC = doEventKey( keyList[i], -1 );

            if( RC )
              goto return_to_caller;

	    break;
	  }

        if( ch )
	  for( int i(0); i<sizeof(remapTable)/sizeof(remapTable[0]); i++ )
	    if( remapTable[i]==ch )
	    {
	      remapTable[i].process();
	      ch = 0;
	      break;
	    }

        if( ch )
          execSchedule( melodies.alarm );

	if( theEnd )
          goto return_to_caller;        // exit to DOS | OS/2

        if( queueDelta )
        {
          queue->scroll( queueDelta );
          queueDelta = 0;
        }

        mailerWait( circleTimer.secsLeft() );
      }

      if( !comPort->isLocked() && comPort->isOnline() )   //check
      {
	log( Msg_InfoDCDSession );
	needAnswer = 2;
	comPort->byDCD();
      }
      else if( !comPort->isLocked() && comPort->haveInput() )
      {
        Buffer *fromModem = comPort->getOffLine( 5000 );

        if( begin( fromModem->str, modemString.connect ) )
        {
          char *up( newStr( fromModem->str ) );

          if( comPort->waitConnect( 0, 
                                    int( comPort->getWaitAnswerTime()*1000l ),
                                    up )==2 )
            connectString = 1;
            
          delete up;
        }
        else if( (miscInfo.state3&HardRing)!=HardRing && 
                 !stricmp( fromModem->str, modemString.ring ) )
          needAnswer = 1;

        if( fromModem->length()>1 && !needAnswer && (miscInfo.state&ModemReport) )
          log( Msg_ModemReport, fromModem->str );

        delete fromModem;
      }
      else if( !comPort->isLocked() && (miscInfo.state3&HardRing)==HardRing &&
               comPort->isRinging() )
        needAnswer = 1;

      static char *oldAnswerDisabled = NULL;

      if( answerDisabled!=oldAnswerDisabled )
      {
        queue->queueName( Msg_WndTitleQue, answerDisabled );
        oldAnswerDisabled = answerDisabled;
      }
    }                      // while circle && !needAnswer && !connectString

    int availAnswer = 1;

    if( needAnswer || connectString )
    {
      comPort->ringSignal();

      if( !answerCurrentlyAvailable )
      {
        if( connectString )
        {
          showAction( Msg_InfoActionHang );
          comPort->hangUp();
        }

        goto NoAnswerPlease;
      }

      if( needAnswer==1 || needAnswer==4 )
      {
	int avail = processLogicalRexxHook( "AnswerAvailable" );

	if( avail==-1 )
	  if( miscInfo.answerTime!=NULL &&
	      !miscInfo.answerTime->isWorked( time( NULL ) ) )
	    avail = 0;

	availAnswer = avail;
      }

      if( availAnswer )
      {
        answerDisabled = NULL;

	if( answer( needAnswer, connectString )==255 )
	{
	  log( Msg_InfoMdmErrAnswer );

          if( processBadModem() )
          {
            if( initModem() )
              goto return_to_caller;
          }
          else
          {
            RC = 1;
            goto return_to_caller;
          }

          if( OSReturnCode )
            goto Done;

	}
	else
        {
          if( OSReturnCode )
            goto Done;

          if( initModem() )
            goto return_to_caller;

          queue->read();
          queue->checkFlavour();
          queue->draw();

          if( miscInfo.state&CallAfterAnswer && circleTimer.isTimeOut() )
          {
            needAnswer = 0;
            goto try_to_call;
          }
        }
      }
      else
      {
NoAnswerPlease:
        answerDisabled = Msg_ScrNoAnswer;
	log( Msg_SesNoAnswer );

        Timer tim( 5000 );

        while( (miscInfo.state3&HardRing)==HardRing && comPort->isRinging() &&
          !tim.isTimeOut() )
          checkPoint();
      }

      needAnswer = 0;
    }
    else
try_to_call:
      if( queue->getCurrentPointer()!=-1 && comPort->getPort() )
      {
        int callAvail = processLogicalRexxHook( "CallAvailable" );

	if( callAvail==-1 )
	  if( miscInfo.callTime!=NULL && !miscInfo.callTime->isWorked( time( NULL ) ) )
	    callAvail = 0;

	if( callAvail )
	{
	  queue->Lock();
	  queue->checkFlavour();

          int queuePriority = queue->getPriority();
          char foundCallable(0);

          queue->queueName( Msg_WndTitleQue, answerDisabled );

	  for( long j(0); j<queue->GetItemsInContainer(); j++ )
          {
            QueueEntry *cur = queue->getCurrent();

            if( cur->getPriority()>=queuePriority && cur->isCallable() )
            {
              foundCallable = 1;
              break;
            }

	    queue->nextEntry();
          }

	  queue->draw();

	  if( foundCallable )
	  {
	    QueueEntry *cur = new QueueEntry( *queue->getCurrent() );
            char thisNodeEmpty = 0;

	    queue->UnLock();

            if( (miscInfo.state&ScanBefore)==ScanBefore )
            {
              cur->rescan();

              if( cur->isEmpty() || !cur->isCallable( 1 ) )
              {
NoCallHoldNode:
                char bbb[65];
                
                cur->toString( bbb, sizeof(bbb)-1 );

                if( cur->isEmpty() )
                  log( Msg_InfoAlreadySent, bbb );
                else
                  log( Msg_InfoNotCallable, bbb );
                  
                queue->updateEntry( *cur );
                queue->needForRescan();
                thisNodeEmpty = 1;
              }
            }
            else
              if( cur->checkHoldState() )
                goto NoCallHoldNode;

            if( !thisNodeEmpty )
              switch( callNode( *cur, 1 ) )
              {
              case 255:
                log( Msg_InfoMdmErrCall );

                if( processBadModem() )
                {
                  if( initModem() )
                    goto return_to_caller;
                }
                else
                {
                  RC = 1;
                  goto return_to_caller;
                }

                if( OSReturnCode )
                  goto Done;

                break;

              case 253:
                if( miscInfo.state&Collision )
                {
                  showAction( Msg_SesCollision );
                  needAnswer = 4;
                }

                if( OSReturnCode )
                  goto Done;

                break;

              default:
                queue->read();
                queue->checkFlavour();
                queue->nextEntry();
                queue->draw();

                if( OSReturnCode )
                  goto Done;
              }

            if( cur->tryExceeded() )
            {
              char buf[65];

              log( Msg_InfoUnDial, cur->toString( buf, sizeof(buf) ) );
            }

	    delete cur;
	  }
	  else				// not found callable
	  {
	    quickDial = 0;
	    queue->UnLock();
	  }
	}
	else				// callAvail
	{
	  log( Msg_SesNoCall );
	  quickDial = 0;
	}
      }                 		// queue empty
      else
	quickDial = 0;
  }

return_to_caller:
  if( exitCount && exitList!=NULL )
  {
    log( Msg_InfoExit );

    for( int i(0); i<exitCount; i++ ) 		// Define StartUp
      doEventKey( exitList[i], -1 );
  }

  return RC;

Done:
  RC = OSReturnCode;
  goto return_to_caller;
}

