#if !defined(__MY_MISC_H)

#define __MY_MISC_H

enum Boolean { True=1, False=0 };

#include "sched.h"
#include "fidoaddr.h"
#include "cursor.h"
#include "window.h"
#include "modem.h"
#include "worktime.h"
#include "pair.h"
#include "log.h"
#include "semaphor.h"
#include "domain.h"
#include "ea.h"
#include "sleeper.h"

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <iostream.h>
#include <share.h>
#include <fcntl.h>
#include <ctype.h>
#include <time.h>

extern const int M_CODE;
extern char *M_COPYRIGHT;
extern char *M_NAME;
extern char *M_NUMBER;
extern char *M_VERSION;

extern char *hexString;
extern char *hexDigit;

extern Cursor *screenCursorPtr;
#define screenCursor (*screenCursorPtr)

extern Color *stdColorPtr,
	     *brdColorPtr,
	     *scrColorPtr,
	     *popUpColorPtr,
	     *terColorPtr;

#define stdColor   (*stdColorPtr)
#define brdColor   (*brdColorPtr)
#define scrColor   (*scrColorPtr)
#define popUpColor (*popUpColorPtr)
#define terColor   (*terColorPtr)

extern LOG *logPtr;

extern Window *wndLogPtr,
	      *wndStatPtr,
	      *wndQueuePtr,
	      *masterScreenPtr,
	      *terminalWindowPtr,
              *statusLineWindowPtr,
              *headLineWindowPtr;

#define wndLog         (*wndLogPtr)
#define wndStat        (*wndStatPtr)
#define wndQueue       (*wndQueuePtr)
#define masterScreen   (*masterScreenPtr)
#define terminalWindow (*terminalWindowPtr)
#define statusLineWindow (*statusLineWindowPtr)
#define headLineWindow (*headLineWindowPtr)

extern volatile char terminalActive,
                     terminalLocalEcho,
                     terminalNeedTranslate,
                     terminalStopOutput;
extern volatile int terminalCapture;
extern short terminalInXLAT[256];
extern short terminalOutXLAT[256];

extern Pair *ansiTab;
extern int ansiHow;
extern Pair ansiTabInternal[];
extern int ansiHowInternal;

extern volatile char systemIsReady;

class SystemInfo
{
public:
  char *name, *sysopName,
       *place, *phone, *baud,
       *flags;

  FIDOAddressList *addr;
  SystemInfo();
  ~SystemInfo();
};

enum SysState
{
  ZDebug =          0x00000001,
  EMSIDebug =       0x00000002,
  ProcessOnce =     0x00000004,
  ModemWindow =     0x00000008,
  Collision =       0x00000010,
  SplitDelete =     0x00000020,
  PasswordAbort =   0x00000040,
  BinkLog =         0x00000080,
  AutoBBS =         0x00000100,
  Full5D =          0x00000200,
  ScanBefore =      0x00000400,
  HydraChat  =      0x00000800,
  ModemReport =     0x00001000,
  NoEA =            0x00002000,
  CallAfterAnswer = 0x00004000,
  TerminalLog     = 0x00008000,
  SplitPKTFromLO  = 0x00010000,
  BinkDRecover    = 0x00020000,
  ExportConfig    = 0x00040000
};

enum SysState1
{
  DiscardPKT       = 0x00000001,
  AdjustTime       = 0x00000002,
  Telebit          = 0x00000004,
  BBSenable        = 0x00000008,
  HoldOnCall       = 0x00000010,
  RetryAfterExtern = 0x00000020,
  StatusLine       = 0x00000040,
  NotAcceptAbort   = 0x00000080
};

enum SysState2
{
  DirZapAvail     =     0x00000001,
  FullLog         =     0x00000002,
  RexxDebug       =     0x00000004,
  SemaphoreBsy    =     0x00000008,
  TaskInLog       =     0x00000010,
  SplitterLong    =     0x00000020,
  SplitDelNoFound =     0x00000040,
  HydraAvail      =     0x00000080,
  ZedZapAvail     =     0x00000100
};

enum SysState3
{
  HPFS        = 0x00000001,
  CheckREQ    = 0x00000002,
  CheckName   = 0x00000004,
  HardRing    = 0x00000008,
  BsyAbort    = 0x00000010,
  FastScan    = 0x00000020,
  ScanOnLine  = 0x00000040
};

enum FreqMode
{
  NoFreq = 0,
  ListedFreq = 1,
  ProtectedFreq = 2,
  AllFreq = 4
};

enum QueueSortMode
{
  QueueSortNoSort = 0,
  QueueSortAddr,
  QueueSortAddrDec,
  QueueSortSize,
  QueueSortSizeDec,
  QueueSortPriority,
  QueueSortPriorityDec,
  QueueSortAge,
  QueueSortAgeDec
};

enum FileCase
{
  FileCaseAsIs = 0,
  FileCaseUpper,
  FileCaseLower,
  FileCaseMixed
};

struct Miscellenous
{
  long taskNumber;
  SysState state;
  SysState1 state1;
  SysState2 state2;
  SysState3 state3;
  int circleMax,
      circleMin,
      rescan,
      rescanPeriod,
      port;
  long baud;
  char *inbound,
       *protectedInbound;
  char *tbox;
  char *flags;
  WorkTime *zmh,
	   *answerTime,
	   *callTime,
	   *freqTime;
  int redial;
  int timeOut;
  long EMSItime;
  int zBlock;
  char *emsiLog;
  char *logName;
  int wndModemX,
      wndModemY,
      wndModemOff;
  int x, y, cx, cy;
  char *schedule,
       *schedComp;
  long minIn,
       minOut;
  int trash;
  unsigned long sessionFreqTime,
		sessionFreqSize,
		dayFreqTime,
		dayFreqSize,
		weekFreqTime,
		weekFreqSize,
		monFreqTime,
		monFreqSize;
  char *freqBase;
  long minFreqBaud;
  char *freqPolicy;
  long splitSize;
  int  splitCircle;
  unsigned long priority;
  long priorityDelta;
  int emptyExit;
  WorkTime *BBSTime;
  long minBBSBaud;
  short maxIn, maxOut;
  int maxTry;
  int logBuffer;
  int bsyAge,
      bsyInterval;
  FreqMode freqMode;
  char *binLogName;
  unsigned long reScanPriority;
  long reScanPriorityDelta;
  char *pipe;
  char *semaphore;
  char *rexxPath;
  char *bsySemaPath;
  char *longBox;
  char *exclude;
  int maxRRing;
  int ringCount;
  int ringWait;
  int checkFlags;
  int checkSched;
  char *remoteName;
  long remotePassword;
  char *exitFlag;
  long maxCost;
  long terminalTimeOut;
  int volume;
  char *wavePath;
  WorkTime *noSound;
  QueueSortMode sortMode;
  int autoInit;
  char *terminalInbound;
  char *sentList;
  WorkTime *callOnlyImmed;
  WorkTime *callOnlyCrash;
  WorkTime *callOnlyDirect;
  WorkTime *callOnlyNormal;
  FileCase fileCase;
  long splitPktSize;
  char separator;
  int timeSyncDelta;
  char *iconMail,
       *iconXMail,
       *iconRIP;
  int holdTime;
  int inFileBufferSize,
      outFileBufferSize;
  char *freqIdx;
  char enableMMPM;
  char addrForm[2];
  int chatTimeout;
  long statusLine;
  char *longOutbound;
  char *saverModule;
  int saverInterval;
  unsigned long compilerPriority;
  long compilerPriorityDelta;
  long splitFileSize;
  long configCheck;
  int sendTry;
};

struct ColorSet
{
  int x;
  int y;
  int isColor;
  unsigned char status;
  unsigned char statusLampOn;
  unsigned char statusLampOff;
  unsigned char border;
  unsigned char action;
  unsigned char actionBold;
  unsigned char abort;
  unsigned char popUp;
  unsigned char popUpItem;
  unsigned char popUpBold;
  unsigned char terminal;
  unsigned char help;
  unsigned char helpBold;
  unsigned char queueCrash;
  unsigned char queueNormal;
  unsigned char queueHold;
  unsigned char queueImmed;
  unsigned char queueDirect;
  unsigned char queuePtr;
  unsigned char modemTo;
  unsigned char modemFrom;
  unsigned char modemCtlTo;
  unsigned char modemCtlFrom;
};

class InfoForBBS
{
public:
  InfoForBBS();

  char *twiceEsc,
       *mailOnly,
       *slowModem,
       *mailNow,
       *loading,
       *goodBye;
};

extern InfoForBBS bbsInfo;

extern char *monthName[12];

extern char pszAppBrake[32];
extern char volatile rexxTransfer;

#define stringLen 65

extern char capName[stringLen];
extern char trnName[stringLen];
extern char terminalNumber[stringLen];
extern int trnUploadType;

extern char pollString[stringLen];
extern int pollFlavour;

extern char dropString[stringLen];
extern int dropFlavour;

extern char sendString[stringLen];
extern char sendFile[stringLen];
extern Flavour sendFlavour;

extern char getString[stringLen];
extern char getFile[stringLen];
extern Flavour getFlavour;

extern Miscellenous miscInfo;
extern SystemInfo systemInfo;
extern ColorSet colorSet;

extern volatile unsigned long systemUsageTime;
extern volatile unsigned long systemUsageBegin;
extern volatile unsigned long systemStartTime;
void registerUsageTime( int );

void makeBeep( int f, int d );
void playMelody( int *a=NULL );
void playWaveFile( char *name );

struct Melodies
{
  Schedule *connect,
           *begSess,
           *endSess,
           *abort,
           *human,
           *error,
           *ring,
           *alarm,
           *start,
           *stop,
           *bell,
           *fail,
           *chat;
};

extern Melodies melodies;
extern char **envPtr;

int processConfig( char *name );

void showAction( char *str );

void putToStatusLine( int c, Color& atr, char *s );
int abortAction( int );
void strItem( ostream& o, register char *s );

void makeTime( char *buf, long rest, int = 0 );
char *strTime( long elapsed, int flag=0 );
char *strBig( unsigned long dsize );
char *strNumber( unsigned long );

char * flavourName( Flavour );
char *realFlavour( Flavour flavour );

void capitalize( char *originName );
void makeFileCase( char *originName );

int makePath( char *path );
int makeFilePath( char *file );

void debugDump();

int stringEnter( int len, char *str, char *prompt );

void showInd( char *n, int pos, int fl );

long makeCircle();
int todayMessage( int how, int max, int limit );

#define strError() strerror( errno )

void bigPause( int secs );
void showDebugName( char *s );

void help( char *, int );

int dataInit();
void dataDeInit();

void mailerPoll();
void mailerGet();
void mailerSend();
void mailerDrop();
void mailerChange();

void initFileInfo( int save );

int sessionLimitExceeded();

void generateFreqAnswer( char *msg, FIDOAddress& aka, char *hisName,
				char *password, void * );

void _Optlink echoSplitter( void * );

int isBundle( char *b );
int begin( register char *orig, register char *patt );
int end( char *orig, char *patt );
int includeStr( register char *pat, register char *where );
char *centreStr( char *s, int len );
void changeBlankWithZero( char *trnName, char *name );
void leaveOneBlank( char *getFile );


void resetTry();
void writeToBinaryLog( char code, void *buf, int len, long stamp=0 );

char * fileName( char * );
void addOneToBigNum( char *p );

extern volatile int OSReturnCode;
extern volatile char needInit;

extern long mainTid;
unsigned long getMyTid();

extern char detachedSession;
extern int daemonMode;

extern char noPMFeatures;
extern char isFullScreen;

extern char *titleMode;
void changeTitle( char *ttl );
void changeIcon();
void setTitleMode( char *mode );
void addTitleMode( char *mode );

int strIsYes( char *str );
int strIsNo( char *str );

FILE *_fsopen( const char *, const char *, int =SH_DENYNO );

#define random(x) (rand()%(x))

void real_flush( FILE *f );

extern int onlineAnswer;
extern int onlineAnswerBaud;
extern int onlineHandle;
extern FIDOAddress *callOneNode;

extern char *brakeLoadPath;

char upperChar( int c );

void resumeClipServer();
void suspendClipServer();

int setFileStamp( char *name, unsigned long stamp, int toLocal );

void openObject( char *obj );
long getTimeZone();

extern int telebitFuckinCorrectionTimes;

struct tm *   gmTime( long * );
struct tm *localTime( long * );

int forceDelete( char * );
int isValidDigit( int ch, char base, char& curDigit, char &left );

int isDigit( int c );
int isSpace( int c );

int putString( FILE *file, char *buf );
int putPairBoth( FILE *file, char *variable, char *value, int fl );

int putPair( FILE *file, char *variable, char *value );
int putPair1( FILE *file, char *variable, char *value );

int putPair( FILE *file, char *variable, short value );
int putPair( FILE *file, char *variable, unsigned long value );

void getQuottedString( FILE *file, Buffer &ret, int delimitter );

void getKeyword( FILE *file, Buffer &ret );
void getValue( FILE *file, Buffer &ret );

unsigned long makePercent( const unsigned long size,
                           const unsigned long current );

unsigned long getMSec();
unsigned long getTimeMark();

void allTrim( char *name );
int pmMessage( char *title, char *msg );

void dataInitMsg( unsigned long how, char *msg );
void dataDeInitMsg( unsigned long how, char *msg );

int getFileSize( int handle );

#endif
