#define INCL_DOSEXCEPTIONS
#define INCL_DOSFILEMGR
#define INCL_DOSDEVIOCTL
#define INCL_DOSDEVICES
#define INCL_DOSPROCESS
#define INCL_DOSERRORS
#include <os2.h>

#include <stdlib.h>
#include <stdio.h>
#include <io.h>
#include <process.h>

#include "except.h"

#include "language.h"

#include "misc.h"
#include "timer.h"
#include "buffer.h"
#include "key.h"
#include "newstr.h"
#include "inbuff.h"
#include "hydra.h"

volatile long modemOs2Rc;

static volatile ULONG curInPointer = 0,
  	        curInSize = 0;

static ULONG minStat = 0x7fffffff,
	     maxStat = 0,
	     cntStat = 0,
	     ttlStat = 0,
             emtStat = 0;

static char inBuffer[ inBufferSize ];

#ifdef Generate_For_ModemAsyncRead

static InputBufferList *inputBufferList;
static volatile inSuspendMode = 0;
static volatile readerNeedForAbort = 0;
static volatile currentlyImRead = 0;

static ULONG readerThread = 0xffffffff;

#endif 					// Generate_For_ModemAsyncRead

void ioctlError( char *name, APIRET rc )
{
  if( !rc )
    return;
    
  log( Msg_ErrIOCtl, name, rc );
  os2Message( rc );
}

void Modem::flushBufferedData()
{
#ifdef Generate_For_ModemAsyncRead
  lockSema();
  inputBufferList->Flush();
  unlockSema();
#endif
  curInPointer = curInSize = 0;
}

void Modem::suspend()
{
#ifdef Generate_For_ModemAsyncRead
  inSuspendMode = 1;

  while( currentlyImRead )
    sleep( 1 );
#endif
}

void Modem::resume()
{
#ifdef Generate_For_ModemAsyncRead
  inSuspendMode = 0;
#endif
}

#ifdef Generate_For_ModemAsyncRead

void Modem::getBuffer()
{
  lockSema();

  curInPointer = curInSize = 0;

  if( inputBufferList->GetItemsInContainer() )
  {
    inputBufferList->Lock();

    InputBuffer& cur = (*inputBufferList)[0];

    curInSize = cur.curInSize;

    if( curInSize )
      memcpy( inBuffer, cur.inBuffer, curInSize );

    inputBufferList->Remove( 0, DeleteModeDelete, LockModeNoNeed );
    inputBufferList->UnLock();
  }

  unlockSema();
}

void _Optlink reader( PVOID q )
{
  registerExceptionHandler
  PULONG handle = (PULONG)q;

  while( !readerNeedForAbort )
  {
    char buf[ inBufferSize ];
    ULONG size;

#else 						//  Generate_For_ModemAsyncRead

void Modem::getBuffer()
{
#endif 						//  Generate_For_ModemAsyncRead

  curInPointer = curInSize = 0;

    register APIRET rc = ERROR_INVALID_HANDLE;

#ifdef Generate_For_ModemAsyncRead
    if( inSuspendMode || *handle==-1 )
#else 						//  Generate_For_ModemAsyncRead
    if( handle==-1 )
#endif 						//  Generate_For_ModemAsyncRead
   {
      checkPoint();
#ifdef Generate_For_ModemAsyncRead
      continue;
#else 						//  Generate_For_ModemAsyncRead
      return;
#endif 						//  Generate_For_ModemAsyncRead
    }
    else
    {
#ifdef Generate_For_ModemAsyncRead
      size = 0;
      currentlyImRead = 1;
      rc = DosRead( *handle, buf,
#else 						//  Generate_For_ModemAsyncRead
      lockSema();
      rc = DosRead( handle, &inBuffer,
#endif 						//  Generate_For_ModemAsyncRead
                    !terminalActive || zmodemActive ||
                                       hydraIsActive ? inBufferSize : 1,
#ifdef Generate_For_ModemAsyncRead
                    &size
#else 						//  Generate_For_ModemAsyncRead
                    (PULONG)&curInSize
#endif 						//  Generate_For_ModemAsyncRead
                  );
#ifdef Generate_For_ModemAsyncRead
      currentlyImRead = 0;
#else
      unlockSema();
#endif 						//  Generate_For_ModemAsyncRead
    }

    if( rc )
      if( rc==ERROR_INTERRUPT || rc==ERROR_INVALID_HANDLE )
#ifdef Generate_For_ModemAsyncRead
        size = 0;
#else 						//  Generate_For_ModemAsyncRead
        curInSize = 0;
#endif 						//  Generate_For_ModemAsyncRead
      else
      {
        log( "~Block BUG~ Modem::get() DosRead==%d curInSize==%d",
             rc,
#ifdef Generate_For_ModemAsyncRead
             size
#else 						//  Generate_For_ModemAsyncRead
             curInSize
#endif 						//  Generate_For_ModemAsyncRead
           );
        os2Message( rc );
      }

#ifdef Generate_For_ModemAsyncRead
    if( !size )
#else 						//  Generate_For_ModemAsyncRead
    if( !curInSize )
#endif 						//  Generate_For_ModemAsyncRead
    {
      emtStat++;
#ifdef Generate_For_ModemAsyncRead
//      checkPoint();
      continue;
#endif 						//  Generate_For_ModemAsyncRead
    }

#ifdef Generate_For_ModemAsyncRead
    inputBufferList->AddAtTail( *new InputBuffer( buf, size ) );

    if( size<minStat )
      minStat = size;

    if( size>maxStat )
      maxStat = size;

    ttlStat += size;
#else 						//  Generate_For_ModemAsyncRead

    if( curInSize<minStat )
      minStat = curInSize;

    if( curInSize>maxStat )
      maxStat = curInSize;

    ttlStat += curInSize;
#endif 						//  Generate_For_ModemAsyncRead
    cntStat++;
#ifdef Generate_For_ModemAsyncRead
  }

  deRegisterExceptionHandler
#endif 						//  Generate_For_ModemAsyncRead
}

int Modem::haveInput()
{
  if( curInPointer < curInSize )
    return 1;

  getBuffer();

  if( curInPointer < curInSize )
    return 1;

#ifdef Generate_For_ModemAsyncRead
  checkPoint();
#endif 						//  Generate_For_ModemAsyncRead
  return 0;
}

int Modem::getByteFromBuffer()
{
  register int r;

  r = int( ((unsigned short)inBuffer[ curInPointer++ ]) & 0xff );

  return r;
}

int Modem::get()
{
  register int r;

  if( curInPointer < curInSize )
    r = getByteFromBuffer();
  else
  {
    getBuffer();

    if( curInPointer < curInSize )
      r = getByteFromBuffer();
    else
    {
//      checkPoint();
      return -1;
    }
  }

  dumpFrom( r );
  return r;
}

static char volatile mayMonitor = 1;

void abortQueueMonitor()
{
  mayMonitor = 0;
}

void _Optlink inputQueueMonitor( void * )
{
  registerExceptionHandler

  DosSetPriority( PRTYS_THREAD, 0, -miscInfo.priorityDelta, 0 );

  Window win( (Z_ScreenWidth-60)/2+1, (Z_ScreenHeight-3)/2+1,
			  60, 3, popUpColor,
			  " inQueueStat ", BS_SINGLELINE, DS_LOWERRIGHT );

  +win;

  char buf[80];

  while( 1 )
  {
    if( !mayMonitor )
      break;

    sprintf( buf, "min=%lu ave=%lu [%lu/%lu:%lu] max=%lu",
             minStat,
             (cntStat ? ttlStat/cntStat : 0),
             ttlStat, cntStat,
             emtStat,
             maxStat );
    win( 1, 1 ).clearLine();
    win.centre( 1, buf );

    sleep( 1000 );
  }

  deRegisterExceptionHandler
}

static DCBINFO originDataDCB;
static ULONG sizeOriginDataDCB = sizeof(originDataDCB),
             size1OriginDataDCB = sizeof(originDataDCB);

Modem::Modem( short port, FOSSILbaud sp,
	      char *ini, char *off, char *ans,
	      char *pref, char *suff, char *after,
              char *abo,
	      short callWt,
	      short answerWt,
              short p, long failsafe,
	      int blockmode, long carrierDelay,
	      char hardware, char breakmode, int _handle,
              char *_onExit, char virtDcd ) :
  virtualDCD( virtDcd ),
  virtualDCDOn( 0 ),
  monitorOn( 0 ),
  portNumber( short(port) ),
  speed( sp ),
  initStr( newStr(ini) ),
  offHookStr( newStr(off) ),
  answerStr( newStr(ans) ),
  dialPref( newStr(pref) ),
  dialSuff( newStr(suff) ),
  afterDial( newStr(after) ),
  abortDial( newStr(abo) ),
  onExit( newStr(_onExit) ),
  callWait( callWt ),
  answerWait( answerWt ),
  pause( p ),
  connectSuff( newStr( "*First*Try*" ) ),
  actualSpeed( 0 ),
  callersId( NULL ),
  failSafe( failsafe ),
  outBuffer( NULL ),
  blockMode( blockmode ),
  curOutPointer( 0 ),
  carrierMonitor( carrierDelay ),
  hardWare( hardware ),
  breakMode( breakmode ),
  name( new char [32] ),
  handle( _handle )
{
  sprintf( name, "\\DEV\\COM%d", portNumber+1 );

  if( initStr==NULL )
    initStr = "ATZ0|";

  if( offHookStr==NULL )
    offHookStr = "v~~^~+++~~~ATH0|";

  if( answerStr==NULL )
    answerStr = "ATA:";

  if( dialPref==NULL )
    dialPref = "ATDP";

  if( dialSuff==NULL )
    dialSuff = ":";

  if( afterDial==NULL )
    afterDial = "~";

  if( abortDial==NULL )
    abortDial = "v~~~^`|";

  if( !speed )
    speed = B2400;

  if( !callWait )
    callWait = 30;

  if( !answerWait )
    answerWait = 30;

  if( !failSafe )
    failSafe = 60000l;

  OutBufferSize = 1024;
  outBuffer = new unsigned char [ OutBufferSize ];

  sema = createMutexSemaphore( "modem\\access" );
  inuse = createMutexSemaphore( "modem\\inuse" );

#ifdef Generate_For_ModemAsyncRead
  inputBufferList = new InputBufferList;
#endif 						//  Generate_For_ModemAsyncRead
}

int Modem::initialize()
{
  if( inuse==0xffffffffl )
  {
    log( Msg_ModemNotInUse );
    return 1;
  }

  if( sema==0xffffffffl )
  {
    log( Msg_ModemNotSema );
    return 1;
  }

  int rc = 0;

  rc = openPort( 1 );

  if( rc )
  {
    log( Msg_ModemNotExist );
    return 1;
  }

  if( setup() )
  {
    log( Msg_ModemBigBaud, actualSpeed );
    return 1;
  }

  lastBaud = actualSpeed;

#ifdef Generate_For_ModemAsyncRead
  readerNeedForAbort = 0;
  readerThread = _beginthread( reader, 0, 40960, (void *)&handle );

  if( readerThread==0xffffffff )
  {
    log( Msg_ErrMdmReader, strError() );
    return 1;
  }

  log( Msg_ThrMdmReader, readerThread );
#endif 						//  Generate_For_ModemAsyncRead
  return 0;
}

Modem::~Modem()
{
  if( handle!=-1 && onExit!=NULL )
    putOffLine( onExit );

#ifdef Generate_For_ModemAsyncRead
  readerNeedForAbort = 1;
  DosWaitThread( &readerThread, DCWW_WAIT );
#endif 						//  Generate_For_ModemAsyncRead

  lockSema();

  if( handle!=-1 )
  {
    size1OriginDataDCB = sizeOriginDataDCB;

    APIRET rc=DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_SETDCBINFO,
                           &originDataDCB, sizeOriginDataDCB, &size1OriginDataDCB,
                           0, 0, 0 );
    ioctlError( "SETDCBINFO(~)", rc );
  }

  unlockSema();

  delete outBuffer;
  closePort();

  deleteMutexSemaphore( sema );
  deleteMutexSemaphore( inuse );
}

void Modem::lockSema()
{
  lockMutexSemaphore( sema );
}

void Modem::unlockSema()
{
  unlockMutexSemaphore( sema );
}

// Low Level

int Modem::setup()
{
  initSpeed();

  struct
  {
    BYTE bits, parity, stop;
  } info = { 8, 0, 0 };

  ULONG size( sizeof( info ) );

  lockSema();

  if( handle!=-1 )
    ioctlError( "SETLINECTRL(setup)",
                DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_SETLINECTRL,
                             &info, size, &size, 0, 0, 0 ) );
  unlockSema();

  state = 0;

  clearin();
  clearout();
  dtr( 1 );
  flowControl( 0 );
  receive( 1 );
  flowControl( 1 );

  return setbaud();
}

void Modem::closePort()
{
  lockSema();

  if( !(portNumber+1) )
  {
    handle = -1;
    unlockSema();
    return;
  }

  if( handle==-1 )
  {
    unlockSema();
    return;
  }

  APIRET rc= DosClose( handle );

  if( rc )
  {
    log( "~BUG~ void Modem::closePort(), unable close a port `%s', handle=%d!",
         name, handle );
    os2Message( rc );
  }
  else
    handle = -1;

  unlockSema();
}

int Modem::openPort( int getDCB )
{
  lockSema();

  APIRET rc = 0;

  if( handle==-1 )
  {
    ULONG action;

    rc = DosOpen( (PSZ)name, (ULONG *)&handle, &action,
                  0,
                  FILE_NORMAL,
                  OPEN_ACTION_OPEN_IF_EXISTS,
                  OPEN_FLAGS_FAIL_ON_ERROR|OPEN_FLAGS_SEQUENTIAL|
                  OPEN_SHARE_DENYNONE|OPEN_ACCESS_READWRITE,
                  0 );
    if( rc )
    {
      log( Msg_ModemUnableOpen, name, rc );
      os2Message( rc );
      handle = -1;
    }
  }

  if( handle!=-1 && getDCB )
  {
    rc = DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_GETDCBINFO,
                      0, 0, 0,
                      &originDataDCB, sizeOriginDataDCB, &size1OriginDataDCB );
    ioctlError( "GETDCBINFO(open)", rc );
  }

  unlockSema();

  if( !rc && !getDCB )
    rc = setup();

  return int( rc );
}

void Modem::ClearOutQue()
{
  lockSema();

  if( handle==-1 )
  {
    unlockSema();
    return;
  }

  BYTE param( 0 );
  ULONG size( sizeof(param) );

  short comData( 0 );
  ULONG comDataSize( sizeof(comData) );

  ioctlError( "FLUSHOUTPUT",
              DosDevIOCtl( handle, IOCTL_GENERAL, DEV_FLUSHOUTPUT,
	                   &param, size, &size,
                           &comData, comDataSize, &comDataSize ) );

  curOutPointer = 0;
  unlockSema();
}

void Modem::clearin()
{
  lockSema();

  if( handle==-1 )
  {
    unlockSema();
    return;
  }

  BYTE param( 0 );
  ULONG size( sizeof(param) );

  short comData( 0 );
  ULONG comDataSize( sizeof(comData) );

  ioctlError( "FLUSHINPUT",
              DosDevIOCtl( handle, IOCTL_GENERAL, DEV_FLUSHINPUT,
	                   &param, size, &size,
	                   &comData, comDataSize, &comDataSize ) );

  unlockSema();
  flushBufferedData();
}

void Modem::do_break( int flag )
{
  lockSema();

  if( handle==-1 )
  {
    unlockSema();
    return;
  }

  ULONG size( 1 );

  short comData( 0 );
  ULONG comDataSize( sizeof(comData) );

  ioctlError( flag ? "SETBREAKON" : "SETBREAKOFF",
              DosDevIOCtl( handle, IOCTL_ASYNC,
	                   flag ? ASYNC_SETBREAKON : ASYNC_SETBREAKOFF,
                           0, size, &size,
                           &comData, comDataSize, &comDataSize ) );
  unlockSema();
}

int Modem::hasInputQueue()
{
  lockSema();

  if( handle==-1 )
  {
    unlockSema();
    return 0;
  }

  RXQUEUE how;
  ULONG howSize( sizeof(how) );

  how.cch = 0;
  
  ioctlError( "GETINQUECOUNT(hasInput)",
              DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_GETINQUECOUNT,
	                   0, 0, 0, &how, howSize, &howSize ) );

  unlockSema();

  return how.cch;
}

int Modem::hasOutputQueue()
{
  lockSema();

  if( handle==-1 )
  {
    unlockSema();
    return 0;
  }

  RXQUEUE how;
  ULONG howSize( sizeof(how) );

  how.cch = 0;
  
  ioctlError( "GETOUTQUECOUNT(hasOutput)",
              DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_GETOUTQUECOUNT,
	                   0, 0, 0, &how, howSize, &howSize ) );

  unlockSema();

  return how.cch;
}

int Modem::sizeInputQueue()
{
  lockSema();

  if( handle==-1 )
  {
    unlockSema();
    return 0;
  }

  RXQUEUE how;
  ULONG howSize( sizeof(how) );

  how.cb = 256;
  
  ioctlError( "GETINQUECOUNT(sizeInput)",
              DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_GETINQUECOUNT,
	                   0, 0, 0, &how, howSize, &howSize ) );

  unlockSema();

  return how.cb;
}

int Modem::sizeOutputQueue()
{
  lockSema();

  if( handle==-1 )
  {
    unlockSema();
    return 0;
  }

  RXQUEUE how;
  ULONG howSize( sizeof(how) );

  how.cb = 256;
  
  ioctlError( "GETOUTQUECOUNT(sizeOutput)",
              DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_GETOUTQUECOUNT,
	                   0, 0, 0, &how, howSize, &howSize ) );

  unlockSema();

  return how.cb;
}

void Modem::stc()
{
  if( handle==-1 || !systemIsReady )
    return;

  int delta=(statusLineWindow.getWidth()-80)/2;

  static BYTE sdOld( 0xff ),
	      rdOld( 0xff );
  register char sd = char( !outEmpty() ),
                rd( 0 );

  if( curInPointer < curInSize )
    rd = 1;
  else
    rd = char( hasInputQueue() ? 1 : 0 );

  if( sd!=sdOld )
    showInd( "SQ", 65+delta, sdOld = sd );

  if( rd!=rdOld )
    showInd( "RQ", 62+delta, rdOld = rd );

  APIRET rc;
  BYTE status = 0;
  ULONG statusSize( sizeof(status) );

  lockSema();

  if( handle!=-1 )
  {
    rc=DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_GETMODEMINPUT,
                    0, 0, 0, &status, statusSize, &statusSize );
    ioctlError( "GETMODEMINPUT(stc)", rc );
  }
  else
    rc = 1;

  unlockSema();

  if( !rc )
  {
    static char octs = char( -2 );
    static char ori  = char( -2 );
    static char odcd = char( -2 );

    register char cts = char( (status&CTS_ON)==CTS_ON );
    register char ri  = char( (status&RI_ON)==RI_ON   );
    register char dcd = char( virtualDCD ? virtualDCDOn
                                         : (status&DCD_ON)==DCD_ON );

    if( octs!=cts )
    {
      octs = cts;
      showInd( "CTS", 48+delta, cts );
    }

    if( ori!=ri )
    {
      ori = ri;
      showInd( "RI",  56+delta, ri );
    }

    if( odcd!=dcd )
    {
      odcd = dcd;
      showInd( virtualDCD ? "VC" : "CD",  59+delta, odcd );
    }
  }

  status = 0;

  lockSema();

  if( handle!=-1 )
  {
    rc = DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_GETMODEMOUTPUT,
                      0, 0, 0, &status, statusSize, &statusSize );
    ioctlError( "GETMODEMOUTPUT(stc)", rc );
  }
  else
    rc = 1;

  unlockSema();

  if( !rc )
  {
    static char orts = char( -2 );
    static char odtr = char( -2 );

    register char rts = char( (status&RTS_ON)==RTS_ON );
    register char dtr = char( (status&DTR_ON)==DTR_ON );

    if( orts!=rts )
    {
      orts = rts;
      showInd( "RTS", 52+delta, rts );
    }

    if( odtr!=dtr )
    {
      odtr = dtr;
      showInd( "DTR", 44+delta, dtr );
    }
  }
}

void Modem::dtr( int flag )
{
  lockSema();

  if( handle==-1 )
  {
    unlockSema();
    return;
  }

  MODEMSTATUS DTRon = { DTR_ON, 0xff },
	      DTRoff = { 0, DTR_OFF };

  short comData( 0 );
  ULONG comDataSize( sizeof(comData) );

  ULONG DTRoffSize( sizeof(DTRoff) );
  ULONG DTRonSize( sizeof(DTRon) );

  if( !flag )
    ioctlError( "SETMODEMCTRL(DTR=OFF)",
                DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_SETMODEMCTRL,
		             &DTRoff, DTRoffSize, &DTRoffSize,
		             &comData, comDataSize, &comDataSize ) );
  else
    ioctlError( "SETMODEMCTRL(DTR=ON)",
                DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_SETMODEMCTRL,
	                     &DTRon, DTRonSize, &DTRonSize,
                             &comData, comDataSize, &comDataSize ) );
  unlockSema();
}

int Modem::setbaud( int needBaud )
{
  initSpeed();

  lockSema();

  if( handle==-1 )
  {
    unlockSema();
    return 0;
  }

  struct
  {
    ULONG baud;
    BYTE frac;
  } baud = { needBaud ? (unsigned long)needBaud :
                        (unsigned long)actualSpeed,
             0
           };
  ULONG baudSize( sizeof( baud ) );
  APIRET rc( DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_EXTSETBAUDRATE,
                          &baud, baudSize, &baudSize,
                          0, 0, 0 ) );
		           
  ioctlError( "EXTSETBAUDRATE", rc );
  unlockSema();
  return int( rc );
}

int Modem::checkOnline()
{
  if( virtualDCDOn )
    return 1;

  lockSema();

  if( handle==-1 )
  {
    unlockSema();
    return 0;
  }

  BYTE status = 0;
  ULONG statusSize( sizeof(status) );

  ioctlError( "GETMODEMINPUT(DCD)",
              DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_GETMODEMINPUT,
                           0, 0, 0, &status, statusSize, &statusSize ) );

  unlockSema();
  return (status&DCD_ON)==DCD_ON;
}

int Modem::putFos( void* p, int l )
{
  lockSema();

  if( handle==-1 )
  {
    unlockSema();
    checkPoint();
    return 0;
  }

  ULONG written;

  modemOs2Rc = DosWrite( handle, p, l, &written );

  unlockSema();

  return int( written );
}

void setTimeOut( DCBINFO& dataDCB )
{
  dataDCB.fbTimeout &= ~MODE_NO_WRITE_TIMEOUT;
  dataDCB.usWriteTimeout = 0;
  dataDCB.usReadTimeout = 0;
  dataDCB.fbTimeout = char( (dataDCB.fbTimeout&(~0x6))|MODE_READ_TIMEOUT );
}

void Modem::ctsrts( int f )
{
  lockSema();

  if( handle==-1 )
  {
    unlockSema();
    return;
  }

  DCBINFO dataDCB;
  ULONG sizeDataDCB = sizeof(dataDCB),
        size1DataDCB = sizeof(dataDCB);

  APIRET rc = DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_GETDCBINFO,
	                   0, 0, 0,
                           &dataDCB, sizeDataDCB, &size1DataDCB );

  ioctlError( "GETDCBINFO(ctsrts)", rc );

  setTimeOut( dataDCB );

  dataDCB.fbCtlHndShake = BYTE( f ? MODE_CTS_HANDSHAKE : 0x00 );
  dataDCB.fbFlowReplace = BYTE( f ? MODE_RTS_HANDSHAKE : 0x00 );

  if( !rc )
  {
    rc = DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_SETDCBINFO,
	              &dataDCB, sizeDataDCB, &size1DataDCB,
	              0, 0, 0 );
    ioctlError( "SETDCBINFO(ctsrts)", rc );
  }

  unlockSema();
}

void Modem::xonxoff( int f )
{
  lockSema();

  if( handle==-1 )
  {
    unlockSema();
    return;
  }

  DCBINFO dataDCB;
  ULONG sizeDataDCB = sizeof(dataDCB),
        size1DataDCB = sizeof(dataDCB);

  APIRET rc = DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_GETDCBINFO,
	                   0, 0, 0,
                           &dataDCB, sizeDataDCB, &size1DataDCB );

  ioctlError( "GETDCBINFO(xonxoff)", rc );

  setTimeOut( dataDCB );

  dataDCB.fbCtlHndShake = 0x00;
  dataDCB.fbFlowReplace = BYTE( f ? (MODE_AUTO_TRANSMIT|MODE_AUTO_RECEIVE)
                                  : 0x00 );

  if( !rc )
  {
    rc = DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_SETDCBINFO,
	              &dataDCB, sizeDataDCB, &size1DataDCB,
                      0, 0, 0 );
    ioctlError( "SETDCBINFO(xonxoff)", rc );
  }

  unlockSema();
}

void Modem::receive( int flag )
{
  lockSema();

  if( handle==-1 )
  {
    unlockSema();
    return;
  }

  MODEMSTATUS RTSon = { RTS_ON, 0xff },
 	      RTSoff = { 0, RTS_OFF };

  short comData( 0 );
  ULONG comDataSize( sizeof(comData) );

  ULONG RTSoffSize( sizeof(RTSoff) );
  ULONG RTSonSize( sizeof(RTSon) );

  if( !flag )
    ioctlError( "SETMODEMCTRL(RTS=OFF)",
                 DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_SETMODEMCTRL,
	                      &RTSoff, RTSoffSize, &RTSoffSize,
                              &comData, comDataSize, &comDataSize ) );
  else
    ioctlError( "SETMODEMCTRL(RTS=ON)",
                DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_SETMODEMCTRL,
	                     &RTSon, RTSonSize, &RTSonSize,
                             &comData, comDataSize, &comDataSize ) );
  unlockSema();
}

int Modem::isRinging()
{
  lockSema();

  if( handle==-1 )
  {
    unlockSema();
    return 0;
  }

  USHORT err = 0;
  ULONG errSize( sizeof(err) );

  ioctlError( "GETMODEMINPUT(RI)",
              DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_GETMODEMINPUT,
                           0, 0, 0, &err, errSize, &errSize ) );

  unlockSema();

  return (err&RI_ON)==RI_ON ? 1 : 0;
}

inline char *yesNo( int fl )
{
  return fl ? "ON" : "OFF";
}

void Modem::logState()
{
  lockSema();

  if( handle==-1 )
  {
    unlockSema();
    return;
  }
    
  LINECONTROL lc;
  USHORT bitrate;
  UCHAR state;
  UCHAR state1;
  ULONG size( sizeof lc );
  
  ioctlError( "GETLINECTRL(logState)",
              DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_GETLINECTRL,
                           0, 0, 0, &lc, size, &size ) );
               
  size = sizeof bitrate;
  
  ioctlError( "GETBAUDRATE(logState)",
              DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_GETBAUDRATE,
                           0, 0, 0, &bitrate, size, &size ) );
               
  log( "ComLineMode: BitRate=0x%x bits=%d parity=0x%x stop=0x%x break=0x%x", 
       bitrate, lc.bDataBits, lc.bParity, lc.bStopBits, lc.fTransBreak );
       
  size = sizeof state;
  
  ioctlError( "GETCOMMSTATUS(logState)",
              DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_GETCOMMSTATUS,
                           0, 0, 0, &state, size, &size ) );
               
  log( "ComStatus1: TW_CTS=%s TW_DSR=%s TW_DCD=%s TW_XON=%s TW_TS_XON=%s TW_W_BREAK=%s TW_TS_IMM=%s RW_DSR=%s",
       yesNo( state&TX_WAITING_FOR_CTS ),
       yesNo( state&TX_WAITING_FOR_DSR ),
       yesNo( state&TX_WAITING_FOR_DCD ),
       yesNo( state&TX_WAITING_FOR_XON ),
       yesNo( state&TX_WAITING_TO_SEND_XON ),
       yesNo( state&TX_WAITING_WHILE_BREAK_ON ),
       yesNo( state&TX_WAITING_TO_SEND_IMM ),
       yesNo( state&TX_WAITING_FOR_DSR ) );
       
  size = sizeof state;
  
  ioctlError( "GETLINESTATUS(logState)",
              DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_GETLINESTATUS,
                           0, 0, 0, &state, size, &size ) );
               
  log( "ComStatus2: WR_QUEUED=%s DATA_TXQ=%s H_TRANS=%s TX_IMM=%s W_ATS_XON=%s W_ATS_XOFF=%s",
       yesNo( state&WRITE_REQUEST_QUEUED ),
       yesNo( state&DATA_IN_TX_QUE ),
       yesNo( state&HARDWARE_TRANSMITTING ),
       yesNo( state&CHAR_READY_TO_SEND_IMM ),
       yesNo( state&WAITING_TO_SEND_XON ),
       yesNo( state&WAITING_TO_SEND_XOFF ) );

  size = sizeof state;
  
  ioctlError( "GETMODEMOUTPUT(logState)",
              DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_GETMODEMOUTPUT,
                           0, 0, 0, &state, size, &size ) );
               
  size = sizeof state1;
  
  ioctlError( "GETMODEMINPUT(logState)",
              DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_GETMODEMINPUT,
                           0, 0, 0, &state1, size, &size ) );
               
  log( "ComModem: DTR=%s RTS=%s CTS=%s DSR=%s DCD=%s RI=%s",
       yesNo( state&DTR_ON ),
       yesNo( state&RTS_ON ),
       yesNo( state1&CTS_ON ),
       yesNo( state1&DSR_ON ),
       yesNo( state1&DCD_ON ),
       yesNo( state1&RI_ON ) );
       
  size = sizeof bitrate;
  
  ioctlError( "GETCOMMERROR(logState)",
              DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_GETCOMMERROR,
                           0, 0, 0, &bitrate, size, &size ) );
               
  log( "ComError: RQ_OVERRUN=%s HARD_OVERRUN=%s PARITY_ERROR=%s FRAMING_ERROR=%s",
       yesNo( bitrate&RX_QUE_OVERRUN ),
       yesNo( bitrate&RX_HARDWARE_OVERRUN ),
       yesNo( bitrate&PARITY_ERROR ),
       yesNo( bitrate&FRAMING_ERROR ) );
       
  size = sizeof bitrate;
  
  ioctlError( "GETCOMMEVENT(logState)",
              DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_GETCOMMEVENT,
                           0, 0, 0, &bitrate, size, &size ) );
               
  log( "ComEvent: RTOUT=%s C_RECV=%s C_SENT=%s C_CTS=%s C_DSR=%s C_DCD=%s BREAK=%s ERR=%s RI=%s",
       yesNo( bitrate&2 ),
       yesNo( bitrate&CHAR_RECEIVED ),
       yesNo( bitrate&LAST_CHAR_SENT ),
       yesNo( bitrate&CTS_CHANGED ),
       yesNo( bitrate&DSR_CHANGED ),
       yesNo( bitrate&DCD_CHANGED ),
       yesNo( bitrate&BREAK_DETECTED ),
       yesNo( bitrate&ERROR_OCCURRED ),
       yesNo( bitrate&RI_DETECTED ) );

  unlockSema();
}

int Modem::isLockedInuse()
{
  return queryMutexSemaphore( inuse );
}

void Modem::lock( char *msg )
{
  if( isLockedInuse() )
    log( Msg_InfoModemLock, msg );

  lockMutexSemaphore( inuse );
}

void Modem::unlock()
{
  unlockMutexSemaphore( inuse );
}

int Modem::isLocked()
{
  return isLockedInuse() || handle==-1;
}
