#if !defined(__PAIR_H)

#define __PAIR_H

#include "class.h"
#include "sched.h"
#include "emsi.h"

class Pair : public BaseObject
{
public:

  enum DoType { doOut, doDo, doModem, doKey, doKeyDo, doSched, doEMSI };
  enum PairRez { NoMatch, InProgress, Completed, Reset };

private:

  union
  {
    char *pattern;
    int key;
  } who;

  char *cur;
  DoType doType;

  union
  {
    char *out;
    void (*doFunc)();
    Schedule *sh;
    EMSI_PACKETS emsi;
  } action;

  int equal( const BaseObject &t ) const;
  void print( ostream& s ) const;
  char *nameOfClass() const;

public:

  Pair();
  Pair( char *ip, char *io );
  Pair( char *ip, char *io, DoType is );
  Pair( char *ip, void (*id)() );
  Pair( int key, char *io );
  Pair( int key, void (*id)() );
  Pair( char *ip, Schedule *ish );
  Pair( char *ip, EMSI_PACKETS em );

  void operator = ( Pair &ini );
  
  int operator [] ( int i );

  int operator == ( int how );
  int operator == ( char *how );

  PairRez compare( char c );
  void reset();
  char *process();
  int matchedLen() const;

  char *getPattern() const { return who.pattern; }
  void (*getFunc())() { return action.doFunc; }
  EMSI_PACKETS getEmsi() const { return action.emsi; }
  Schedule *getSchedule() const;
  DoType getType() const { return doType; }
  char *getOut() const { return action.out; }
};

extern Pair *connectList;
extern int connectCount;
extern int currentConnectItem;

extern Pair *handShakeList;
extern int handShakeCount;
extern Pair *EMSIuserAction;

#endif
