#if !defined(__QUEUE_H)

#define __QUEUE_H

#include "fidoaddr.h"
#include "misc.h"
#include "filetype.h"
#include "semaphor.h"
#include "shmem.h"
#include "enable.h"

struct Traffic
{
  uLong
	holdMail, crashMail, normalMail,
	immedMail, directMail,
	holdXMail, crashXMail, normalXMail,
	immedXMail, directXMail,
	holdFiles, crashFiles, normalFiles,
	immedFiles, directFiles,
	request;
  long age;
  Flavour poll;
  unsigned long holdTo;
  char busy;
};

enum ScanType
{
  ScanTypeUnknown=0,
  ScanTypeWithFiles=1,
  ScanTypeWithoutFiles=2
};

class QueueEntry : public NodeConfig
{
  Traffic traf;

  char mapLine,
       iMenable,
       iAcceptable,
       costFault,
       imCur;
  short tryOK,
	tryBad;

  char *tryLast;
  char ours;

  int equal( const BaseObject &t ) const;
  void print( ostream& s ) const;
  char *nameOfClass() const;

protected:
  long rescanLO( ScanType, char *path, char *name,
                 uLong& mail, uLong& xmail, uLong& files, long created );

  void processHold( char *, char * );
  void processLO( ScanType type, char *path, char *name,
                  Flavour myFlavour, long created );

  void clear();
  void addTraffic( char *rexxRez );
  void setAge( char *, long );
  void setMailTraf( Flavour fl, long size, long created );

  int existIn( EnableList *list );
  void infoShowNdl( void * );
  int checkHoldState( char *, char * );

public:

  QueueEntry( const NodeConfig& cfg );
  QueueEntry( const QueueEntry & );
  virtual ~QueueEntry();

  void addTraffic( unsigned long * );
  void addTraffic( Traffic& );

  void preRescan();
  void postRescan();

  int equal( const QueueEntry &t ) const;
  int less( const QueueEntry &t ) const;

  void processShortBink( ScanType type, char *name, char *path,
                         long size, long created );
  void processLongBinkName( ScanType type, 
                            char *path, char *bink, char *name,
                            long size, long create );
  void processTBox( ScanType, char *name );
  void rescanPath( ScanType, char *path, Flavour fl );
  void rescanTBox( ScanType );
  void rescanLongBox( ScanType );
  void rescanCfg( ScanType );
  void rescanBink( ScanType );
  void rescanLongBink( ScanType );
  void rescan();
  int checkHoldState();

  void draw( int line, QueueEntry *cur );
  void show();
  void info();

  Flavour getQueuedFlavour();
  Flavour getTrafficFlavour() const;
  int isCallable( int =0 );
  int isFlavourCallable();

  char getPollMode() const;
  void setPollMode( Flavour mode=normalFlavour );

  int isMapped() const { return mapLine; }
  int isEmpty() const;
  int isAcceptable() const { return iAcceptable; }
  int isOurs() const { return ours; }

  uLong holdSize() const;
  uLong normalSize() const;
  uLong crashSize() const;
  uLong immedSize() const;
  uLong directSize() const;

  void loadTry();
  void saveTry( int bad, int ok );
  void setTryLast( char *msg );

  int tryExceeded() const;

  long getMail() const;
  long getFiles() const;

  void unMap();

  uLong getImmedMail() const  { return traf.immedMail; }
  uLong getDirectMail() const { return traf.directMail; }
  uLong getHoldMail() const   { return traf.holdMail; }
  uLong getCrashMail() const  { return traf.crashMail; }
  uLong getNormalMail() const { return traf.normalMail; }

  uLong getImmedXMail() const  { return traf.immedXMail; }
  uLong getDirectXMail() const { return traf.directXMail; }
  uLong getHoldXMail() const   { return traf.holdXMail; }
  uLong getCrashXMail() const  { return traf.crashXMail; }
  uLong getNormalXMail() const { return traf.normalXMail; }

  uLong getImmedFiles() const  { return traf.immedFiles; }
  uLong getDirectFiles() const { return traf.directFiles; }
  uLong getHoldFiles() const   { return traf.holdFiles; }
  uLong getCrashFiles() const  { return traf.crashFiles; }
  uLong getNormalFiles() const { return traf.normalFiles; }

  uLong getRequest() const { return traf.request; }
  long getAge() const { return traf.age; }
  Flavour getPoll() const { return traf.poll; }

  Traffic *getTrafficPointer() { return &this->traf; }

  void format( char *buf, int flag=0 );
  char *getTryLast() const { return tryLast; }

  int getPriority() const;

  unsigned long getHoldTo() const { return traf.holdTo; }
  void setHoldTo( unsigned long delta );
};

struct SharedQueueHeader
{
  long lastScaned;
  long scanStart;
  long count;
  char inScan,
       needScan;
};

class Queue : public Collection		//QueueEntry
{
  int current,
      scrollCurrent;
  char first;
  int wasCurrent,
      wasScrollCurrent;
  Flavour flavour;
  int priority;

  char *lastWas,
       *lastWasUsed,
       lastIn;
  unsigned long lastRead;
  static unsigned long globalSemaphore;
  static SharedQueueHeader *sharedQueueHeader;
  static SharedMemory *shMem;

protected:

  QueueEntry *createEntry( uShort, uShort, uShort, uShort,
                           char *, Domain * );
  void rescanBinkEntry( uShort, uShort, uShort, uShort, char *, Domain *,
                        char *, char *, long, long );
  void rescanBinkZone( int, char *, Domain * );
  void rescanBinkPoint( uShort, uShort, uShort, char *, char *, Domain * );

  void adjust();
  void redrawLastConnect();
  void addIfNeedOrDelete( QueueEntry* );
  QueueEntry* checkQEntry( QueueEntry* );

public:

  Queue( SemaphoreMode sema );
  ~Queue();

  friend ostream& operator << ( ostream& o, const Queue& c );

  QueueEntry* Find( QueueEntry& item, LockMode lock=LockModeNeed ) const;
  QueueEntry& operator [] ( const int pos ) const;

  void Remove( QueueEntry& item, DeleteMode del=DeleteModeDefault,
               LockMode lock=LockModeNeed );
  void Remove( int item, DeleteMode del=DeleteModeDefault,
               LockMode lock=LockModeNeed );

  void add( QueueEntry& system );

  static int initStatic( char * );
  static void needForRescan();

  void rescanAllType();

  void nextEntry();
  void sort();
  void draw();
  void read();
  void write();
  void checkFlavour();
  
  void rescanQueue();
  void rescanBink();
  void rescanLongBink();
  void rescanCfg();
  void rescanTBox();
  void rescanLongBox();
  void updateEntry( QueueEntry& );

  QueueEntry *getCurrent() const;

  int getCurrentPointer() const { return current; }

  int getPriority() const { return priority; }

  Flavour getFlavour() const { return flavour; }

  void setLast( char *w, char *w1, char i );
  void getLast();
  void view();
  void queueName( char *msg, char *dis );
  void scroll( int );

  friend void _Optlink rescanServer( void * );
};

class QueueIterator : public CollectionIterator 	//QueueEntry
{
public:
  QueueIterator( Queue& ini );

  QueueEntry& operator () () const;
};

class QueueEntryList : public Collection		//QueueEntry
{
public:
  QueueEntryList();

  QueueEntry* Find( QueueEntry& item, LockMode lock=LockModeNeed ) const;
  QueueEntry& operator [] ( const int pos ) const;
};

class QueueEntryIterator : public CollectionIterator
{
public:
  QueueEntryIterator( QueueEntryList& ini );
  
  QueueEntry& operator () () const;
};

class Try
{
  char *last;
  short ok;
  short bad;
  Buffer rest;
  FILE *file;
  char *full;
  char changed;
  unsigned long okTime,
                lastTime,
                holdTo,
                line;

  void saveTime( int& rc, char *key, unsigned long time );
  unsigned long decode( char *str, char last, char *msg );

public:
  Try( char *path, int flag );
  ~Try();

  void setInfo( int bad, int ok, char *l );

  short getOk() const { return ok; }
  short getBad() const { return bad; }
  unsigned long getOkTime() const { return okTime; }
  unsigned long getLastTime() const { return lastTime; }
  unsigned long getHoldTo() const { return holdTo; }
  char *getLast() const { return last; }
};

extern Queue *queue;
extern long sharedMemoryQueueSize;
extern char *answerDisabled;

struct SharedQueueEntry
{
  Address addr;
  Traffic traf;
};

#endif
