#include <string.h>

#include "language.h"

#include "misc.h"
#include "rexx.h"
#include "rexxhook.h"

RexxHook::RexxHook( char *_h, char *_p ) :
  hook( newStr(_h) ),
  proc( newStr(_p) )
{}

RexxHook::~RexxHook()
{
  delete hook;
  delete proc;
}

int RexxHook::equal( const BaseObject &t ) const
{
  return !stricmp( hook, (*(const RexxHook *)&t).hook);
}

void RexxHook::print( ostream& s ) const
{
  s << hook << "->" << proc;
}

char *RexxHook::nameOfClass() const
{
  return "RexxHook";
}

HookList::HookList() :
  Collection( 5, 5, SemaphoreModeNoNeed )
{
}

RexxHook* HookList::Find( RexxHook& item, LockMode lock ) const
{
  return (RexxHook*)Collection::Find( item, lock );
}
    
RexxHook& HookList::operator [] ( const int pos ) const
{
  return *(RexxHook *)&(Collection::operator[]( pos ));
}

HookIterator::HookIterator( HookList& ini ) :
  CollectionIterator( *(Collection *)&ini )
{
}

RexxHook& HookIterator::operator () () const
{
  return *(RexxHook *)&(CollectionIterator::operator()());
}

char *getRexxHook( char *name )
{
  if( rexxHookList==NULL )
    return NULL;

  RexxHook hook( name, NULL );
  RexxHook *test = rexxHookList->Find( hook );

  if( test==NULL )
    return NULL;

  return test->getProc();
}

char *processRexxHookBoth( int needLog,
                           char *name, char *a1, char *a2, char *a3,
                           char *a4, char *a5, char *a6,
                           int needResult )
{
  char *proc = getRexxHook( name );

  if( proc==NULL )
    return NULL;

  char *out( NULL );

  if( CallRexx( proc, &out, a1, a2, a3, a4, a5, a6, needLog ) )
    if( needResult )
      return out;

  delete out;
  return NULL;
}

int processLogicalRexxHook( char *name,
                            char *a1, char *a2, char *a3,
                            char *a4, char *a5, char *a6 )
{
  char *out=processRexxHook( name, a1, a2, a3, a4, a5, a6, 1 );
  int ret;

  if( out==NULL || !stricmp( out, "YourSelf" ) )
    ret = -1;
  else if( !stricmp( out, "Yes" ) )
    ret = 1;
  else if( !stricmp( out, "No" ) )
    ret = 0;
  else
  {
    log( Msg_RexxErrReturned, name, out );
    ret = -1;
  }

  delete out;
  return ret;
}

char *processRexxHook( char *name,
                       char *a1, char *a2, char *a3,
                       char *a4, char *a5, char *a6,
                       int needResult )
{
  return processRexxHookBoth( 1, name, a1, a2, a3, a4, a5, a6, needResult );
}

char *processRexxHookNoLog( char *name,
                            char *a1, char *a2, char *a3,
                            char *a4, char *a5, char *a6,
                            int needResult )
{
  return processRexxHookBoth( 0, name, a1, a2, a3, a4, a5, a6, needResult );
}
