#define INCL_DOSSEMAPHORES
#define INCL_DOSDATETIME
#define INCL_DOSPROCESS
#define INCL_DOSEXCEPTIONS
#include <os2.h>

#include <io.h>
#include <share.h>
#include <fcntl.h>
#include <time.h>

#include "except.h"

#include "language.h"

#include "misc.h"
#include "sched.h"
#include "queue.h"
#include "session.h"
#include "semaphor.h"

void pleaseDraw( char *str1, char *str2, char& needLatest )
{
  static char *old1 = NULL;
  static char *old2 = NULL;

  if( old1==NULL           || old2==NULL           ||
      str1==NULL           || str2==NULL           ||
      strcmp( old1, str1 ) || strcmp( old2, str2 ) )
    needLatest = 1;

  if( !needLatest || sessionStart || terminalActive || rexxTransfer )
    return;

  if( str1==NULL )
    str1 = newStr( Msg_ScrEvenNone );

  if( str2==NULL )
    str2 = newStr( Msg_ScrEvenNone );

  delete old1;
  delete old2;

  old1 = newStr( str1 );
  old2 = newStr( str2 );

  Color bold( colorSet.actionBold );
  int len = strlen( Msg_ScrEven );

  wndStat.drawTransact( 2, 7, stdColor, Msg_ScrEven );

  if( strlen( str1 )>wndStat.getWidth()-len-1 )
    str1[ wndStat.getWidth()-len-1 ] = 0;

  wndStat.drawTransact( len+2, 7, bold, str1 );

  wndStat.drawTransact( 2, 8, stdColor, Msg_ScrEvenTime );
  len = strlen( Msg_ScrEvenTime );

  if( strlen( str2 )>wndStat.getWidth()-len-1 )
    str2[ wndStat.getWidth()-len-1 ] = 0;

  wndStat.drawTransact( len+2, 8, bold, str2 );

  needLatest = 0;
}

static volatile abortScheduleFlag = 0;
static HEV sema;

char *makeEventTime( Schedule& sh )
{
  char buf[32],
       *str2 = new char [ 128 ];
  *str2 = 0;

  if( sh.dayOfMonth )
  {
    sprintf( buf, "%d-", sh.dayOfMonth );
    strcat( str2, buf );
  }

  if( sh.dayOfWeek )
  {
    sprintf( buf, "%d.", sh.dayOfWeek );
    strcat( str2, buf );
  }

  int sec(  int( sh.start%60 ) ),
      mins( int( (sh.start/60)%60 ) ),
      hour( int( sh.start/60/60 ) );

  sprintf( buf, "%02d:%02d", hour, mins );
  strcat( str2, buf );

  if( sec )
  {
    sprintf( buf, ":%02d", sec );
    strcat( str2, buf );
  }

  return str2;
}

void _Optlink scheduleProcess( void * )
{
  registerExceptionHandler

  DosSetPriority( PRTYS_THREAD, 0, -miscInfo.priorityDelta, 0 );

  int RC;
  char *str1 = NULL,
       *str2 = NULL;
  char needLatest = 0;

  RC = int( createEventSemaphore( &sema, "ScheduleProcess" ) );

  if( RC )
  {
    log( Msg_ErrUnableCreateSchedSema, RC );
    os2Message( RC );
    log( Msg_ErrSchedulerFailed );
    return;
  }

  while( 1 )
  {
    if( abortScheduleFlag )
    {
      deRegisterExceptionHandler
      return;
    }

    pleaseDraw( str1, str2, needLatest );

    time_t tim = time( NULL );
    struct tm *tm = localTime( &tim );
    SchedDate dt = {
                     short( tm->tm_year+1900 ), 
                     char(  tm->tm_mday      ),
                     char(  tm->tm_mon+1     ),
		     char(  tm->tm_wday      )
                   };
    char doFlag = 0;

    if( memcmp( &dt, &schedDate, sizeof(dt) ) )
    {
      memcpy( &schedDate, &dt, sizeof(dt) );

      for( int i(0); i<schedCount; i++ )
        schedList[i].done = 0;

      saveEvents();
      schedCurrent = 0;
    }

    long startSched( schedCurrent );
    long start = (tm->tm_hour*60l+tm->tm_min)*60+tm->tm_sec;

    while( 1 )
    {
      if( !schedList[schedCurrent].done )
      {
        char need( 1 );

        if( schedList[schedCurrent].dayOfMonth )
          if( schedList[schedCurrent].dayOfMonth != tm->tm_mday )
            need = 0;

        if( schedList[schedCurrent].dayOfWeek )
          if( schedList[schedCurrent].dayOfWeek != 
              (tm->tm_wday ? tm->tm_wday : 7 ) )
            need = 0;

        if( need )
        {
          if( start < schedList[schedCurrent].start )
            break;

          if( schedList[schedCurrent].type==SchedTypeForce ||
              ( start >= schedList[schedCurrent].start &&
	        start <= schedList[schedCurrent].start+miscInfo.trash ) )
	  {
	    schedList[schedCurrent].done = 1;

            char *startAt = makeEventTime( schedList[schedCurrent] );

	    log( Msg_InfoEventStart, startAt, schedList[schedCurrent].name );
            delete startAt;

            doFlag = 1;
	    RC = doEventKey( schedList[schedCurrent], queue->getCurrentPointer() );
	    saveEvents();

	    if( RC )
            {
              if( RC>OSReturnCode )
                OSReturnCode = RC;

              deRegisterExceptionHandler
              return;
            }                    //if( RC )
	  }                      //if( time )
        }                        //if( need )
      }                          //if( !done )

      if( ++schedCurrent>=schedCount )
        schedCurrent = 0;

      if( startSched==schedCurrent )
      {
        schedCurrent = 0;
        break;
      }
    }                            //for( ;; )

    delete str1;
    str1 = newStr( schedList[schedCurrent].name );

    delete str2;
    str2 = makeEventTime( schedList[schedCurrent] );

    pleaseDraw( str1, str2, needLatest );

    if( !doFlag )
    {
      HTIMER timer;

      RC = int( DosAsyncTimer( miscInfo.checkSched*1000, (HSEM)sema, &timer ) );

      if( RC )
      {
        log( Msg_ErrUnableSchedAsync, RC );
        os2Message( RC );
        log( Msg_ErrSchedulerFailed );
        return;
      }

      RC = int( DosWaitEventSem( sema, (ULONG)SEM_INDEFINITE_WAIT ) );

      if( RC )
      {
        log( Msg_ErrUnableWaitSchedSema, RC );
        os2Message( RC );
        log( Msg_ErrSchedulerFailed );
        return;
      }

      ULONG cnt;

      RC = int( DosResetEventSem( sema, &cnt ) );

      if( RC )
      {
        log( Msg_ErrUnableResetSchedSema, RC );
        os2Message( RC );
      }
    }
  }
}

void abortScheduleProcess()
{
  abortScheduleFlag = 1;

  APIRET rc = DosPostEventSem( sema );

  if( rc )
  {
    log( Msg_ErrUnablePostSchedSema, rc );
    os2Message( rc );
  }
}
