#ifndef __SEMAPHORES_H

#define __SEMAPHORES_H

#include <string.h>

#include "class.h"
#include "newstr.h"

unsigned long createEventSemaphore( unsigned long *sema, char *tag );
void          deleteEventSemaphore( unsigned long sema );

unsigned long createQueue( unsigned long *sema, char *tag,
                           char **queueName=0 );
void          deleteQueue( unsigned long *sema );

int checkUsage( char *name );

unsigned long createMutexSemaphore( char *tag );
void          deleteMutexSemaphore( unsigned long sema );
int           queryMutexSemaphore( unsigned long sema );
char         *findMutexSemaphoreName( unsigned long sema );
void          lockMutexSemaphore( unsigned long sema );
void          unlockMutexSemaphore( unsigned long sema );
int           createNamedMutex( char *name, unsigned long *sema );

void brakeSemaphoreCreate( char *name );
void brakeSemaphoreRemove( char *name );
void brakeSemaphorePost(   char *name );
void brakeSemaphoreReset(  char *name );
void brakeSemaphoreWait(   char *name );

int busySemaphoreCreate( char *name );
int busySemaphoreRemove( char *name );
int busySemaphoreExist(  char *name );

class ResourceItem : public BaseObject
{
private:
  char *name;
  unsigned long handle;

  int equal( const BaseObject &t ) const;
  void print( ostream& s ) const;
  char *nameOfClass() const;

public:

  ResourceItem( unsigned long _h, char *_n );
  ~ResourceItem();

  char *getName() const { return name; }
  unsigned long getHandle() const { return handle; }
};

class ResourceList : public Collection
{
public:
  ResourceList();

  ResourceItem* Find( ResourceItem& item, LockMode lock=LockModeNeed ) const;
  ResourceItem& operator [] ( const int pos ) const;

  char *findResourceName( unsigned long handle );
};

class ResourceIterator : public CollectionIterator
{
public:
  ResourceIterator( ResourceList& ini );

  ResourceItem& operator () () const;
};

extern ResourceList mutexSemList,
                    eventSemList,
                    queueList;

void addResourceItem( ResourceList &list, unsigned long handle,
                             char *name );
void detachResourceItem( ResourceList &list, unsigned long handle );

#endif
