#if !defined(__MAILER_H)

#define __MAILER_H

#include "queue.h"

enum Compatibility
{

// Compatibility codes:

  ZMO = 0x00000001,         // Plain Zmodem
  ZAP = 0x00000002,         // ZedZap
  DZA = 0x00000004,         // DirectZedZap
  JAN = 0x00000008,         // Janus
  HYD = 0x00000010,         // Hydra
  NCP = 0x00000020,         // No compatible protocols (failure).
  NRQ = 0x00000040,         // No file requests accepted by this system.
  ARC = 0x00000080,         // ARCmail 0.60-capable, as defined by the FTSC.
  XMA = 0x00000100,         // Supports other forms of compressed mail.
  FNC = 0x00000200,         // Filename conversion. This indicates that any
	   		    //  transmitted files must follow the MS-DOS
			    //  restrictions of an eight character file name
			    //  followed by a three character extension;
			    //  eg. FILENAME.EXT

// Link - Calling system options:

  PUA = 0x00000400,         // Pickup mail for all presented addresses.
  PUP = 0x00000800,         // Pickup mail for primary address only.
  NPU = 0x00001000,         // No mail pickup desired.

// Link - Answering system options:

  HAT  = 0x00002000,         // Hold ALL traffic.
  HXT  = 0x00004000,         // Hold compressed mail traffic.
  HRQ  = 0x00008000,         // Hold file requests (not processed at this time).
  RH1  = 0x00010000,         // Hydra: Send request first
  COPY = 0x80000000          // Copy to directory
};

int callNode( QueueEntry&, int );
int answer( int, int );

int emsiHandShake( QueueEntryList **akaList, Compatibility myCode,
			  Compatibility& hisCode, QueueEntry* called,
			  char **password );

extern FIDOAddress *primaryAka;

extern char *hisPassword,
	    *mailerProductCode,
	    *mailerName,
	    *mailerVersion,
	    *mailerSerialNumber,
	    *hisName,
	    *hisPlace,
	    *hisSysOp,
	    *hisPhone,
	    *hisBaud,
	    *hisFlags,
            *mailerOwner,
            *sessionCost;

extern int remoteMailerCode;

extern unsigned long hisTime,
	             hisMOH;

extern long hisTZ;

extern unsigned long ourMail,
	             ourFiles,
	             hisMail,
	             hisFiles;

extern char bbsAtThisCall,
	    sessionProtected,
	    sessionListed,
            wasBusyEncountred;

extern QueueEntryList *illegalAka;

#pragma pack(1)
struct SessionInfo
{
  unsigned short zone, net, node, point;
  char type;
  char protocol;
  char crc;
  long elapsed;
  long netSend, netRecv,
       xmSend,  xmRecv,
       filesSend, filesRecv;
  char domain[17];
};
#pragma pack()

extern SessionInfo sessionBinaryInfo;
extern long volatile sessionStart;

int sendAka( QueueEntry& aka, int wazoo, int& fsent,
             Compatibility hisCode, char **msg,
             char *password, int answer, FILE *xfer, int needSendFreq );

int CopyFileToPath( char *fname, char *alias, char *path );

#endif
