#define INCL_DOS
#define INCL_DOSERRORS
#include <os2.h>

#include <string.h>

#include "misc.h"
#include "language.h"
#include "shmem.h"

SharedMemory::SharedMemory( char *nm, long sz ) :
  name( newStr( nm ) ),
  size( sz ),
  allocated( 0 )
{
  APIRET rc;
  int cnt(0);

  if( !size )
    size = 60*1024;

TryAgain:
  rc = DosGetNamedSharedMem( (PPVOID)&base, name,
                             PAG_READ|PAG_WRITE ); //|PAG_GUARD

  ++cnt;

  if( rc )
    if( rc!=ERROR_FILE_NOT_FOUND || cnt>1 )
    {
      log( Msg_ErrUnableGetShMem, name, rc );
      os2Message( rc );
    }
    else
    {
      rc = DosAllocSharedMem( (PPVOID)&base, name, size+sizeof(long),
                              PAG_READ|PAG_WRITE|PAG_COMMIT ); //|PAG_GUARD

      if( rc )
        if( rc!=ERROR_ALREADY_EXISTS )
        {
          log( Msg_ErrUnableAllocShMem, name, rc );
          os2Message( rc );
        }
        else
          goto TryAgain;
      else
        allocated = 1;
    }

  if( !rc )
  {
    if( allocated )
      *(long *)base = size;
    else
      size = *(long *)base;

    base = ((char *)base)+sizeof(long);
    current = base;
    end = base+size;
  }
  else
    current = base = end = 0;
}

SharedMemory::~SharedMemory()
{
  delete name;

  APIRET rc = DosFreeMem( base );

  if( rc )
  {
    log( Msg_ErrUnableFreeShMem, name, rc );
    os2Message( rc );
  }
}

void SharedMemory::seek( int p )
{
  if( p>=size )
    log( "~BUG~ SharedMemory::seek pointer out of range ptr=%lu size=%lu",
         p, size );
  else
    current = base+p;
}

int SharedMemory::write( void *ptr, long sz )
{
  if( current+sz>=end )
    return 1;

  memcpy( current, ptr, (size_t)sz );
  current += sz;
  return 0;
}

int SharedMemory::read( void *ptr, long sz )
{
  if( current+sz>=end )
    return 1;

  memcpy( ptr, current, (size_t)sz );
  current += sz;
  return 0;
}
