#include <process.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <io.h>
#include <share.h>
#include <fcntl.h>
#include <string.h>
#include <sys\stat.h>
#include <sys\types.h>
#include <ctype.h>
#include <builtin.h>

#define INCL_DOSEXCEPTIONS
#define INCL_DOSPROCESS
#define INCL_DOSERRORS
#define INCL_DOSFILEMGR
#define INCL_DOSSEMAPHORES
#include <os2.h>

#include "except.h"

#include "language.h"

#include "fidoaddr.h"
#include "misc.h"
#include "filetype.h"
#include "32base.h"
#include "direct.h"
#include "ea.h"
#include "zmodem.h"

void makeDay( char *day )
{
  static const char *days[7] = { "SU", "MO", "TH", "WE", "TU", "FR", "SA" };
  int i;

  if( *day )
    switch( day[2] )
    {
    case '9':

      day[2] = 'A';
      break;

    case 'Z':
    case 'z':

      for( i=0; i<7; i++ )
	if( !strnicmp( days[i], day, 2 ) )
	  break;

      if( i==7 )
      {
	log( Msg_SplHeap );
        return;
      }

      if( ++i==7 )
	i = 0;

      strcpy( day, days[i] );
      strcat( day, "0" );
      break;

    default:

      day[2]++;

    }
  else
  {
    time_t tim = time( NULL );
    struct tm *tm = localTime( &tim );

    strcpy( day, days[ tm->tm_wday ] );
    strcat( day, "0" );
  }
}

char *makeName( char *tBox, char *bundle )
{
  char *full( new char [ strlen(tBox)+strlen(bundle)+2 ] ),
	drive   [_MAX_DRIVE],
	dir     [_MAX_DIR],
	name    [_MAX_FNAME],
	day     [_MAX_EXT],
	firstDay[_MAX_EXT],
	firstFlag( 1 );

  _splitpath( bundle, drive, dir, name, day );
  strcpy( day, day+1 );
  strcpy( firstDay, day );

  for( ;; )
  {
    strcpy( full, tBox );
    strcat( full, name );
    strcat( full, "." );
    strcat( full, day );

    if( access( full, 0 ) )
      break;

    makeDay( day );

    if( firstFlag )
      firstFlag = 0;
    else
      if( !stricmp( firstDay, day ) )
      {
	delete full;
	full = NULL;
	break;
      }
  }

  return full;
}

char *getLongSuffix( char flavourChar, char *msg, char alt )
{
  char ch( toupper(flavourChar) );

  if( ch==alt )
    return flavourString( normalFlavour );

  switch( ch )
  {
  case 'I':
    return flavourString( immedFlavour );

  case 'C':
    return flavourString( crashFlavour );

  case 'D':
    return flavourString( directFlavour );

  case 'N':
    return flavourString( normalFlavour );

  default:
    log( "~SPLITTER~ %s unknown flavour 0x%02x -> hold", msg, flavourChar );

  // fall

  case 'H':
    break;
  }

  return flavourString( holdFlavour );
}

char move( char *name, char *newName, char flag, char *adr, long created,
           int& movedCount, int& movedSize, size_t fileSize )
{
  char needPostProcess( 1 ),
       rc( 0 );

  capitalize( name );
  capitalize( newName );
  makeFilePath( newName );

  if( flag=='#' && miscInfo.state&SplitDelete )
    flag='^';

  if( flag!='#' && flag!='^' )
  {
unable_to_rename:
    APIRET RC = DosCopy( (PSZ)name, (PSZ)newName, 0 );

    if( RC )
    {
      log( Msg_SplUnableCopy, name, newName, adr, RC );
      os2Message( RC );
    }
    else
      rc = 1;
  }
  else
  {
    if( rename( name, newName ) )
      goto unable_to_rename;

    if( flag=='^' )
      needPostProcess = 0;

    rc = 1;
  }

  if( rc )
  {
    if( needPostProcess )
      postProcessAttach( flag, name );

    setFileStamp( newName, created, 0 );

    movedCount++;
    movedSize += fileSize;
  }

  return rc;
}

char move1( char *name, char *newName, char flag,
            long created, char *adr,
            int& movedCount, int& movedSize, size_t fileSize )
{
  capitalize( name );
  capitalize( newName );
  makePath( newName );

  char *forName( newName+strlen(newName) );
  APIRET rc;
  long mark( getTimeMark() );

  for( int i=100000; i; --i )
  {
    sprintf( forName, "%08x.Pkt", mark++ );
    capitalize( forName );

    rc = DosCopy( (PSZ)name, (PSZ)newName, 0 );

    if( !rc )
      break;
  }

  if( !i )
  {
    log( Msg_SplUnableCopy, name, newName, adr, rc );
    os2Message( rc );
    return 0;
  }

  movedCount++;
  movedSize += fileSize;

  postProcessAttach( flag, name );

  setFileStamp( newName, created, 0 );

  return 1;
}

char moveFile( char *name, char *tBox, char flag,
 	       char flavourChar, long created,
 	       char longNames, char *adr,
               int& movedCount, int& movedSize, size_t fileSize )
{
  char flavour[128]( "." );

  if( longNames )
    strcat( flavour, getLongSuffix( flavourChar, "moveFile", 'F' ) );
  else
    *flavour = flavourChar;

  char *file( fileName( name ) ),
       *newName = (char *)_alloca( strlen(flavour)+
                                   strlen(tBox)+
                                   strlen(file)+
                                   3 );

  strcpy( newName, tBox );
  strcat( newName, flavour );
  strcat( newName, "\\" );
  strcat( newName, file );

  return move( name, newName, flag, adr, created, movedCount, movedSize, fileSize );
}

char moveEcho( char *name, char *tBox, char flag,
 	       char flavourChar, long created,
 	       char longNames, char *adr,
               int& movedCount, int& movedSize, size_t fileSize )
{
  char flavour[128]( "." );

  if( longNames )
    strcat( flavour, getLongSuffix( flavourChar, "moveEcho", 'F' ) );
  else
    *flavour = flavourChar;

  char *boxName = (char *)_alloca( strlen(flavour)+strlen(tBox)+3 );

  strcpy( boxName, tBox );
  strcat( boxName, flavour );
  strcat( boxName, "\\" );

  char *newName = makeName( boxName, name );
  char rc( 0 );

  if( newName==NULL )
    log( Msg_SplUnable, fileName( name ), adr );
  else
    rc = move( name, newName, flag, adr, created,
               movedCount, movedSize, fileSize ); 

  delete newName;
  return rc;
}

char movePkt( char *name, char *tBox, char flag,
 	      char flavourChar, long created,
 	      char longNames, char *adr,
              int& movedCount, int& movedSize, size_t fileSize )
{
  char flavour[128]( "." );

  if( longNames )
    strcat( flavour, getLongSuffix( flavourChar, "movePkt", 'F' ) );
  else
    *flavour = flavourChar;

  char *newName = (char *)_alloca( strlen(flavour)+strlen(tBox)+2 );

  strcpy( newName, tBox );
  strcat( newName, flavour );
  strcat( newName, "\\" );

  return move1( name, newName, flag, created, adr,
                movedCount, movedSize, fileSize );
}

struct LoLineArgs
{
  int& filesCount;
  int& movedSize;
  int& movedCount;
  char longNames;
  int& illegal;
  int& badRemoved;
  char *adr;
  char *tBox;
  char flavour;
};

int processLoLine( char *name, char flag, void *args )
{
  FileType type;

  msgOfFileType( name, &type );

  switch( type )
  {
  case FileXMail:
    if( !miscInfo.splitSize )
     return VERY_SPEC_COND;

    break;

  case FileMail:
    if( !(miscInfo.state&SplitPKTFromLO) || !miscInfo.splitPktSize )
       return VERY_SPEC_COND;

    break;

  default:
    if( miscInfo.splitFileSize )
       return VERY_SPEC_COND;
  }
  
  LoLineArgs *p( (LoLineArgs *)args );
  struct stat st;
  int ret( VERY_SPEC_COND );
  
  if( stat( name, &st ) )
  {
    ++p->illegal;

    if( errno==ENOENT )
    {
      if( (miscInfo.state2&SplitDelNoFound)==SplitDelNoFound )
      {
        ret = SPEC_COND;
        ++p->badRemoved;
      }
  
      log( Msg_SplNotFound, name, p->adr,
           ret==SPEC_COND ? Msg_SplNotFoundDeleted : Msg_SplNotFoundNotDeleted );
    }
    else
      log( Msg_SplStatFail, name, p->adr, strError() );

    return ret;
  }

  if( type==FileXMail )				//FileXMail
  {
    p->filesCount++;
  
    if( st.st_size>miscInfo.splitSize )
      if( moveEcho( name, p->tBox, flag, p->flavour,
                    st.st_mtime, p->longNames, p->adr,
                    p->movedCount, p->movedSize, st.st_size ) )
        ret = OK;
  }
  else if( type==FileMail )			//FileMail
  {
    p->filesCount++;

    if( st.st_size>miscInfo.splitPktSize )
      if( movePkt( name, p->tBox, flag, p->flavour,
                   st.st_mtime, p->longNames, p->adr,
                   p->movedCount, p->movedSize, st.st_size ) )
        ret = OK;
  }
  else 						//FileFile
  {
    p->filesCount++;
  
    if( st.st_size<miscInfo.splitFileSize )
      if( moveFile( name, p->tBox, flag, p->flavour,
                    st.st_mtime, p->longNames, p->adr,
                    p->movedCount, p->movedSize, st.st_size ) )
        ret = OK;
  }

  return ret;
}


void processBundle( char *path, char *name,
		    uShort zone, uShort net,
		    uShort node, uShort point,
		    int& filesCount,
		    int& movedSize,
		    int& movedCount,
		    char longNames,
                    int& illegal, int& badRemoved, Domain *dom,
                    Flavour flavourOrig )
{
  if( !miscInfo.splitSize )
    return;

  char *fullName = (char *)_alloca( strlen(path)+strlen(name)+1 );
  struct stat st;

  strcpy( fullName, path );
  strcat( fullName, name );
  capitalize( fullName );

  if( stat( fullName, &st ) || !st.st_size )
    return;

  FIDOAddress adr( zone, net, node, point, dom->getName() );
  char buf[65],
       *pref,
       *suff;

  if( longNames )
  {
    pref = miscInfo.longBox;
    suff = adr.getLongName();
  }
  else
  {
    pref = miscInfo.tbox;
    suff = adr.getTBoxName();
  }

  adr.toString( buf, 64 );

  if( pref==NULL || suff==NULL )
  {
    delete suff;
    log( Msg_SplUnableProcess, buf );
    return;
  }

  char *tBox = (char *)_alloca( strlen(pref)+strlen(suff)+1 );

  strcpy( tBox, pref );
  strcat( tBox, suff );
  capitalize( tBox );

  char flavour;

  if( flavourOrig==unknownFlavour )
    flavour = name[9];
  else
  {
    char fl[16];

    FIDOAddress::updateLoName( flavourOrig, fl );
    flavour = fl[1];
  }

  if( adr.makeBsy( 1 ) )
    log( Msg_SplBusy, buf );
  else
  {
    FLO xlo( fullName );
    LoLineArgs args =
    {
      filesCount,
      movedSize,
      movedCount,
      longNames,
      illegal,
      badRemoved,
      buf,
      tBox,
      flavour
    };

    xlo.process( 0, processLoLine, &args );

    adr.makeBsy( 0 );
  }

  delete suff;
}

void processPktBundle( char *path, char *name,
	   	       uShort zone, uShort net,
		       uShort node, uShort point,
		       int& filesCount,
		       int& movedSize,
		       int& movedCount,
		       char longNames, Domain *dom,
                       Flavour flavourOrig )
{
  if( !miscInfo.splitPktSize )
    return;

  char *fullName = (char *)_alloca( strlen(path)+strlen(name)+1 );
  struct stat st;

  strcpy( fullName, path );
  strcat( fullName, name );
  capitalize( fullName );

  if( stat( fullName, &st ) || st.st_size<miscInfo.splitPktSize )
    return;

  FIDOAddress adr( zone, net, node, point, dom->getName() );
  char buf[65],
       *pref,
       *suff;

  if( longNames )
  {
    pref = miscInfo.longBox;
    suff = adr.getLongName();
  }
  else
  {
    pref = miscInfo.tbox;
    suff = adr.getTBoxName();
  }

  adr.toString( buf, 64 );

  if( pref==NULL || suff==NULL )
  {
    delete suff;
    log( Msg_SplUnableProcess, buf );
    return;
  }

  char *tBox = (char *)_alloca( strlen(pref)+strlen(suff)+32 );

  strcpy( tBox, pref );
  strcat( tBox, suff );

  char flavour[128]( "." );

  if( longNames )
    if( flavourOrig!=unknownFlavour )
      strcat( flavour, flavourString( flavourOrig ) );
    else
      strcat( flavour, getLongSuffix( name[9], "splitPktBundle", 'O' ) );
  else
  {
    char flav;

    if( longNames )
    {
      FIDOAddress::updateLoName( flavourOrig, flavour );
      flav = flavour[1];
      flavour[1] = 0;
    }
    else
      flav = name[9];

    flav = toupper( flav );

    if( flav!='O' && flav!='N' )
      *flavour = flav;
    else
      *flavour = 0;
  }

  strcat( tBox, flavour );
  strcat( tBox, "\\" );

  if( adr.makeBsy( 1 ) )
    log( Msg_SplBusy, buf );
  else
  {
    move1( fullName, tBox, '^', st.st_mtime, buf,
           movedCount, movedSize, st.st_size );
    adr.makeBsy( 0 );
  }

  filesCount++;

  delete suff;
}

void processPoint( uShort zone, uShort net, uShort node,
		   char *fullPath, char *nodePath,
		   int& filesCount, int& movedSize,
		   int& movedCount, char longNames,
                   int& illegal, int& badRemoved, Domain *dom )
{
  char *mask( "\\*" ),
       *path = (char *)_alloca( strlen(fullPath)+strlen(nodePath)+strlen(mask)+1 ),
       *full = (char *)_alloca( strlen(fullPath)+strlen(nodePath)+2 );

  strcpy( path, fullPath );
  strcat( path, nodePath );
  strcat( path, mask );

  strcpy( full, fullPath );
  strcat( full, nodePath );
  strcat( full, "\\" );

  for( Directory rc( path ); rc; rc++ )
  {
    uShort point( 0 ), fict( 0 );
    char *name = rc.getName();

    if( strlen( name )==12 && split16base( name, fict, point ) )
      switch( isBundle( name ) )
      {
      case 2:
	processBundle( full, name, zone, net, node, point,
		       filesCount, movedSize, movedCount, longNames,
                       illegal, badRemoved, dom, unknownFlavour );
        break;

      case 1:
	processPktBundle( full, name, zone, net, node, point,
		          filesCount, movedSize, movedCount, longNames, dom,
                          unknownFlavour );
        break;
      }
  }
}

void processZone( int outLen, char *zoneStr,
		  int& filesCount, int& movedSize,
		  int& movedCount, Domain *dom,
 		  char longNames,
                  int& illegal, int& badRemoved )
{
  uShort zone( 0 );
  char ext[_MAX_EXT];

  {                            //  ᪮ ᯮ짮 fnsplit
    char drive[_MAX_DRIVE],
	 dir[_MAX_DIR],
	 file[_MAX_FNAME];

    _splitpath( zoneStr, drive, dir, file, ext );

    if( *ext )
    {
      if( val16base( ext+1, zone ) )
	return;
    }
    else
      zone = dom->getDefZone();
  }

  char mask[]( "\\*" ),
    *path = (char *)_alloca( outLen+strlen(ext)+sizeof( mask ) ),
    *full = (char *)_alloca( outLen+strlen(ext)+20 ),
    *outbound( dom->getOutbound() );

  strcpy( path, outbound );
  strcpy( path+outLen-1, ext );
  strcat( path, mask );

  strcpy( full, outbound );
  strcpy( full+outLen-1, ext );
  strcat( full, "\\" );

  for( Directory rc( path,
                     SearchAttributes( MayBeDirectory|MayBeReadOnly|MayBeArchived ) );
                 rc; rc++ )
  {
    char *name = rc.getName();
    uShort net( 0 ), node( 0 );

    if( strlen( name )==12 && split16base( name, net, node ) )
      if( (rc.getAttrib()&MayBeDirectory)==MayBeDirectory &&
          !stricmp( name+8, ".pnt" ) )
	processPoint( zone, net, node, full, name,
		      filesCount, movedSize, movedCount,
                      longNames,
                      illegal, badRemoved, dom );
      else
	switch( isBundle( name ) )
        {
        case 2:
	  processBundle( full, name, zone, net, node, 0,
			 filesCount, movedSize, movedCount,
                         longNames,
                         illegal, badRemoved, dom, unknownFlavour );
          break;

        case 1:
	  processPktBundle( full, name, zone, net, node, 0,
		   	    filesCount, movedSize, movedCount,
                            longNames, dom, unknownFlavour );
          break;
        }
  }

  delete outbound;
}

void processLongSplit( int& filesCount, int& movedSize,
                       int& movedCount, char longNames,
                       int& illegal, int& badRemoved )
{
  if( miscInfo.longOutbound==NULL )
    return;

  static char mask[]( "*" );
  char *path = (char *)_alloca( strlen(miscInfo.longOutbound)+sizeof(mask) );

  strcpy( path, miscInfo.longOutbound );
  strcat( path, mask );

  for( Directory rc( path, SearchAttributes( MayBeReadOnly|MayBeArchived ) );
      rc; rc++ )
  {
    char *name = newStr( rc.getName() );
    uShort zone, net, node, point;
    char *domain;
    LongNameType type;
    Flavour fl;

    if( splitLongName( name, zone, net, node, point, &domain, fl, type ) )
    {
      FIDOAddress cur( zone, net, node, point, domain );

      if( type==mailLongName )
      {
        if( fl==unknownFlavour )
          log( "~SPLITTER~ processLongSplit: fl==unknownFlavour (mail)" );
        else
          processPktBundle( rc.getPath(), rc.getName(), zone, net, node, point,
                            filesCount, movedSize, movedCount,
                            longNames, cur.getDomainPtr(), fl );
      }
      else if( type==listLongName )
      {
        if( fl==unknownFlavour )
          log( "~SPLITTER~ processLongSplit: fl==unknownFlavour (list)" );
        else
          processBundle( rc.getPath(), rc.getName(), zone, net, node, point,
                         filesCount, movedSize, movedCount,
                         longNames,
                         illegal, badRemoved, cur.getDomainPtr(), fl );
      }
    }

    delete name;
  }
}


void processSplit( char longNames )
{
  long start( getTimeMark() );
  int filesCount( 0 ),
      movedSize( 0 ),
      movedCount( 0 ),
      illegal( 0 ),
      badRemoved( 0 );

  static char mask[]( ".*" );

  for( DomainIterator list( *outboundInfo.domains ); list; list++ )
  {
    Domain *dom = &list();
    char *outbound( dom->getOutbound() );
    int outLen( strlen(outbound) );
    char *path( new char[ outLen+sizeof(mask) ] );

    strcpy( path, outbound );
    strcpy( path+outLen-1, mask );

    for( Directory rc( path,
                       SearchAttributes( MustBeDirectory|MayBeReadOnly|MayBeArchived ) );
                   rc; rc++ )
    {
      char *name = rc.getName();

      if( (rc.getAttrib()&MayBeDirectory)==MayBeDirectory &&
          strcmp( name, "." ) && strcmp( name, ".." ) )
	  processZone( outLen, name, filesCount, movedSize, movedCount,
                       dom, longNames, illegal, badRemoved );
    }

    delete outbound;
    delete path;
  }

  processLongSplit( filesCount, movedSize, movedCount, longNames,
                    illegal, badRemoved );

  if( movedCount || illegal )
  {
    char *size = strBig( movedSize );

    start = getTimeMark() - start;

    if( illegal )
      log( Msg_SplBundStat1, illegal, badRemoved );

    log( Msg_SplBundStat,
	 start/60, start%60, filesCount, movedCount, size );

    delete size;
  }
}

// -------------------------- Manage

static HEV doSplit = 0xffffffff;
static char volatile notInProgress = 1;
static char volatile mayRun = 1;
static long splitterTID = -1;

void _Optlink echoSplitter( void *ppp )
{
  registerExceptionHandler

  DosSetPriority( PRTYS_THREAD, 0, -miscInfo.priorityDelta, 0 );

  ULONG cnt;
  APIRET rc = DosResetEventSem( doSplit, &cnt );

  for( ;; )
  {
    DosWaitEventSem( doSplit, miscInfo.splitCircle*1000 );

    if( !mayRun )
    {
      deRegisterExceptionHandler
      return;
    }

    rc = DosResetEventSem( doSplit, &cnt );

    if( rc && rc!=ERROR_ALREADY_RESET )
    {
      log( Msg_SemSplNotReseted, rc );
      os2Message( rc );
    }

    notInProgress = 0;
    processSplit( (char)(long)ppp );
    notInProgress = 1;
  }
}

int isSplitterActive()
{
  return !notInProgress;
}

void splitterForce()
{
  if( doSplit==0xffffffff )
    return;

  APIRET rc = DosPostEventSem( doSplit );

  if( rc && rc!=ERROR_ALREADY_POSTED )
  {
    log( Msg_SemSplNotPosted, rc );
    os2Message( rc );
  }
}

void splitterDone()
{
  mayRun = 0;
  splitterForce();
  DosWaitThread( (PTID)&splitterTID, DCWW_WAIT );
  deleteEventSemaphore( doSplit );
  doSplit = 0xffffffff;
}

void startSplitter()
{
  if( !miscInfo.splitSize && !miscInfo.splitPktSize && !miscInfo.splitFileSize )
    return;

  if( checkUsage( "Task\\Splitter.Is.Active" ) )
  {
    log( Msg_ErrActiveSplitter );
    return;
  }

  long longNames = (miscInfo.state2&SplitterLong)==SplitterLong ? 1 : 0;

  if( longNames )
  {
    if( miscInfo.longBox==NULL )
    {
      log( Msg_SplTerminate, "LongBoxes" );
      return;
    }
  }
  else
    if( miscInfo.tbox==NULL )
    {
      log( Msg_SplTerminate, "TBoxes" );
      return;
    }

  APIRET rc = createEventSemaphore( &doSplit, "DoSplit" );

  if( rc )
  {
    log( Msg_SplErrBegin, rc );
    os2Message( rc );
    return;
  }

  splitterForce();

  if( miscInfo.splitCircle<1 )
    miscInfo.splitCircle = 1;

  splitterTID = _beginthread( echoSplitter, 0, 40960, (void *)longNames );

  if( splitterTID==-1 )
    log( Msg_ErrEchoSpl, strError() );
  else
    log( Msg_ThrEchoSpl, splitterTID );
}
