#define INCL_DOSPROCESS
#define INCL_DOSEXCEPTIONS
#define INCL_DOSDATETIME
#define INCL_WINSHELLDATA
#include <os2.h>

#include <stdlib.h>
#include <ctype.h>
#include <io.h>
#include <fcntl.h>
#include <share.h>
#include <process.h>
#include <sys\stat.h>
#include <conio.h>
#include <builtin.h>

#include "language.h"

#include "except.h"

#include "misc.h"
#include "fidoaddr.h"
#include "nodelist.h"
#include "timer.h"
#include "pair.h"
#include "buffer.h"
#include "key.h"
#include "zmodem.h"
#include "hydra.h"
#include "zdialog.h"
#include "execute.h"
#include "rexx.h"
#include "rexxhook.h"
#include "dinamic.h"

static int oldSecs = 0, oldMove = 0;
static Window *pauseWnd = NULL;
static int oldX = 17,
	   oldY = 10;

volatile char terminalLocalEcho = 0,
              terminalNeedTranslate = 1,
              terminalStopOutput = 0;

static volatile char wasTimedOut( 0 );
static Timer timeOut;

static volatile int inSuspendMode = 0;
static volatile int inReading = 0;
static volatile int needForAbort = 0;

void suspendReader()
{
  inSuspendMode = 1;

  while( inReading )
    checkPoint();
}

void resumeReader()
{
  inSuspendMode = 0;
}

void terminalWriteToScreen( int chr );

void _Optlink terminalReader( void * )
{
  registerExceptionHandler

  long connected( 0 );
  char wasRinging = 0;

  inSuspendMode = 0;
  inReading = 0;
  needForAbort = 0;

  while( 1 )
  {
    if( inSuspendMode )
    {
      checkPoint();
      continue;
    }

    if( needForAbort )
      break;

    if( comPort->isOnline() )
    {
      if( !connected )
      {
        execSchedule( melodies.connect );
        connected = getTimeMark();
        addTitleMode( Msg_TtlConnected );
        changeTitle( NULL );
      }
    }
    else if( connected )
    {
      execSchedule( melodies.endSess );
      setTitleMode( Msg_TtlTerminal );
      changeTitle( NULL );

      char *str = strTime( getTimeMark()-connected, 0 );

      log( Msg_InfoWasOnline, str );
      delete str;
      connected=0;
//      timeOut.startTimer( miscInfo.terminalTimeOut*1000 );
    }
    else if( miscInfo.terminalTimeOut && timeOut.isTimeOut() )
    {
      wasTimedOut = 1;
      terminalActive = 0;
      continue;
    }

    if( (miscInfo.state3&HardRing)==HardRing )
      if( comPort->isRinging() )
      {
        if( !wasRinging )
        {
          comPort->ringSignal();
          wasRinging = 1;
        }
      }
      else
        wasRinging = 0;

    if( inSuspendMode )
    {
      checkPoint();
      continue;
    }

    inReading = 1;

    register int chr( comPort->get() );

    inReading = 0;

    if( chr!=-1 && !terminalStopOutput )
      terminalWriteToScreen( chr );
  }

  deRegisterExceptionHandler
}

void messageBox( char *fmti ... )
{
  va_list argptr;
  char buf[ 4096 ];

  va_start( argptr, fmti );
  vsprintf( buf, fmti, argptr );
  va_end( argptr );

  Z_InformUser( Msg_ScrError, buf, Msg_ScrContinue, NULL );
}

void waitIsTerm( int secs )
{
  if( !oldSecs )
    oldMove = secs;

  if( oldSecs!=secs )
  {
    char *msg = (char *)_alloca( strlen(Msg_TrmNextTry)+32 );

    sprintf( msg, Msg_TrmNextTry, secs );
    oldSecs = secs;
    (*pauseWnd)(1,2).clearLine();
    pauseWnd->centre( 2, msg  );

    if( secs+1 < oldMove )
    {
      oldMove = secs;
      oldX=random(Z_ScreenWidth-pauseWnd->getWidth()-4)+1;
      oldY=random(Z_ScreenHeight-pauseWnd->getHeight()-3)+1;
      pauseWnd->move( oldX, oldY );
    }
  }
}

int termPause()
{
  if( !miscInfo.redial )
    return 0;

  Timer t( long(miscInfo.redial)*1000l );
  oldSecs = 0 ;

  char buf[32];

  sprintf( buf, "%d", miscInfo.redial );

  pauseWnd = new Window( oldX, oldY,
                         strlen(Msg_TrmNextTry)+strlen(buf)+4, 5,
                         popUpColor,
			 Msg_TrmPause, BS_SINGLELINE, DS_LOWERRIGHT );

  +*pauseWnd;
  waitIsTerm( t.secsLeft() );

  int done( 0 );

  while( !t.isTimeOut() && !comPort->haveInput() && !done )
  {
    switch( abortAction( 1 ) )
    {
    case 0x20:
      break;

    case 27:
      done = 1;
      break;

    default:
      waitIsTerm( t.secsLeft() );
    }
  }

  delete pauseWnd;
  return done;
}

void termCmdDial()
{
  initFileInfo( 0 );

  int i( stringEnter( sizeof(terminalNumber), terminalNumber,
		      Msg_TrmAddrPhone ) || !*terminalNumber );

  terminalWindow.setPhysicalCursor( 1 );

  if( i )
    return;

  initFileInfo( 1 );

  if( terminalNumber[0]=='@' )
  {
    suspendReader();
    doTrivialRexx( terminalNumber+1 );
    resumeReader();
    return;
  }

  char *phone( NULL ),
       *pref( NULL );
  int wait( 0 );

  if( terminalNumber[0]=='#' )
    phone = newStr( terminalNumber+1 );
  else
  {
    NodeConfig *entered( new NodeConfig( terminalNumber ) );

    if( !entered->isValid() )
      delete entered;
    else
    {
      if( systemList!=NULL )
      {
	NodeConfig *found( systemList->Find( *entered ) );

	if( found!=NULL )
	{
          delete entered;
          entered = new NodeConfig( found );
	}
      }

      wait = entered->getWaitCarrier();
      pref = newStr( entered->getDial() );
      phone = entered->getPhone();

      if( phone!=NULL )
        phone = entered->getTranslatedPhone( entered->getPhone() );

      delete entered;
    }
  }

  if( phone==NULL )
  {
    messageBox( Msg_TrmUnkPhone, terminalNumber );
    return;
  }

  -screenCursor;

  suspendReader();

  for( int done(0); !done; )
    switch( comPort->dial( phone, pref, NULL, wait, 1 ) )
    {
    case 2:
    case 3:
      done = 1;
      break;

    case 7:
      break;

    default:
      if( termPause() )
	done = 1;
    }

  delete phone;
  delete pref;

  resumeReader();
  +screenCursor;
  sleep( 500 );
}

volatile char terminalActive = 0;
volatile int terminalCapture = 0;
short terminalInXLAT[256];
short terminalOutXLAT[256];

char *virtualInput = NULL;

const ansiEscPars = 10;

static int ansiEscSeqFlag = 0;
static int ansiEscP[ansiEscPars];
static int& ansiEscP1 = ansiEscP[0];
static int& ansiEscP2 = ansiEscP[1];
static Buffer ansiEscSeqSave;

static int savedX = 1,
	   savedY = 1;
static int savedXwork = 0,
	   savedYwork = 0;
Buffer matchedSymbols;

void ansiEscSeq()
{
  ansiEscSeqFlag = 1;
  memset( ansiEscP, 0, sizeof(ansiEscP) );
  ansiEscSeqSave.reset();
  ansiEscSeqSave.putInto( 033 );
  ansiEscSeqSave.putInto( '[' );
  terminalActive = 2;
}

void termTab()
{
  register int x( terminalWindow.getCurrentX() );
  static char str[ 10 ];

  for( register int xn( 1 ); x>=xn; xn += 8 ) {}

  for( int ch = 0; x<xn; x++ )
    str[ch++] = ' ';

  str[ch]=0;
  terminalWindow << str;
}

void termClear()
{
  --terminalWindow;
  terminalWindow( 1, 1 );
}

static char *termInbound;

inline void getTermInbound()
{
  termInbound = miscInfo.terminalInbound==NULL ?
                miscInfo.inbound : miscInfo.terminalInbound;
}

void downZModem( int )
{
  getTermInbound();
  changeTitle( Msg_TtlDownLoad );
  log( Msg_TrmGetZ );
  suspendReader();

  ReceivedFileList list;

  log( Msg_TrmGetZres,
       Get_Zmodem( termInbound, list, Zmodem )==OK ? "" : Msg_TrmGetZres1 );

  resumeReader();
  execSchedule( melodies.abort );
  changeTitle( NULL );
}

void downZModem1()
{
  downZModem( 0 );
}

void downHydra( int )
{
  getTermInbound();
  changeTitle( Msg_TtlDownLoad );
  log( Msg_TrmGetH );

  ReceivedFileList list;

  HydraInit( 0, 0, 0, termInbound, list, 0, NULL );
  suspendReader();
  log( Msg_TrmGetZres,
     Hydra( NULL, NULL )==XFER_ABORT ? Msg_TrmGetZres1 : "" );
  resumeReader();
  execSchedule( melodies.abort );
  HydraDeInit();
  changeTitle( NULL );
}

void termDownLoad()
{
  if( !comPort->isOnline() )  //check
    messageBox( Msg_TrmNCDGet );
  else
  {
    MENU_ITEM downMenuItems[] =
    {
      "ZModem", MIF_CALLFUNC, k_Alt_Z, 0, 1, 1, downZModem, "",
      "Hydra ", MIF_CALLFUNC, k_Alt_H, 0, 1, 2, downHydra, "",
      NULL, 0, 0, 0, 0, 0, NULL, ""
    };

    Color buttonHig( colorSet.popUpBold );

    MENU downMenu =
    {
      (Z_ScreenWidth-10)/2+1, (Z_ScreenHeight-4)/2+1, 4, 10,
      8,
      MMF_VERTMENU,
      popUpColor.fore, popUpColor.back,
      buttonHig.fore, buttonHig.back,
      buttonHig.fore, buttonHig.back,
      popUpColor.fore, popUpColor.back,
      1, 1, 1,
      BS_SINGLELINE, popUpColor.fore, popUpColor.back,
      DS_LOWERRIGHT, 0x07,
      "",
      downMenuItems
    };

    -screenCursor;
    Z_HandleMenu( &downMenu );
    +screenCursor;
  }
}

void termIgnore()
{
  terminalActive = 2;
}

static int termSavedX = 1,
           termSavedY = 1;

void termSaveXY()
{
  terminalActive = 2;
  termSavedX = terminalWindow.getCurrentX();
  termSavedY = terminalWindow.getCurrentY();
}

void termRestXY()
{
  terminalActive = 2;
  terminalWindow( termSavedX, termSavedY );
  terminalWindow.setPhysicalCursor( 1 );
}

void termBell()
{
  execSchedule( melodies.bell );
}

void termBack()
{
  int x = terminalWindow.getCurrentX() - 1;

  if( x<=0 )
    x = 1;

  terminalWindow( x, terminalWindow.getCurrentY() );
}

Pair ansiTabInternal[] =
{
  Pair( "\033[", ansiEscSeq ),
  Pair( "\014", termClear ),
  Pair( "\t", termTab ),
  Pair( "\7", termBell ),
  Pair( "\b", termBack ),
  Pair( "\033#6", termIgnore ),
  Pair( "\033#5", termIgnore ),
  Pair( "\033#4", termIgnore ),
  Pair( "\033#3", termIgnore ),
  Pair( "\0338", termRestXY ),
  Pair( "\0337", termSaveXY ),
  Pair( "\033Z", termIgnore ),
  Pair( "\033D", termIgnore ),
  Pair( "\033M", termIgnore ),
  Pair( "\033=", termIgnore ),
  Pair( "\033>", termIgnore ),
  Pair( "\033E", termIgnore ),
  Pair( "\033H", termIgnore ),
  Pair( "\033(A", termIgnore ),
  Pair( "\033(B", termIgnore ),
  Pair( "\033)A", termIgnore ),
  Pair( "\033)B", termIgnore ),
  Pair( "\033(0", termIgnore ),
  Pair( "\033)0", termIgnore ),
  Pair( "\033(1", termIgnore ),
  Pair( "\033)1", termIgnore ),
  Pair( "\033(2", termIgnore ),
  Pair( "\033)2", termIgnore )
//  Pair( "\030B00", downZModem1 )
};

int ansiHowInternal = sizeof( ansiTabInternal )/sizeof( Pair );

Pair *ansiTab = ansiTabInternal;
int ansiHow = ansiHowInternal;

void resetMatched()
{
  matchedSymbols.reset();

  for( int j(0); j<ansiHow; j++ )
    ansiTab[j].reset();
}

void saveXY()
{
  savedXwork = terminalWindow.getCurrentX();
  savedYwork = terminalWindow.getCurrentY();
}

void restXY()
{
  terminalWindow( savedXwork, savedYwork );
}

int ansiEscSeqTreat( int ch )
{
  ansiEscSeqSave.putInto( (char)ch );

  int i, j;
  Color makedColor( terminalWindow.getColor() );

  switch( ch )
  {
  case 'P':		// ESC[<n>P character delete
  case '@':		// ESC[<n>@ character insert
  case 'M':		// ESC[<n>M line delete
  case 'L':		// ESC[<n>L line insert
  case 'r':		// ESC[<tr>;<br>r set scrolling region
  case 'g':		// ESC[<n>g n==0 tab clear, n==3 tab clear all
  case 'l':		// ESC[<n>l terminal mode reset
  case 'h':		// ESC[<n>h terminal mode set
  case 'x':		// ESC[{sol}x terminal parameters request ???
  case 'c':		// ESC[c teransmit device attributes
//    goto illegalSequence;
    break;

  case 'm':

    for( j=0; j<ansiEscSeqFlag; j++ )
      switch( ansiEscP[j] )
      {
      case 0:
        makedColor = terColor;
        break;

      case 1:
        makedColor.fore |= 0x8;
        break;

      case 2:
        makedColor.fore &= ~0x8;
        break;

      case 3:						//italics
        makedColor.fore = Color( (makedColor.fore&0x8) | BLUE );
        break;

      case 4:						//underline
        makedColor.fore = Color( (makedColor.fore&0x8) | BLUE );
        break;

      case 5:
      case 6:						//fast blink
        makedColor.back |= 0x8;
        break;

      case 7:
        makedColor.back = Color( terColor.fore&0x7 );
        makedColor.fore = Color( terColor.back&0x7 );
        break;

      case 8:
        makedColor.fore = Color( terColor.back&0x7 );
        makedColor.back = Color( terColor.back&0x7 );
        break;

      case 30:
        makedColor.fore = Color( (makedColor.fore&0x8) | BLACK );
        break;

      case 31:
        makedColor.fore = Color( (makedColor.fore&0x8) | RED );
        break;

      case 32:
        makedColor.fore = Color( (makedColor.fore&0x8) | GREEN );
        break;

      case 33:
        makedColor.fore = Color( (makedColor.fore&0x8) | BROWN );
        break;

      case 34:
        makedColor.fore = Color( (makedColor.fore&0x8) | BLUE );
        break;

      case 35:
        makedColor.fore = Color( (makedColor.fore&0x8) | MAGENTA );
        break;

      case 36:
        makedColor.fore = Color( (makedColor.fore&0x8) | CYAN );
        break;

      case 37:
        makedColor.fore = Color( (makedColor.fore&0x8) | WHITE );
        break;

      case 40:
        makedColor.back = Color( (makedColor.back&0x8) | BLACK );
        break;

      case 41:
        makedColor.back = Color( (makedColor.back&0x8) | RED );
        break;

      case 42:
        makedColor.back = Color( (makedColor.back&0x8) | GREEN );
        break;

      case 43:
        makedColor.back = Color( (makedColor.back&0x8) | BROWN );
        break;

      case 44:
        makedColor.back = Color( (makedColor.back&0x8) | BLUE );
        break;

      case 45:
        makedColor.back = Color( (makedColor.back&0x8) | MAGENTA );
        break;

      case 46:
        makedColor.back = Color( (makedColor.back&0x8) | CYAN );
        break;

      case 47:
        makedColor.back = Color( (makedColor.back&0x8) | WHITE );
        break;

      default:				//ignore illegal
 //       goto illegalSequence;
        break;
      }

    terminalWindow = makedColor;
    break;

  case 's':
    savedX = terminalWindow.getCurrentX();
    savedY = terminalWindow.getCurrentY();
    break;

  case 'u':
    terminalWindow( savedX, savedY );
    break;

  case 'K':
    if( !ansiEscP1 )
      terminalWindow.clearLine();
    else
      goto illegalSequence;
    break;

  case 'J':
    switch( ansiEscP1 )
    {
    case 0:
      saveXY();
      terminalWindow.clearLine();
      for( i = savedYwork+1; i<=terminalWindow.getHeight(); i++ )
	terminalWindow( i ).clearLine();
      restXY();
      break;

    case 1:
      saveXY();
      for( i = 1; i<savedYwork; i++ )
	terminalWindow( i ).clearLine();

      terminalWindow( savedYwork );

      for( i=1; i<savedXwork; i++ )
	terminalWindow << char( ' ' );
      break;

    case 2:
      --terminalWindow;
      terminalWindow( 1, 1 );
      break;

    default:
      goto illegalSequence;
    }
    break;

  case 'H':
  case 'f':
    if( !ansiEscP1 )
      ansiEscP1 = 1;

    if( ansiEscP1>terminalWindow.getHeight() )
      goto illegalSequence;

    if( !ansiEscP2 )
      ansiEscP2 = 1;

    if( ansiEscP2>terminalWindow.getWidth() )
      goto illegalSequence;

    terminalWindow( ansiEscP2, ansiEscP1 );
    break;

  case 'A':
    if( !ansiEscP1 )
      ansiEscP1 = 1;

    ansiEscP1 = terminalWindow.getCurrentY() - ansiEscP1;

    while( ansiEscP1<=0 )
      ansiEscP1 += terminalWindow.getHeight();

    terminalWindow( terminalWindow.getCurrentX(), ansiEscP1 );
    break;

  case 'B':
    if( !ansiEscP1 )
      ansiEscP1 = 1;

    ansiEscP1 = terminalWindow.getCurrentY() + ansiEscP1;

    while( ansiEscP1>terminalWindow.getHeight() )
      ansiEscP1 -= terminalWindow.getHeight();

    terminalWindow( terminalWindow.getCurrentX(), ansiEscP1 );
    break;

  case 'C':
    if( !ansiEscP1 )
      ansiEscP1 = 1;

    ansiEscP1 = terminalWindow.getCurrentX() + ansiEscP1;

    while( ansiEscP1>terminalWindow.getWidth() )
      ansiEscP1 -= terminalWindow.getWidth();

    terminalWindow( ansiEscP1, terminalWindow.getCurrentY() );
    break;

  case 'D':
    if( !ansiEscP1 )
      ansiEscP1 = 1;

    ansiEscP1 = terminalWindow.getCurrentX() - ansiEscP1;

    while( ansiEscP1<=0 )
      ansiEscP1 += terminalWindow.getWidth();

    terminalWindow( ansiEscP1, terminalWindow.getCurrentY() );
    break;

  case 'n':
    if( ansiEscP1==6 )
    {
      static char buf[16];

      sprintf( buf, "%c[%d;%dR", 27,
               terminalWindow.getCurrentY(), terminalWindow.getCurrentX() );
      virtualInput = buf;
    }
    else
      goto illegalSequence;
    break;

  case ';':
    if( ++ansiEscSeqFlag==ansiEscPars+1 )
      goto illegalSequence;

    return 0;

  default:
    if( isdigit(ch) )
      if( ansiEscSeqFlag!=ansiEscPars+1 )
      {
	ansiEscP[ ansiEscSeqFlag-1 ] = ansiEscP[ ansiEscSeqFlag-1 ]*10 + (ch-'0');
	return 0;
      }

illegalSequence:
    ansiEscSeqFlag = 0;
    return 1;
  }

  ansiEscSeqFlag = 0;
  return 0;
}

void outToScreenOneChar( int ch )
{
  if( terminalWindowPtr==NULL )
    return;

  if( ansiEscSeqFlag )
  {
    if( ansiEscSeqTreat( ch ) )
      for( int i(0); i<ansiEscSeqSave.length(); i++ )
	terminalWindow << char(ansiEscSeqSave[i]);
  }
  else
    terminalWindow << char(ch);
}

void outToScreen( int ch )
{
  if( terminalWindowPtr==NULL )
    return;

  if( terminalNeedTranslate && ch<256 )
  {
    register int n( terminalInXLAT[ ch ] );

    if( n!=-1 )
      ch = n;
  }

  Pair::PairRez ret( Pair::NoMatch );

  for( int i(0); i<ansiHow; i++ )
    switch( ansiTab[i].compare( (char)ch ) )
    {
    case Pair::Completed:
      ret = Pair::Completed;
      goto endOfSearch;

    case Pair::Reset:
      {
	register int how;

	for( int j(0), maxLen(0); j<ansiHow; j++ )
	  if( (how=ansiTab[j].matchedLen()) > maxLen )
	    maxLen = how;

	if( maxLen<matchedSymbols.length() )
	{
	  how = matchedSymbols.length()-maxLen;

	  for( j=0; j<how; j++ )
	    outToScreenOneChar( matchedSymbols[j] );

	  matchedSymbols.leftShift( how );
	}
      }
      break;

    case Pair::InProgress:
      if( ret != Pair::InProgress )
      {
	matchedSymbols.putInto( (char)ch );
	ret = Pair::InProgress;
      }
      break;

    case Pair::NoMatch:
      break;

    default:
      messageBox( "ansiTab[i].compare( ch )   Pair::PairRez" );
    }

endOfSearch:                   // 室  for  ⠡

  switch( ret )
  {
  case Pair::NoMatch:
   for( i=0; i<matchedSymbols.length(); i++ )
     outToScreenOneChar( matchedSymbols[i] );

   outToScreenOneChar( ch );
   resetMatched();
   break;

  case Pair::Completed:
    {
      char *input( ansiTab[i].process() );

      if( input!=NULL )
	virtualInput = input;
    }
    resetMatched();
    break;
  }

}

void termCmdHelp()
{
  -screenCursor;
  help( Msg_ScrTerminalHelp, 0 );
  +screenCursor;
}

void termCmdDone()
{
  terminalActive = 0;
}

void termCmdHang()
{
  log( Msg_TrmHangUp );
  suspendReader();
  comPort->hangUp();
  resumeReader();
}

void termCmdClr()
{
  --terminalWindow;
}

void termCmdInit()
{
  log( Msg_TrmInit );
  suspendReader();
  comPort->init();
  resumeReader();
}

void termCmdEcho()
{
  terminalLocalEcho ^= 1;
}

void termCmdCap()
{
  if( !terminalCapture )
  {
    initFileInfo( 0 );

    int i( stringEnter( sizeof(capName), capName, Msg_TrmCapFileName ) || !*capName );

    terminalWindow.setPhysicalCursor( 1 );

    if( i )
      terminalCapture = 0;
    else
    {
      capitalize( capName );

      initFileInfo( 1 );

      int mode( O_WRONLY|O_BINARY );

      if( access( capName, 0 ) )
      {
	mode |= O_CREAT;
	makeFilePath( capName );
      }

      terminalCapture = sopen( capName, mode|O_NOINHERIT, SH_DENYWR, S_IWRITE );

      if( !terminalCapture )
      {
	int old( terminalCapture );

	terminalCapture = dup( terminalCapture );
	close( old );
      }

      if( terminalCapture==-1 )
      {
	terminalCapture = 0;
	messageBox( Msg_ErrUnableOpen, capName, strError() );
      }
      else
	lseek( terminalCapture, 0, SEEK_END );
    }
  }
  else
  {
    close( terminalCapture );
    terminalCapture = 0;
  }

}

int stringUpload( int len, char *str, char *prompt, int& wazoo )
{
  initFileInfo( 0 );

  char *work = (char *)_alloca( len+1 );

  memset( work, 0, len+1 );
  strcpy( work, str );

  Color buttonTxt( colorSet.popUpItem ),
	buttonHig( colorSet.popUpBold );

  wazoo = trnUploadType;

  DLG_CONTROL Control[] =
  {
    { "-> ",
      CSF_DATAENTRY, CSS_ACTIVATE|CSS_CALLNOTHING,
      k_Alt_F,
      0,
      2, 2,
      1, (len<<8)|33,
      buttonTxt.fore, buttonTxt.back,
      buttonTxt.fore, buttonTxt.back,
      work, "", NULL
    },
    {
      Msg_ScrPrmtOK, CSF_BUTTON, CSS_ACTIVATE|CSS_EXIT,
      k_Alt_D,
      1,
      6, 5,
      1, 12,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      NULL, "", NULL
    },
    {
      Msg_ScrPrmtCancel, CSF_BUTTON, CSS_ACTIVATE|CSS_EXIT,
      k_Esc,
      1,
      6, 7,
      1, 12,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      NULL, "", NULL
    },
    {
      " Zmodem ", CSF_ONEOFBUTTON, CSS_ACTIVATE|CSS_EXIT,
      k_Alt_Z,
      1,
      24, 4,
      1, Zmodem,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      &wazoo, "", NULL
    },
    {
      " ZedZap ", CSF_ONEOFBUTTON, CSS_ACTIVATE|CSS_EXIT,
      k_Alt_E,
      2,
      24, 5,
      1, ZedZap,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      &wazoo, "", NULL
    },
    {
      " DirZap ", CSF_ONEOFBUTTON, CSS_ACTIVATE|CSS_EXIT,
      k_Alt_R,
      3,
      24, 6,
      1, DirZap,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      &wazoo, "", NULL
    },
    {
      " Hydra  ", CSF_ONEOFBUTTON, CSS_ACTIVATE|CSS_EXIT,
      k_Alt_H,
      1,
      24, 7,
      1, HydraSes,
      buttonTxt.fore, buttonTxt.back,
      buttonHig.fore, buttonHig.back,
      &wazoo, "", NULL
    },
    {
      NULL, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, "", NULL
    }
  };

  DIALOG dial =
  {
    (Z_ScreenWidth-40)/2+1, (Z_ScreenHeight-10)/2+1, 10, 40, DBF_NOHINT,
    (int(popUpColor)<<8)|0x20,
    popUpColor.fore, popUpColor.back,
    1, 1, 1,
    BS_SINGLELINE, popUpColor.fore, popUpColor.back,
    DS_LOWERRIGHT, 0x07,
    prompt,
    NULL,
    Control
  };

  int rc;

  suspendReader();

  while( 1 )
  {
    rc = Z_HandleDialog( &dial );
    strncpy( str, work, len );

    for( int i( strlen(str)-1 ); i>=0 && str[i]==' '; i-- ) {}

    str[++i]=0;

    if( rc==1 && !*str )
      Z_InformUser( Msg_ScrError, Msg_TrmEmpty, Msg_ScrContinue, NULL );
    else
      break;
  }

  resumeReader();

  terminalWindow.setPhysicalCursor( 1 );

  if( rc==1 )
  {
    trnUploadType = wazoo;
    initFileInfo( 1 );
  }

  return rc!=1;
}

void termUpLoad()
{
  if( !comPort->isOnline() )  //check
  {
    messageBox( Msg_TrmNCDPut );
    execSchedule( melodies.abort );
    return;
  }

  int wazoo;

  if( stringUpload( sizeof(trnName), trnName, Msg_TrmUpLoadName, wazoo ) || !*trnName )
    return;
  else
  {
    makeFileCase( trnName );
    -screenCursor;

    changeTitle( Msg_TtlUpLoad );

    log( Msg_TrmUpLoadWith, wazoo==HydraSes ? Msg_TrmUpLoadWithHydra
                                            : Msg_TrmUpLoadWithZModem,
         trnName );

    suspendReader();

    if( wazoo!=HydraSes )
      if( ZSend_Batch( trnName, wazoo ) != ERROR )
        log( Msg_TrmUpLoadSucc );
      else
        log( Msg_TrmUpLoadUnSucc );
    else
    {
      char *inbound = miscInfo.terminalInbound==NULL ?
                      miscInfo.inbound : miscInfo.terminalInbound;
      ReceivedFileList in;

      if( HydraBatch( trnName, inbound, in )==XFER_ABORT )
        log( Msg_TrmUpLoadUnSucc );
      else
        log( Msg_TrmUpLoadSucc );
    }

    resumeReader();
    changeTitle( NULL );

    +screenCursor;
  }

  execSchedule( melodies.abort );
}

extern ULONG iniFileHandle;

void termRexx()
{
  char string[65]( ".Brk" );

  if( !noPMFeatures && !detachedSession )
    BrakePrfQueryProfileString( iniFileHandle, pszAppBrake,
	 		       (PSZ)"TerminalRexxProcedure",
			       NULL,
			       (PVOID)string,
			       (ULONG)sizeof( string ) );

  int i( stringEnter( sizeof(string), string, Msg_ScrQueKbdARTtl ) );

  capitalize( string );

  if( !i )
  {
    if( !noPMFeatures && !detachedSession )
      BrakePrfWriteProfileString( iniFileHandle, pszAppBrake,
			          (PSZ)"TerminalRexxProcedure",
			          (PSZ)string );

    FIDOAddress *fAddr( defaultAddress );

    doParamsRexx( *fAddr, 0, NULL, string, -1, NULL );
  }
}


void termTrans()
{
  terminalNeedTranslate ^= 1;
}

void termShell()
{
  char *cmd[2];

  cmd[1] = NULL;

  cmd[0] = getenv( "COMSPEC" );

  Cursor *cursor = new Cursor( (int)detachedSession );

  -*cursor;

  if( cmd[0]==NULL )
    cmd[0] = "Cmd.Exe";

  suspendReader();
  suspendClipServer();
  executeCommand( P_WAIT, cmd, RunOption(RunOptClear|RunOptPort), 0, 0, 0 );
  resumeClipServer();
  resumeReader();

  +*cursor;
  delete cursor;

  +screenCursor;
  terminalWindow.setPhysicalCursor();
}

void termPauseOut()
{
  terminalStopOutput ^= 2;

  while( !keyPressed() )
    checkPoint();

  keyGet();
  terminalStopOutput ^= 2;
}

void termStopOut()
{
  terminalStopOutput ^= 1;
}

static Pair remapTable[] =
{
  Pair( key_F1,        "\033OP" ),    // PF1            F1
  Pair( key_F2,        "\033OQ" ),    // PF2            F2
  Pair( key_F3,        "\033Ow" ),    // keypad '7'     F3
  Pair( key_F4,        "\033Ox" ),    // keypad '8'     F4
  Pair( key_F5,        "\033Ot" ),    // keypad '4'     F5
  Pair( key_F6,        "\033Ou" ),    // keypad '5'     F6
  Pair( key_F7,        "\033Oq" ),    // keypad '1'     F7
  Pair( key_F8,        "\033Or" ),    // keypad '2'     F8
  Pair( key_F9,        "\033Op" ),    // keypad '0'     F9
  Pair( key_F10,       "\033Op" ),    // keypad '0'     F10
  Pair( key_UpArr,     "\033[A" ),    // Up Arrow       Up Arrow
  Pair( key_LtArr,     "\033[D" ),    // Left Arrow     Left Arrow
  Pair( key_RtArr,     "\033[C" ),    // Right Arrow    Right Arrow
  Pair( key_DnArr,     "\033[B" ),    // Down Arrow     Down Arrow
  Pair( key_Shift_F1,  "\033OR" ),    // PF3            Shift-F1
  Pair( key_Shift_F2,  "\033OS" ),    // PF4            Shift-F2
  Pair( key_Shift_F3,  "\033Oy" ),    // keypad '9'     Shift-F3
  Pair( key_Shift_F4,  "\033Om" ),    // keypad '-'     Shift-F4
  Pair( key_Shift_F5,  "\033Ov" ),    // keypad '6'     Shift-F5
  Pair( key_Shift_F6,  "\033Ol" ),    // keypad ','     Shift-F6
  Pair( key_Shift_F7,  "\033Os" ),    // keypad '3'     Shift-F7
  Pair( key_Shift_F8,  "\033OM" ),    // keypad 'Enter' Shift-F8
  Pair( key_Shift_F9,  "\033On" ),    // keypad '.'     Shift-F9
  Pair( key_Shift_F10, "\033OM" ),    // keypad 'Enter' Shift-F10

  Pair( key_PgUp,   termUpLoad  ),    // PgUp
  Pair( key_PgDn,   termDownLoad  ),  // PgDn
  Pair( key_Alt_D,  termCmdDial ),    // Alt-D
  Pair( key_Ins,    termCmdHelp ),    // Alt-=
  Pair( key_Alt_X,  termCmdDone ),    // Alt-X
  Pair( key_Alt_H,  termCmdHang ),    // Alt-H
  Pair( key_Alt_C,  termCmdClr ),     // Alt-C
  Pair( key_Alt_I,  termCmdInit ),    // Alt-I
  Pair( key_Alt_E,  termCmdEcho ),    // Alt-E
  Pair( key_Alt_T,  termTrans ),      // Alt-T
  Pair( key_Alt_L,  termCmdCap ),     // Alt-L
  Pair( key_Alt_O,  termShell ),      // Alt-O
  Pair( key_Alt_R,  termRexx ),       // Alt-R
  Pair( key_Ctrl_S, termPauseOut ),   // Ctrl-S
  Pair( key_Ctrl_O, termStopOut )     // Ctrl-O

};

void terminalWriteToScreen( int chr )
{
  if( terminalWindowPtr==NULL )
    return;

  if( terminalCapture )
    write( terminalCapture, &chr, 1 );

  outToScreen( chr );
  terminalWindow.setPhysicalCursor();
}

void terminalEmulator()
{
  if( !comPort->getPort() )
  {
    log( Msg_TermNoModem );
    return;
  }

  if( detachedSession )
  {
    log( Msg_TermDetached );
    return;
  }

  setTitleMode( Msg_TtlTerminal );
  changeTitle( NULL );

  if( getRexxHook( "Terminal" )!=NULL )
  {
    processRexxHook( "Terminal" );
    return;
  }

  comPort->lock( Msg_InfoModemLockTerminal );
  log( Msg_TrmStart );

  long startTerminal( getTimeMark() );
  long beginTerminal( time( 0 ) );

  +terminalWindow;
  terminalWindow.pop();
  -wndLog;
  -wndStat;
  -wndQueue;
  -masterScreen;

  if( (miscInfo.state1&StatusLine)==StatusLine )
    -headLineWindow;

  if( (miscInfo.state&ModemWindow)==ModemWindow  )
  {
    -*wndModemFrame;
    -wndModem;
  }

  terminalActive = 1;

  resetMatched();

  virtualInput = NULL;
  ansiEscSeqFlag = 0;
  savedX = 1;
  savedY = 1;
  terminalLocalEcho = 0;
  terminalNeedTranslate = 1;
  terminalStopOutput = 0;

  +screenCursor;

  terminalWindow.setPhysicalCursor( 1 );
  wasTimedOut = 0;
  timeOut.startTimer( miscInfo.terminalTimeOut*1000 );

  int readerTid = _beginthread( terminalReader, 0, 40960, 0 );

  if( readerTid==-1 )
  {
    terminalActive = 0;
    log( Msg_ErrUnableStartTrmRdrThr, strError() );
  }

  while( terminalActive )
  {
    unsigned ch( 0 );

    if( virtualInput!=NULL )
    {
      ch = *virtualInput;
      virtualInput++;

      if( !*virtualInput )
	virtualInput = NULL;

      checkPoint();
    }

    if( !ch && keyPressed() )
      ch = keyGet();

    if( !ch )
    {
      checkPoint();
      continue;
    }

    for( int i(0); i<sizeof(remapTable)/sizeof(remapTable[0]); i++ )
      if( remapTable[i]==ch )
      {
        suspendReader();
        virtualInput = remapTable[i].process();
        resumeReader();
        terminalWindow.setPhysicalCursor( 1 );
        break;
      }

    if( i==sizeof(remapTable)/sizeof(remapTable[0]) )
    {
      ch &= 0xff;

      register int n( terminalOutXLAT[ ch ] );

      if( terminalNeedTranslate && n!=-1 )
        ch = n;

      comPort->put(ch);

      if( terminalLocalEcho && !terminalStopOutput )
      {
        if( ch<256 )
        {
          n = terminalInXLAT[ ch ];

          if( n!=-1 )
            ch = n;
        }

        terminalWriteToScreen( ch );
      }					//terminalLocalEcho
    }					//i==sizeof(remapTable)/sizeof(remapTable[0])
    
    timeOut.startTimer( miscInfo.terminalTimeOut*1000 );
  }

  if( readerTid!=-1 )
  {
    needForAbort = 1;
    DosWaitThread( (PTID)&readerTid, DCWW_WAIT );
  }

  if( terminalCapture )
  {
    close( terminalCapture );
    terminalCapture = 0;
  }

  -screenCursor;

  +wndLog;
  +wndStat;
  +wndQueue;
  +masterScreen;

  if( (miscInfo.state1&StatusLine)==StatusLine )
    +headLineWindow;

  terminalWindow.push();
  -terminalWindow;

  if( (miscInfo.state&ModemWindow)==ModemWindow  )
  {
    +*wndModemFrame;
    +wndModem;
  }

  comPort->unlock();

  long endTerm( getTimeMark()-startTerminal );

  writeToBinaryLog( 2, &endTerm, sizeof(startTerminal), beginTerminal );

  setTitleMode( NULL );
  changeTitle( NULL );

  if( wasTimedOut )
    log( Msg_InfoTerminalAbort );
  else
    log( Msg_TrmDone );

  zInCount = 0;
  zOutCount = 0;
  zInBytes = 0;
  zOutBytes = 0;
}
