#ifndef __XMODEM_H

#define __XMODEM_H

#include "timer.h"

typedef struct
{
  long control;			// These must always start with a long!
  TRANS options;		// Transfer options
  short result;			// Result from last operation
  short sub_results;		// Extra result codes
  Timer T1;			// General purpose timer
  Timer T2;			// General purpose timer
  short Window;			// SEAlink window size
  long SendBLK;			// Current block to be sent
  long NextBLK;			// Next block we will try to send
  long ACKBLK;			// Block that was last ACK'd
  long LastBlk;			// Last block in file
  long ARBLK;			// Used in ACK Check calculations
  long WriteBLK;		// Block number to write to file
  long filelen;			// Length of file being sent
  long curr_byte;		// Current byte offset of sending or receiving
  long prev_bytes;		// Bytes that we are resyncing over
  long total_blocks;		// Total number of blocks in file to be received
  long resync_block;		// Block number we received to resync to
  short NumNAK;			// Number of NAK's received this block
  short ACKsRcvd;		// Number of ACK's received since file start
  short ACKST;			// Current state of the ack/nak state variable
  short tries;			// Number of tries thus far
  short goodfile;		// 0 if file was bad, 1 if file was good
  short datalen;		// Length of data in this block
  short recblock;		// Block number received
  short sent_ACK;		// Whether or not we sent an ACK already
  short tot_errs;		// Total number of errors
  unsigned char ARBLK8;		// 8 bit value of ARBLK
  unsigned char blocknum;	// 8 bit value of SendBLK
  unsigned char check;		// checksum value 
  unsigned char save_header;	// Received header from first block
  short CHR;			// General purpose receive character
  
  union				// File date and time in Telink or SEAlink format
  {
    struct
    {
      unsigned short time;
      unsigned short date;
    } twowords;
    struct
    {
      unsigned long timedate;
    } oneword;
  } save_filetime;
  
  char received_name[20];	// Received filename from Telink or SEAlink
  char m7name[12];		// Filename in Modem7 format
  char sending_program[16];	// Sending program name
  char *filename;		// The filename to be sent or received
  char *path;			// Just the path to the file to be sent/received
  char *fptr;			// Pointer into character fields
  char *temp_name;		// Temporary name for receiving 
  FILE *file_pointer;		// The pointer for read/write/seek operations
  
  unsigned char header;
  unsigned char block_num;
  unsigned char block_num_comp;
  unsigned char data[128];
  unsigned char data_check[2];
} XMARGS, *XMARGSP;

void XSSetVars(  XMARGSP );
int  XSInit(     XMARGSP );
int  XSEnd(      XMARGSP );
int  XSXmtStart( XMARGSP );
int  XSXmTeStrt( XMARGSP );
int  XSCheckACK( XMARGSP );
int  XSSendBlk(  XMARGSP );
int  XSWaitEnd(  XMARGSP );

typedef struct
{
  char *state_name;
  int (*state_func)( XMARGSP );
} XSTATES, *XSTATEP;

XSTATES Xmodem_Sender[] =
{
  { "XSInit", XSInit     },
  { "XSEnd",  XSEnd      },
  { "XS0",    XSXmtStart },
  { "XS0T",   XSXmTeStrt },
  { "XS1",    XSCheckACK },
  { "XS2",    XSSendBlk  },
  { "XS3",    XSWaitEnd  }
};

int ACInit(     XMARGSP );
int ACEnd(      XMARGSP );
int ACChkRcvd(  XMARGSP );
int ACSLCheck(  XMARGSP );
int ACSLVerify( XMARGSP );
int ACSLACKNAK( XMARGSP );
int ACXMCheck(  XMARGSP );
int ACSLOCheck( XMARGSP );
int ACSL1Check( XMARGSP );
int ACACKNAK(   XMARGSP );
int ACXMACK(    XMARGSP );
int ACXMNAK(    XMARGSP );
int ACRESYNC(   XMARGSP );

typedef struct
{
  char *state_name;
  int (*state_func)( XMARGSP );
} ASTATES, *ASTATEP;

ASTATES ACKNAK_Check[] =
{
  { "ACInit", ACInit     },
  { "ACEnd",  ACEnd      },
  { "AC0",    ACChkRcvd  },
  { "AC1",    ACSLCheck  },
  { "AC2",    ACSLVerify },
  { "AC3",    ACSLACKNAK },
  { "AC4",    ACXMCheck  },
  { "AC5",    ACSLOCheck },
  { "AC6",    ACSL1Check },
  { "AC7",    ACACKNAK   },
  { "AC8",    ACXMACK    },
  { "AC9",    ACXMNAK    },
  { "AC10",   ACRESYNC   }
};

#endif