#if !defined( _ZINTER_H )

#define _ZINTER_H

#include "window.h"
#include "zmodem.h"
#include "recover.h"
#include "crc.h"
#include "ascii.h"

#define KSIZE           1024
#define WAZOOMAX        8192

typedef unsigned char byte;
typedef unsigned short word;

#define TxBufSize WAZOOMAX+128

extern byte *Txbuf;                              // Pointer to xmit buffer

extern char Rxhdr[];                             // Received header
extern char Txhdr[];                             // Transmitted header

extern long *RXlong;
extern long *TXlong;

extern long Strtpos;                             // Starting byte position of
						 //   download
extern long Rxpos;                      /* Received file position                           */
extern int Crc32;                                /* Display flag indicating
						  * 32 bit CRC being used */
extern int Znulls;                               /* # of nulls to send at
						  * beginning of ZDATA hdr     */

extern int Rxtimeout;                            /* Tenths of seconds to wait
						  * for something          */
extern int Rxframeind;                           /* ZBIN ZBIN32,ZHEX type of
						  * frame received */

extern int z_size;                              // current block size

extern char *Upload_path;            // Dest. path of file being received
extern char *realname;               // real filename

extern char zmodemActive;
extern int waZooType;

extern char zReceiveMode;
extern char holdForSkipFile;

extern long zTransferStart;
extern long zTransferCpsAfter;

char *headerName( int type );
char *headerType( int type );

void zModemFNC( char *name, int flag );

int zGetTimeOut();

int Zabort();
unsigned long zfree( char *drive );

inline void Z_PutLongIntoHeader( long posit )
{
  *((long *) Txhdr) = posit;
}

inline long _Z_PullLongFromHeader( unsigned char *hdr )
{
  return *((long *) hdr);
}

int RZ_ReceiveData( byte *, int, int =0 );
int RZ_32ReceiveData( byte *, int, int =0 );
int RZ_InitReceiver( void );
int RZ_ReceiveBatch( ReceivedFileList& );
int RZ_ReceiveFile( ReceivedFileList& );
int RZ_GetHeader( void );
int RZ_SaveToDisk( long * );
void RZ_AckBibi( void );

void send_can();
int Z_GetZDL();
void ZS_SendBinaryHeader( unsigned short, byte * );
void ZS_32SendBinaryHeader( unsigned short, byte * );
int ZS_SendData( byte *, int, unsigned short, int =0 );
int ZS_32SendData( byte *, int, unsigned short, int =0 );
void ZS_SendByte( byte );
int ZS_GetReceiverInfo( void );
int ZS_SendFile( int, int );
int ZS_SendFileData( int );
int ZS_SyncWithReceiver( int );
void ZS_EndSend( void );
int _Z_GetBinaryHeader( unsigned char * );
int _Z_32GetBinaryHeader( unsigned char * );
int _Z_GetHexHeader( unsigned char * );
int _Z_GetHex( void );
int _Z_TimedRead( void );
int Z_UncorkTransmitter();
int GetTimedByte( long msec );
int Z_GetHeader( byte *hdr );
void Z_SendHexHeader( unsigned int type, register unsigned char *hdr );
void Z_PutString( register unsigned char *s );

inline int Z_GetByte( int tenths ) { return GetTimedByte( tenths*10 ); }

/*--------------------------------------------------------------------------*/
/* Routines from N_Zmodem.H ...                                             */
/*--------------------------------------------------------------------------*/

#define ZPAD            '*'
#define ZDLE            030
#define ZDLEE           (ZDLE^0100)
#define ZBIN            'A'
#define ZHEX            'B'
#define ZBIN32          'C'

/*--------------------------------------------------------------------------*/
/* Frame types (see array "frametypes" in zm.c)                             */
/*--------------------------------------------------------------------------*/
#define ZRQINIT         0
#define ZRINIT          1
#define ZSINIT          2
#define ZACK            3
#define ZFILE           4
#define ZSKIP           5
#define ZNAK            6
#define ZABORT          7
#define ZFIN            8
#define ZRPOS           9
#define ZDATA           10
#define ZEOF            11
#define ZFERR           12
#define ZCRC            13
#define ZCHALLENGE      14
#define ZCOMPL          15
#define ZCAN            16
#define ZFREECNT        17
#define ZCOMMAND        18
#define ZSTDERR         19

/*--------------------------------------------------------------------------*/
/* ZDLE sequences                                                           */
/*--------------------------------------------------------------------------*/
#define ZCRCE           'h'
#define ZCRCG           'i'
#define ZCRCQ           'j'
#define ZCRCW           'k'
#define ZRUB0           'l'
#define ZRUB1           'm'

/*--------------------------------------------------------------------------*/
/* Z_GetZDL return values (internal)                                        */
/* -1 is general error, -2 is timeout                                       */
/*--------------------------------------------------------------------------*/
#define GOTOR           0400
#define GOTCRCE         (ZCRCE|GOTOR)
#define GOTCRCG         (ZCRCG|GOTOR)
#define GOTCRCQ         (ZCRCQ|GOTOR)
#define GOTCRCW         (ZCRCW|GOTOR)
#define GOTCAN          (GOTOR|030)

/*--------------------------------------------------------------------------*/
/* Byte positions within header array                                       */
/*--------------------------------------------------------------------------*/
#define ZF0             3
#define ZF1             2
#define ZF2             1
#define ZF3             0
#define ZP0             0
#define ZP1             1
#define ZP2             2
#define ZP3             3

/*--------------------------------------------------------------------------*/
/* Bit Masks for ZRINIT flags byte ZF0                                      */
/*--------------------------------------------------------------------------*/
#define CANFDX          01
#define CANOVIO         02
#define CANBRK          04
#define CANCRY          010
#define CANLZW          020
#define CANFC32         040

/*--------------------------------------------------------------------------*/
/* PARAMETERS FOR ZFILE FRAME...                                            */
/*--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------*/
/* Conversion options one of these in ZF0                                   */
/*--------------------------------------------------------------------------*/
#define ZCBIN           1
#define ZCNL            2
#define ZCRESUM         3

/*--------------------------------------------------------------------------*/
/* Management options, one of these in ZF1                                  */
/*--------------------------------------------------------------------------*/
#define ZMNEW           1
#define ZMCRC           2
#define ZMAPND          3
#define ZMCLOB          4
#define ZMSPARS         5
#define ZMDIFF          6
#define ZMPROT          7

/*--------------------------------------------------------------------------*/
/* Transport options, one of these in ZF2                                   */
/*--------------------------------------------------------------------------*/
#define ZTLZW           1
#define ZTCRYPT         2
#define ZTRLE           3

/*--------------------------------------------------------------------------*/
/* Parameters for ZCOMMAND frame ZF0 (otherwise 0)                          */
/*--------------------------------------------------------------------------*/
#define ZCACK1          1

/*--------------------------------------------------------------------------*/
/* Miscellaneous definitions                                                */
/*--------------------------------------------------------------------------*/
#define XON             ('Q'&037)
#define XOFF            ('S'&037)
#define CPMEOF          ('Z'&037)

#define RXBINARY        FALSE
#define RXASCII         FALSE
#define LZCONV          0
#define LZMANAG         0
#define LZTRANS         0
#define SECSPERDAY      (24L*60L*60L)

#endif
