//#define FILE_MANAGER

#define INCL_DOSEXCEPTIONS
#define INCL_DOSPROCESS
#define INCL_DOSQUEUES
#include <os2.h>

#include <process.h>
#include <share.h>
#include <fcntl.h>
#include <stdio.h>
#include <io.h>
#include <ctype.h>
#include <stdlib.h>
#include <time.h>

#include "except.h"

#include "language.h"

#include "misc.h"
#include "timer.h"
#include "filetype.h"
#include "session.h"
#include "process.h"

#include "zdialog.h"
#include "zinter.h"

/*--------------------------------------------------------------------------*/
/* Private declarations                                                     */
/*--------------------------------------------------------------------------*/

static FILE *Outfile;                // Handle of file being received
static int Tryzhdrtype;              // Hdr type to send for Last rx close
static char isBinary;                // Current file is binary mode
static char EOFseen;                 // indicates cpm eof (^Z) was received
static char Zconv;                   // ZMODEM file conversion request
static int RxCount;                  // Count of data bytes received
static long Filestart;               // File offset we started this xfer from
static long rxbytes;                 // received bytes

/* Parameters for ZSINIT frame */

#define ZATTNLEN 32
static char Attn[ZATTNLEN + 1];      // String rx sends to tx on err

// ---------------------- FILE I/O -------------------------------

#ifdef FILE_MANAGER

static int writeTid = -1;
static unsigned long writeQueue = 0xffffffffl;
static int wasError = 0;

#endif

#ifndef FILE_MANAGER
int saveFileData( unsigned int count, byte *buf )
#else
void realOS2saveFileData( unsigned int count, byte *buf )
#endif
{
  static byte lastsent = 0;
  int rc = OK;

  if( isBinary )
  {
    if( fwrite( buf, 1, count, Outfile ) != count )
      goto oops;

#ifdef FILE_MANAGER
    log( "?saveFileData: written %lu bytes from %08x", count, buf );
#endif
  }
  else
  {
    if( EOFseen )
      goto done;

    for( register byte *p( buf ); count > 0; count-- )
    {
      if( *p == CPMEOF )
      {
	EOFseen = TRUE;
	goto done;
      }

      if( *p == '\n' )
      {
	if( lastsent != '\r' && putc( '\r', Outfile ) == EOF )
	  goto oops;
      }
      else
      {
	if( lastsent == '\r' && putc( '\n', Outfile ) == EOF )
	  goto oops;
      }

      if( putc( (lastsent = *p++), Outfile ) == EOF)
	goto oops;
    }

  }
done:

#ifndef FILE_MANAGER
//  checkPoint();
#endif

#ifdef FILE_MANAGER
  if( rc!=OK )
    wasError = 1;
  return;
#else
  return rc;
#endif

oops:
  log( Msg_ErrUnableWrite, Filename, strError() );
  rc = ERROR;
  goto done;
}

#ifdef FILE_MANAGER

int saveFileData( unsigned int count, byte *buf )
{
  if( wasError )
    return ERROR;

  DosWriteQueue( writeQueue, 0, count, buf, 0 );

  Txbuf = new unsigned char [ TxBufSize ];

//  checkPoint();
  return OK;
}

void _Optlink writeServer( void * )
{
  registerExceptionHandler

  DosSetPriority( PRTYS_THREAD, 0, -miscInfo.priorityDelta, 0 );

  for( int done=0; !done; )
  {
    REQUESTDATA req;
    ULONG len;
    BYTE prior;
    PBYTE buf;

    req.ulData = 0;

    APIRET rc = DosReadQueue( writeQueue, &req, &len, (PPVOID)&buf,
			      0, DCWW_WAIT, &prior, 0 );

    switch( req.ulData )
    {
    case 255:
      done = 1;
      break;

    case 0:
      if( rc || buf==NULL || !len)
	break;

      realOS2saveFileData( len, (byte *)buf );

      delete buf;
    }
  }

  deRegisterExceptionHandler
}

#endif

void completeFileIO()
{
#ifdef FILE_MANAGER

  if( writeQueue!=0xffffffff && writeTid!=-1 )
  {
    DosWriteQueue( writeQueue, 255, 1, completeFileIO, 0 );
    DosWaitThread( (PTID)&writeTid, DCWW_WAIT );
    deleteQueue( &writeQueue );
    writeTid = -1;
  }

  wasError = 0;

#endif

  if( Outfile!=NULL )
    fclose( Outfile );

  Outfile = NULL;
}

FILE *initFileIO( byte *p )
{
  FILE *ret = _fsopen( Filename, (char *)p, SH_DENYWR );

  if( ret == NULL )
    return NULL;

  if( miscInfo.inFileBufferSize )
    setvbuf( ret, NULL, _IOFBF, miscInfo.inFileBufferSize*1024 );

#ifdef FILE_MANAGER

  APIRET rc;

  if( rc=createQueue( &writeQueue, "ZReceive" ), rc )
  {
    log( Msg_ZRErrManagerQue, rc );
    os2Message( rc );

not_opened:
    writeQueue = 0xffffffff;
    fclose( ret );
    return NULL;
  }

  if( (writeTid=_beginthread( writeServer, 0, 40960, 0 ))==-1 )
  {
    log( Msg_ZRErrManager, strError() );

    deleteQueue( &writeQueue );
    goto not_opened;
  }

  wasError = 0;
#endif

  return ret;
}

// ---------------------- End FILE I/O ---------------------------

static int needSkipAll = 0;

/*--------------------------------------------------------------------------*/
/* GET ZMODEM                                                               */
/* Receive a batch of files.                                                */
/* returns OK(0) for good xfer, ERROR (-1) for bad                          */
/*--------------------------------------------------------------------------*/
int Get_Zmodem( char *rcvpath, ReceivedFileList& xferinfo, int wazoo, int skipAll )
{
  log( "?Get_Zmodem started" );

  Crc32 = 0;

  zReceiveMode = 1;
  zmodemActive = 1;
  timeDelimiter = '/';

  needSkipAll = skipAll;
  waZooType = wazoo;

  infoDisplayStart();

  *Attn = 0;
  Outfile = NULL;
  realname = NULL;
  filetime = 0;
  Crc32 = 0;

  Rxtimeout = zGetTimeOut();

  log( "?Get_Zmodem: Timeout=%d", Rxtimeout*10 );
  log( "?Get_Zmodem: sesCpsMinimum=%d", sesCpsMinimum );

  Upload_path = new char [ strlen(rcvpath)+2 ];
  strcpy( Upload_path, rcvpath );

  if( Upload_path[ strlen(Upload_path)-1 ] != '\\' && Upload_path[ strlen(Upload_path)-1 ] != '/' )
    strcat( Upload_path, "\\" );

  zRecStart = getTimeMark();

  Tryzhdrtype = ZRINIT;

  int i( RZ_InitReceiver() );

  log( "?Get_Zmodem: RZ_InitReceiver -> %s", headerName( i ) );

  switch( i )
  {
  case ZFILE:
    if( RZ_ReceiveBatch( xferinfo )!=OK )
      goto Abort;

  // fall

  case ZCOMPL:
    i = OK;               //john: TRUE
    goto Done;

  case USER_ABORT:
    i = ERROR;
    goto Done;

  // fall

  default:
Abort:
    comPort->clearout();
    comPort->clearin();
    send_can();                          // transmit at least 10 cans

    Timer t( 2000 );                     // wait no more than 2 seconds

    while( !t.isTimeOut() && !comPort->outEmpty() && comPort->isOnline())
      checkPoint();

    completeFileIO();

    i = ERROR;  //john: 0
  }

Done:

  infoDisplayEnd();

  delete Upload_path;

  delete realname;
  realname = NULL;

  zmodemActive = 0;
  zReceiveMode = 0;

  return i;
}

inline void RZ_ReceiveDataDisplayPos( int dataCounter )
{
  infoDisplayPos( rxbytes+dataCounter );
}

/*--------------------------------------------------------------------------*/
/* RZ RECEIVE DATA                                                          */
/* Receive array buf of max length with ending ZDLE sequence                */
/* and CRC.  Returns the ending character or error code.                    */
/*--------------------------------------------------------------------------*/
int RZ_ReceiveData( register byte *buf, register int length, int disp )
{
  register int c;
  register word crc;
  char *endpos;
  int d;
  int dataCounter = 0;

  switch( Zabort() )
  {
  case 27:
    return USER_ABORT;

  case 's':
    return USER_SKIP;
  }

  if( Rxframeind == ZBIN32 )
    return RZ_32ReceiveData( buf, length, disp );

  crc = 0;
  RxCount = 0;
  buf[0] = buf[1] = 0;
  endpos = (char *)buf + length;

  while( (char *)buf <= endpos )
  {
    if( (c = Z_GetZDL()) & ~0xFF )
CRCfoo:
      switch( c )
      {
      case GOTCRCE:
      case GOTCRCG:
      case GOTCRCQ:
      case GOTCRCW:
				      /*-----------------------------------*/
				      /* C R C s                           */
				      /*-----------------------------------*/
        log( "?RZ_ReceiveData: frame %s len=%d at %lu", headerName( c ),
                        dataCounter, rxbytes );

	crc = word(Z_UpdateCRC( ((d = c) & 0xFF), crc ));

	if( (c = Z_GetZDL() ) & ~0xFF )
	  goto CRCfoo;

	crc = word(Z_UpdateCRC( c, crc ));

	if( (c = Z_GetZDL()) & ~0xFF )
	  goto CRCfoo;

	crc = word(Z_UpdateCRC( c, crc ));

	if( crc & 0xFFFF )
	{
          log( Msg_ZErrDataCRC, int(rxbytes) );
	  return ERROR;
	}

	RxCount = length - (int) (endpos - (char *) buf);
	return d;

      case GOTCAN:
				      /*-----------------------------------*/
				      /* Cancel                            */
				      /*-----------------------------------*/
        log( Msg_ZInfoCanceledRemote );
	return ZCAN;

      case TIMEOUT:
				      /*-----------------------------------*/
				      /* Timeout                           */
				      /*-----------------------------------*/
        log( Msg_ZErrDataTimeOut );
	return c;

      case ERROR:
      case USER_ABORT:
        return c;

      case RCDO:
				      /*-----------------------------------*/
				      /* No carrier                        */
				      /*-----------------------------------*/
        log( Msg_ZErrCarrierLost );
	comPort->clearin();
	return c;

      default:
				      /*-----------------------------------*/
				      /* Something bizarre                 */
				      /*-----------------------------------*/
        log( "?Z-BUG: not this case handled by RZ_ReceiveData: %d", c );
	comPort->clearin();
	return c;
      }

      *buf++ = (unsigned char) c;
      crc = word(Z_UpdateCRC( c, crc ));

      dataCounter++;

      if( disp )
        RZ_ReceiveDataDisplayPos( dataCounter );
    }

  log( Msg_ZErrTooLong, int(rxbytes) );
  return ERROR;
}

/*--------------------------------------------------------------------------*/
/* RZ RECEIVE DATA with 32 bit CRC                                          */
/* Receive array buf of max length with ending ZDLE sequence                */
/* and CRC.  Returns the ending character or error code.                    */
/*--------------------------------------------------------------------------*/
int RZ_32ReceiveData( register byte *buf, register int length, int disp )
{
  register int c;
  unsigned long crc;
  char *endpos;
  int d;
  int dataCounter=0;

  crc = 0xFFFFFFFFl;
  RxCount = 0;
  buf[0] = buf[1] = 0;
  endpos = (char *)buf + length;

  while( (char *)buf <= endpos )
  {
    if( (c = Z_GetZDL()) & ~0xFF )
CRCfoo:
      switch( c )
      {
      case GOTCRCE:
      case GOTCRCG:
      case GOTCRCQ:
      case GOTCRCW:
				      /*-----------------------------------*/
				      /* C R C s                           */
				      /*-----------------------------------*/
        log( "?RZ_32ReceiveData: frame %s len=%d at %lu", headerName( c ),
                        dataCounter, rxbytes );

	d = c;
	c &= 0377;
	crc = Z_32UpdateCRC( c, crc );
	if( (c = Z_GetZDL()) & ~0xFF )
	  goto CRCfoo;

	crc = Z_32UpdateCRC( c, crc );
	if( (c = Z_GetZDL()) & ~0xFF )
	  goto CRCfoo;

	crc = Z_32UpdateCRC( c, crc );
	if( (c = Z_GetZDL()) & ~0xFF )
	  goto CRCfoo;

	crc = Z_32UpdateCRC( c, crc );
	if( (c = Z_GetZDL()) & ~0xFF )
	  goto CRCfoo;

	crc = Z_32UpdateCRC( c, crc );
	if( crc != 0xDEBB20E3l )
	{
          log( Msg_ZErrDataCRC, int(rxbytes) );
	  return ERROR;
	}

	RxCount = length - (int) (endpos - (char *)buf);
	return d;

      case GOTCAN:
				      /*-----------------------------------*/
				      /* Cancel                            */
				      /*-----------------------------------*/
        log( Msg_ZInfoCanceledRemote );
	return ZCAN;

      case TIMEOUT:
				      /*-----------------------------------*/
				      /* Timeout                           */
				      /*-----------------------------------*/
        log( Msg_ZErrDataTimeOut );
	return c;

      case ERROR:
      case USER_ABORT:
        return c;

      case RCDO:
				      /*-----------------------------------*/
				      /* No carrier                        */
				      /*-----------------------------------*/
        log( Msg_ZErrCarrierLost );
	comPort->clearin();
	return c;

      default:
				      /*-----------------------------------*/
				      /* Something bizarre                 */
				      /*-----------------------------------*/
        log( "?Z-BUG: not this case handled by RZ_32ReceiveData: %d", c );
	comPort->clearin();
	return c;
      }

    *buf++ = (unsigned char) c;
    crc = Z_32UpdateCRC( c, crc );

    dataCounter++;

    if( disp )
      RZ_ReceiveDataDisplayPos( dataCounter );
  }

  log( Msg_ZErrTooLong, int(rxbytes) );
  return ERROR;
}

/*--------------------------------------------------------------------------*/
/* RZ INIT RECEIVER                                                         */
/* Initialize for Zmodem receive attempt, try to activate Zmodem sender     */
/* Handles ZSINIT, ZFREECNT, and ZCOMMAND frames                            */
/*                                                                          */
/* Return codes:                                                            */
/*    ZFILE .... Zmodem filename received                                   */
/*    ZCOMPL ... transaction finished                                       */
/*    ERROR .... any other condition                                        */
/*--------------------------------------------------------------------------*/
int RZ_InitReceiver()
{
  register int n;
  int errors = 0;
  char *sptr = NULL;

  infoDisplayMsg( Msg_ZScrWait, Msg_ZScrReceive );
  log( "?RZ_InitReceiver started" );

  char first = 1;

  for( n = 12; --n >= 0; )
  {
      /*--------------------------------------------------------------*/
      /* Set buffer length (0=unlimited, don't wait).                 */
      /* Also set capability flags                                    */
      /*--------------------------------------------------------------*/

    if( !first )
      infoDisplayTry( n );
    else
      first = 0;

    Z_PutLongIntoHeader( 0L );
    Txhdr[ZF0] = CANFC32 | CANFDX | CANOVIO;
    Z_SendHexHeader( Tryzhdrtype, (byte *)Txhdr );

    if( Tryzhdrtype == ZSKIP || Tryzhdrtype == ZFERR )
    {
      Tryzhdrtype = ZRINIT;
      Z_SendHexHeader( Tryzhdrtype, (byte *)Txhdr );
    }

AGAIN:
    int c = Z_GetHeader( (byte *)Rxhdr );

    switch( c )
    {
    case ZFILE:
      Zconv = Rxhdr[ZF0];
      Tryzhdrtype = ZRINIT;
      Crc32 = Rxframeind == ZBIN32;

      {
        register int c1 = RZ_ReceiveData( Txbuf, WAZOOMAX );

        switch( c1 )
        {
        case GOTCRCW:
          return ZFILE;

        case USER_ABORT:
          log( Msg_ZInfoESC );
          return USER_ABORT;
        }
      }

      Z_SendHexHeader( ZNAK, (byte *)Txhdr );

      if( --n < 0 )
      {
	sptr = "ZFILE";
	goto Err;
      }

      goto AGAIN;

    case ZSINIT:
      {
        register int c1 = RZ_ReceiveData( (byte *) Attn, ZATTNLEN );

        switch( c1 )
        {
        case GOTCRCW:
          Z_PutLongIntoHeader( 1L );
          Z_SendHexHeader( ZACK, (byte *)Txhdr );
          break;

        case USER_ABORT:
          log( Msg_ZInfoESC );
          return USER_ABORT;

        default:
          Z_SendHexHeader( ZNAK, (byte *)Txhdr );
        }
      }

      if( !*Attn )					//more new stuff
	strcpy( Attn, "**" );

      if( --n < 0 )
      {
	sptr = "ZSINIT";
	goto Err;
      }

      goto AGAIN;

    case ZFREECNT:			      //     ???
//      Z_PutLongIntoHeader( zfree( Upload_path ) );
      Z_PutLongIntoHeader( 0x7fffffffl );      // p   ;)
      Z_SendHexHeader( ZACK, (byte *)Txhdr );
      goto AGAIN;

    case ZCOMMAND:
	    /*-----------------------------------------*/
	    /* Paranoia is good for you...             */
	    /* Ignore command from remote, but lie and */
	    /* say we did the command ok.              */
	    /*-----------------------------------------*/
      {
        register int c1 = RZ_ReceiveData( Txbuf, WAZOOMAX );

        switch( c1 )
        {
        case GOTCRCW:
          log( Msg_ZInfoParanoia );
  	  Txbuf[50]=0;
          log( Msg_ZInfoIgnoreCMD, Txbuf);
	  Z_PutLongIntoHeader( 0L );

	  do
	  {
	    Z_SendHexHeader (ZCOMPL, (byte *)Txhdr);
            c1 = Z_GetHeader( (byte *)Rxhdr );

            if( c1==USER_ABORT )
            {
              log( Msg_ZInfoESC );
              return USER_ABORT;
            }

	  } while( ++errors < 10 && c1 != ZFIN );

	  RZ_AckBibi();
	  return ZCOMPL;

        case USER_ABORT:
          log( Msg_ZInfoESC );
          return USER_ABORT;

        default:
          Z_SendHexHeader( ZNAK, (byte *)Txhdr );
        }
      }

      if( --n < 0 )
      {
	sptr = "CMD";
	goto Err;
      }

      goto AGAIN;

    case ZCOMPL:
      if( --n < 0 )
      {
	sptr = "COMPL";
	goto Err;
      }
      goto AGAIN;

    case ZFIN:
      RZ_AckBibi();
      return ZCOMPL;

    case USER_ABORT:
      log( Msg_ZInfoESC );
      return USER_ABORT;

    case ZCAN:
      sptr =Msg_ZInfoCanceledRemote;
      goto Err;

    case RCDO:
      sptr = Msg_ZErrCarrierLost;
      comPort->clearin();
      goto Err;

    case ZRQINIT:
    case ZNAK:
    case ZRINIT:
    case ZEOF:
    case TIMEOUT:
    case ERROR:
      break;

    default:
//      sptr = "~BUG~";
      log( "?Z-BUG: not this case handled by RZ_InitReceiver: %d", c );
//      goto Err;
    }
  }

  sptr = Msg_ZErrDataTimeOut;

Err:
  log( Msg_ZErrRecInitFailed, sptr );
  return ERROR;
}

/*--------------------------------------------------------------------------*/
/* RZFILES                                                                  */
/* Receive a batch of files using ZMODEM protocol                           */
/*--------------------------------------------------------------------------*/
int RZ_ReceiveBatch( ReceivedFileList& xferinfo )
{
  register int c;
  log( "?RZ_ReceiveBatch started" );

  for( ;; )
    switch( c = RZ_ReceiveFile( xferinfo ) )
    {
    case ZEOF:
      endReceiveRecover( OK );

      if( !terminalActive && !rexxTransfer )
        processImmediate( realname );

    /* fallthrough */

    case ZFERR:
    case ZSKIP:
      switch( RZ_InitReceiver() )
      {
      case ZCOMPL:
	return OK;

      case ZFILE:
	break;

      case ERROR:
	return ERROR;

      case USER_ABORT:
	return USER_ABORT;

      default:
        log( "?Z-BUG: not this case handled by RZ_ReceiveBatch: %d", c );
	return ERROR;
      }
      break;

    default:
      completeFileIO();
      endReceiveRecover( ERROR );

      return c;
    }
}

/*--------------------------------------------------------------------------*/
/* RZ RECEIVE FILE                                                          */
/* Receive one file; assumes file name frame is preloaded in Txbuf          */
/*--------------------------------------------------------------------------*/
int RZ_ReceiveFile( ReceivedFileList& xferinfo )
{
  char *sptr=NULL;
  EOFseen = FALSE;

  log( "?RZ_ReceiveFile started" );
  register int c = RZ_GetHeader();

  if( c == ZSKIP )
    return ( Tryzhdrtype = ZSKIP );

  if( c == ERROR || c == ZFERR )
    return ( Tryzhdrtype = ZFERR );

  const Nstart = 40;

  int n( Nstart );

  rxbytes = Filestart;

  int needSkip;

  infoDisplayName( Crc32 ? 32 : 16, isProtocol( waZooType ), wasFileName,
		   file_length, isBinary ? "" : " Ascii", NULL, 
                   isProtocolShort( waZooType ) );
  infoDisplayPos( Filestart );
  infoDisplayMarkTime( Filestart );

  FileType inFileType;
  msgOfFileType( realname, &inFileType );

  zTransferStart = getTimeMark();
  timeDelimiter = '/';

  if( Resume_WaZOO )
  {
    char *ss = strNumber( Filestart );

    log( Msg_ZReceiveFrom, ss );
    delete ss;
  }

  const long zrposTry = 10000;
  Timer zrposResend( 0 );
  zrposResend.endTimer();

  for( ;; )
  {
    if( zrposResend.isTimeOut() )
    {
send_zrpos:
      Z_PutLongIntoHeader( rxbytes );
      Z_SendHexHeader( ZRPOS, (byte *)Txhdr );
    }

NxtHdr:
    switch( c = Z_GetHeader( (byte *)Rxhdr ) )
    {
    case USER_ABORT:
      goto userAbort;

    case USER_SKIP:
      goto SkipThisFile;

    case ZDATA:
      if( Rxpos != rxbytes )
      {
        log( Msg_ZErrBadPos, Rxpos, rxbytes );

        if( zrposResend.isTimeOut() )
        {
need_zrpos:
	  Z_PutString( (byte *)Attn );
          infoDisplayTry( n );
          zrposResend.startTimer( zrposTry );
          goto send_zrpos;
        }
	continue;
      }

MoreData:
      switch( c = RZ_ReceiveData( Txbuf, WAZOOMAX, 1 ) )
      {
      case USER_ABORT:
        goto userAbort;

      case USER_SKIP:
        goto SkipThisFile;

      case ZCAN:
      case RCDO:
	goto Err;

      case ERROR:
				      /*-----------------------*/
				      /* CRC error             */
				      /*-----------------------*/
	if( --n < 0 )
	{
	  sptr = Msg_ZErrManyError;
	  goto Err;
	}
        goto need_zrpos;

      case TIMEOUT:
	if( --n < 0 )
	{
	  sptr = Msg_ZErrDataTimeOut;
	  goto Err;
	}
        goto need_zrpos;

      case GOTCRCW:
				      /*-----------------------*/
				      /* End of frame          */
				      /*-----------------------*/
	n = Nstart;

	needSkip = RZ_SaveToDisk( &rxbytes );

        switch( needSkip )
        {
        case USER_ABORT:
          goto userAbort;

        case ERROR:
	  goto ERROR_disk;

        case ERROR_ABORT:
	  goto abortSession;

	case ZSKIP:
	  goto SkipThisFile;
        }

	Z_PutLongIntoHeader( rxbytes );
	Z_SendHexHeader( ZACK, (byte *)Txhdr );
	goto NxtHdr;

      case GOTCRCQ:
				      /*-----------------------*/
				      /* Zack expected         */
				      /*-----------------------*/
	n = Nstart;

	needSkip = RZ_SaveToDisk( &rxbytes );

        switch( needSkip )
        {
        case USER_ABORT:
          goto userAbort;

        case ERROR:
	  goto ERROR_disk;

        case ERROR_ABORT:
	  goto abortSession;

	case ZSKIP:
	  goto SkipThisFile;
        }

	Z_PutLongIntoHeader( rxbytes );
	Z_SendHexHeader( ZACK, (byte *)Txhdr );
	goto MoreData;

      case GOTCRCG:
				      /*-----------------------*/
				      /* Non-stop              */
				      /*-----------------------*/
	n = Nstart;

	needSkip = RZ_SaveToDisk( &rxbytes );

        switch( needSkip )
        {
        case USER_ABORT:
          goto userAbort;

        case ERROR:
	  goto ERROR_disk;

        case ERROR_ABORT:
	  goto abortSession;

	case ZSKIP:
	  goto SkipThisFile;
        }

	goto MoreData;

      case GOTCRCE:
				      /*-----------------------*/
				      /* Header to follow      */
				      /*-----------------------*/
	n = Nstart;

	needSkip = RZ_SaveToDisk( &rxbytes );

        switch( needSkip )
        {
        case USER_ABORT:
          goto userAbort;

        case ERROR:
	  goto ERROR_disk;

        case ERROR_ABORT:
	  goto abortSession;

	case ZSKIP:
	  goto SkipThisFile;
        }

	goto NxtHdr;
      }

    case ZNAK:
      if( --n < 0 )
      {
	sptr = Msg_ZErrJunkBlock;
	goto Err;
      }
      goto need_zrpos;

    case TIMEOUT:
      if( --n < 0 )
      {
	sptr = Msg_ZErrJunkBlock;
	goto Err;
      }
//      goto NxtHdr;		//zrpos???
      goto need_zrpos;

    case ZFILE:
				 /*-----------------------------------------*/
				 /* Sender didn't see our ZRPOS yet         */
				 /*-----------------------------------------*/
      if( RZ_ReceiveData( Txbuf, WAZOOMAX )==USER_ABORT )
        goto userAbort;

      goto need_zrpos;

    case ZEOF:
				/*------------------------------------------*/
				/* End of the file                          */
				/* Ignore EOF if it's at wrong place; force */
				/* a timeout because the eof might have     */
				/* gone out before we sent our ZRPOS        */
				/*------------------------------------------*/
      if( Rxpos != rxbytes )
      {
        log( Msg_ZErrUnExpEOF );
	goto NxtHdr;
      }

      xferinfo.AddAtTail( *new ReceivedFile( realname ) );

end_message:

      zInCount++;
      zInBytes += rxbytes-Filestart;
      switch( inFileType )
      {
      case FileMail:
	sessionBinaryInfo.netRecv += rxbytes-Filestart;
	break;

      case FileXMail:
	sessionBinaryInfo.xmRecv += rxbytes-Filestart;
	break;

      default:
	sessionBinaryInfo.filesRecv += rxbytes-Filestart;
	break;
      }

      infoDisplayDone( rxbytes-Filestart, Filestart, 1 );

      completeFileIO();

      goto Done;

    case ERROR:
				 /*-----------------------------------------*/
				 /* Too much garbage in header search error */
				 /*-----------------------------------------*/
      if( --n < 0 )
      {
	sptr = Msg_ZErrHdtGarbage;
	goto Err;
      }
      goto NxtHdr;

    case ZSKIP:
      goto Done;

    case RCDO:
      sptr = Msg_ZErrCarrierLost;
      goto Err;

    case ZCAN:
      sptr = Msg_ZInfoCanceledRemote;
      goto Err;

    default:
      log( "?Z-BUG: not this case handled by RZ_ReceiveFile: %d", c );
      sptr = Msg_ZErrDebries;
      goto Err;
    }
  }

Err:
  comPort->clearin();
  if( sptr!=NULL )
    log( Msg_ZErrReceiveFail, sptr );

abortSession:
  c = ERROR;
  goto end_message;

ERROR_disk:
  c = ZFERR;

Done:
  return c;

SkipThisFile:
  {
    log( Msg_ZInfoAltS );

    comPort->clearout();
    Z_PutString( (byte *)Attn );

    Timer t( 2000 );

    while( comPort->haveInput() && comPort->isOnline() && !t.isTimeOut() )
      comPort->clearin();

    if( !terminalActive && remoteMailerCode!=0xe0 )	// !t-mail & !terminal
    {
      log( "?RZ_ReceiveFile: skip on fly - ZSKIP" );
      Z_PutLongIntoHeader( 0 );
      Z_SendHexHeader( ZSKIP, (byte *)Txhdr );
      c = ZEOF;
      goto end_message;
    }
    else
    {
      log( "?RZ_ReceiveFile: skip on fly - ZRPOS" );
      rxbytes = file_length;
      infoDisplayTry( n );
      zrposResend.startTimer( zrposTry );
      goto send_zrpos;
    }
  }

userAbort:
  log( Msg_ZInfoESC );
  c = USER_ABORT;
  goto end_message;
}

/*--------------------------------------------------------------------------*/
/* RZ GET HEADER                                                            */
/* Process incoming file information header                                 */
/*--------------------------------------------------------------------------*/
int RZ_GetHeader()
{
  log( "?RZ_GetHeader started" );
  holdForSkipFile = 0;

   /*--------------------------------------------------------------------*/
   /* Setup the transfer mode                                            */
   /*--------------------------------------------------------------------*/
  isBinary = (char) ((!RXBINARY && Zconv == ZCNL) ? 0 : 1);
  Resume_WaZOO = 0;

   /*--------------------------------------------------------------------*/
   /* Extract and verify filesize, if given.                             */
   /* Reject file if not at least 10K free                               */
   /*--------------------------------------------------------------------*/
  file_length = 0L;
  filetime = 0L;
  char *fileinfo( (char *)Txbuf + 1 + strlen ((char *)Txbuf) );

  if( *fileinfo )
    sscanf( fileinfo, "%ld %lo", &file_length, &filetime );

  log( "?RZ_GetHeader: len=%lu tim=%lu", file_length, filetime );

  if( (miscInfo.state1&AdjustTime)==AdjustTime )
  {
    filetime &= ~1;
    log( "?RZ_GetHeader: MSDOS adjust tim=%lu", filetime );
  }

   /*--------------------------------------------------------------------*/
   /* Get and/or fix filename for uploaded file                          */
   /*--------------------------------------------------------------------*/

  delete Filename;
  Filename = new char [ strlen(Upload_path)+_MAX_PATH ];
  strcpy( Filename, Upload_path );

  byte *ourname = (byte *)( Filename + strlen( Filename ) );
  byte *p = Txbuf + strlen( (char *)Txbuf ) - 1; // Find transmitted simple filename

  while( p >= Txbuf && *p != '\\' && *p != '/' && *p != ':' )
    p--;

  strcpy( (char *)ourname, (char *)p+1 );            // Start w/ our path & their name

  zModemFNC( (char *)ourname, (miscInfo.state3&HPFS)!=HPFS );

  log( "?RZ_GetHeader: found `%s'", Filename );

  if( needSkipAll )
  {
    log( "?RZ_GetHeader: needSkipAll is active" );
    return ZSKIP;
  }

  int rc( startReceiveRecover( &p ) );

  log( "?startReceiveRecover RC=%d", rc );

  if( rc!=OK )
    return rc;

  makeFileCase( Filename );
  makeFilePath( Filename );

  if( (Outfile = initFileIO( p ))==NULL )
  {
    log( Msg_ErrUnableOpen, Filename, strError() );
    return ZFERR;
  }

  if( isatty( fileno( Outfile ) ) )
  {
    log( Msg_ZErrIsDev, fileName( Filename ) );
    completeFileIO();
    return ZFERR;
  }

  Filestart = (Resume_WaZOO) ? filelength( fileno( Outfile ) ) : 0L;
  Strtpos = Filestart;
  fseek( Outfile, Filestart, SEEK_SET );

  long diskfree = zfree( Upload_path );

  if( ((unsigned long)file_length)+10240u > diskfree && 
      ((unsigned long)file_length)+10240u-Filestart > diskfree )
  {
    log( Msg_ZErrNoSpace, Txbuf, zfree( Upload_path ), file_length-Filestart );
    completeFileIO();
    return ZFERR;
  }

  log( "?RZ_GetHeader OK completed" );
  return OK;
}

/*--------------------------------------------------------------------------*/
/* RZ SAVE TO DISK                                                          */
/* Writes the received file data to the output file.                        */
/* If in ASCII mode, stops writing at first ^Z, and converts all            */
/*   solo CR's or LF's to CR/LF pairs.                                      */
/*--------------------------------------------------------------------------*/
int RZ_SaveToDisk( long *rxbytes )
{
  if( holdForSkipFile )
  {
    holdForSkipFile = 0;
    return ZSKIP;
  }

  switch( Zabort() )
  {
  case 27:
    return USER_ABORT;

  case 's':
    return ZSKIP;
  }

  infoDisplayBlock( RxCount );
  infoDisplayPos( (*rxbytes)+RxCount );

  int cps = infoDisplayCps();

  if( !terminalActive && !rexxTransfer )
  {
    if(	getTimeMark() > zTransferStart+zTransferCpsAfter )
    {
      timeDelimiter = '|';

      if( sesCpsMinimum && cps<sesCpsMinimum )
      {
        log( Msg_ZInfoLowCPS );
        return ERROR_ABORT;
      }
    }

    if( sessionLimitExceeded() )
      return ERROR_ABORT;
  }

  int rc = saveFileData( RxCount, Txbuf );

  *rxbytes += RxCount;

  return rc;
}

/*--------------------------------------------------------------------------*/
/* RZ ACK BIBI                                                              */
/* Ack a ZFIN packet, let byegones be byegones                              */
/*--------------------------------------------------------------------------*/
void RZ_AckBibi()
{
  log( "?RZ_AckBibi started" );

  for( register int n( 4 ); --n; )
  {
    Z_PutLongIntoHeader( 0L );
    Z_SendHexHeader( ZFIN, (byte *)Txhdr );

    switch( Z_GetByte( 200 ) )
    {
      case 'O':
	Z_GetByte( 100 );                // Discard 2nd 'O'

      case TIMEOUT:
      case RCDO:
	return;
    }
  }
}
