/*
  All rights reserved.
  Copyright (c) 1989-97 by John Gladkih & Bad data, Inc.

  Original version for SysV, 1989 by John Gladkih.
  Port to OS/2 (for IBM icc) 1997 by John Gladkih.

  You may distrubute and change this text as you want.
  You don't remove my copiright. ;-)
*/

/* Un-comment next line with #define if you want to compile for UNIX. */
/* In this case you not want to compile courses.c, courses.c          */
/* is part of OS/2 port                                               */

/* #define UNIX                                                       */

#ifdef UNIX
#include <sys/times.h>
#include <curses.h>
#define cursor( y, x, c )	move( 23-(x), y ), addch( c ), refresh()
#define display( x )		addstr( x ), refresh()
#define line( y, x, str )	move( 23-(x), y ), addstr( str ), refresh()
#else
#include "curses.h"
#define cursor( y, x, c )	move( 23-(x), y ), addch( c )
#define display( x )		addstr( x )
#define line( y, x, str )	move( 23-(x), y ), addstr( str )
#endif
#include <string.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>

#include <sys/types.h>

#define XMAX			79
#define YMAX			22
#define LIMIT			253
#define SNAKE			254
#define MANGUST			255
#define EMPTY			0
#define SNAKE_BODY		'#'
#define SNAKE_HEAD		'*'
#define MANGUST_LOCATION	'$'
#define FREE_LOCATION		' '
#define TABLE_SIZE		sizeof table
#define TOTAL_RECORDS		15
#define NAME_LENGTH		20

typedef char *STRING;

struct segment
{
  struct segment *next,
                 *back;
  int x,
      y;
};

typedef struct segment *POINTER;
typedef unsigned char CONTROL;

POINTER head,
        tail;
CONTROL field[ XMAX+1 ][ YMAX+1 ],
        game_over,
        redraw_screen,
        key_is_pressed;
long score,
     old_score,
     length,
     old_length,
     t_krolik,
     t_mangust,
     dx,
     dy,
     player_class;
STRING class_str = "Press: A - novice player, B - non-bad, C - good, D - super";
int save_symbol;
time_t s_tim,
       time();
time_t time_mark;

#ifdef UNIX
struct tms t_b;
#endif

struct game_parameters_type
{
  float mangust,
        krolik;
  int shtraf,
      initial_length;
  time_t delay_time,
         interrupt_time;
};
struct game_parameters_type game_parameters[4] =
{
#ifdef UNIX
  { 0.008, 0.013, 3, 6,  7, 5 },
  { 0.011, 0.012, 5, 10, 4, 6 },
  { 0.014, 0.011, 7, 14, 3, 7 },
  { 0.017, 0.010, 9, 18, 2, 8 }
#else
  { 0.008, 0.013, 3, 6, 100, 5 },
  { 0.011, 0.012, 5, 10, 80, 6 },
  { 0.014, 0.011, 7, 14, 60, 7 },
  { 0.017, 0.010, 9, 18, 40, 8 }
#endif
};
FILE *table_file;
typedef struct
{
  long score;
  int length;
  CONTROL end_flag;
  char name[NAME_LENGTH+1];
} TABLE_TYPE;
STRING blank = "                                     ";
long ix = 33434,
     iy = 23345,
     iz = 5235;
double fmod();

#define random()                                           \
         fmod( (double)(ix = 171*ix % 30269)/30269. +      \
               (double)(iy = 172*iy % 30307)/30307. +      \
               (double)(iz = 170*iz % 30323)/30323., 1. )

void delay_game( void ) /* for bsd-like system you may use select() */
{
#ifdef UNIX
  for( times(&t_b), s_tim=t_b.tms_utime+t_b.tms_stime; times(&t_b),
       t_b.tms_utime+t_b.tms_stime<s_tim+
       game_parameters[ player_class ].delay_time; );
#else
  os2delay( game_parameters[ player_class ].delay_time );
#endif
}

void get_symbol( void )
{
  save_symbol = getch();
  key_is_pressed = TRUE;

  switch( save_symbol )
  {
  case ' ':
    redraw_screen = TRUE;
    break;

  case KEY_LEFT:
    dx = -1;
    dy = 0;
    break;

  case KEY_UP:
    dx = 0;
    dy = 1;
    break;

  case KEY_RIGHT:
    dx = 1;
    dy = 0;
    break;

  case KEY_DOWN:
    dx = 0;
    dy = -1;
    break;

  case 0105:
  case 0145:
    clear();
    endwin();
    exit( 0 );

  default:
    break;
  }
}

STRING dec( long num )
{
  int i,
      j;
  static char out[12];

  i = 11;

  do
    out[i--] = num%10 + 060;
  while( num/=10 );

  for( j=0; i<11; )
    out[ j++ ] = out[ ++i ];

  out[ j ] = 0;

  return out;
}

int wait_while_not_key_pressed( STRING delimiters )
{
  CONTROL entered = FALSE;
  int i;

  nodelay( stdscr, FALSE );

  do
  {
    get_symbol();
    save_symbol -= save_symbol<'a'||save_symbol>'z' ? 0 : 040;

    for( i=0; delimiters[i] && !entered; i++ )
      entered = delimiters[i]==save_symbol;

  } while( !entered);

  nodelay( stdscr, TRUE );

  return save_symbol;
}

void copy( void *dst, void *src, long len )
{
  long i;

  for( i=0; i<len; i++ )
    ((char *)dst)[i] = ((char *)src)[i];
}

void error( void )
{
  if( ferror(table_file) )
  {
    fclose( table_file );
    clear();
    attron( A_BOLD );
    line( 20, 20, "Sorry, table file I/O error" );
    attroff( A_BOLD );
    beep();
    table_file = NULL;
  }
}

void step( CONTROL previous )
{
  POINTER work;
  CONTROL recurse;
  int i;

  switch( previous )
  {
  case MANGUST:
    score -= game_parameters[ player_class ].shtraf;
    t_mangust--;
    length -= game_parameters[player_class].shtraf;

    if( length>=game_parameters[ player_class ].initial_length )
    {
      work = tail;
      tail = work->back->back;
      tail->next = NULL;
      cursor( work->x, work->y, FREE_LOCATION );
      cursor( work->back->x, work->back->y, FREE_LOCATION );
      field[ work->x ][ work->y ] = field[ work->back->x ][ work->back->y ] = EMPTY;
      free( work->back );
      free( work );
      line( 56, 23, "      " );
      line( 56, 23, dec(t_mangust) );
    }
    else
      game_over = 1;

    break;

  case LIMIT:
    game_over = 2;
    break;

  case SNAKE:
    game_over = 3;
    break;

  case EMPTY:
    break;

  default:
    t_krolik--;
    score += previous;
    length += previous;
    line( 73, 23, "      " );
    line( 73, 23, dec(t_krolik) );

    for( i=0; i<previous && !game_over; i++ )
    {
      get_symbol();
      work = (POINTER)malloc( sizeof(struct segment) );
      work->x = head->x+dx;
      work->y = head->y+dy;
      work->back = NULL;
      work->next = head;
      head->back = work;
      head = work;
      cursor( head->next->x, head->next->y, SNAKE_BODY );
      cursor( head->x, head->y, SNAKE_HEAD );
      recurse = field[ head->x ][ head->y ];
      field[ head->x ][ head->y ] = SNAKE;
      delay_game();
      step( recurse );
    }
  }

  if( time(0)>=time_mark )
  {
    score++;
    key_is_pressed = FALSE;
    time_mark = time(0)+game_parameters[ player_class ].interrupt_time;
  }

  if( old_score!=score )
  {
    if( score<0 )
      score = 0;

    line( 9, 23, "      " );
    line( 9, 23, dec(score) );
    old_score = score;
  }

  if( old_length!=length )
  {
    line( 26, 23, "      " );
    line( 26, 23, dec(length) );
    old_length = length;
  }
}

main()
{
  int i,
      j;
  CONTROL previous_field_value;
  POINTER work,
          temporary;
  STRING s_t = "+-------+-------+-------+",
         s_p = "|       |       |       |",
         s_i = "+-------+-------+-------+",
         s_d = "+-------+-------+-------+",
         end_mangust = "Killed by mangust",
         end_self = "Self death",
         end_limit = "Out of border",
         file_name = ".snake.scores";
  TABLE_TYPE table[ 4 ][ TOTAL_RECORDS ],
             used[ TOTAL_RECORDS+1 ],
             tmp;

  for( s_tim=time(0)&0x1fff; s_tim; s_tim-- )
    random();

  initscr();
  flushinp();
  cbreak();
  noecho();
  keypad( stdscr, TRUE );
  nodelay( stdscr, TRUE );
  line( 33, 23, "' S N A K E'" );
  line( 0, 21, "Copyright (c) 1989 by Bad Data Inc." );
  line( 0, 20, "All rights reserved." );
  line( 0, 18, "Produced by John Gladkih" );
  line( 1, 13, "Space - refresh" );
  line( 5, 11, "E - panic" );
  line( 5, 8, "$ - mangust" );
  line( 1, 6, "Digit - rabbit" );
  line( 8, 0, class_str );
  line( 45, 17, s_t );
  line( 45, 16, s_p );
  line( 45, 15, s_p );
  line( 45, 14, s_p );
  line( 45, 13, s_p );
  line( 45, 12, s_i );
  line( 45, 11, s_p );
  line( 45, 10, s_p );
  line( 45, 9, s_p );
  line( 45, 8, s_p );
  line( 45, 7, s_i );
  line( 45, 6, s_p );
  line( 45, 5, s_p );
  line( 45, 4, s_p );
  line( 45, 3, s_p );
  line( 45, 2, s_d );
  cursor( 57, 4, 'v' );
  cursor( 57, 5, '|' );
  cursor( 66, 10, '>' );
  cursor( 65, 10, '-' );
  cursor( 49, 10, '<' );
  cursor( 50, 10, '-' );
  cursor( 57, 15, '^' );
  cursor( 57, 14, '|' );
  beep();
  player_class = wait_while_not_key_pressed( "ABCD" )-0101;

  for( ;; )
  {
    dx = dy = game_over = 0;
    redraw_screen = TRUE;
    score = old_score = 0;
    length = old_length =game_parameters[ player_class ].initial_length;
    t_krolik = t_mangust = 0;
    head = NULL;

    for( i=0; i<=XMAX; i++ )
      for( j=0; j<=YMAX; j++ )
        if( !i || !j || i==XMAX || j==YMAX )
          field[i][j] = LIMIT;
        else
          field[i][j] = EMPTY;

    for( i=length; i!=0; i-- )
    {
      work = (POINTER)malloc( sizeof(struct segment ) );
      field[ work->x=XMAX/2-length/2+i][work->y=YMAX/2] = SNAKE;
      work->next = head;
      work->back = NULL;

      if( head==NULL )
        tail = work;
      else
        head->back = work;

      head = work;
    }

    while( !game_over )
    {
      get_symbol();

      if( redraw_screen )
      {
        clear();

        for( i=XMAX-1; i; i-- )
        {
          cursor( i, 0, '-' );
          cursor( i, YMAX, '-' );

          if( i<YMAX )
          {
            cursor( 0, i, '|' );
            cursor( XMAX, i, '|' );
          }
        }

        cursor( 0, 0, '+' );
        cursor( XMAX, 0, '+' );
        cursor( 0, YMAX, '+' );
        cursor( XMAX, YMAX, '+' );

        for( i=1; i<XMAX-1; i++ )
          for( j=1; j<YMAX-1; j++ )
            switch( field[i][j] )
            {
            case SNAKE:
              if( head->x==i && head->y==j )
                cursor( i, j, SNAKE_HEAD );
              else
                cursor( i, j, SNAKE_BODY );
              break;

            case MANGUST:
              cursor( i, j, MANGUST_LOCATION );
              break;

            case EMPTY:
              break;

            default:
              cursor( i, j, field[i][j]+060 );
              break;
            }

        line( 1, 23, "Score :" );
        line( 17, 23, "Length :" );
        line( 45, 23, "Mangusts :" );
        line( 63, 23, "Rabbits :" );
        line( 9, 23, dec(score) );
        line( 26, 23, dec(length) );
        line( 56, 23, dec(t_mangust) );
        line( 73, 23, dec(t_krolik) );
        attron( A_REVERSE );
        cursor( 37, 23, player_class+0101 );
        attroff( A_REVERSE );
        redraw_screen = FALSE;

        if( !(dx||dy) )
        {
          line( 29, 0, " To start press arrow key " );

          do
            get_symbol();
          while( !dx && ! dy );

          line( 29, 0, "--------------------------" );
          key_is_pressed = FALSE;
          time_mark = time(0)+game_parameters[ player_class ].interrupt_time;
        }
      }			/* redraw screen */

      delay_game();
      work = tail;
      (tail=tail->back)->next = NULL;
      work->back = NULL;
      work->next = head;
      i = work->x;
      j = work->y;
      work->x = head->x+dx;
      work->y = head->y+dy;
      head->back = work;
      head = work;
      cursor( i, j, FREE_LOCATION );
      cursor( head->next->x, head->next->y, SNAKE_BODY );
      cursor( head->x, head->y, SNAKE_HEAD );
      previous_field_value = field[ head->x ][ head->y ];
      field[ head->x ][ head->y ] = SNAKE;
      field[i][j] = EMPTY;
      step( previous_field_value );

      if( !game_over )
      {
        if( random()<=game_parameters[player_class].mangust &&
            field[ i=(int)(random()*(double)(XMAX-1)+1.)]
                 [ j=(int)(random()*(double)(YMAX-1)+1.)] == EMPTY )
        {
          line( 56, 23, dec(++t_mangust) );
          field[i][j] = MANGUST;
          cursor( i, j, MANGUST_LOCATION );
        }

        if( random()<=game_parameters[player_class].krolik &&
            field[ i=(int)(random()*(double)(XMAX-1)+1.)]
                 [ j=(int)(random()*(double)(YMAX-1)+1.)] == EMPTY )
        {
          line( 73, 23, dec(++t_krolik) );
          field[i][j] = (int)(random()*8.+1.);
          cursor( i, j, field[i][j]+060 );
          beep();
        }
      }
    }

    move( 0, 45 );
    clrtoeol();
    attron( A_BOLD );

    switch( game_over )
    {
    case 1:
      line( 45, 23, end_mangust );
      break;

    case 3:
      line( 45, 23, end_self );
      break;

    case 2:
      line( 45, 23, end_limit );
      break;
    }

    attroff( A_BOLD );
    beep();

    for( work=head; work!=NULL; work=temporary )
    {
      temporary = work->next;
      free( work );
    }
    {
      char ferr;

      if( (table_file=fopen( file_name, "rb" ))!=NULL )
      {
        fread( table, TABLE_SIZE, 1, table_file );
        ferr = ferror( table_file ) || feof( table_file );
        fclose( table_file );
      }
      else
        ferr = TRUE;

      if( ferr )
        for( j=0; j<4; j++ )
          for( i=0; i<TOTAL_RECORDS; i++ )
          {
            table[j][i].name[0] = 0;
            table[j][i].score = 0;
            table[j][i].end_flag = 4;
            table[j][i].length = -1;
          }
    }

    if( table[player_class][TOTAL_RECORDS-1].score>=score )
    {
      attron( A_REVERSE );
      line( 20, 10, blank );
      line( 20, 11, " Too low score for writting to table " );
      line( 20, 12, blank );
      attroff( A_REVERSE );
      beep();
    }
    else
    {
      copy( used, &table[player_class][0], sizeof(TABLE_TYPE)*TOTAL_RECORDS );
      used[TOTAL_RECORDS].score = score;
      used[TOTAL_RECORDS].end_flag = game_over;
      used[TOTAL_RECORDS].length = length;

      for( i=0; i<NAME_LENGTH; used[TOTAL_RECORDS].name[i++]=0 );

      for( i=TOTAL_RECORDS-1; i>=0; i-- )
        for( j=0; j<=i; j++ )
          if( used[j].score<used[j+1].score ||
              (used[j].score==used[j+1].score && used[j].length<used[j+1].length) )
          {
            copy( &tmp, used+j, sizeof(TABLE_TYPE) );
            copy( used+j, used+j+1, sizeof(TABLE_TYPE) );
            copy( used+j+1, &tmp, sizeof(TABLE_TYPE) );
          }

      clear();
      line( 31, 22, "Table of class " );
      cursor( 46, 22, player_class+0101 );
      line( 15, 20, "Name              Score      Length" );

      for( i=0; i<TOTAL_RECORDS; i++ )
      {
        if( used[i].score==score && used[i].length==length &&
            (!used[i].name[0]) )
        {
          attron( A_BOLD );
          j = i;
        }

        line( 1, 18-i, dec( i+1 ) );

        if( used[i].end_flag!=4 )
        {
          line(  4, 18-i, used[i].name );
          line( 33, 18-i, dec( used[i].score ) );
          line( 44, 18-i, dec( used[i].length ) );
        }

        switch( used[i].end_flag )
        {
        case 1:
          line( 55, 18-i, end_mangust );
          break;

        case 3:
          line( 55, 18-i, end_self );
          break;

        case 2:
          line( 55, 18-i, end_limit );
          break;

        case 4:
          line( 55, 18-i, "Empty slot" );
          break;

        default:;
        }

        if( used[i].score==score && used[i].length==length )
          attroff( A_BOLD );

      }

      i = 0;
      attron( A_UNDERLINE );
      line( 25, 1, "Insert your name to table !" );
      attroff( A_UNDERLINE );
      beep();
      move( 23-(18-j), 4+i );
      nodelay( stdscr, FALSE );
      attron( A_BOLD );

      do
      {
        save_symbol = getch()&0x7f;

        if( save_symbol==010 || save_symbol==0177 )
          if( i>0 )
          {
            used[j].name[i--] = 0;
            cursor( 4+i, 18-j, ' ' );
            move( 23-(18-j), 4+i );
          }
          else
            beep();
        else
          if( save_symbol>037 && save_symbol<=0176 )
          {
            cursor( 4+i, 18-j, save_symbol );
            used[j].name[i++] = save_symbol;
          }
          else
            beep();

      } while( i<NAME_LENGTH && save_symbol!='\n' && save_symbol!='\r' );

      attroff( A_BOLD );
      nodelay( stdscr, TRUE );
      copy( &table[ player_class ][ 0 ], used, sizeof(TABLE_TYPE)*TOTAL_RECORDS );

      if( (table_file=fopen( file_name, "wb" ))==NULL )
      {
        attron( A_REVERSE );
        line( 20, 10, blank );
        line( 20, 11, "    Alas, table not accessible!    " );
        line( 20, 12, blank );
        attroff( A_REVERSE );
      }
      else
      {
        fwrite( table, TABLE_SIZE, 1, table_file );
        error();
      }

      attroff( A_REVERSE );
      clear();

      if( table_file!=NULL )
      {
        fclose( table_file );
        attron( A_REVERSE );
        line( 30, 8, "                 " );
        line( 30, 9, " Table updated ! " );
        line( 30, 10, "                 " );
        attroff( A_REVERSE );
      }
    }

    line( 8, 0, class_str );

    nodelay( stdscr, FALSE );

    if( (i=wait_while_not_key_pressed( " ABCD" )) != ' ' )
      player_class = i - 0101;

    nodelay( stdscr, TRUE );
  }
}
