/************************ SWin ***************************\
 *
 *	Copyright (C) 1996 by Dmitry Zavalishin
 *
 *	Module 	: control window classes impl.
 *
 *      $Log: controls.c $
 *      Revision 1.3  1996/10/03 07:43:57  dz
 *      buttons
 *
 *      Revision 1.2  1996/09/20 14:40:18  dz
 *      update
 *
 *      Revision 1.1  1996/09/12 01:12:51  dz
 *      Initial revision
 *
 *
 *      
 *
\*/

#include "sw_controls.h"
#include "sw_device.h" // flush



void sw_control::label( const char *text, sw_color color )
    {
    del_label();
    l = new sw_label( w, sw_point( x() - strlen(text), y() ), text, color );
    }

void sw_control::del_label()
    {
    if( l ) delete l;
    l = 0;
    }



  // -------------------------------------------------------------------------

sw_label::sw_label(
                   sw_abstract_win &parent, sw_point pos,
                   const char *_text, sw_color color
                  )
    :
color_v(color),
sw_control( parent, sw_rect(pos,sw_size(strlen(_text),1))  )
    {
    text_v = 0;
    len_v = 0;
    text(_text);
    focusable_v = No; // Don't try to set focus to us
    }

sw_label::sw_label( sw_abstract_win &parent, sw_color color ) :
color_v(color), sw_control( parent, sw_rect(0,0,1,1)  )
    {
    text_v = 0;
    len_v = 0;
    }


void sw_label::text( const char *s )
    {
    delete [] text_v;
    len_v = strlen(s);
    text_v = new char[len_v*2];
    if(len_v-xs()) resize(sw_size(len_v-xs(),0));
    for( int i = 0; i < len_v*2; i++ )
        {
        text_v[i++] = *s ? *s++ : ' ';
        text_v[i] = color_v.value();
        }
    refresh(sw_rect(0,0,len_v,1));
    }

sw_label::~sw_label()
    {
    delete [] text_v;
    }
        
void sw_label::refresh(const sw_rect&a)
    {
    if( text_v == 0 ) return;
    sw_rect r;
      //if( r.intersection(a, sw_rect(0,0,len_v,1) ) )
      //  put( r, text_v+(r.x()*2) );
    put( sw_rect(0,0,len_v,1), text_v );
    }



  // -------------------------------------------------------------------------

void sw_abs_button::button_press()
    {
    state(1);
    device.flush();
    //sw_sleep(128); BUG!! Here we must send us a 'release' message somehow?
    state(0);
    }


  // -------------------------------------------------------------------------

sw_onechar_button::sw_onechar_button
    (
     sw_abstract_win &parent, sw_point pos,
     sw_color color, char _c1, char _c2
    ) : sw_abs_button( parent, sw_rect(pos,sw_size(1,1)) )
    {
    c1[0] = _c1;
    c2[0] = _c2;
    c1[1] = color.value();
    c2[1] = color.value();
    }

void sw_onechar_button::refresh(const sw_rect&)
    {
    put( sw_rect(0,0,1,1), state_v ? c2 : c1 );
    }
    

  // -------------------------------------------------------------------------

sw_text_button::sw_text_button
    (
     sw_abstract_win &parent, sw_point pos, const string &text,
     sw_color c1, sw_color cs, sw_color c2
    )
    :
sw_abs_button( parent, sw_rect(pos,sw_size(text.length()+1,2)) ),
c1_v(c1), cs_v(cs), c2_v(c2),
text_v( text ), l( text.length() ),
lower_line_1('',text.length()), lower_line_2(' ',text.length()+1)
    {
    }

void sw_text_button::refresh(const sw_rect&)
    {
    if( state_v )
        {
        put( sw_point(1,0), in_focus ? cs_v : c1_v, text_v );
        put( sw_point(0,0), c2_v, " " );
        put( sw_point(0,1), c2_v, lower_line_2 );
        }
    else
        {
        put( sw_point(0,0), in_focus ? cs_v : c1_v, text_v );
        put( sw_point(l,0), c2_v, "" );
        put( sw_point(0,1), c2_v, " " );
        put( sw_point(1,1), c2_v, lower_line_1 );
        }
    }


  // -------------------------------------------------------------------------


