/************************ SWin ***************************\
 *
 *	Copyright (C) 1996 by Infinity Soft
 *
 *	Module 	:	Exceptions
 *
 *      $Log: sw_except.h $
 *      Revision 1.3  1996/09/30 14:18:41  dz
 *      log control written
 *
 *      Revision 1.2  1996/09/09 14:53:22  dz
 *      update
 *
 *      Revision 1.1  1996/09/09 01:47:48  dz
 *      Initial revision
 *
 *
 *
 *
\*/



#ifndef	SW_EXCEPT_H
#define	SW_EXCEPT_H

#include "strng.h"
//#include "compiler.h"
//#include "log.h"

//typedef const char *   exc_s;
typedef string   exc_s;

class _Export sw_Exception
    {
    public:
        exc_s   where;   // Func. name, file
        exc_s   cause;   // Description of what happened
        exc_s   data;    // Offending data
        
        sw_Exception( exc_s a_where, exc_s a_cause, exc_s a_data )
            : where(a_where), cause( a_cause ), data( a_data )
            {}

        sw_Exception( exc_s a_where, exc_s a_cause )
            : where( a_where ), cause( a_cause )
            {}
        
        void print_error();
/*
            {
            error
                ( 
                 EI_None, "Exception in %s: %s (%s)",
                 where.c_str(), cause.c_str(), data.c_str()
                );
            }
*/
    };

//
//  ⠫ .  ᫮.   ,
// ᫨ ,    -   .
//

class _Export sw_Fail : public sw_Exception
    {
    public:
        sw_Fail( exc_s a_where, exc_s a_cause, exc_s a_data = "")
            : sw_Exception( a_where, a_cause, a_data )
            { }
    };


class _Export sw_Errno : public sw_Exception
    {
    public:
        sw_Errno( const string & wh, const string & wa, long e )
            : sw_Exception( wh, wa )
            {
            char es[100];
            sprintf( es, "%ld", e );
            data = es;
            }
    };

//
// 㬥 㭪樨 -   
//

// ਬ:
//  throw sw_Invalid_Argument
//     ( __FILE__, "Bad address domain part", string(ap) );


class _Export sw_Invalid_Argument : public sw_Fail
    {
    public:
        sw_Invalid_Argument(exc_s a_where, exc_s a_cause, exc_s a_data )
            : sw_Fail( a_where, a_cause, a_data )
            { }
    };

//
//  ⠯  ᠭ  ,  ᠭ  ⠪
//

// ਬ:
//  throw sw_Invalid_Setup
//     ( "", "", "" );

class _Export sw_Invalid_Setup : public sw_Fail
    {
    public:
        sw_Invalid_Setup(exc_s a_where, exc_s a_cause, exc_s a_data )
            : sw_Fail( a_where, a_cause, a_data )
            { }
    };

//
//  㤮⢮७ 譥 ( ⭮襭  ࠬ) ᫮ -
//  䠩, ᡮ ᪮, etc
//

// ਬ:
//  throw sw_Invalid_Condition
//     ( __FILE__, "returned error", "" );

class _Export sw_Invalid_Condition : public sw_Fail
    {
    public:
        sw_Invalid_Condition(exc_s a_where, exc_s a_cause, exc_s a_data )
            : sw_Fail( a_where, a_cause, a_data )
            { }
    };

//
// ꥪ   ⮬ ﭨ,   뢠 ᥩ ⮤.
//

// ਬ:
//  throw sw_Invalid_Object
//     ( __FILE__, "object is not ready for operation", "" );

class _Export sw_Invalid_Object : public sw_Fail
    {
    public:
        sw_Invalid_Object(exc_s a_where, exc_s a_cause, exc_s a_data )
            : sw_Fail( a_where, a_cause, a_data )
            { }
    };

//
//  䠩,   ,  ᯨ᪠ -  饬, 䨭.
//

// ਬ:
//  throw sw_EOF
//     ( __FILE__, "end of file", "" );

class _Export sw_EOF : public sw_Fail
    {
    public:
        sw_EOF(exc_s a_where, exc_s a_cause, exc_s a_data )
            : sw_Fail( a_where, a_cause, a_data )
            { }
    };


// --------------------------------------------------------------------------

//
//  檨  㤭० .
//  ⭮ ⢠   ᢥ稢.
//

class _Export sw_Error : public sw_Exception
    {
    public:
        sw_Error(exc_s a_where, exc_s a_cause, exc_s a_data )
            : sw_Exception( a_where, a_cause, a_data )
            { }
    };





//
//  !! ࠧ.
//

// ਬ:
//  throw sw_Out_Of_Memory
//     ( __FILE__, "go buy large HD", "" );

class _Export sw_Out_Of_Memory : public sw_Error
    {
    public:
        sw_Out_Of_Memory(exc_s a_where )
            : sw_Error( a_where, "Out of memory", "" )
            { }
    };

//
// ७ 訡. ࠧ.
//

// ਬ:
//  throw sw_Internal_Error
//     ( __FILE__, "crazy", "" );

class _Export sw_Internal_Error : public sw_Error
    {
    public:
        sw_Internal_Error(exc_s a_where, exc_s a_cause, exc_s a_data )
            : sw_Error( a_where, a_cause, a_data )
            { }
    };


//
// ^Break.  㪠  ⮫쪮 (?)  Stop_U1_On_Signal()
//

// ਬ:
//  throw sw_Break( __FILE__ );

class _Export sw_Break : public sw_Error
    {
    public:
        sw_Break(exc_s a_where, exc_s a_cause )
            : sw_Error( a_where, a_cause, "" )
            { }
    };





#endif // SW_EXCEPT_H



