#define INCL_DOSQUEUES
#define INCL_DOSFILEMGR
#define INCL_DOSERRORS
#define INCL_DOSPROCESS
#define INCL_DOSMISC
#define INCL_DOSEXCEPTIONS
#include <os2.h>

#include <process.h>
#include <builtin.h>

#include "except.h"
#include "misc.h"
#include "timer.h"

#include "language.h"

static Window *wnd;

void _Optlink readTh( void *h )
{
  registerExceptionHandler

  while( comPort->isOnline() )
  {
    char str[401];

    if( !comPort->promptString( NULL, str, sizeof(str)-1, 0 ) )
    {
      ULONG writ(0);

      strcat( str, "\n" );
      comPort->put( '\r' );
      comPort->put( '\n' );

      for( char *p=str; *p; p+=writ )
        DosWrite( long(h), p, strlen(p), &writ );

      (*wnd) << str << char('\r');
    }
  }

  deRegisterExceptionHandler
}

void _Optlink writeTh( void *h )
{
  registerExceptionHandler

  while( comPort->isOnline() )
  {
    ULONG rea=0;
    char ch;

    DosRead( long(h), &ch, 1, &rea );

    if( rea )
    {
      comPort->putWait( ch, 1 );
      (*wnd) << char(ch);

      if( ch=='\n' )
      {
        comPort->putWait( '\r', 1 );
        (*wnd) << char('\r');
      }
    }
  }

  deRegisterExceptionHandler
}

int login()
{
  int tryCnt=0;

  for( Timer t( 120000l ); !t.isTimeOut() && ++tryCnt!=4; )
  {
    char   name[256] = {0};
    char passwd[256] = {0};

    if( comPort->promptString( "\r\nName ? ", name, sizeof(name), t.rest() ) )
      break;

    if( comPort->promptString( "\r\nPassword ? ", passwd, sizeof(passwd), t.rest(), 1 ) )
      break;

    for( long crc(0), i(0); passwd[i]; crc = (crc << 1) ^ passwd[i++] ) continue;

    crc &= 0x7fffffff;

    if( strcmp( miscInfo.remoteName, name ) || miscInfo.remotePassword!=crc )
    {
      char msg[] = "\r\n\n*** ACCESS DENIED ***\r\n";

      log( Msg_ErrLoginDenied, name, passwd );
      comPort->putOnLine( msg, sizeof(msg)-1, t.rest() );
    }
    else
      return 0;
  }

  if( t.isTimeOut() )
    log( Msg_ErrLoginTimeOut );

  if( comPort->isOnline() )  //check
    log( Msg_ErrLoginNoCarrier );

  return 1;
}

void internalLogin()
{
  if( miscInfo.remoteName==NULL || !miscInfo.remotePassword )
  {
    char msg[] = "\r\n\n*** ACCESS NOT AVAILABLE ***\r\n";

    log( Msg_LoginNotAvail );
    comPort->putOnLine( msg, sizeof(msg)-1, 5000 );
    sleep( 1000 );
    return;
  }

  log( Msg_LoginRequested );
  showAction( Msg_ScrLoginRequested );

  if( login() )
    return;

  APIRET rc;
  HFILE readHandle,
        writeHandle,
        readHandle1,
        writeHandle1;

  if( rc=DosCreatePipe( &readHandle, &writeHandle, 512*1024 ), rc )
  {
    log( Msg_ErrLoginPipe, rc );
    os2Message( rc );
    return;
  }

  if( rc=DosCreatePipe( &readHandle1, &writeHandle1, 512*1024 ), rc )
  {
    DosClose( readHandle );
    DosClose( writeHandle );
    log( Msg_ErrLoginPipe, rc );
    os2Message( rc );
    return;
  }

  suspendClipServer();

  HFILE consoleHandle0 = 0xffffffffl,
        consoleHandle1 = 0xffffffffl,
        consoleHandle2 = 0xffffffffl;

  DosDupHandle( 0, &consoleHandle0 );
  DosDupHandle( 1, &consoleHandle1 );
  DosDupHandle( 2, &consoleHandle2 );

  HFILE hand1 = 1,
        hand2 = 2,
        hand0 = 0;

  DosDupHandle( writeHandle, &hand1 );
  DosDupHandle( writeHandle, &hand2 );
  DosDupHandle( readHandle1, &hand0 );

  long tid  = _beginthread( writeTh, 0, 40960, (void *)readHandle );

  if( tid==-1 )
    log( Msg_ErrLoginServ2, strError() );

  long tid1 = _beginthread( readTh, 0, 40960, (void *)writeHandle1 );

  if( tid1==-1 )
    log( Msg_ErrLoginServ1, strError() );

  long pid(-1);

  const failureObjectSize( 256 );

  char *failureObject = (char *)_alloca( failureObjectSize );
  *failureObject = 0;
  RESULTCODES ret;
  char *cmd,
       *nn = getenv( "COMSPEC" );

  if( nn==NULL )
    nn = "CMD.EXE";

  cmd = (char *)_alloca( strlen(nn)+2 );
  strcpy( cmd, nn );
  cmd[ strlen(cmd)+1 ] = 0;
  capitalize( cmd );

  log( Msg_LogExecute, cmd );

  Color remCl( WHITE, BLACK );

  wnd = new Window( 1, 1, Z_ScreenWidth, Z_ScreenHeight, remCl  );
  +(*wnd);

  rc = DosExecPgm( failureObject, failureObjectSize,
                   EXEC_ASYNC, cmd, (PSZ)0, &ret, cmd );

  if( !rc )
    pid = ret.codeTerminate;

  if( pid==-1 )
  {
    log( Msg_LogExecErr, rc, failureObject );
    os2Message( rc );
  }

  if( tid!=-1 && tid1!=-1 && pid!=-1 )
    while( 1 )
    {
      if( !comPort->isOnline() )
      {
        log( Msg_ErrLoginNoCarrier );
        break;
      }

      PID completed;

      rc = DosWaitChild( DCWA_PROCESS, DCWW_NOWAIT, &ret, &completed, pid );

      if( rc != ERROR_CHILD_NOT_COMPLETE )
      {
        log( Msg_ErrLoginNoChild, rc );
        os2Message( rc );
        break;
      }

      sleep( 1000 );
    }

  DosKillThread( tid );
  DosKillThread( tid1 );

  delete wnd;
  DosKillProcess( DKP_PROCESSTREE, pid );

  DosClose( readHandle );
  DosClose( writeHandle );
  DosClose( readHandle1 );
  DosClose( writeHandle1 );

  comPort->lockSema();

  DosWaitThread( (PULONG)&tid,  DCWW_WAIT );
  DosWaitThread( (PULONG)&tid1, DCWW_WAIT );

  comPort->unlockSema();

  hand1 = 1;
  hand2 = 2;
  hand0 = 0;
  DosDupHandle( consoleHandle0, &hand0 );
  DosDupHandle( consoleHandle1, &hand1 );
  DosDupHandle( consoleHandle2, &hand2 );
  resumeClipServer();
}
