#if !defined(_MY_MODEM_H)

#define _MY_MODEM_H

#include <stddef.h>
#include <string.h>
#include <iostream.h>

#include "timer.h"
#include "buffer.h"
#include "window.h"
#include "class.h"

struct ModemStrings
{
  char *ring,
       *busy,
       *ok,
       *connect,
       *nodialtone,
       *rring,
       *nocarrier,
       *error,
       *voice,
       *noanswer,
       *noconnect;
  char *statCmd,
       *statStop;
  char *callersId;
};

enum FOSSILbaud { B300    = 1, B600   =  2, B1200   = 3,  B2400  = 4,
		  B4800   = 5, B9600  =  6, B19200  = 7,  B38400 = 8,
                  B57600  = 9, B76800 = 10, B115200 = 11, B230400 = 12,
                  B345600 = 13 };

class Modem
{
  long lastBaud;
  long tranx;
  int  speed;

  char *name;

  short portNumber;
  long handle;

  int  state;

  char hardWare,
       breakMode;
  char monitorOn;

  char virtualDCD,
       virtualDCDOn;

  long failSafe,
       carrierMonitor;

  int OutBufferSize;

  unsigned char *outBuffer;

  int curOutPointer;

  int blockMode;

  char *initStr, *offHookStr, *answerStr;
  char *dialSuff, *dialPref, *afterDial;
  char *abortDial, *onExit;

  short callWait,
        answerWait,
        pause;

  long actualSpeed;
  char *connectSuff;
  char *callersId;
  char arq;
  char halfDuplex;
  char locked;

protected:
  int waitConnect( int dialingMode = 0, int wait = 0 );
  void getRing();
  void noCallersId();

public:
  Modem( short port=-1, FOSSILbaud speed=B19200,
	 char *ini=NULL, char *off=NULL, char *ans=NULL,
	 char *pref=NULL, char *suff=NULL, char *after=NULL,
         char *term=NULL,
	 short callWt=60,
	 short answerWt=60,
         short pause=0, long failsafe=60000l,
	 int blockmode=0, long carrierDelay=0, char hardware=1,
	 char breakmode=0, int handle=-1, char *onExit=NULL, char virlualDcd=0 );
  ~Modem();

   friend ostream& operator << ( ostream& o, const Modem& c );
   virtual int initialize();

// Low Level

  virtual void flushBufferedData();
  virtual void closePort();
  virtual int openPort( int =0 );

  virtual int setup();

  virtual void initSpeed();

  virtual int  ok()                  { return !state; }
  virtual void dtr(int f );
  virtual int setbaud( int b=0 );
  virtual void xonxoff(int f );
  virtual void ctsrts(int f );
  virtual void do_break( int f );
  virtual void flowControl( int f )  { hardWare ? ctsrts( f ) : xonxoff( f ); }
  virtual void com_kick()            { flowControl(0); flowControl(1); }

  virtual int checkOnline();

  virtual void ClearOutQue();
  virtual void clearin();
  virtual int isOnline();
  virtual void clearout();

  virtual int putFos(const void* p, int l);    // write buffer; return actual count
  virtual int putFos(int c) { return putFos( (void *)&c, 1 ); }  // write char nowait; return count

  virtual int put( int c );
  virtual int get();
  virtual int putWait( int c, int dcd=1 );
  virtual int putMessage( char * );

  virtual void byDCD();

  virtual void getBuffer();
  virtual int bufferByte( int );
  virtual int unBufferBytes();

  virtual Buffer *getOffLine( int time );
  virtual void stripEcho( int time );

  virtual int haveInput();
  virtual int putOffLine( char *buf, char *ok=NULL, char echo=1 );
  virtual int putOnLine( char *buf, int len, long time );

  virtual void receive( int fl );

  virtual void suspend();
  virtual void resume();

// LOCKs

  void lock() { locked=1; }
  void unlock() { locked=0; }
  int isLocked() { return locked || handle==-1; }

// info

  virtual void stc();

  virtual int hasInputQueue();
  virtual int hasOutputQueue();
  virtual int getBreakMode() const { return breakMode; }
  virtual int getHardWareMode() const { return hardWare; }
  virtual short getPort() const { return short(portNumber+1); }
  virtual int getHandle() const { return handle; }
  virtual int getWaitCallTime() const { return callWait; }
  virtual int getWaitAnswerTime() const { return answerWait; }
  virtual char *getPrefix() const { return dialPref; }
  virtual char *getSuffix() const { return dialSuff; }
  virtual char *getCallersId() const { return callersId; }
  virtual int outEmpty() { return !hasOutputQueue(); }
  virtual int isRinging();
  virtual void ringSignal();
  virtual void statistics();

  virtual long getSpeed()
  { 
    if( !actualSpeed )
      initSpeed();

    return actualSpeed;
  }

  virtual char *getSuff() const { return connectSuff; }
  virtual int isOpened() const { return handle!=-1; }
  virtual char isArq() const { return arq; }
  virtual char isHalfDuplex() const { return halfDuplex; }
  virtual long getTranx() { return tranx; }
  long getLastBaud() { return lastBaud; }

// High Level

  virtual int initOnce();
  virtual int init();
  virtual int hangUp();
  virtual int offHook( char * =(char *)(void *)0 );
  virtual int waitOk( char * =(char *)(void *)0 );
  virtual void setConnectString( char *str );
  virtual void processConnect( int& done, int dialingMode, Timer& t,
                               int fromMailer, Buffer *fromModem );
  virtual int dial( char *number, char *pref=NULL,
		           char *suff=NULL, int wait=0, int blAvail=0 );
  virtual int answer( int );
  virtual int promptString( char *pr, char *st, int len,
                                   long secs, int flag=0 );
  virtual void setConnectInfo( int );
};

class FictionModem : public Modem
{
  void getRing() {}
public:
  FictionModem() : Modem() {}

  int initialize() { return 0; }
  friend ostream& operator << ( ostream& o, const FictionModem& c )
    { o << "No modem" << endl; return o; }
  void closePort() {}
  int openPort( int =0 ) { return 0; }
  int setup() { return 0; }
  void initSpeed() {}
  int ok() { return 1; }
  void dtr( int ) {}
  int setbaud( int =0 ) { return 0; }
  void xonxoff( int ) {}
  void ctsrts( int ) {}
  void do_break( int ) {}
  void flowControl( int ) {}
  void com_kick() {}
  int checkOnline() { return 0; }
  void ClearOutQue() {}
  void clearin() {}
  int isOnline() { return 0; }
  void clearout() {}

  int putFos( const void*, int h ) { return h; }
  int putFos( int ) { return 1; }

  int put( int c ) { return 1; }
  int get() { return 0; }
  int putWait( int, int ) { return 1; }
  int putMessage( char * ) { return 0; }

  void byDCD() {}

  void getBuffer() {}
  int bufferByte( int ) { return  0; }
  int unBufferBytes() { return 0; }

  Buffer * getOffLine( int ) { return new Buffer; }
  void stripEcho( int ) {}

  int haveInput() { return 0; }
  int putOffLine( char *, char * = NULL, char =1 ) { return 0; }
  int putOnLine( char *, int, long ) { return 0; }

  void receive( int ) {}

  virtual void suspend() {}
  virtual void resume() {}

// info

  int hasInputQueue() { return 0; }
  int hasOutputQueue() { return 0; }
  int getBreakMode() const { return 0; }
  int getHardWareMode() const { return 1; }
  short getPort() const { return 0; }
  int getHandle() const { return 0; }
  int getWaitCallTime() const { return 0; }
  int getWaitAnswerTime() const { return 0; }
  char *getPrefix() const { return ""; }
  char *getSuffix() const { return ""; }
  char *getCallersId() const { return 0; }
  int isRinging() { return 0; }
  void ringSignal() {}
  virtual void statistics() {}

  long getSpeed() { return 300; }
  char *getSuff() const { return "None"; }
  int isOpened() const { return 1; }
  char isArq() const { return 0; }
  char isHalfDuplex() const { return 0; }

  long getTranx() { return 0; }
  long getLastBaud() { return 0; }

// High Level

  int initOnce() { return 0; }
  int init() { return 0; }
  int hangUp() { return 0; }
  int offHook( char * =(char *)(void *)0 ) { return 0; }
  int waitOk( char * =(char *)(void *)0 ) { return 0; }
  void setConnectString( char * ) {}
  void processConnect( int&, int, Timer&, int, Buffer * ) {}
  int dial( char *,
                   char * =NULL,
                   char * =NULL,
                   int =0,
                   int =0 )
    { return 1; }

  int answer( int )  { return 1; }
  virtual int promptString( char *, char *, int, long, int =0 )
    { return 1; }
  virtual void setConnectInfo( int ) {}
};

extern Modem *comPort;

struct DefinePort
{
  FOSSILbaud speed;
  char *init,
       *answer,
       *hangup,
       *pref,
       *suff,
       *after,
       *dialterm,
       *onExit;
  long failSafe;
  int callWait,
      answerWait,
      pause,
      blockMode,
      carrierDelay,
      hardwareFlow,
      breakMode,
      virtualDCD;
};

extern DefinePort portInfo;
extern ModemStrings modemString;

class ModemString
{
private:
  char *str;

public:

  ModemString( char *f ) : str( newStr( f ) ) {}

  ~ModemString() { delete str; }

  int operator == ( const ModemString &t ) const
    { return !stricmp( str, t.str ); }

  friend ostream& operator << ( ostream& o, const ModemString& c )
  {
    o << "{Str:" << c.str << char('}');
    return o;
  }

  char *getString() const { return str; }
};

class ModemStringList : public OrdinalCollection<ModemString>
{
public:
  ModemStringList() : OrdinalCollection<ModemString>( 5, 5, SemaphoreModeNoNeed )
  {}
};

class ModemStringIterator : public OrdinalCollectionIterator<ModemString>
{
public:
  ModemStringIterator( ModemStringList& ini ) :
    OrdinalCollectionIterator<ModemString>( (OrdinalCollection<ModemString>&)ini )
  {}
};

extern ModemStringList *ignoreString;
extern ModemStringList *badCounterString;

extern char  zmodemActive;

extern Window *wndModemPtr;
extern Window *wndModemFrame;
extern volatile long wndModemLast;

#define wndModem     (*wndModemPtr)

void controlChar( int ch, int to );

inline void dumpTo( int ch, int fl=0 )
{
  controlChar( fl ? ch : ch&0xff, 1 );
}

inline void dumpFrom( int ch, int fl=0 )
{
  controlChar( fl ? ch : ch&0xff, 0 );
}

int processBadModem();
extern volatile long modemOs2Rc;

#endif
