#define INCL_DOSEXCEPTIONS
#define INCL_DOSFILEMGR
#define INCL_DOSDEVIOCTL
#define INCL_DOSDEVICES
#define INCL_DOSPROCESS
#define INCL_DOSERRORS
#include <os2.h>

#include <stdlib.h>
#include <stdio.h>
#include <io.h>
#include <process.h>

#include "except.h"

#include "language.h"

#include "misc.h"
#include "timer.h"
#include "buffer.h"
#include "key.h"
#include "newstr.h"

static volatile ULONG curInPointer = 0,
  	        curInSize = 0;

static ULONG minStat = 0x7fffffff,
	     maxStat = 0,
	     cntStat = 0,
	     ttlStat = 0,
             emtStat = 0;

static const inBufferSize = 1024;

static char inBuffer[ inBufferSize ];

volatile long modemOs2Rc;

void Modem::flushBufferedData()
{
  curInPointer = curInSize = 0;
}

void Modem::suspend()
{
}

void Modem::resume()
{
}

void Modem::getBuffer()
{
  curInPointer = curInSize = 0;

  if( handle==-1 )
    return;

  register int how;

  if( (miscInfo.state1&BlockedRead)==BlockedRead )
    how = inBufferSize;
  else
  {
    how = comPort->hasInputQueue();

    if( !how )
    {
      emtStat++;
      return;
    }

    if( how>inBufferSize )
      how = inBufferSize;
  }

  register APIRET rc = DosRead( handle, inBuffer, how, (PULONG)&curInSize );

  if( rc )
    if( rc==ERROR_INTERRUPT || rc==ERROR_INVALID_HANDLE )
      curInSize = 0;
    else
    {
      log( "~Block BUG~ Modem::get() DosRead==%d curInSize==%d", rc, curInSize );
      os2Message( rc );
    }

  if( !curInSize )
  {
    emtStat++;
    return;
  }

  if( curInSize<minStat )
    minStat = curInSize;

  if( curInSize>maxStat )
    maxStat = curInSize;

  ttlStat += curInSize;
  cntStat++;
}

int Modem::haveInput()
{
  if( curInPointer < curInSize )
    return 1;

  getBuffer();

  if( curInPointer < curInSize )
    return 1;

  checkPoint();
  return 0;
}

int Modem::get()
{
  register int r;

  if( curInPointer < curInSize )
    r = int( ((unsigned short)inBuffer[ curInPointer++ ]) & 0xff );
  else
  {
    getBuffer();

    if( curInPointer < curInSize )
      r = int( ((unsigned short)inBuffer[ curInPointer++ ]) & 0xff );
    else
    {
      checkPoint();
      return -1;
    }
  }

  dumpFrom( r );
  return r;
}

static char volatile mayMonitor = 1;

void abortQueueMonitor()
{
  mayMonitor = 0;
}

void _Optlink inputQueueMonitor( void * )
{
  registerExceptionHandler

  DosSetPriority( PRTYS_THREAD, 0, -miscInfo.priorityDelta, 0 );

  Window win( (Z_ScreenWidth-60)/2+1, (Z_ScreenHeight-3)/2+1,
			  60, 3, popUpColor,
			  " inQueueStat ", BS_SINGLELINE, DS_LOWERRIGHT );

  +win;

  char buf[80];

  for( ;; )
  {
    if( !mayMonitor )
      break;

    sprintf( buf, "min=%lu ave=%lu [%lu/%lu:%lu] max=%lu",
             minStat,
             (cntStat ? ttlStat/cntStat : 0),
             ttlStat, cntStat,
             emtStat,
             maxStat );
    win( 1, 1 ).clearLine();
    win.centre( 1, buf );

    DosSleep( 1000 );
  }

  deRegisterExceptionHandler
}

static DCBINFO originDataDCB;
static ULONG sizeOriginDataDCB = sizeof(originDataDCB),
             size1OriginDataDCB = sizeof(originDataDCB);

Modem::Modem( short port, FOSSILbaud sp,
	      char *ini, char *off, char *ans,
	      char *pref, char *suff, char *after,
              char *abo,
	      short callWt,
	      short answerWt,
              short p, long failsafe,
	      int blockmode, long carrierDelay,
	      char hardware, char breakmode, int _handle,
              char *_onExit ) :
  inUse( 0 ),
  monitorOn( 0 ),
  locked( 0 ),
  portNumber( short(port) ),
  speed( sp ),
  initStr( newStr(ini) ),
  offHookStr( newStr(off) ),
  answerStr( newStr(ans) ),
  dialPref( newStr(pref) ),
  dialSuff( newStr(suff) ),
  afterDial( newStr(after) ),
  abortDial( newStr(abo) ),
  onExit( newStr(_onExit) ),
  callWait( callWt ),
  answerWait( answerWt ),
  pause( p ),
  connectSuff( newStr( "*First*Try*" ) ),
  actualSpeed( 0 ),
  failSafe( failsafe ),
  outBuffer( NULL ),
  blockMode( blockmode ),
  curOutPointer( 0 ),
  carrierMonitor( carrierDelay ),
  hardWare( hardware ),
  breakMode( breakmode ),
  name( new char [32] ),
  handle( _handle )
{
  sprintf( name, "\\DEV\\COM%d", portNumber+1 );

  if( initStr==NULL )
    initStr = "ATZ0|~~~";

  if( offHookStr==NULL )
    offHookStr = "v~~^~+++~~~ATH0|";

  if( answerStr==NULL )
    answerStr = "ATA|";

  if( dialPref==NULL )
    dialPref = "ATDP";

  if( dialSuff==NULL )
    dialSuff = "|";

  if( afterDial==NULL )
    afterDial = "~";

  if( abortDial==NULL )
    abortDial = "v~~~^`|";

  if( !speed )
    speed = B2400;

  if( !callWait )
    callWait = 30;

  if( !answerWait )
    answerWait = 30;

  if( !failSafe )
    failSafe = 60000l;

  OutBufferSize = 1024;

  outBuffer = new unsigned char [ OutBufferSize ];
}

int Modem::initialize()
{
  int rc = 0;

  rc = openPort( 1 );

  if( rc )
  {
    log( Msg_ModemNotExist );
    return 1;
  }

  if( setup() )
  {
    log( Msg_ModemBigBaud, actualSpeed );
    return 1;
  }

  return 0;
}

Modem::~Modem()
{
  if( handle!=-1 )
  {
    putOffLine( onExit );
    size1OriginDataDCB = sizeOriginDataDCB;

    APIRET rc =DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_SETDCBINFO,
                            &originDataDCB, sizeOriginDataDCB, &size1OriginDataDCB,
                            0, 0, 0 );
    if( rc )
    {
      log( Msg_ErrUnableSetDCB, rc );
      os2Message( rc );
    }
  }

  delete outBuffer;
  closePort();
}

// Low Level

int Modem::setup()
{
  initSpeed();

  struct
  {
    BYTE bits, parity, stop;
  } info = { 8, 0, 0 };

  ULONG size( sizeof( info ) );

  DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_SETLINECTRL,
	       &info, size, &size, 0, 0, 0 );
  state = 0;

  clearin();
  clearout();
  dtr( 1 );
  flowControl( 1 );
  receive( 1 );

  return setbaud( speed );
}

void dumpHandles( ostream *o );

void Modem::closePort()
{
  if( !(portNumber+1) )
  {
    handle = -1;
    return;
  }

  if( handle==-1 )
    log( "~BUG~ void Modem::closePort(), port `%s' already closed!", name );

Again:

  APIRET rc= DosClose( handle );

  if( rc )
  {
    log( "~BUG~ void Modem::closePort(), unable close a port `%s', handle=%d!",
         name, handle );
    os2Message( rc );
  }
  else
  {
    USHORT err = 0;
    ULONG errSize( sizeof(err) );
    APIRET rc=DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_GETMODEMINPUT,
                           0, 0, 0, &err, errSize, &errSize );

    if( !rc )
    {
      log( "~BUG~ void Modem::closePort(), port `%s' not closed! try again.", name );
      goto Again;
    }

    handle = -1;
  }
}

int Modem::openPort( int getDCB )
{
  APIRET rc = 0;

  if( handle==-1 )
  {
    ULONG action;

    rc = DosOpen( (PSZ)name, (ULONG *)&handle, &action,
                  0,
                  FILE_NORMAL,
                  OPEN_ACTION_OPEN_IF_EXISTS,
                  OPEN_FLAGS_FAIL_ON_ERROR|OPEN_FLAGS_SEQUENTIAL|
                  OPEN_SHARE_DENYNONE|OPEN_ACCESS_READWRITE,
                  0 );
    if( rc )
    {
      log( Msg_ModemUnableOpen, name, rc );
      os2Message( rc );
      handle = -1;
    }
  }

  if( handle!=-1 && getDCB )
  {
    rc = DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_GETDCBINFO,
                      0, 0, 0,
                      &originDataDCB, sizeOriginDataDCB, &size1OriginDataDCB );
    if( rc )
    {
      log( Msg_ErrUnableGetDCB, rc );
      os2Message( rc );
    }
  }

  if( !rc )
    rc = setup();

  return rc;
}

void Modem::ClearOutQue()
{
  if( handle==-1 )
    return;

  BYTE param( 0 );
  ULONG size( sizeof(param) );

  short comData( 0 );
  ULONG comDataSize( sizeof(comData) );

  DosDevIOCtl( handle, IOCTL_GENERAL, DEV_FLUSHOUTPUT,
	       &param, size, &size,
	       &comData, comDataSize, &comDataSize );

  curOutPointer = 0;
}

void Modem::clearin()
{
  if( handle==-1 )
    return;

  BYTE param( 0 );
  ULONG size( sizeof(param) );

  short comData( 0 );
  ULONG comDataSize( sizeof(comData) );

  DosDevIOCtl( handle, IOCTL_GENERAL, DEV_FLUSHINPUT,
	       &param, size, &size,
	       &comData, comDataSize, &comDataSize );

  curInPointer = curInSize = 0;
}

void Modem::do_break( int flag )
{
  if( handle==-1 )
    return;

  ULONG size( 1 );

  short comData( 0 );
  ULONG comDataSize( sizeof(comData) );

  DosDevIOCtl( handle, IOCTL_ASYNC,
	       flag ? ASYNC_SETBREAKON : ASYNC_SETBREAKOFF,
	       0, size, &size,
	       &comData, comDataSize, &comDataSize );
}

int Modem::hasInputQueue()
{
  if( handle==-1 )
    return 0;

  RXQUEUE how;
  ULONG howSize( sizeof(how) );

  how.cch = 0;
  DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_GETINQUECOUNT,
	       0, 0, 0, &how, howSize, &howSize );

  return how.cch;
}

int Modem::hasOutputQueue()
{
  if( handle==-1 )
    return 0;

  RXQUEUE how;
  ULONG howSize( sizeof(how) );

  how.cch = 0;
  DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_GETOUTQUECOUNT,
	       0, 0, 0, &how, howSize, &howSize );

  return how.cch;
}

void Modem::stc()
{
  if( handle==-1 || !systemIsReady )
    return;

  int delta=(statusLineWindow.getWidth()-80)/2;

  static BYTE sdOld( 0xff ),
	      rdOld( 0xff );
  register char sd = char( !outEmpty() ),
                rd( 0 );

  if( curInPointer < curInSize )
    rd = 1;
  else
    rd = char( hasInputQueue() ? 1 : 0 );

  if( sd!=sdOld )
    showInd( "SQ", 65+delta, sdOld = sd );

  if( rd!=rdOld )
    showInd( "RQ", 62+delta, rdOld = rd );

  APIRET rc;
  BYTE status = 0;
  ULONG statusSize( sizeof(status) );

  if( handle!=-1 )
    rc=DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_GETMODEMINPUT,
                    0, 0, 0, &status, statusSize, &statusSize );

  else
    rc = 1;

  if( !rc )
  {
    static char octs = -2;
    static char ori  = -2;
    static char odcd = -2;

    register char cts = char( (status&CTS_ON)==CTS_ON );
    register char ri  = char( (status&RI_ON)==RI_ON   );
    register char dcd = char( (status&DCD_ON)==DCD_ON );

    if( octs!=cts )
    {
      octs = cts;
      showInd( "CTS", 48+delta, cts );
    }

    if( ori!=ri )
    {
      ori = ri;
      showInd( "RI",  56+delta, ri );
    }

    if( odcd!=dcd )
    {
      odcd = dcd;
      showInd( "CD",  59+delta, odcd );
    }
  }

  status = 0;

  if( handle!=-1 )
    rc = DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_GETMODEMOUTPUT,
                      0, 0, 0, &status, statusSize, &statusSize );
  else
    rc = 1;

  if( !rc )
  {
    static char orts = -2;
    static char odtr = -2;

    register char rts = char( (status&RTS_ON)==RTS_ON );
    register char dtr = char( (status&DTR_ON)==DTR_ON );

    if( orts!=rts )
    {
      orts = rts;
      showInd( "RTS", 52+delta, rts );
    }

    if( odtr!=dtr )
    {
      odtr = dtr;
      showInd( "DTR", 44+delta, dtr );
    }
  }
}

void Modem::dtr( int flag )
{
  if( handle==-1 )
    return;

  MODEMSTATUS DTRon = { DTR_ON, 0xff },
	      DTRoff = { 0, DTR_OFF };

  short comData( 0 );
  ULONG comDataSize( sizeof(comData) );

  ULONG DTRoffSize( sizeof(DTRoff) );
  ULONG DTRonSize( sizeof(DTRon) );

  if( !flag )
    DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_SETMODEMCTRL,
		 &DTRoff, DTRoffSize, &DTRoffSize,
		 &comData, comDataSize, &comDataSize );
  else
    DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_SETMODEMCTRL,
		 &DTRon, DTRonSize, &DTRonSize,
		 &comData, comDataSize, &comDataSize );
}

int Modem::setbaud( int )
{
  initSpeed();

  struct
  {
    ULONG   baud;
    BYTE   frac;
  } baud = { (unsigned long)actualSpeed, 0 };

  ULONG baudSize( sizeof( baud ) );

  return DosDevIOCtl( handle, IOCTL_ASYNC, 0x43,
		      &baud, baudSize, &baudSize,
		      0, 0, 0 );
}

int Modem::checkOnline()
{
  if( handle==-1 )
    return 0;

  BYTE status = 0;
  ULONG statusSize( sizeof(status) );

  DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_GETMODEMINPUT,
	       0, 0, 0, &status, statusSize, &statusSize );

  return (status&DCD_ON)==DCD_ON;
}

int Modem::putFos( const void* p, int l )
{
  if( handle==-1 )
  {
    checkPoint();
    return 0;
  }

  ULONG written;

  modemOs2Rc = DosWrite( handle, (void *)p, l, &written );

  if( written!=l )
    checkPoint();

  return written;
}

static void setTimeOut( DCBINFO& dataDCB )
{
  dataDCB.fbTimeout &= ~MODE_NO_WRITE_TIMEOUT;
  dataDCB.usWriteTimeout = 10;

  if( (miscInfo.state1&BlockedRead)==BlockedRead )
  {
    dataDCB.usReadTimeout = 10;
    dataDCB.fbTimeout = char( (dataDCB.fbTimeout&(~0x6))|MODE_READ_TIMEOUT );
//  dataDCB.usReadTimeout = 100;
//  dataDCB.fbTimeout = char( (dataDCB.fbTimeout&(~0x6))|MODE_WAIT_READ_TIMEOUT );
//  dataDCB.fbTimeout = char( dataDCB.fbTimeout|MODE_NOWAIT_READ_TIMEOUT );
  }
}

void Modem::ctsrts( int f )
{
  if( handle==-1 )
    return;

  DCBINFO dataDCB;
  ULONG sizeDataDCB = sizeof(dataDCB),
        size1DataDCB = sizeof(dataDCB);

  APIRET rc = DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_GETDCBINFO,
	                   0, 0, 0,
                           &dataDCB, sizeDataDCB, &size1DataDCB );

  if( rc )
  {
    log( Msg_ErrUnableGetDCB, rc );
    os2Message( rc );
  }

  setTimeOut( dataDCB );

  dataDCB.fbCtlHndShake = BYTE( f ? MODE_CTS_HANDSHAKE : 0x00 );
  dataDCB.fbFlowReplace = BYTE( f ? MODE_RTS_HANDSHAKE : 0x00 );

  if( !rc )
  {
    rc = DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_SETDCBINFO,
	              &dataDCB, sizeDataDCB, &size1DataDCB,
	              0, 0, 0 );

    if( rc )
    {
      log( Msg_ErrUnableSetDCB, rc );
      os2Message( rc );
    }
  }
}

void Modem::xonxoff( int f )
{
  if( handle==-1 )
    return;

  DCBINFO dataDCB;
  ULONG sizeDataDCB = sizeof(dataDCB),
        size1DataDCB = sizeof(dataDCB);

  APIRET rc = DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_GETDCBINFO,
	                   0, 0, 0,
                           &dataDCB, sizeDataDCB, &size1DataDCB );

  if( rc )
  {
    log( Msg_ErrUnableGetDCB, rc );
    os2Message( rc );
  }

  setTimeOut( dataDCB );

  dataDCB.fbCtlHndShake = 0x00;
  dataDCB.fbFlowReplace = BYTE( f ? (MODE_AUTO_TRANSMIT|MODE_AUTO_RECEIVE)
                                  : 0x00 );

  if( !rc )
  {
    rc = DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_SETDCBINFO,
	              &dataDCB, sizeDataDCB, &size1DataDCB,
                      0, 0, 0 );
    if( rc )
    {
      log( Msg_ErrUnableSetDCB, rc );
      os2Message( rc );
    }
  }
}

void Modem::receive( int flag )
{
  if( handle==-1 )
    return;

  MODEMSTATUS RTSon = { RTS_ON, 0xff },
 	      RTSoff = { 0, RTS_OFF };

  short comData( 0 );
  ULONG comDataSize( sizeof(comData) );

  ULONG RTSoffSize( sizeof(RTSoff) );
  ULONG RTSonSize( sizeof(RTSon) );

  if( !flag )
  {
    flowControl( 0 );
    DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_SETMODEMCTRL,
		 &RTSoff, RTSoffSize, &RTSoffSize,
		 &comData, comDataSize, &comDataSize );
  }
  else
  {
    DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_SETMODEMCTRL,
		 &RTSon, RTSonSize, &RTSonSize,
		 &comData, comDataSize, &comDataSize );
    flowControl( 1 );
  }
}

int Modem::isRinging()
{
  if( handle==-1 )
    return 0;

  USHORT err = 0;
  ULONG errSize( sizeof(err) );

  DosDevIOCtl( handle, IOCTL_ASYNC, ASYNC_GETMODEMINPUT,
               0, 0, 0, &err, errSize, &errSize );

  return (err&RI_ON)==RI_ON ? 1 : 0;
}
