#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <dos.h>
#include <stdlib.h>
#include <stdarg.h>
#include <time.h>

#include "busy.h"

#ifdef __OS2__
#define TIMEZONE _timezone
#else
#define TIMEZONE timezone
#endif

void brakeDump( int level, unsigned long stampIni, char *fmt ... )
{
  if( level>dumpBrakeBlg )
    return;

  char buf[512];
  time_t stamp( stampIni );
  char *buf1 = ctime( &stamp );
  va_list argptr;

  buf1[ strlen(buf1)-1 ]=0;
  va_start( argptr, fmt );
  vsprintf( buf, fmt, argptr );
  va_end( argptr );
  
  fprintf( stderr, "%s %s\n", buf1, buf );
}

void brakeParser()
{
  if( getenv( "TZ" )==NULL )
  {
    printf( "*Fatal* TZ environiment variable wanted.\n"
	    "        Sorry! Unable to continue. :-(\n" );
    returnToDos( 1 );
  }

  tzset();

  struct SessionInfo
  {
    unsigned short zone, net, node, point;
    char type;
    char protocol;
    char crc;
    unsigned long elapsed;
    unsigned long netSend, netRecv,
		  xmSend,  xmRecv,
		  filesSend, filesRecv;
    char domain[17];
  };

  SessionInfo sesInfo;

  struct
  {
    unsigned long stamp;
    unsigned char code;
  } header;

  long elapsed;

  while( !feof(logFile) && !ferror(logFile) )
  {
    memset( &header, 0, sizeof( header ) );

    int len = fread( &header, 1, sizeof( header ), logFile );
    progressBar();

    if( len!=sizeof( header ) )
      goto need_abort;

    switch( header.code )
    {
    case 5:
    case 6:
      if( fread( &sesInfo, 1, sizeof(sesInfo), logFile )!=sizeof(sesInfo) )
      {
	printf( "*Error* Incomplete input file at record %lu\n", logLines );

need_abort:
	header.code = 255;
	return;
      }
      break;

    case 7:
      break;

    case 254:
      if( fread( &elapsed, 1, sizeof(char), logFile )!=sizeof(char) )
        goto need_abort;
      break;

    default:
      if( fread( &elapsed, 1, sizeof(elapsed), logFile )!=sizeof(elapsed) )
        goto need_abort;
    }

    logLines++;


    if( header.stamp<=hystHeader.date )
    {
      skipLines++;
      startTime = header.stamp;
      continue;
    }

    struct tm *tmm = localtime( (time_t *)&header.stamp );
    strftime( logEnd, sizeof(logEnd), "%d %b %H:%M:%S", tmm );
    logEnd[ sizeof(logEnd)-1 ] = 0;

    if( !*logStart )
      strcpy( logStart, logEnd );

    if( header.stamp>endTime )
      endTime = header.stamp;

    if( !startTime )
      startTime = endTime;

    stamp = endTime;
    actionStarted = stamp;

    static char buf[32],
		*protocolName[4] = { "Zmodem", "ZedZap", "DirZap", "Hydra" };
    index curNode;
    char *work;

    switch( header.code )
    {
    case 255:
      if( ferror( logFile ) )
	printf( "*Error* Can't read file at record %lu\n", logLines );
      break;

    case 7: 			// Empty stamp
      brakeDump( 2, header.stamp, "Alive" );
      break;
      
    case 254: 			// version
      brakeDump( 2, header.stamp, "Version:  %d", char(elapsed) );
      break;

    case 0:			// Answer
      totalAnswering += registerTime( elapsed, Answering );
      inSessions++;
      brakeDump( 2, header.stamp, "Answer:   %d", elapsed );
      break;

    case 1:			// Dial
      totalDialing += registerTime( elapsed, Dialing );
      outSessions++;
      brakeDump( 2, header.stamp, "Dialing:  %d", elapsed );
      break;

    case 2:			// terminal
      totalTerminal += registerTime( elapsed, terminalCalled );
      brakeDump( 2, header.stamp, "Terminal: %d", elapsed );
      break;

    case 3:			// bbs
      totalBBS += registerTime( elapsed, BBS );
      brakeDump( 2, header.stamp, "BBS:      %d", elapsed );
      break;

    case 4:			// external
      totalExternal += registerTime( elapsed, Executing );
      brakeDump( 2, header.stamp, "External: %d", elapsed );
      break;

    case 5:			// income session
      work = "In";
      goto session;
      
    case 6:			// outcome session
      work = "Out";
session:
      char bbProt[16],
           *bbProto;

      if( (unsigned)sesInfo.protocol
          <
          (unsigned)sizeof( protocolName )/sizeof( protocolName[0] ) )
        bbProto = protocolName[ sesInfo.protocol ];
      else
      {
        sprintf( bbProt, "?%d?", sesInfo.protocol );
        bbProto = bbProt;
      }

      brakeDump( 1, header.stamp, 
                "%s%-3s "
//                "%s "
                "%d:%02d:%02d %d:%d/%d.%d%s %s-%d %d/%d %d/%d %d/%d", 
                dumpBrakeBlg==1 ? "" : "Session:  ",
                work,
/*
                 sesInfo.type==0 ? "FTS-1" : 
                 sesInfo.type==1 ? "WaZOO" : 
                 sesInfo.type==2 ? "EMSI" : "?",
*/
                sesInfo.elapsed/60/60,
                (sesInfo.elapsed/60)%60,
                sesInfo.elapsed%60,
                sesInfo.zone, sesInfo.net, sesInfo.node, sesInfo.point, sesInfo.domain,
                bbProto, sesInfo.crc,
                sesInfo.netSend, sesInfo.netRecv,
                sesInfo.xmSend, sesInfo.xmRecv,
                sesInfo.filesSend, sesInfo.filesRecv
               );
      
      for( curNode=0; curNode<totalNodes; curNode++ )
	if( nodesInfo[curNode]->point==sesInfo.point &&
	    nodesInfo[curNode]->node==sesInfo.node &&
	    nodesInfo[curNode]->net==sesInfo.net &&
	    nodesInfo[curNode]->zone==sesInfo.zone && 
	    !stricmp( nodesInfo[curNode]->domain, sesInfo.domain )
          )
	  break;

      if( curNode==totalNodes )
      {

	if( totalNodes==MaxItems )
	{
	  printf( "*Fatal* Too many nodes! Support up to %u nodes.\n"
		  "        Sorry! Program aborted at record %lu :(\n",
		  MaxItems, logLines );
	  returnToDos( 1 );
	}

	curNode=totalNodes++;

	nodesInfo = (NodeInfo **)realloc( nodesInfo, totalNodes*sizeof(NodeInfo **) );
	checkPtr( nodesInfo );

	NodeInfo *newNode( new NodeInfo );
	checkPtr( newNode );

	memset( newNode, 0, sizeof(NodeInfo) );
	nodesInfo[curNode] = newNode;

	newNode->domain = new char[ strlen(sesInfo.domain) +1 ];
	checkPtr( newNode->domain );
	strcpy( newNode->domain, sesInfo.domain );

	newNode->zone=sesInfo.zone;
	newNode->net=sesInfo.net;
	newNode->node=sesInfo.node;
	newNode->point=sesInfo.point;
      }

      if( header.code==5 )
      {
	totalIncome += registerTime( sesInfo.elapsed, Incoming );
	nodesInfo[curNode]->income += sesInfo.elapsed;
	inHand++;
	nodesInfo[curNode]->inSession++;
      }
      else
      {
	totalOutcome += registerTime( sesInfo.elapsed, Outcoming );
	nodesInfo[curNode]->outcome += sesInfo.elapsed;
	outHand++;
	nodesInfo[curNode]->outSession++;
      }

      totalBytesIn += sesInfo.netRecv+sesInfo.xmRecv+sesInfo.filesRecv;
      inMail += sesInfo.netRecv;
      inXmail += sesInfo.xmRecv;
      inFiles += sesInfo.filesRecv;

      totalBytesOut += sesInfo.netSend+sesInfo.xmSend+sesInfo.filesSend;
      outMail += sesInfo.netSend;
      outXmail += sesInfo.xmSend;
      outFiles += sesInfo.filesSend;

      nodesInfo[curNode]->received += sesInfo.netRecv+sesInfo.xmRecv+sesInfo.filesRecv;
      nodesInfo[curNode]->inMail += sesInfo.netRecv;
      nodesInfo[curNode]->inXmail += sesInfo.xmRecv;
      nodesInfo[curNode]->inFiles += sesInfo.filesRecv;

      nodesInfo[curNode]->sent += sesInfo.netSend+sesInfo.xmSend+sesInfo.filesSend;
      nodesInfo[curNode]->outMail += sesInfo.netSend;
      nodesInfo[curNode]->outXmail += sesInfo.xmSend;
      nodesInfo[curNode]->outFiles += sesInfo.filesSend;

      nodesInfo[curNode]->ioTime += sesInfo.elapsed;

      sprintf( buf, "%s-%d", bbProto, int( sesInfo.crc ) );
      registerIO( buf );

      switch( sesInfo.type )
      {
      case 2:
	sesEMSI++;
	break;

      case 1:
	sesWaZOO++;
	break;

      case 0:
	sesFTS0001++;
	break;

      default:
	printf( "*Warning* Unknown session type at record %lu\n", logLines );
	sesXZ++;
      }
      break;

    default:
      printf( "*Error* Illegal record type %d at record %lu\n", int(header.code), logLines );
    }
  }

  progressBar();
}

