#include <errno.h>
#include <fcntl.h>
#include <io.h>
#include <share.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "jammb.h"
#include "jamsub.h"

#include "busy.h"

void makeJam( char *path, FILE *file,
	      char *from, char *to, char *subj,
	      char *addr, char *toAddr ){

  printf( "Creating JAM-base message\n" );

  fflush( file );
  rewind( file );

  int txtLen = int( filelength( fileno( file ) ) );

  unsigned short zone(0), net(0), node(0), point(0);
  char *domain,
       *pntKludge1( NULL ),
       *pntKludge2( NULL ),
       *intlKludge( NULL ),
       *msgId( NULL );

  checkAddress( addr, zone, net, node, point, &domain, 1 );

  unsigned short toZone(0), toNet(0), toNode(0), toPoint(0);

  if( toAddr!=NULL )
    checkAddress( toAddr, toZone, toNet, toNode, toPoint, &domain, 1 );

  if( point )
  {
    pntKludge1 = new char [16];
    checkPtr( pntKludge1 );
    sprintf( pntKludge1, "FMPT %u", point );
  }

  if( toPoint )
  {
    pntKludge2 = new char [16];
    checkPtr( pntKludge2 );
    sprintf( pntKludge2, "TOPT %u", toPoint );
  }

  if( toAddr!=NULL )
  {
    char *klName( "INTL " );
    intlKludge = new char [ strlen( addr )+strlen( toAddr )+
			    strlen( klName )+1+3 ];

    checkPtr( intlKludge );

    strcpy( intlKludge, klName );
    strcat( intlKludge, toAddr );
    strcat( intlKludge,  " " );
    strcat( intlKludge, addr );
  }

  msgId = new char[ strlen(addr)+10 ];

  sprintf( msgId, "%s %08lx", addr, time(NULL) );

  JAMAPIREC JamRec;
  UINT32 SubFldPos;

  if( !JAMsysInitApiRec( &JamRec, path, 2048 ))
  {
    printf( "Not enough memory for JAM-base operation\n" );

jam_error:
    completeCode = 12;
    delete pntKludge1;
    delete pntKludge2;
    delete intlKludge;
    delete msgId;
    return;
  }

  if( !JAMmbOpen( &JamRec ))
  {
    if( JamRec.Errno != ENOENT )
    {
jam_error_1:
      IOerror( path );
      JAMsysDeinitApiRec( &JamRec );
      goto jam_error;
    }

    if( !JAMmbCreate( &JamRec ) )
      goto jam_error_1;
  }

  JamMsgInit( &JamRec );
  SubFldPos = 0;

  JamMsgAddSFldStr( &JamRec, JAMSFLD_SENDERNAME, from, &SubFldPos );
  JamMsgAddSFldStr( &JamRec, JAMSFLD_RECVRNAME, to, &SubFldPos );
  JamMsgAddSFldStr( &JamRec, JAMSFLD_SUBJECT, subj, &SubFldPos );

  JamRec.Hdr.TimesRead = 0;
  JamRec.Hdr.Cost      = 0;

  JamRec.Hdr.DateWritten = JAMsysTime( NULL );
  JamRec.Hdr.DateProcessed = 0;

  JamRec.Hdr.Attribute = MSG_LOCAL |
                         (toAddr==NULL ? MSG_TYPEECHO : MSG_TYPENET );

  JamMsgAddSFldStr( &JamRec, JAMSFLD_OADDRESS, addr, &SubFldPos );

  if( toAddr!=NULL )
    JamMsgAddSFldStr( &JamRec, JAMSFLD_DADDRESS, toAddr, &SubFldPos );

  JamMsgAddSFldStr( &JamRec, JAMSFLD_MSGID, msgId, &SubFldPos );
  JamMsgAddSFldStr( &JamRec, JAMSFLD_PID,
                    "Vagabondo"
#if defined(__OS2__)
                    "/2"
#endif
                    , &SubFldPos );

  if( intlKludge!=NULL )
    JamMsgAddSFldStr( &JamRec, JAMSFLD_FTSKLUDGE, intlKludge, &SubFldPos );

  if( pntKludge1!=NULL )
    JamMsgAddSFldStr( &JamRec, JAMSFLD_FTSKLUDGE, pntKludge1, &SubFldPos );

  if( pntKludge2!=NULL )
    JamMsgAddSFldStr( &JamRec, JAMSFLD_FTSKLUDGE, pntKludge2, &SubFldPos );

  char *MsgText( new char[ txtLen+1 ] );

  checkPtr( MsgText );

  txtLen = fread( MsgText, 1, txtLen, file );

  if( ferror( file ) )
  {
    printf( "Can`t read work file `%s'\n", statName );
    completeCode = 12;
  }
  else
  {
    MsgText[ txtLen ] = 0;
    JamRec.Hdr.SubfieldLen = SubFldPos;
    JamRec.Hdr.TxtLen = txtLen;

    for( int i(0); i<txtLen; i++ )
      if( MsgText[i]=='\n' )
         MsgText[i]='\r';

    if( !JamMsgWrite( &JamRec, ( CHAR8 * ) MsgText ) )
    {
      delete MsgText;
      goto jam_error_1;
    }

    JamMsgDeinit( &JamRec );
    JAMmbClose( &JamRec );
    JAMsysDeinitApiRec( &JamRec );
  }

  delete MsgText;
  delete pntKludge1;
  delete pntKludge2;
  delete intlKludge;
  delete msgId;

}
