#include <stdio.h>
#include <string.h>
#include <io.h>
#include <time.h>

#ifndef __EMX__
#include <mem.h>
#include <alloc.h>
#include <dir.h>
#endif

#ifdef __EMX__
#include <sys/types.h>
#include <dirent.h>
#endif

#include <dos.h>
#include <stdlib.h>

#include "busy.h"

#ifndef __EMX__
long getMaxMsgNumber( char *msg_path )
{
 char mask[]( "*.msg" ),
      *msg_dir = new char [ strlen(msg_path)+sizeof(mask) ];

 strcpy( msg_dir, msg_path );
 strcat( msg_dir, mask );

 struct ffblk ff;
 long size=0, 
      readed=findfirst( msg_dir, &ff, FA_ARCH );

 if( readed )
  size = -1;
 else
  while( readed==0 )
  {
   long cur=atol( ff.ff_name );

   if( size<cur )
    size = cur;

   readed=findnext( &ff );
  }

 delete msg_dir;
 return size;
}
#endif

#ifdef __EMX__
long getMaxMsgNumber( char *msg_path )
{
  char *dirName = (char *)alloca( strlen(msg_path)+1 );

  strcpy( dirName, msg_path );
  dirName[ strlen( dirName)-1 ] = 0;

  DIR *dir = opendir( dirName );

  if( dir==NULL )
    return -1;

  long maxNum(0);

  for( ;; )
  {
    dirent *ent = readdir( dir );
    char *e;
    
    if( ent==NULL )
      break;
      
    long cur = strtol( ent->d_name, &e, 10 );

    if( !stricmp( e, ".msg" ) && cur>maxNum )
      maxNum = cur;
  }

  closedir( dir );
  return maxNum;
}
#endif

struct FIDO_msg {
 char msg_from[36],
      msg_to[36],
      subj[72],
      msg_date[20];
 short times_read,
       dest_node,
       orig_node,
       cost,
       orig_net,
       dest_net;
 long date_written,
      date_arrived;
 short reply;
 unsigned short attributes;    /* Bit 0    Private
				1    Crash
				2    Message read
				3    Message sent
				4    File-attach
				5    Forwarded
				6    Orphan
				7    Kill/sent
				8    Local message
				9    Hold
			       10    ???
			       11    File-request
			       12    Return-receipt requested
			       13    Return-receipt
			       14    Audit-trail requested
			       15    Update-request  */
 short unreply;
};

void makeMsg( char *path, FILE *file,
	      char *from, char *to, char *subj,
	      char *addr, char *toAddr ){

 fflush( file );

 char *msg_dir( new char [ strlen( path ) + 2 ] ),
      *name( new char [ strlen( path ) + 20 ] );

 checkPtr( msg_dir );
 checkPtr( name );

 strcpy( msg_dir, path );

 if( msg_dir[ strlen( msg_dir )-1 ] != '\\' )
  strcat( msg_dir, "\\" );

 long maxMsgNum = getMaxMsgNumber( msg_dir );

 if( maxMsgNum==-1 ){
  IOerror(msg_dir);
  completeCode = 12;
  delete msg_dir;
  delete name;
  return;
 }

 sprintf( name, "%s%d.msg", msg_dir, int((maxMsgNum+1)&0xffff) );
 printf( "Creating Fido msg-base message `%s'\n", name );

 FILE *in( fopen( name,"wb") );

 if( in==NULL ){
  IOerror(name);
  completeCode = 12;
  delete msg_dir;
  delete name;
  return;
 }

 struct time t;
 struct date d;

 gettime( &t );
 getdate( &d );

 char *text( new char [ 256 ] );

 checkPtr( text );

 unsigned short zone(0), net(0), node(0), point(0);
 char *domain,
      *pntKludge1( new char [32] ),
      *pntKludge2( new char [32] ),
      *intlKludge( NULL );

 checkPtr( pntKludge1 );
 checkPtr( pntKludge2 );

 checkAddress( addr, zone, net, node, point, &domain, 1 );

 unsigned short toZone(0), toNet(0), toNode(0), toPoint(0);

 if( toAddr!=NULL )
   checkAddress( toAddr, toZone, toNet, toNode, toPoint, &domain, 1 );

 if( point )
   sprintf( pntKludge1, "\01FMPT %u\r\n", point );

 if( toPoint )
   sprintf( pntKludge2, "\01TOPT %u\r\n", toPoint );

 if( toAddr!=NULL )
 {
   char *klName( "\01INTL " );
   intlKludge = new char [ strlen( addr )+strlen( toAddr )+
			   strlen( klName )+1+3 ];

   checkPtr( intlKludge );

   strcpy( intlKludge, klName );
   strcat( intlKludge, toAddr );
   strcat( intlKludge,  " " );
   strcat( intlKludge, addr );
   strcat( intlKludge,  "\r\n" );
 }

 sprintf( text,
  "%s"
  "%s"
  "%s"
  "\01MSGID: %s %08lx\r\n"
  "\01PID: Vagabondo"
#if defined(__OS2__)
  "/2"
#endif
  "\r\n",
  intlKludge!=NULL ? intlKludge : "",
  point ? pntKludge1 : "",
  toPoint ? pntKludge2 : "",
  addr,
  time( NULL) );

 delete pntKludge1;
 delete pntKludge2;

 if( intlKludge!=NULL )
   delete intlKludge;

 FIDO_msg *msg( new FIDO_msg );

 checkPtr( msg );

 memset( msg, 0, sizeof(FIDO_msg) );

 strncpy( msg->msg_from, from, 36 );
 strncpy( msg->msg_to, to, 36 );
 strncpy( msg->subj, subj, 72 );

 sprintf( msg->msg_date, "%02d %3s %02d  %02d:%02d:%02d",
  d.da_day, monthName[d.da_mon-1], d.da_year%100,
  t.ti_hour, t.ti_min, t.ti_sec );

 msg->times_read=0;
 msg->orig_node=node;
 msg->dest_node=toNode;
 msg->cost=1;
 msg->orig_net=net;
 msg->dest_net=toNet;
 msg->date_written=0;
 msg->date_arrived=0;
 msg->reply=0;
 msg->attributes=0x0100;
 msg->unreply=0;

 if( fwrite( msg, 1, sizeof(FIDO_msg), in )!=sizeof(FIDO_msg) ){
err_write:
  IOerror("Can't write to *.msg base");
  completeCode = 12;
  delete msg_dir;
  delete name;
  delete text;
  delete msg;
  fclose( in );
  return;
 }

 if( fwrite( text, 1, strlen(text), in )!=strlen(text) )
   goto err_write;

 rewind( file );

 for( ;; )
 {
   int readed = getc( file );

   if( readed==EOF )
     break;
   else
     if( readed=='\n' )
     {
       if( putc( '\r', in )==EOF )
	 goto err_write;
     }
     else
       if( putc( (int)readed, in )==EOF )
	 goto err_write;

 }

 if( putc( 0, in )==EOF )
   goto err_write;

 fclose( in );

 delete msg_dir;
 delete name;
 delete text;
 delete msg;
}
